/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    FixMMStimetags
* 
* MODULE:	
*
* LANGUAGE:	C++
*
* FUNCTION:
*
* PURPOSE:      Calculates first (J) and second (JxB/ne) terms of Ohms law based
*               on B field, ion and electron moments and the E-field.
* 
* DESCRIPTION:
*
* ARGUMENTS:
*    IN:
*     VECTOR_TIME_SERIES    Bfield, Vion, Velec, Efield   
*     SCALAR_TIME_SERIES    Densities
*     TIME_INTERVAL         Time                 time to calculate curl or div for
* 
*    OUT:
*     VECTOR_TIME SERIES    J, JxB/ne
*
* GLOBALS:
*
* RETURN:
*    TYPE:     QplugReturnStatus
*    VALUES:   see qplug_if.h
*
*
* ALGORITHM:   See chap ISSI Auroral Physics, Eq 3.14 or any plasma physics text book
*
* HISTORY:	SEH, 21 Sep 2016, Firts version
*
***********************************************************************/

#include "DvObj.h"
#include "qplug_if.h"
#include "qdutil.h"


extern  "C" QplugReturnStatus  DvCentreTimetags  (vector <DvObject_var>&, vector <DvObject_var>&);


using namespace QSAS;

QplugReturnStatus DvCentreTimetags (vector <DvObject_var>& input, vector <DvObject_var>& outputs)
{

    DvString out_name = GetSlotText(0);
    DvObject_var out_ptr;
    if (out_name.blank() || out_name.empty()){
        out_ptr = input[0];  // work on WL object
    } 
    else {
        // work on copy
        out_ptr = new DvObject(input[0]);
    }

  DvObject_var tt_in = out_ptr->get_xref("DEPEND_0");
  DvObject_var tt = new DvObject(tt_in);
  
  if (!tt.is_ok() || !tt->get_xref("Delta_plus").is_ok()) {
    QplugAppendTextDisplay("No such attribute");
    return QPLUG_SUCCESS;
  }

  double dp=tt->get_xref("Delta_plus")->asDouble();
  double dm=tt->get_xref("Delta_minus")->asDouble();

  double dt = fabs(dp-dm)/2.0;
  if (dt < 1e-6) {
    QplugAppendTextDisplay("Already centered");
    return QPLUG_SUCCESS;
  }

  printf ("Deltas = %g   %g  %g \n",dm,dp,dt);
  DvObject_var delta_var = new DvObject(dt);
  tt+=delta_var;
  tt->change_xref("DELTA_PLUS",delta_var);
  tt->change_xref("DELTA_MINUS",delta_var);
  out_ptr->change_xref("DEPEND_0",tt);
  
  if(out_ptr->get_id() != input[0]->get_id()) outputs.push_back(out_ptr);
  
  QplugAppendTextDisplay("Timetags centered");
  return QPLUG_SUCCESS;
}


	

