// ------------------------------------------------------------------
// ISSI QSAS PlugIn : deHoffman-Teller analysis
// 
// This plugin takes a magnetic field vector and a plasma velocity
// vector as inputs, and calculates the deHoffman-Teller velocity
//
// Required inputs :
// 	bfield		- vector time series with magnetic field
//	velocity	- vector time series with plasma velocity
//
// Optional inputs (leave empty or '0' if not used):
//      density         - scalar time series with plasma density
//      t_par           - scalar time series with paralell plasma temperature
//      t_perp          - scalar time series with perpendicular plasma temperature
//      normal          - vector unit vector with boundary normal to dot inth V_HT
//
// Outputs
//      E_ht            - vector time series with Vht x B
//      E_C             - vector time series with V x B
//
//      Also, the following outputs are valid if a density is given :
//      V_rest          - vector time series with Vr = V - Vht 
//      V_alfven        - vector time series with the Alfven velocity 
//
//
// SEH, 19 Oct, 2001
//      29 May, 2002 - added a normal as input and V_ht * n
//      12 Aug, 2002 - added density as optional input
//      05 Sep, 2002 - temporary bug fix : return all WS outputs
//      15 Jan, 2003 - V_rest as output also without density + ccoeff for Va/Vrest
//      12 Dec, 2003 - modified for QSAS_2.0
//      21 Jan, 2004 - time interval output + QString for output window
//      30 Mar, 2004 - pressure correction introduced
//      05 May, 2004 - use qdos classes and methods -> remove 'minvert.c' dependence
//      06 May, 2004 - incorporated AJA's input checking  + Doxygen info
//      18 Jun, 2004 - logfile possible
//      30 Jan, 2005 - corrected bug in pressure anisotropy
//      13 May, 2005 - catch sigularity in alpha calculation; if B^2==0, or alpha >> 1, set alpha = 0
//      21 Jan, 2008 - hack to allow Themis temperatur tensor in addition to T_par, T_perp
//      22 Jan, 2008 - removed (comment out) calls to minvar
//	    27 Jan, 2008 - temporary removed QT dependencies
//      30 Jan, 2008 - reinserted QT dependencies + added hack for Themis SI conversion n = [10^6>m^-3]
//      04 June 2015 - migrate to DVOS
//      03 Nov, 2016 - fixed bug in VA calculation for cases w/o alpha correction
//      12 Nov, 2016 - workaround to fix DV issues - see line 706
//      15 Nov, 2016 - fixed timetag issues + cleaned objects prior to cc- and slope calculation
//
//
// Doxygen info :
//
/* -*- mode: C++ ; c-file-style: "stroustrup" */
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include <qstring.h>
#include <qdatetime.h>
#ifndef DV_GAP_WIDTH 
#define DV_GAP_WIDTH "join gap or boxcar width"
#endif
#ifndef DV_BOX_WIDTH 
#define DV_BOX_WIDTH "join gap or boxcar width"
#endif

double QSAS_vecsum2(DvObject_var Vector);
double QSAS_vecsum2(DvObject_var VectorSeq, long index);
double minimizeD(int M, DvObject_var V, DvObject_var vm, DvObject_var Bm);
DvObject_var V_HTcalc(long nestsize, DvObject_var tt, DvObject_var &B, DvObject_var &V, DvObject_var &K0c,     
		      DvObject_var &K1, DvObject_var &K2, DvObject_var &KV,  DvObject_var &KVT);

DvObject_var str2Vec(DvString s);
DvObject_var aveVectorOver(DvObject_var VectorSeq, long sta, long sto);
double aveScalarOver(DvObject_var ScalarSeq, long sta, long sto);
double correlationCoeff(DvObject_var, DvObject_var,int);
double slope(DvObject_var, DvObject_var,int);

extern "C" QplugReturnStatus  DvHofTel(vector <DvObject_var>&, vector <DvObject_var>&);

using namespace QSAS;




QplugReturnStatus  DvHofTel  ( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
// ---------------------------------------------------------------------------
/*!
  \brief Perform a deHoffmann-Teller analysis and return the results as QSAS_objects and outputs 
  \brief performs a deHoffmann-Teller analysis
  \param call_list - pointer to a list of input objects
*/
//
// SEH, 15 Aug, 2002
//      06 May, 2004
//
// ---------------------------------------------------------------------------
{ 
  QTime t;
  t.start();                     // start clock

  long    nestsize;
  char    txtbuf[255];
  QplugAppendTextDisplay("HofTel> Version 23 Nov 2016 1730LT");
 

  // ---- get density if supplied -> 3 instead of 2 series to join ----------

  DvObject_var density =  new DvObject(inputs[2]);
  if (density.is_ok() ) {
    if (density->not_dbl() || density->arraySize() != 1 ){
      QplugAppendTextDisplay("density object not scalar double\n");
    }
    else{
      QplugAppendTextDisplay("HofTel> Density supplied.. ");
    }
  }
  else {
    QplugAppendTextDisplay("HofTel> Density not supplied.. ");
  }


 
  // ---- get Pressure if supplied -> 5 instead of 2 series to join ----------

  bool alpha = false; 
  DvObject_var temp_par =  new DvObject(inputs[4]);
  DvObject_var temp_perp =  new DvObject(inputs[5]);
  DvObject_var temp_tensor =   new DvObject(inputs[4]);
  if (temp_par->arraySize() == 1 && temp_perp->arraySize() == 1 && temp_par->asInt() != 0 && temp_perp->asInt() != 0 ) {
    QplugAppendTextDisplay("HofTel> Temp supplied as t_par and t_perp.. ");
    alpha = true;
  } 
  else if (temp_tensor->arraySize() > 1) {
    QplugAppendTextDisplay("HofTel> Temp supplied as tensor.. ");
    alpha = true;
  }  
  else {
    QplugAppendTextDisplay("HofTel> Temp not supplied.. ");
  }

  QplugAppendTextDisplay("HofTel> Letting DVOS join on the fly ");


  DvObject_var  density_nest;
  DvObject_var  Tpar_nest;
  DvObject_var  Tperp_nest;
  DvObject_var  tt_out;                                  // tt of masked V and B
  DvObject_var  normal;                                  // the normal to be dotted into Vht

  //  ---- get Time interval and mask ----------------

  printf("HT : get time interval \n");
  fflush(stdout);
  DvEvent start_end = inputs[3]->getTimeRange();


  //  ---- get B-field -------------------------------

  printf("HT : get masked B-field \n"); fflush(stdout);
 
  DvObject_var  bfield_nest = new DvObject(inputs[0]);

  if(bfield_nest.is_nil())
    {
      QplugAppendTextDisplay("B is nil");
      return QPLUG_FAILURE;
    }
 
  // added by AJA to ensure B field units are correct (nT)
//  bfield_nest = bfield_nest->subSequence(start_end);  
  bfield_nest = bfield_nest->changeUnitsTo("1.0e-9>T", "nT");

 
  if( bfield_nest.is_nil() ){
    QplugAppendTextDisplay("ERROR: no B field data in selected interval");
    return QPLUG_FAILURE;
  }

  nestsize = bfield_nest->seqSize();
  printf("HT : Nestsize B = %ld\n",nestsize);
  fflush(stdout);
  
  //  ---- get Velocity (or E-field) ----------------------

  printf("HT : get velocity or E-field \n"); fflush(stdout);

  DvObject_var velocity_nest = new DvObject(inputs[1]);
  velocity_nest = velocity_nest->subSequence(start_end);
  DvObject_var nested_E;

  // added by AJA to ensure velocity units are correct (km s^-1)
  if (velocity_nest->sameBaseSI("1.0e3>m s^-1")){
    // base units the same, ensure conversion is too
    QplugAppendTextDisplay("HofTel> Input is plasma velocity...");
    velocity_nest = velocity_nest->changeUnitsTo("1.0e3>m s^-1");
    nested_E = velocity_nest; // for diagnostic print only
  }
  else if (velocity_nest->sameBaseSI("1.0e-3>V m^-1")){
    // --- SEH : input is E-field -> generate V from E x B ----
    QplugAppendTextDisplay("HofTel> Input is electric field...");
    nested_E = velocity_nest->changeUnitsTo("1.0e-3>V m^-1");

    velocity_nest =  nested_E->vec(bfield_nest);
    velocity_nest /= bfield_nest->power(2);
    *velocity_nest *=  1000.;
  }
  else {
    QplugAppendTextDisplay("ERROR: cannot correct units of velocity");
    return QPLUG_FAILURE;
  }
 

  if( velocity_nest->nError() > 0 ){
    QplugAppendTextDisplay("ERROR: no V or E data in selected interval");
    QplugAppendTextDisplay(velocity_nest->getError());
    return QPLUG_FAILURE;
  }
 
  nestsize = velocity_nest->seqSize();
  printf("HT : Nestsize V = %ld\n",nestsize);
  fflush(stdout);

  tt_out = velocity_nest->getTimeTags();
  double deltat, avg_width;
  tt_out->isRegular(deltat); // need spacing but not regular tt
  avg_width = 2*deltat;    // nyquist
 
  tt_out->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  tt_out->change_xref(DV_GAP_WIDTH, avg_width);
  tt_out->change_xref(DV_GAP_OPTION, DV_LINEAR);

  bfield_nest->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  bfield_nest->change_xref(DV_GAP_WIDTH, avg_width);
  bfield_nest->change_xref(DV_BOX_WIDTH, avg_width);
  bfield_nest->change_xref(DV_GAP_OPTION, DV_LINEAR);

  velocity_nest->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  velocity_nest->change_xref(DV_BOX_WIDTH, avg_width);
  velocity_nest->change_xref(DV_GAP_WIDTH, avg_width);
  velocity_nest->change_xref(DV_GAP_OPTION, DV_LINEAR);

  bfield_nest = bfield_nest->Join(tt_out); // do this explicitly to avoid repeated joins
  velocity_nest = velocity_nest->Join(tt_out); // do this explicitly to avoid repeated joins




  // ---- get density if supplied -------

  printf("HT : get density if supplied \n");
  fflush(stdout);
  if (density.is_ok()) {
     density_nest->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
     density_nest->change_xref(DV_GAP_WIDTH, avg_width);
     density_nest->change_xref(DV_GAP_OPTION, DV_LINEAR);
     density_nest = density->Join(tt_out);
	
    // -- added by AJA to ensure density units are correct (1/cc) --
    density_nest = density_nest->changeUnitsTo("1.0e6>m^-3"); // DvString understands 1.0e6 or 10^6
	
  
    if( density_nest->nError() > 0 ){
      QplugAppendTextDisplay("ERROR: no density data in selected interval");
      QplugAppendTextDisplay(density_nest->getError());
      return QPLUG_FAILURE;
    }
    nestsize = density_nest->seqSize();
    printf("HT : Nestsize rho = %ld\n",nestsize); fflush(stdout);
  }



  // ---- get pressure if supplied -------

  printf("HT : get temperature if supplied \n"); fflush(stdout);
  if (alpha) {
    if (temp_tensor->arraySize() == 3) {
   
      // for Themis P tensor diagonal
      temp_tensor = temp_tensor->Join(tt_out);
      temp_tensor = temp_tensor->changeUnitsTo("1.602e-19>J");
	 
      // -- SEH 20 Jan, 2008 : extract tpar and t_perp from Themis magt3
      double inv_k_boltzmann = 1.602e-19/1.38054e-17; // [MK/ev]
	 
      Tpar_nest = temp_tensor->getArrayComponent(2);
      Tperp_nest = temp_tensor->getArrayComponent(1);
	   
    } 
    else {    

      Tpar_nest = temp_par->Join(tt_out);	 
      Tpar_nest = Tpar_nest->changeUnitsTo("1.0e6>K");
      Tperp_nest = temp_perp->Join(tt_out);	 
      Tperp_nest = Tperp_nest->changeUnitsTo("1.0e6>K");
	 	
 
      if (Tperp_nest->nError() > 0 ) {
	alpha= false;
     	QplugAppendTextDisplay("ERROR: no Tperp data in selected interval");
     	QplugAppendTextDisplay(Tperp_nest->getError());
      }
	 
      if (Tpar_nest->nError() > 0 ) {
	alpha= false;
     	QplugAppendTextDisplay("ERROR: no Tpar data in selected interval");
     	QplugAppendTextDisplay(Tpar_nest->getError());
      }
	     
      nestsize = Tperp_nest->seqSize();
      printf("HT : Nestsize temp = %ld\n",nestsize); fflush(stdout);
    }

  }


  // ---- get normal if supplied ----------

  printf("HT : get normal if supplied \n"); fflush(stdout);

  normal = inputs[6];
  if (normal->not_str() && !(normal->totalSize() == 1 && normal->asInt() == 0) ) {
    printf("HT : Normal supplied as QSAS object...\n"); fflush(stdout);
  }
  else if (normal->is_str() && !normal->str().empty()) {
    DvString s = normal->str();
    normal = str2Vec(s);

    printf("HT : Normal supplied as string...[%g %g %g]\n",normal->asDouble(0), normal->asDouble(1), normal->asDouble(2));
  }
  else {
    normal = new DvObject(0.0, DIM3);
    printf("HT : Normal not supplied...\n");
  }
     
  fflush(stdout);

 
  // --- declare various objects and call Vht function ----

  printf("HT : declare HT objects \n");
  fflush(stdout);
  DvObject_var Vht_x_B = new DvObject(0.0, DIM3, nestsize);
  DvObject_var V_x_B   = new DvObject(0.0, DIM3, nestsize);
  DvObject_var V_rest   = new DvObject(0.0, DIM3, nestsize);
  DvObject_var V_alf = new DvObject(0.0, DIM3, nestsize);  
 
  DvObject_var K0c  = new DvObject(0.0, DIM33);
  DvObject_var K1 = new DvObject(0.0, DIM33);
  DvObject_var K2 = new DvObject(0.0, DIM33);
  DvObject_var KV = new DvObject(0.0, DIM3);
  DvObject_var KVT = new DvObject(0.0, DIM3);
  DvObject_var V_ht0 = new DvObject(0.0, DIM3);
  DvObject_var a_ht  = new DvObject(0.0, DIM3);

 
  V_x_B->change_xref(CATDESC,"E_C = V_x_B convection electric field");
  Vht_x_B->change_xref(CATDESC,"E_HT = Vht_x_B residual electric field");

  DvObject_var V_ht = V_HTcalc(nestsize, tt_out, bfield_nest, velocity_nest, K0c, K1, K2, KV, KVT);


  // ----- Accelerating HT frame - see blue book sec 9.5 ----------
 
  printf("HT : accelerating HT frame \n"); fflush(stdout);
  vector <size_t> dims66;
  dims66.push_back(6);
  dims66.push_back(6);
  DvObject_var K66c = new DvObject( 0.0, dims66);


  for (int i=0; i<3; i++) for (int j=0; j<3; j++) {
    K66c->dbl(0,i,j) = K0c->dbl(0,i,j);
    K66c->dbl(0,i,j+3) = K1->dbl(0,i,j);
    K66c->dbl(0,i+3,j) = K1->dbl(0,i,j);
    K66c->dbl(0,i+3,j+3) = K2->dbl(0,i,j); 
  }

  DvObject_var Kinv66 = K66c->inverse();
 
  for (int i=0; i<3; i++) { 
    V_ht0->dbl(i) = Kinv66->dbl(0,i,0) * KV->dbl(0) + Kinv66->dbl(0,i,1) * KV->dbl(1) + Kinv66->dbl(0,i,2) * KV->dbl(2) + Kinv66->dbl(0,i,3) * KVT->dbl(0) + Kinv66->dbl(0,i,4) * KVT->dbl(1) + Kinv66->dbl(0,i,5) * KVT->dbl(2);
    a_ht->dbl(i) = Kinv66->dbl(0,i+3,0) * KV->dbl(0) + Kinv66->dbl(0,i+3,1) * KV->dbl(1) + Kinv66->dbl(0,i+3,2) * KV->dbl(2) + Kinv66->dbl(0,i+3,3) * KVT->dbl(0) + Kinv66->dbl(0,i+3,4) * KVT->dbl(1) + Kinv66->dbl(0,i+3,5) * KVT->dbl(2);
  }


  // ----- calc the quantities dotted into the normal --------------

  printf("HT : calc n dot values \n");
  fflush(stdout);
  double V_ht_n=0.0, V_ht0_n=0.0, a_ht_n=0.0;

  V_ht_n   = V_ht->dbl(0) * normal->asDouble(0)  + V_ht->dbl(1) * normal->asDouble(1)  + V_ht->dbl(2) * normal->asDouble(2);
  V_ht0_n  = V_ht0->dbl(0) * normal->asDouble(0) + V_ht0->dbl(1) * normal->asDouble(1) + V_ht0->dbl(2) * normal->asDouble(2);
  a_ht_n   = a_ht->dbl(0) * normal->asDouble(0)  + a_ht->dbl(1) * normal->asDouble(1)  + a_ht->dbl(2) * normal->asDouble(2);


  // ---- error analysis - see 9.4.1 in blue book -------------


  printf("HT : error analysis \n");
  fflush(stdout);
  DvObject_var Suv = new DvObject(0.0, DIM33);
  double sigma_n=0.00;
  double  dvht = 1.0E12 * minimizeD(nestsize, V_ht, velocity_nest, bfield_nest);  // 10^12 doe to nT^2 and km^/s^2 units 

  DvObject_var K0invc = K0c->inverse();

  for (int i=0; i<3; i++) for (int j=0; j<3; j++) Suv->dbl(0,i,j) = K0invc->dbl(0,i,j) *  dvht / (2.0*(double)nestsize - 3.0);
  for (int i=0; i<3; i++) for (int j=0; j<3; j++) sigma_n += normal->asDouble(i) * Suv->dbl(0,i,j) * normal->asDouble(j);


  // ---- send some output to QSAS window -------

  printf("HT : send outputs \n");
  fflush(stdout);
  DvTime t_start = tt_out->time(0);
  DvTime t_stop = tt_out->time(nestsize-1);
  DvTime t_center = t_start + (t_stop - t_start) / 2.;

  QString Status;
  sprintf(txtbuf,"HofTel > t = %s to %s (%ld samples)\n",t_start.getISOstring().c_str()+11,t_stop.getISOstring().c_str()+11,nestsize);
  Status += txtbuf;
  sprintf(txtbuf,"HofTel > Param                 Value                   dot normal:\n");
  Status += txtbuf;
  sprintf(txtbuf,"HofTel > V_HT  = [%10.4f %10.4f %10.4f]  (%9.4f +/- %9.4f) km/s\n",V_ht->dbl(0), V_ht->dbl(1), V_ht->dbl(2), V_ht_n, sqrt(sigma_n));
  Status += txtbuf;
  sprintf(txtbuf,"HofTel > V_HT0 = [%10.4f %10.4f %10.4f]  (%9.4f) km/s\n",V_ht0->dbl(0), V_ht0->dbl(1), V_ht0->dbl(2), V_ht0_n);
  Status += txtbuf;
  sprintf(txtbuf,"HofTel > a_HT  = [%10.4f %10.4f %10.4f]  (%9.4f) km/s^2\n",a_ht->dbl(0), a_ht->dbl(1), a_ht->dbl(2), a_ht_n);
  Status += txtbuf;

  // ----- SEH 12 Jul 2004 : Want to compare V-Vht and V_magnetospheath ----------------

  if (density.is_ok()) {
    long upper = 1+(long) nestsize/10;
    upper = (upper > 5) ? 5 : upper;
    double n0 = aveScalarOver(density_nest,0,upper);                          // ave density on one side
    double n1 = aveScalarOver(density_nest,nestsize-upper-1,nestsize-1);      // ave density on other side
    long start = (n0 > n1) ? 0 : nestsize-upper-1;
    long stop = start+upper;

    fflush(stdout);
   
    DvObject_var Vsh = aveVectorOver(velocity_nest,start,stop);
   
    for (int i=0; i<3; i++) Vsh->dbl(i) -= V_ht->dbl(i);
    sprintf(txtbuf,"\nHofTel > Vsh  = [%10.4f %10.4f %10.4f]  km/s  ",Vsh->dbl(0), Vsh->dbl(1), Vsh->dbl(2));
    Status += txtbuf;
    sprintf(txtbuf,"(average of V samples %ld-%ld - Vht)\n",start,stop);
    Status += txtbuf;
  }

  //  --- generate the two outputs Vht x B and V x B ---------------------


  double cc = 1.0e-03;          // conversion due to B in nT and V in km/s
  double mu_0 = 1.25664e-06;    // modified mu_0 (due to B in nT and dens in 1/ccm)
  double rho,m_p= 1.67e-27;     // rho = mass dens with proton_mass m_p
  double alfa = 0.00;           // for pressure anisotropy
  double sum_alfa=0.0;          // to calc average alpha correction
  double kb = 1.38054e-23;      // boltzmann const
  double bb = 0.0;              // temporary variable to hold B^2


  DvTime timetag=tt_out->time(0);
 
  for (int i=0; i<nestsize; i++) {
    Vht_x_B->dbl(i,0) = cc * (bfield_nest->dbl(i,1) * V_ht->dbl(2)  - bfield_nest->dbl(i,2) * V_ht->dbl(1));
    Vht_x_B->dbl(i,1) = cc * (bfield_nest->dbl(i,2) * V_ht->dbl(0)  - bfield_nest->dbl(i,0) * V_ht->dbl(2));
    Vht_x_B->dbl(i,2) = cc * (bfield_nest->dbl(i,0) * V_ht->dbl(1)  - bfield_nest->dbl(i,1) * V_ht->dbl(0));


    V_x_B->dbl(i,0) = cc * (bfield_nest->dbl(i,1) * velocity_nest->dbl(i,2)  -  bfield_nest->dbl(i,2) * velocity_nest->dbl(i,1));
    V_x_B->dbl(i,1) = cc * (bfield_nest->dbl(i,2) * velocity_nest->dbl(i,0)  -  bfield_nest->dbl(i,0) * velocity_nest->dbl(i,2));
    V_x_B->dbl(i,2) = cc * (bfield_nest->dbl(i,0) * velocity_nest->dbl(i,1)  -  bfield_nest->dbl(i,1) * velocity_nest->dbl(i,0));

    for (int j=0; j<3; j++) V_rest->dbl(i,j) =  velocity_nest->dbl(i,j) - V_ht->dbl(j);

    if (density.is_ok()) {
      timetag = tt_out->time(i);
      rho = density_nest->dbl(i) * 1.e06 * m_p;
      if (alpha) {
	bb = 1e-18*QSAS_vecsum2(bfield_nest,i);                                              // B^2 in [T]
	alfa = mu_0 * 1e12 * density_nest->dbl(i) * kb * (fabs(Tpar_nest->dbl(i)) - fabs(Tperp_nest->dbl(i))) / bb;     // 1e12 = cm^3 -> m^3 and MK -> K
	if (bb < 1e-31) alfa = 0.00;
	sum_alfa+=alfa;
      }
      for (int j=0; j<3; j++)  {
	 if(rho < 1.e-30) {
	    // limit to speed of light
	    double scale = 3.e5 / sqrt( QSAS_vecsum2(bfield_nest,i) );
	    V_alf->dbl(i,j) = bfield_nest->dbl(i,j) * scale;
	  }
	  else {
	    V_alf->dbl(i,j)  = sqrt(1.0-alfa) * 1e-03 * bfield_nest->dbl(i,j) * 1e-09 / sqrt(mu_0 * rho);
          }
       }
    }
  } 
  
 
  printf("HT : checking frame \n");
  fflush(stdout);
  DvString frame_in = inputs[0]->getXrefText("Frame");
  if( frame_in.empty()) 
    frame_in = inputs[1]->getXrefText("Frame");
  if( frame_in.empty()) 
    frame_in = "vector>???_xyz";


  V_x_B->change_xref(FRAME,frame_in);
  V_x_B->change_xref(UNITS,"mV/m");
  V_x_B->change_xref(SI_CONVERSION,"0.001>V m^-1");
  V_x_B->change_xref(LABLAXIS,"Ec");
  V_x_B->change_xref(DEPEND_0, tt_out);

  Vht_x_B->copy_xrefs_from(V_x_B);
  Vht_x_B->change_xref(LABLAXIS,"E_Ht");



  // ------ Calculate correlation - se first eq on p.228 in blue book ------


  printf("HT : correlation coeff & slope \n");
  fflush(stdout);
  DvObject_var nullvec=new DvObject(0.0, DIM3);
  double minD0 =  minimizeD(nestsize, nullvec, velocity_nest, bfield_nest);
  double minDVht = minimizeD(nestsize, V_ht, velocity_nest, bfield_nest);
  double HTcoeff = sqrt(1.0 - minDVht/minD0);
  sprintf(txtbuf,"\nCorrelation coeffecient (E_HT/E_c) =%10.6f  (after Blue Book, first Eq, p228)\n",HTcoeff);
  Status += txtbuf;

  HTcoeff = correlationCoeff(V_x_B, Vht_x_B, nestsize);
  sprintf(txtbuf,"Correlation coeffecient (E_HT/E_c) =%10.6f  (after classic correlation coeff)\n",HTcoeff);
  Status += txtbuf;

  double HTslope = slope(Vht_x_B, V_x_B, nestsize);
 
  sprintf(txtbuf,"Regression line slope (E_HT/E_C)   =%10.6f\n\n",HTslope);
  Status += txtbuf;
  if (density.is_ok()) {
    double VAcoeff = correlationCoeff(V_alf, V_rest, nestsize);
    if (alpha) {
      sprintf(txtbuf,"Va with alpha correction, <alpha> = %5g)\n",sum_alfa/nestsize);
      Status += txtbuf;
    }
    sprintf(txtbuf,"Correlation coeffecient (Va/Vrest) =%10.6f  (after classic correlation coeff)\n",VAcoeff);
    Status += txtbuf;
    double VAslope = slope(V_alf,V_rest,nestsize);
    sprintf(txtbuf,"Regression line slope (Va/Vrest)   =%10.6f\n",VAslope);
    Status += txtbuf;

    DvObject_var MetaVAslope=new DvObject(VAslope);
    DvObject_var MetaVAcoeff=new DvObject(VAcoeff);
    MetaVAslope->change_xref(CATDESC,"Regression line slope (Va/Vrest)");
    MetaVAcoeff->change_xref(CATDESC,"Correlation coeffecient (Va/Vrest)");
    V_alf->change_xref("Slope",MetaVAslope);
    V_alf->change_xref("Ccoeff",MetaVAcoeff);

  }
  QplugAppendTextDisplay(Status.toLatin1().data());


  V_ht->change_xref(FRAME,frame_in);
  V_ht->change_xref(UNITS,"km/s");
  V_ht->change_xref(SI_CONVERSION,"1e3>m s^-1");
  Vht_x_B->change_xref("V_HT", V_ht);

  DvObject_var MetaHTslope=new DvObject(HTslope);
  MetaHTslope->change_xref(CATDESC,"Regression line slope (E_HT/E_C)");
  DvObject_var MetaHTcoeff=new DvObject(HTcoeff);
  MetaHTcoeff->change_xref(CATDESC,"Correlation coeffecient (E_HT/E_c)");
  Vht_x_B->change_xref("Slope", MetaHTslope);
  Vht_x_B->change_xref("Ccoeff", MetaHTcoeff);




  // ----- Push V_conv x B and V_ht x B back on WL --------------

  printf("HT : push back \n");
  outputs.push_back(Vht_x_B);
  outputs.push_back(V_x_B);

  if (density.is_ok()) {
    V_rest->change_xref(FRAME, frame_in);
    V_rest->change_xref(UNITS,"km/s");
    V_rest->change_xref(SI_CONVERSION,"1.0e3 > m s^-1");
    V_rest->change_xref(DEPEND_0, tt_out);
    outputs.push_back(V_rest);

    V_alf->copy_xrefs_from(V_rest);
    V_alf->change_xref(LABLAXIS,"V_Alfven");
    outputs.push_back(V_alf);
  }
  else{
    // be safe and put null object on argument list
    // this would allow these possibly unused outputs 
    // to not be at the end of output list
   
    DvObject_var empty1 = new DvObject();
    DvObject_var empty2 = new DvObject();
    outputs.push_back(empty1);
    outputs.push_back(empty2);
  } 

 
  qDebug("HT : Time elapsed : %d\n", t.elapsed() ); // prints the number of msecs elapsed

  return QPLUG_SUCCESS;
}






double minimizeD(int M, DvObject_var V, DvObject_var Vm, DvObject_var Bm)
// -----------------------------------------------------------------------------------------
/*!
  \brief Calculate the least square residue between two vectors
  \return the least square residue
  \param M - number of elements in the input vectors
  \param V - vector time series 
  \param vm - Vector time series 
  \param Bm - vector time series 

  Calculates the residual from Eq 9.8 in blue ISSI book

*/
//
// SEH, Nov, 2001
//      May, 2002
//      May, 2004,  using qdos classes rather than own VEC structures
//
// V,vm [km/s], B [nT]
//
// -----------------------------------------------------------------------------------------
{
  int i;
  double E_abs, Dsum=0.0;
  DvObject_var Em = new DvObject(0.0, DIM3);

  for (i=0; i<M; i++) {
    Em->dbl(0) = ((Vm->dbl(i,1) - V->dbl(1))*Bm->dbl(i,2) - (Vm->dbl(i,2) - V->dbl(2))*Bm->dbl(i,1));
    Em->dbl(1) = ((Vm->dbl(i,0) - V->dbl(0))*Bm->dbl(i,2) - (Vm->dbl(i,2) - V->dbl(2))*Bm->dbl(i,0));
    Em->dbl(2) = ((Vm->dbl(i,1) - V->dbl(1))*Bm->dbl(i,0) - (Vm->dbl(i,0) - V->dbl(0))*Bm->dbl(i,1));

    E_abs = 1.0E-12 * QSAS_vecsum2(Em);
    Dsum = Dsum + E_abs;
  }
  return (Dsum / (double) M);   // in units of (V/m)^2
}





DvObject_var V_HTcalc(long nestsize, DvObject_var tt, DvObject_var &B, DvObject_var &V,
		      DvObject_var &K0, DvObject_var &K1, DvObject_var &K2, DvObject_var &KV, DvObject_var &KVT)
// ---------------------------------------------------------------------------
/*!
  \brief calculates the deHoffmann-Teller velocity
  \return a vector with the deHoffmann-Teller velocity in the same frame as the input data
  \param nestsize - number of elements in the B-field and Velcity vectors
  \param tt - time tags - used for the accellerated HT frame
  \param B - vector time series with magnetic field
  \param V - vector time series with plasma velocity
  \param K0 - 3x3 projection matrix - see Blue ISSI book, p 226
  \param K1 - 3x3 projection matrix - see Blue ISSI book, p 240
  \param K2 - 3x3 projection matrix - see Blue ISSI book, p 240
  \param KV - vector 
  \param KVT - vector 

  This routine calculates a deHoffmann-Teller velocity and fills in the various
  projection matrices as given by the ISSI book on Multi-spacecraft analysis
  (ISSI SR-001)

*/
//
// SEH, Nov, 2001
//      May, 2002
//      Dec, 2003 Modified for Qsas 2. More to be done (use more qdos classes)
//      May, 2004 use more qdos classes and methods, e.g. invert
//      May, 2004 Doxygen docs
//      Nov, 2016 DVOS version
//
// -----------------------------------------------------------------------------------------
{
  
  // modified to use timetags in case of data gaps.
  DvTime t0 = tt->time(0);
  
  double time;

  DvObject_var V_ht = new DvObject(0.0, DIM3);
  DvObject_var B_k = new DvObject(0.0, DIM3);

  // K - 3x3 projection matrix - see Blue ISSI book, p 226

  DvObject_var K = new DvObject(0.0, DIM33);

  for(int k=0; k<nestsize; k++)  {         


    time = tt->time(k) - t0;

    for (int j=0; j<3; j++) {
      B_k->dbl(j) = B->dbl(k,j);
    }
    double bb = QSAS_vecsum2(B_k);

    for (int i=0; i<3; i++) {
      for (int j=0; j<3; j++) {
	K->dbl(0,i,j) = (double)(i==j) * bb - (B_k->dbl(i)*B_k->dbl(j) );
      }
    }
    
    for (int i=0; i<3; i++) {
      for (int j=0; j<3; j++) {
	K0->dbl(0,i,j) += K->dbl(0,i,j);
	K1->dbl(0,i,j) += K->dbl(0,i,j) * time;
	K2->dbl(0,i,j) += K->dbl(0,i,j) * time * time;
	KV->dbl(i) += K->dbl(0,i,j) * V->dbl(k,j);
	KVT->dbl(i) += K->dbl(0,i,j) * V->dbl(k,j) *time;           
      }
    } 

  }

  *KV  /= (double) nestsize;
  *KVT /= (double) nestsize;
  *K0  /= (double) nestsize;                                   // K0 == <K>
  *K1  /= (double) nestsize;
  *K2  /= (double) nestsize;

 
  DvObject_var Kinv = K0->inverse();

  for (int i=0; i<3; i++) {
    V_ht->dbl(i) = Kinv->dbl(0,i,0) * KV->dbl(0) + Kinv->dbl(0,i,1) * KV->dbl(1) + Kinv->dbl(0,i,2) * KV->dbl(2); 
  }

  return V_ht;
}


double QSAS_vecsum2(DvObject_var Vector)
// ---------------------------------------------------------------------------
/*!
  \brief Calculates the length of a vector
  \return the norm^2 of a vector
  \param Vector - vector 

*/
//
// SEH, 05 May 2004
//
// ---------------------------------------------------------------------------
{
  double sum=0.0;

  for (int i=0; i<3; i++) sum += Vector->dbl(i) * Vector->dbl(i);

  return sum;
}



double QSAS_vecsum2(DvObject_var VectorSeq, long index)
// ---------------------------------------------------------------------------
/*!
  \brief Calculates the length of a vector
  \return the norm^2 of a vector TS
  \param VectorSeq - vector series 
  \param index - select element in vector series to calculate norm^2
  
  Overloaded of below

*/
//
// SEH, 05 May 2004
//
// ---------------------------------------------------------------------------
{
  double sum=0.0;

  for (int i=0; i<3; i++) sum += VectorSeq->dbl(index,i) * VectorSeq->dbl(index,i);

  return sum;
}


DvObject_var aveVectorOver(DvObject_var VectorSeq, long start, long stop)
// ---------------------------------------------------------------------------
/*!
  \brief Calculates the average of a vector between two indices
  \return a Vector with the average
  \param VectorSeq - vector series 
  \param start, stop - indices to average over
*/
//
// SEH, 05 May 2004
//
// ---------------------------------------------------------------------------
{

  DvObject_var v = new DvObject(0.0, DIM3);

  for (int i=start; i<stop; i++) 
    for (int j=0; j<3; j++) v->dbl(j) += VectorSeq->dbl(i,j);

  *v /= (double) abs(stop - start );

  return v;
}



double aveScalarOver(DvObject_var ScalarSeq, long start, long stop)
// ---------------------------------------------------------------------------
/*!
  \brief Calculates the average of a scalar between two indices
  \return a Scalar with the average
  \param ScalarSeq - vector series 
  \param start, stop - indices to average over

  use overloading ?
*/
//
// SEH, 12 Jul 2004
//
// ---------------------------------------------------------------------------
{

  double s = 0;
  for (int i=start; i<stop; i++)  s += ScalarSeq->dbl(i);

  s /= (double) abs(stop - start);

  return s;
}



// -------------------------------------------------------------------
/*!
  \brief Parses the a text and converts text to a Vector
  \return  QSAS matrix with the parsed values of the string
  \param s - a string of the form '[2.00 -23.0  9.0]'
*/
//
// SEH, Mar 27, 2003
//
// -------------------------------------------------------------------
DvObject_var str2Vec(DvString s)
{
  DvObject_var v = new DvObject(0.0, DIM3);
  s.strip('[');
  s.strip(']');
  s=s.trimmed(); // remove surrounding white space
  v->dbl(0) = s.before(' ').toDouble();
  s=s.after(' ');
  s=s.trimmed();
  v->dbl(1) = s.before(' ').toDouble();
  s=s.after(' ');
  s=s.trimmed();
  v->dbl(2) = s.toDouble();
	
  return(v);
}




double slope(DvObject_var X_inU, DvObject_var Y_inU, int nU)
// ---------------------------------------------------------------------------
//
//  calculates the regression line, i.e. the coeff b in y(x) = a + bx
//
//
// SEH, 15 Aug, 2002
//      21 Nov 2003, grabbed from HofTel + converted to return a double
//      15 Nov 2016, clean inputs before calculations
//
// ---------------------------------------------------------------------------

{
  DvMask msk;
  bool OK = X_inU->getCleanMask(msk);
  DvMask msk_y;
  OK += Y_inU->getCleanMask(msk_y);

  DvObject_var X_in = new DvObject(X_inU); // apply_mask does modify the object itself
  DvObject_var Y_in = new DvObject(Y_inU);
  if(OK){
    msk += msk_y; // adds records to remove
    X_in->apply_mask(msk);
    Y_in->apply_mask(msk);
  }

  int n=X_in->seqSize();
  int numpts = 3*n;

  vector<double> Xc(numpts, 0.0);
  vector<double> Yc(numpts, 0.0);
  vector<double> Tc(numpts, 0.0);
  
  for (int i=0; i<n; i++) {
    Xc[i] = X_in->dbl(i,0);
    Xc[i+n] = X_in->dbl(i,1);
    Xc[i+2*n] = X_in->dbl(i,2);
    Yc[i] = Y_in->dbl(i,0);
    Yc[i+n] = Y_in->dbl(i,1);
    Yc[i+2*n] = Y_in->dbl(i,2);
  }
  
  double sum_X = 0.0;
  double sum_Y = 0.0;
  for(int i=0; i< numpts; i++){
  	sum_X += Xc[i];
  	sum_Y += Yc[i];
  }
  double ave_X = sum_X / (double) (numpts);
  double ave_Y = sum_Y / (double) (numpts);
  double  b=0.0, tt=0.0;

  for (int i=0; i<numpts; i++) {
    Tc[i] = Xc[i] - ave_X;
    b += Tc[i] * Yc[i];
    tt += Tc[i] * Tc[i];
  }

  b /= tt;

  if (n != nU) {
    char txtbuf[255];
    sprintf(txtbuf,"HofTel> slope data cleaned: n_in = %d, n_used = %d",nU,n);
    QplugAppendTextDisplay(txtbuf);
    printf("%s\n",txtbuf);
  }
  return(b);
}



double correlationCoeff(DvObject_var X_inU, DvObject_var Y_inU, int nU)
// ---------------------------------------------------------------------------
//
//  calculates the correlation coefficient
//
//
// SEH, 15 Jan 2003 
//      21 Nov 2003, grabbed from HofTel + converted to return a double 
//      15 Nov 2016, clean inputs before calculations
//
// ---------------------------------------------------------------------------
{

  DvMask msk;
  bool OK = X_inU->getCleanMask(msk);
  DvMask msk_y;
  OK += Y_inU->getCleanMask(msk_y);

  DvObject_var X_in = new DvObject(X_inU); // apply_mask does modify the object itself
  DvObject_var Y_in = new DvObject(Y_inU);
  if(OK){
    msk += msk_y; // adds records to remove
    X_in->apply_mask(msk);
    Y_in->apply_mask(msk);
  }
  
  int n=X_in->seqSize();
  int numpts = 3*n;
  double sum_upper=0.0, sum_xx=0.0, sum_yy=0.0;
 
  vector<double> Xc(numpts, 0.0);
  vector<double> Yc(numpts, 0.0);
  vector<double> Tc(numpts, 0.0);


  for (int i=0; i<n; i++) {
    Xc[i] = X_in->dbl(i,0);
    Xc[i+n] = X_in->dbl(i,1);
    Xc[i+2*n] = X_in->dbl(i,2);
    Yc[i] = Y_in->dbl(i,0);
    Yc[i+n] = Y_in->dbl(i,1);
    Yc[i+2*n] = Y_in->dbl(i,2);
  }

  double sum_x=0;
  for(int i=0; i<numpts; i++) sum_x += Xc[i]; 
  sum_x /= (double) numpts;
  double sum_y=0;
  for(int i=0; i<numpts; i++) sum_y += Yc[i]; 
  sum_y /= (double) numpts;

  for (int i=0; i<numpts; i++) {
    sum_upper += (Xc[i] - sum_x) * (Yc[i] - sum_y);
    sum_xx += (Xc[i] - sum_x)*(Xc[i] - sum_x);
    sum_yy += (Yc[i] - sum_y)*(Yc[i] - sum_y);
  }

  if (n != nU) {
    char txtbuf[255];
    sprintf(txtbuf,"HofTel> CC data cleaned: n_in = %d, n_used = %d",nU,n);
    QplugAppendTextDisplay(txtbuf);
    printf("%s\n",txtbuf);
  }

  return (double) (sum_upper / (sqrt(sum_xx) * sqrt(sum_yy)));
}

