/****************************************************************************
							LICENSE

  Copyright (C) 2010 Eugen Sorbalo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  A copy of the GNU General Public License is available on the internet at:

  http://www.gnu.org/copyleft/gpl.html

  or you can write to:

  The Free Software Foundation, Inc.
  675 Mass Ave
  Cambridge, MA 02139, USA
****************************************************************************/

/**
 * \file NablaEstimation.cpp
 * \author: eugenatjacobs
 *
 * \brief Implementation of EugClusterAnalysis class (the gradient estimate part)
 *
 *
 */

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include "ClusterCommons.h"


//------- class CEstimateNormal -----------------------------
EugClusterAnalysis::CEstimateNormalComponentFor3SpacecraftMethod::CEstimateNormalComponentFor3SpacecraftMethod(const unsigned short option)
{
	CEstimateNormalComponentFor3SpacecraftMethod();
	set_option(option);
}

void EugClusterAnalysis::CEstimateNormalComponentFor3SpacecraftMethod::set_option(const unsigned short option)
{
	switch (option) {
		case 1: option_ = kGradientParallel;break;
		case 2: option_ = kGradientPerpendicular;break;
		case 3: option_ = kStationarity;break;
		default: option_ = kNone;
	}
}

int EugClusterAnalysis::CEstimateNormalComponentFor3SpacecraftMethod::computeSlopeOfTimeSeries(DvObject_var &g)
{
	//If it is not stationarity assumption, we still don't have any error.
	if ((option_ == kStationarity) && (slope_.is_nil())) {
		int nestsize = g->seqSize();
		unsigned short nr_spacecrafts = (unsigned short)g->Dims()[0];
		unsigned short nr_dimensions = (unsigned short)g->Dims()[1];
		
		vector <size_t> dims;
		dims.push_back(nr_dimensions);
		//Compute g_mc (the field in the spacecraft constellation mass center):
		DvObject_var V_mc = new DvObject(0.0, dims, nestsize);
		slope_ = new DvObject(0.0, dims, nestsize);
		
		for (size_t coord = 1; coord < nr_dimensions; coord++) {
			for (size_t t = 0; t < nestsize; t++) {
				for (size_t satellite = 0; satellite < 3; satellite++)
					V_mc->dbl(t,coord) += (double) g->dbl(t,satellite,coord);
				V_mc->dbl(t,coord) /= 3.0;

			}
			
			//Linear fit:
			const unsigned int linear_fit_window = 15;
			double time_window = linear_fit_window * 4, half_time_window = time_window / 2; //4 seconds

			double s_t = 0, s_t2 = 0, s_g = 0, s_tg = 0, current_time;
			DvObject_var time_tags = g->getTimeTags();
			
			int t_back = 0, t_current = 0, t_front = 0; //intems from t_back to t_front will be used for slope computation.
			for (t_current = 0; t_current < nestsize; t_current++) {
				while ( (time_tags->time(t_front) - time_tags->time(t_current) < half_time_window) && (t_front+1 < nestsize)) {
					current_time = time_tags->time(t_front) - time_tags->time(0); //everything is done relatively to the first
					s_t += current_time;
					s_t2 += current_time * current_time;
					s_g += V_mc->dbl(t_front,coord);
					s_tg += current_time * V_mc->dbl(t_front,coord);
					t_front++;
				}
				while (time_tags->time(t_front) - time_tags->time(t_back) > time_window) {
					current_time = time_tags->time(t_back) - time_tags->time(0); //everything is done relatively to the first
					s_t -= current_time;
					s_t2 -= current_time * current_time;
					s_g -= V_mc->dbl(t_back,coord);
					s_tg -= current_time * V_mc->dbl(t_back,coord);
					t_back++;
				}
				slope_->dbl(t_current,coord) = ((t_front - t_back + 1) * s_tg - s_g * s_t) /
											((t_front - t_back + 1) * s_t2 - s_t * s_t);
			}
		}
		slope_->copy_xrefs_from(g);

	}
	return 0;
}

double EugClusterAnalysis::CEstimateNormalComponentFor3SpacecraftMethod::get_slope(const int t, const unsigned long coord) const
{
	if (slope_.is_nil()) return 0.0;
	return slope_->dbl(t,coord);
}

// -------- EugClusterAnalysis ----------------------------

Status::ReturnStatus EugClusterAnalysis::Compute4PointsEstimate()
{
	Status::ReturnStatus result = Status::kOk;
	if (nr_satellites_ != 4) {
		status.set_message("This function is for 4 satellites only.");
		result = Status::kWrongInput;
		return result;
	}
	else if (normal_TS_.is_nil() || vol_TS_.is_nil()) {
		status.set_message("The Normal and the Volume are required but at least one of them is null.");
		result = Status::kWrongInput;
		return result;
	}

	vector<size_t> dims;
	dims.push_back(nr_satellites_);
	dims.push_back(3);
	
	DvObject_var K  = new DvObject(0.0, dims); // Reciprocal vector - see Blue book, sec 12.4
	DvObject_var Vs = new DvObject(0.0, dims);
	DvObject_var GradMagn;
	DvObject_var Grad;

	grad_magn_TS_ = new DvObject(0.0, DIM3, nr_measurements_); //if the value is scalar => it is equal to its magnitude.
	if (nr_coordinates_ == 1) {
		//For a scalar field, we don't have any curls, divergents,
		grad_TS_ = new DvObject(0.0, DIM3, nr_measurements_);
		Grad = new DvObject(0.0, DIM3);
	}
	else {
		grad_TS_ = new DvObject(0.0, DIM33, nr_measurements_);
		Grad     = new DvObject(0.0, DIM33);
		GradMagn = new DvObject(0.0, DIM3);
	}


	int t;
	bool volumeError = false;
	
//#pragma omp parallel for private(t)
	for (t = 0; t < nr_measurements_; t++) {
		//1. Make sure the variables are initialized:

		//2. Compute reciprocal vectors (a function of spacecraft positions only).

		// The following lines compute the so-called volumetric tensor. I leave it here for future applications.
		//		DvObject_var R = new QdRMatrix(0.0, DIM33); // Volumetric tensor - see Blue book sec 12.2.1 + 14.2.1
		//		for (short i = 0; i < 3; i++)
		//			for (short j = 0; j < 3; j++) {
		//				for (unsigned short sc = 0; sc < nr_satellites_; sc++)
		//					(*R)[i][j] += (double) (*position_TS_)[t][sc][i] * (double) (*position_TS_)[t][sc][j];
		//			}

		vol_TS_->dbl(t) = ComputeReciprocal4Spacecrafts(t, K); //the output is the total volume of the tetrahedron.
		
		if(vol_TS_->dbl(t) < 1e-6) volumeError = true;
		
		for (size_t sc = 0; sc < nr_satellites_; sc++) {
			for (size_t coord = 0; coord < nr_coordinates_; coord++) {
				Vs->dbl(0,sc,coord) = field_TS_->dbl(t,sc,coord);
			}
		}

		if (nr_coordinates_ == 1) {
			computeGradScalar(Vs, K, Grad);
			for (short i = 0; i < 3; i++) {
				grad_magn_TS_->dbl(t,i) = Grad->dbl(i);
				grad_TS_->dbl(t,i) = Grad->dbl(i);
			}
		}
		else {
			computeGradMagnitude(Vs, K, GradMagn);
			computeGradVec(Vs, K, Grad);
			for (size_t i = 0; i < 3; i++) {
				grad_magn_TS_->dbl(t,i) = GradMagn->dbl(i);
				for (size_t j = 0; j < 3; j++)
					grad_TS_->dbl(t,i,j) = Grad->dbl(0,i,j);
			}
		}
	}
	if (result == Status::kOk)
		Status::PrintMessage(false, "NablaEstimator: DEBUG: break 3");

	if(volumeError) vol_TS_->error("Tetrahedron Volume becomes too small in interval, results dubious");
	
	return result;
}

Status::ReturnStatus EugClusterAnalysis::Compute3PointsEstimate()
{
	Status::ReturnStatus result = Status::kOk;
	if (nr_satellites_ != 3) {
		status.set_message("This function is for 3 satellites only.");
		result = Status::kWrongInput;
		return result;
	}
	else if (normal_TS_.is_nil() || vol_TS_.is_nil()) {
		status.set_message("The Normal and the Volume are required but at least one of them is null.");
		result = Status::kWrongInput;
		return result;
	}

	vector<size_t> dims;
	dims.push_back(nr_satellites_);
	dims.push_back(3);
	
	DvObject_var K = new DvObject(0.0, dims);  // Reciprocals - see Vogt et al. 2009, formula (12)

	vector<size_t> dimkv;
	dimkv.push_back(kVectorRank);
	DvObject_var Normal = new DvObject(0.0, dimkv);  //Normal to the plane of spacecrafts.

	vector<size_t> dimvs;
	dimkv.push_back(nr_satellites_);
	dimkv.push_back(nr_coordinates_);
	DvObject_var Vs = new DvObject(0.0, dimkv);
	
	DvObject_var Grad;

	if (nr_coordinates_ == 1){
		grad_TS_ = new DvObject(0.0, DIM3, nr_measurements_);
		Grad     = new DvObject(0.0, DIM3);
	}
	else{
		grad_TS_ = new DvObject(0.0, DIM33, nr_measurements_);
		Grad     = new DvObject(0.0, DIM33);
	}

	//temporary variables:
	DvObject_var qtemp = new DvObject(0.0, DIM3);
	double dgdn = 0.0;


	estimator.computeSlopeOfTimeSeries(field_TS_);
	

	for (size_t t=0; t<nr_measurements_; t++) {
		//1. Make sure the variables are initialized:

		//2. Compute reciprocal vectors (a function of spacecraft positions only).
		for (size_t i=0; i<3; i++) Normal->dbl(i) = normal_TS_->dbl(t,i);
		//the output is the total volume of the tetrahedron:
		computeReciprocal3Spacecrafts(t, Normal, vol_TS_->dbl(t), K);
						
		for (size_t sc=0; sc<nr_satellites_; sc++)
			for (size_t i=0; i<nr_coordinates_; i++)
				Vs->dbl(0,sc,i) = field_TS_->dbl(t,sc,i);
		
		dgdn = 0.0;
		if (nr_coordinates_ == 1){
			computeGradScalar(Vs, K, Grad);
			computeEstimateDgDn(Grad, Normal, estimator, t, 0, dgdn);
			for (size_t i=0; i<3; i++)
			{
				Grad->dbl(i) += Normal->dbl(i) * dgdn;
				grad_TS_->dbl(t,i) = Grad->dbl(i);
			}
		}
		else{

			computeGradVec(Vs,K, Grad);
			
			for (size_t coord = 0; coord<3; coord++)
			{
				for (size_t jj = 0; jj<3; jj++) {
					qtemp->dbl(jj) = Grad->dbl(0,coord,jj);
				}
				computeEstimateDgDn(qtemp, Normal, estimator, t, coord, dgdn);
				for (size_t jj = 0; jj<3; jj++) Grad->dbl(0,coord,jj) += Normal->dbl(jj) * dgdn;
			}

			for (size_t i=0; i<3; i++)
				for (size_t j=0; j<3; j++){
					grad_TS_->dbl(t,i,j) = Grad->dbl(0,i,j);
				}
		}
	}
	return result;
}

Status::ReturnStatus EugClusterAnalysis::ComputeGradientEstimate(signed short nr_spacecrafts /*= -1*/)
{
	Status::ReturnStatus result = Status::kOk;
	if (field_TS_.is_nil() || position_TS_.is_nil()) {
		status.set_message("No input for the gradient estimate.");
		result = Status::kNoInput;
	}
	else {
		nr_measurements_ = field_TS_->seqSize();
		if (nr_measurements_ < 1) {
			status.set_message("No input for the gradient estimate.");
			result = Status::kNoInput;
		}
		else if (nr_measurements_ != position_TS_->seqSize()) {
			status.set_message("Position array and measurements array sizes do not match.");
			result = Status::kWrongInput;
		}
		else {
			nr_satellites_  = static_cast<unsigned short>(field_TS_->Dims()[0]);
			nr_coordinates_ = static_cast<unsigned short>(field_TS_->Dims()[1]);
			if ((nr_coordinates_ != 1) && (nr_coordinates_ != 3)) {
				status.set_message("Only scalars or 3 dimensional vectors are accepted.");
				result = Status::kWrongInput;
			}
			else if (nr_satellites_ < 3) {
				status.set_message("Data from at least 3 satellites are required.");
				result = Status::kWrongInput;
			}
			else if (nr_spacecrafts == -1) {
				nr_satellites_ = std::min((unsigned short)4, nr_satellites_);
			}
			else if ((nr_spacecrafts != 3) && (nr_spacecrafts != 4)) {
				status.set_message("Currently only 3 or 4 satellites methods are supported.");
				result = Status::kWrongInput;
			}
			else {
				nr_satellites_ = nr_spacecrafts;
			}
		}
	}
	if (result == Status::kOk) {
		computeNormalToSurfaceOfFirst3Sat(nr_measurements_, vol_TS_, normal_TS_); //valid for all methods with at least 3 satellites.
		result = (nr_satellites_ == 3)? Compute3PointsEstimate() : Compute4PointsEstimate();
	}
	return result;
}

Status::ReturnStatus EugClusterAnalysis::ComputeCurlAndDiv(double curl_scale_factor /*= 1*/)
{
	Status::ReturnStatus result = Status::kOk;
	if (grad_TS_.is_nil()) {
		result = Status::kNoInput;
	}
	else if (nr_coordinates_ != 3) {
		result = Status::kVectorFieldRequired;
	}
	else {
		div_TS_  = new DvObject(0.0, nr_measurements_);
		vector <size_t> dims;
		dims.push_back(kVectorRank);
		curl_TS_ = new DvObject(0.0, dims, nr_measurements_);
		int c[] = { 0, 1, 2, 0, 1 };
		for (size_t t = 0; t < nr_measurements_; t++) {
			for (size_t coord = 0; coord < kVectorRank; coord++) {
				//taken from the IDL code of atmuso:
				div_TS_->dbl(t) += grad_TS_->dbl(t,coord,coord);
				curl_TS_->dbl(t,c[coord]) = (grad_TS_->dbl(t, c[coord + 2], c[coord + 1])
						- grad_TS_->dbl(t, c[coord + 1], c[coord + 2])) / curl_scale_factor;
			}
		}
	}
	return result;
}

Status::ReturnStatus EugClusterAnalysis::ComputeBarycenter()
{
	Status::ReturnStatus result = Status::kOk;

	//cout << "Hello" << endl;
	//Compute Barycenter
	if (!field_TS_.is_nil()) {

		vector <size_t> dims;
		dims.push_back(kVectorRank);
		bary_coordinates_TS_ = new DvObject(0.0, dims, nr_measurements_);
		for (int t = 0; t < nr_measurements_; t++)
			for (unsigned short satellite = 0; satellite < nr_satellites_; satellite++)
			{
				for (unsigned short coord = 0; coord < kVectorRank; coord++)
					bary_coordinates_TS_->dbl(t,coord) += position_TS_->dbl(t,satellite,coord) / nr_satellites_;
			}
		if (nr_coordinates_ == kVectorRank) {
		
			vector <size_t> dimc;
			dimc.push_back(nr_coordinates_);
			bary_vector_TS_ = new DvObject(0.0, dimc, nr_measurements_);
			for (int t = 0; t < nr_measurements_; t++)
				for (unsigned short satellite = 0; satellite < nr_satellites_; satellite++)
				{
					for (unsigned short coord = 0; coord < nr_coordinates_; coord++)
						bary_vector_TS_->dbl(t,coord) += field_TS_->dbl(t,satellite,coord) / nr_satellites_;
				}
		}
		else if (nr_coordinates_ == kScalarRank) {
			bary_scalar_TS_ = new DvObject(0.0, nr_measurements_);
			for (int t = 0; t < nr_measurements_; t++)
				for (unsigned short satellite = 0; satellite < nr_satellites_; satellite++)
				{
					for (unsigned short coord = 0; coord < nr_coordinates_; coord++)
						bary_scalar_TS_->dbl(t) += field_TS_->dbl(t,satellite,coord) / nr_satellites_;
				}
		}

	}
	else {
		result = Status::kNoInput;
	}
	return result;
}

double NormalizeToTen(DvObject_var &matr_TS, DvObject_var &destination)
{
	double result = 1.0;
	if (matr_TS.is_ok()) {
		int nr_measurements = matr_TS->seqSize();
		if (nr_measurements > 0) {
			unsigned short nr_coordinates = static_cast<unsigned short>(matr_TS->Dims()[0]);
			destination = new DvObject(0.0, DIM3, nr_measurements);
			//1. Determine the maximum
			double maxim = matr_TS->dbl(0,0);
			for (int t = 0; t< nr_measurements; ++t)
				for (unsigned short coord = 0; coord < nr_coordinates; ++coord)
					maxim = max(matr_TS->dbl(t,coord), maxim);

			//2.Normalize
			result = maxim / 10;
			for (int t = 0; t< nr_measurements; ++t)
				for (unsigned short coord = 0; coord < nr_coordinates; ++coord)
					destination->dbl(t,coord) = matr_TS->dbl(t,coord) / result;
		}
	}
	return result;
}

Status::ReturnStatus EugClusterAnalysis::Postprocess()
{
	Status::ReturnStatus result = Status::kOk;
	if (grad_TS_.is_nil() || curl_TS_.is_nil()) {
		result = Status::kNoInput;
	}
	else if (nr_coordinates_ == kVectorRank) {

		Status::PrintMessage(true, "Computing MinVar of the curl\n");

		//  ----- do a MinVar of Curl Vector vector -------
		char evalsevecs[80];
		DvString str("Eigenvalues, eigenvectors of MVAJ are :\n");
		const int constrained = 0;


		DvObject_var temp_curl;
		double normal_const = NormalizeToTen(curl_TS_, temp_curl);
		normal_const = normal_const * normal_const;
		DvObject_var nullObj = new DvObject();
		
		DvMinVariance minvar(temp_curl, nullObj); // minvariance call
		rotation_matrix_ = minvar.eigenvectors();
		DvObject_var Eigenvalues = minvar.eigenvalues();
		
		if (rotation_matrix_->dbl(0,0,0) < 0)
			*rotation_matrix_ *= -1.0;

		for (size_t i = 0; i < 3; i++) {
			printf("%10.5g  [%10.6g, %10.6g, %10.6g ] v= %10.5g\n", Eigenvalues->dbl(i) * normal_const,
					rotation_matrix_->dbl(0,i,0), rotation_matrix_->dbl(0,i,1), rotation_matrix_->dbl(0,i,2), Eigenvalues->dbl(i));
			sprintf(evalsevecs, "%10.6g  [%10.6g, %10.6g, %10.6g ] v=%10.6g \n", Eigenvalues->dbl(i) * normal_const,
					rotation_matrix_->dbl(0,i,0), rotation_matrix_->dbl(0,i,1), rotation_matrix_->dbl(0,i,2), Eigenvalues->dbl(i));
			str += evalsevecs;
		} // end for
		fflush(stdout);
		// ----- rotate J into the MinVar system -------


		double mu0 = 1.25664e-6;
		curl_TS_MVA_ = new DvObject(0.0, DIM3, nr_measurements_);
		bary_TS_MVA_ = new DvObject(0.0, DIM3, nr_measurements_);
		velo_TS_MVA_ = new DvObject(0.0, DIM3, nr_measurements_);
		if (bary_vector_TS_.is_ok()) {
			curl_TS_MVA_ = rotation_matrix_ * curl_TS_;
			bary_TS_MVA_ = rotation_matrix_ * bary_vector_TS_;
		}

		DvObject_var Jint = new DvObject(0.0, DIM3);
		DvObject_var Jave = new DvObject(0.0, DIM3);
		DvObject_var Jint_GSE = new DvObject(0.0, DIM3);
		DvObject_var Jave_GSE = new DvObject(0.0, DIM3);

		DvObject_var Bint = new DvObject(0.0, DIM3);
		DvObject_var v_n = new DvObject(0.0, DIM3);
		DvObject_var deltaB = new DvObject(0.0, DIM3);
		double nf = (double) nr_measurements_;

		char l_txtbuf[150] = "";
		for (size_t i = 0; i < 3; i++) {
			deltaB->dbl(i) = bary_TS_MVA_->dbl(nr_measurements_ - 1,i) - bary_TS_MVA_->dbl(0,i);
			sprintf(l_txtbuf, "\nDelta B_%d = %f", (int)i, deltaB->dbl(i));
			str += l_txtbuf;
		}

		for (int k = 0; k < nr_measurements_; k++) {
			for (size_t i = 0; i < 3; i++) {
				Bint->dbl(i) += bary_TS_MVA_->dbl(k,i);
				Jint->dbl(i) += curl_TS_MVA_->dbl(k,i) * deltat_;
				Jave->dbl(i) += curl_TS_MVA_->dbl(k,i);
				Jint_GSE->dbl(i) += curl_TS_->dbl(k,i) * deltat_;
				Jave_GSE->dbl(i) += curl_TS_->dbl(k,i);
			}
		}

		double bave = rotation_matrix_->dbl(0,constrained,0) * Bint->dbl(0) / nf
				+ rotation_matrix_->dbl(0,constrained,1) * Bint->dbl(1) / nf
				+ rotation_matrix_->dbl(0,constrained,2) * Bint->dbl(2) / nf;

		double df = rotation_matrix_->dbl(constrained,0) * Jave_GSE->dbl(0) / nf
				  + rotation_matrix_->dbl(constrained,1) * Jave_GSE->dbl(1) / nf
				  + rotation_matrix_->dbl(constrained,2) * Jave_GSE->dbl(2) / nf;

		printf("<J> =  %g  %g  %g  (%g)\n", Jave->dbl(0) / nf, Jave->dbl(1) / nf, Jave->dbl(2) / nf, df);
		sprintf(l_txtbuf, "\n\n<J> = [ %g  %g  %g] (%g along n_MVAJ)\n", Jave->dbl(0) / nf, Jave->dbl(1)
				/ nf, Jave->dbl(2) / nf, df);
		str += l_txtbuf;

		sprintf(l_txtbuf, "\n\n<Jgse> = [ %g  %g  %g] (%g along n_MVAJ)\n", Jave_GSE->dbl(0) / nf,
				Jave_GSE->dbl(1) / nf, Jave_GSE->dbl(2) / nf, df);
		str += l_txtbuf;

		double maxJ = maxVal(curl_TS_);
		sprintf(l_txtbuf, "J_max = %10.3g\n", maxJ);
		str += l_txtbuf;

		printf("<B> =  %g  %g  %g  (%g)\n",  Bint->dbl(0) / nf, Bint->dbl(1) / nf, Bint->dbl(2) / nf, bave);
		sprintf(l_txtbuf, "<B> = [ %g  %g  %g] (%g along n_MVAJ)\n\n", Bint->dbl(0) / nf, Bint->dbl(1)
				/ nf, Bint->dbl(2) / nf, bave);
		str += l_txtbuf;
		fflush(stdout);

		if (bary_vector_TS_.is_ok()) {
			for (int k = 1; k < nr_measurements_; k++) {
				velo_TS_MVA_->dbl(k,0) = (bary_vector_TS_->dbl(k - 1,1) - bary_vector_TS_->dbl(k,1)) * 1e-12 / (deltat_
						* mu0 * curl_TS_MVA_->dbl(k,2));
				velo_TS_MVA_->dbl(k,1) = (bary_vector_TS_->dbl(k - 1,2) - bary_vector_TS_->dbl(k,2)) * 1e-12 / (deltat_
						* mu0 * curl_TS_MVA_->dbl(k,1));
				velo_TS_MVA_->dbl(k,2) = (velo_TS_MVA_->dbl(k,0) + velo_TS_MVA_->dbl(k,1)) / 2.0; //avg of above
				//   printf("k = %d, %f  %f  -> %f  \n",k, (double)(*velo_TS_MVA_)[k][0],(double)(*velo_TS_MVA_)[k][1],(double)(*velo_TS_MVA_)[k][2] );
			}
		}
		if (constrained == 1) {
			v_n->dbl(0) =  deltaB->dbl(0) * 1e-9 / (mu0 * Jint->dbl(2));
			v_n->dbl(1) = -deltaB->dbl(2) * 1e-9 / (mu0 * Jint->dbl(0));
			v_n->dbl(2) = (v_n->dbl(0) + v_n->dbl(1)) / 2.0; //avg of above
		}
		else {
			v_n->dbl(0) = deltaB->dbl(1) * 1e-9 / (mu0 * Jint->dbl(2));
			v_n->dbl(1) = -deltaB->dbl(2) * 1e-9 / (mu0 * Jint->dbl(1));
			v_n->dbl(2) = (v_n->dbl(0) + v_n->dbl(1)) / 2.0;
		}

		Status::PrintMessage(true, "--\n");
		for (size_t i = 0; i < 2; i++) {
			//   printf("J_%d =  %g\n",i,(double)(*Jint)[i]);
			sprintf(l_txtbuf, "V_%d    = %g\n", (int)i, v_n->dbl(i));
			printf("%s\n", l_txtbuf);
			str += l_txtbuf;
		}
		sprintf(l_txtbuf, "V_ave = %g\n", v_n->dbl(2));
		str += l_txtbuf;

		// ---- do MVA of barycentric B-field -----------
		DvObject_var RmatxB = new DvObject(0.0, DIM33);
		DvObject_var Rmatx_tmp = new DvObject(0.0, DIM33);
		str += "\nEigenvals and Eigenvecs for MVAB_bary :\n";
		normal_const = NormalizeToTen(bary_vector_TS_, temp_curl);
		normal_const = normal_const * normal_const;
		if (bary_vector_TS_.is_ok()) {
			DvMinVariance minvar(temp_curl, nullObj); // minvariance call
			RmatxB = minvar.eigenvectors();
			Eigenvalues = minvar.eigenvalues();
			
		}
		if (RmatxB->dbl(0,0,0) < 0)
			*RmatxB *= -1.0;

		for (size_t i = 0; i < 3; i++) {
			printf("%10.5g  [%10.6g, %10.6g, %10.6g ]\n", Eigenvalues->dbl(i) * normal_const, RmatxB->dbl(0,i,0),
					RmatxB->dbl(0,i,1), RmatxB->dbl(0,i,2));
			sprintf(evalsevecs, "%10.6g  [%10.6g, %10.6g, %10.6g ]\n", Eigenvalues->dbl(i) * normal_const,
					RmatxB->dbl(0,i,0), RmatxB->dbl(0,i,1), RmatxB->dbl(0,i,2));
			str += evalsevecs;
		} // end for
		fflush(stdout);
		
		bave = RmatxB->dbl(0,constrained,0) * Bint->dbl(0) / nf + RmatxB->dbl(0,constrained,1)
				* Bint->dbl(1) / nf + RmatxB->dbl(0,constrained,2) * Bint->dbl(2) / nf;
		df = RmatxB->dbl(0,constrained,0) * Jave->dbl(0) / nf + RmatxB->dbl(0,constrained,1)
				* Jave->dbl(1) / nf + RmatxB->dbl(0,constrained,2) * Jave->dbl(2) / nf;

		sprintf(l_txtbuf, "\n\n<J> = [ %g  %g  %g] (%g along n_MVAB_bary)\n", Jave->dbl(0) / nf,
				Jave->dbl(1) / nf, Jave->dbl(2) / nf, df);
		str += l_txtbuf;

		printf("<B> =  %g  %g  %g  (%g)\n",  Bint->dbl(0) / nf, Bint->dbl(1) / nf, Bint->dbl(2) / nf, bave);
		fflush(stdout);
		sprintf(l_txtbuf, "<B> = [ %g  %g  %g] (%g along n_MVAB_bary)\n\n", Bint->dbl(0) / nf,
				Bint->dbl(1) / nf, Bint->dbl(2) / nf, bave);
		str += l_txtbuf;
		Status::PrintMessage(true, str.c_str());
	}
	return result;
}
// --------------------------------------------

double EugClusterAnalysis::ComputeReciprocal4Spacecrafts(int t, DvObject_var &K)
{
	int c[] = { 0, 1, 2, 0, 1 };
	int i, j, sc, c1 = 0, c2 = 1, c3 = 2, c4 = 3;
	double totvol = 0.0, vol[] = { 0.0, 0.0, 0.0, 0.0 };

	// Reciprocal vectors (a function of positions only) according to Eq. (14.7) in Blue Book
	for (i = 0; i < 3; i++) {
		K->dbl(0,c1,i) = (position_TS_->dbl(t,c3,c[i + 1]) - position_TS_->dbl(t,c2,c[i + 1]))
				* (position_TS_->dbl(t,c4,c[i + 2]) - position_TS_->dbl(t,c2,c[i + 2]))
				- (position_TS_->dbl(t,c3,c[i + 2]) - position_TS_->dbl(t,c2,c[i + 2]))
						* (position_TS_->dbl(t,c4,c[i + 1]) - position_TS_->dbl(t,c2,c[i + 1]));
		K->dbl(0,c2,i) = -((position_TS_->dbl(t,c4,c[i + 1]) - position_TS_->dbl(t,c3,c[i + 1]))
				* (position_TS_->dbl(t,c3,c[i + 2]) - position_TS_->dbl(t,c1,c[i + 2]))
				- (position_TS_->dbl(t,c4,c[i + 2]) - position_TS_->dbl(t,c3,c[i + 2]))
						* (position_TS_->dbl(t,c3,c[i + 1]) - position_TS_->dbl(t,c1,c[i + 1])));
		K->dbl(0,c3,i) = (position_TS_->dbl(t,c4,c[i + 1]) - position_TS_->dbl(t,c1,c[i + 1]))
				* (position_TS_->dbl(t,c4,c[i + 2]) - position_TS_->dbl(t,c2,c[i + 2]))
				- (position_TS_->dbl(t,c4,c[i + 2]) - position_TS_->dbl(t,c1,c[i + 2]))
						* (position_TS_->dbl(t,c4,c[i + 1]) - position_TS_->dbl(t,c2,c[i + 1]));
		K->dbl(0,c4,i) = (position_TS_->dbl(t,c2,c[i + 1]) - position_TS_->dbl(t,c1,c[i + 1]))
				* (position_TS_->dbl(t,c3,c[i + 2]) - position_TS_->dbl(t,c1,c[i + 2]))
				- (position_TS_->dbl(t,c2,c[i + 2]) - position_TS_->dbl(t,c1,c[i + 2]))
						* (position_TS_->dbl(t,c3,c[i + 1]) - position_TS_->dbl(t,c1,c[i + 1]));
	}


	//Normally, the volume must be same. But there are always some round-off errors.
	vol[c1] = -(position_TS_->dbl(t,c2,0) - position_TS_->dbl(t,c1,0)) * K->dbl(0,c1,0)
			- (position_TS_->dbl(t,c2,1) - position_TS_->dbl(t,c1,1)) * K->dbl(0,c1,1)
			- (position_TS_->dbl(t,c2,2) - position_TS_->dbl(t,c1,2)) * K->dbl(0,c1,2);
	vol[c2] = -(position_TS_->dbl(t,c3,0) - position_TS_->dbl(t,c2,0)) * K->dbl(0,c2,0)
			- (position_TS_->dbl(t,c3,1) - position_TS_->dbl(t,c2,1)) * K->dbl(0,c2,1)
			- (position_TS_->dbl(t,c3,2) - position_TS_->dbl(t,c2,2)) * K->dbl(0,c2,2);
	vol[c3] = -(position_TS_->dbl(t,c4,0) - position_TS_->dbl(t,c3,0)) * K->dbl(0,c3,0)
			- (position_TS_->dbl(t,c4,1) - position_TS_->dbl(t,c3,1)) * K->dbl(0,c3,1)
			- (position_TS_->dbl(t,c4,2) - position_TS_->dbl(t,c3,2)) * K->dbl(0,c3,2);
	vol[c4] = +(position_TS_->dbl(t,c4,0) - position_TS_->dbl(t,c1,0)) * K->dbl(0,c4,0)
			+ (position_TS_->dbl(t,c4,1) - position_TS_->dbl(t,c1,1)) * K->dbl(0,c4,1)
			+ (position_TS_->dbl(t,c4,2) - position_TS_->dbl(t,c1,2)) * K->dbl(0,c4,2);

	for (sc = 0; sc < 4; sc++) {
		for (i = 0; i < 3; i++) {
			K->dbl(0,sc,i) /= vol[sc];
            totvol += 0.25 * fabs(vol[sc]);
		}
	}
	return totvol;
}

void EugClusterAnalysis::computeReciprocal3Spacecrafts(const int t, DvObject_var &normal, const double surface, DvObject_var &K)
{
	//Assume that DvObject_var P = new DvObject(0.0, Dim(nr_spacecrafts,3), nestsize);
	//Assume that: DvObject_var normal = new DvObject(0.0, DIM3);
	//Assume that: DvObject_var K = new DvObject(0.0, Dim(nr_spacecrafts,3));

	int c[]={0,1,2,0,1};  //to compute the cross product: cyclic permutations

	//Formula (12), but normal is normalized here.
	for (unsigned short sat = 0; sat < 3; sat++)
		for (unsigned short i = 0; i < 3; i++) //coordinate
			K->dbl(0,sat,i) = (normal->dbl(c[i + 1]) *
					(position_TS_->dbl(t,c[sat + 2],c[i + 2]) - position_TS_->dbl(t,c[sat + 1],c[i + 2]))
					-  normal->dbl(c[i + 2]) * (position_TS_->dbl(t,c[sat + 2],c[i + 1])
					- position_TS_->dbl(t,c[sat + 1],c[i + 1]))) / surface;
}

void EugClusterAnalysis::computeNormalToSurfaceOfFirst3Sat(int nr_position_measurements,
		DvObject_var &surface_TS, DvObject_var &normal_TS)
{
	int nr_meas = min((int)position_TS_->seqSize(), nr_position_measurements);  //for safety reasons.
	//Assume that normal_TS_ was not pre-allocated:
    vector<size_t>dims;
    dims.push_back(kVectorRank);
	normal_TS = new DvObject(0.0, dims, nr_meas);
    DvObject_var normal = new DvObject(0.0, dims);
    
    dims.push_back(kVectorRank); // now rank 2

    //normal coordinate as viewed from each satellite (temp. variable).
	DvObject_var Ns = new DvObject(0.0, dims);
	surface_TS = new DvObject(0.0, nr_meas);

	int c[]={0,1,2,0,1};  //to compute the cross product: cyclic permutations

	for (int t=0; t<nr_meas; t++) {

		for (int sat=0; sat<3; sat++)  //we always have at least 3 satellites.
			for (int i=0; i<3; i++) {  //coordinate (position is always a 3 vector)
				Ns->dbl(0,sat,i) =	(position_TS_->dbl(t,c[sat+1],c[i+1]) - position_TS_->dbl(t,c[sat+0],c[i+1])) *
								(position_TS_->dbl(t,c[sat+2],c[i+2]) - position_TS_->dbl(t,c[sat+0],c[i+2])) -
								(position_TS_->dbl(t,c[sat+1],c[i+2]) - position_TS_->dbl(t,c[sat+0],c[i+2])) *
								(position_TS_->dbl(t,c[sat+2],c[i+1]) - position_TS_->dbl(t,c[sat+0],c[i+1]));
			}
		//	Average over normals viewed from different points:
		for (short i=0; i<3; i++)  //coordinate
			normal->dbl(i) = (Ns->dbl(0,0,i) + Ns->dbl(0,1,i) + Ns->dbl(0,2,i))/3.;
		
        surface_TS->dbl(t) = sqrt(computeSquareVectorNorm(normal));
		for (int i=0; i<3; i++)  //coordinate
			normal_TS->dbl(t,i) = normal->dbl(i) / surface_TS->dbl(t);
	}
}

double EugClusterAnalysis::maxVal(DvObject_var &V)
{
	double sumvec, maxval = 0.0;
	int n = V->seqSize();
	for (unsigned int t = 0; t < n; t++) {
		sumvec = V->dbl(t,0) * V->dbl(t,0) + V->dbl(t,1) * V->dbl(t,1) + V->dbl(t,2) * V->dbl(t,2);
		if (sumvec > maxval)
			maxval = sumvec;
	}
	return sqrt(maxval);
}

double EugClusterAnalysis::sum_SCs(DvObject_var &V1, DvObject_var &V2, int coordV1, int coordV2, unsigned short nr_spacecrafts)
{
	double sum = 0.0;

	for (int i = 0; i < nr_spacecrafts; i++) {
		sum += V1->dbl(0,i,coordV1) * V2->dbl(0,i,coordV2);
	}
	return (sum);
}

double EugClusterAnalysis::DivergenceVec(DvObject_var &V, DvObject_var &R)
{
	// see Eq. 14.16 in Blue Book
	return (sum_SCs(V, R, 0, 0, nr_satellites_) + sum_SCs(V, R, 1, 1, nr_satellites_) + sum_SCs(V, R, 2, 2, nr_satellites_));
}

void EugClusterAnalysis::CurlVec(DvObject_var &V, DvObject_var &R, const unsigned short nr_spacecrafts, DvObject_var &Curl)
{
	// see Eq. 14.17 in Blue Book
	// SEH, 2 May 2005; flipped

	//Assume that the last variable is: DvObject_var Curl = new DvObject(0.0, DIM3);
	Curl->dbl(0) = (-sum_SCs(V, R, 1, 2, nr_spacecrafts) + sum_SCs(V, R, 2, 1, nr_spacecrafts));
	Curl->dbl(1) = (-sum_SCs(V, R, 2, 0, nr_spacecrafts) + sum_SCs(V, R, 0, 2, nr_spacecrafts));
	Curl->dbl(2) = (-sum_SCs(V, R, 0, 1, nr_spacecrafts) + sum_SCs(V, R, 1, 0, nr_spacecrafts));
}

void EugClusterAnalysis::computeGradVec(DvObject_var &V, DvObject_var &R, DvObject_var &Grad)
{
	//Gradient of a vector is a 3x3 matrix.
	//Assume that: DvObject_var Grad = new DvObject(0.0, DIM33 );

	for (size_t coord = 0; coord < 3; coord++) {
		for (size_t j = 0; j < 3; j++) {
			Grad->dbl(0,coord,j) = sum_SCs(V, R, coord, j, nr_satellites_);
		}
	}
}

void EugClusterAnalysis::computeGradScalar(DvObject_var &Vs, DvObject_var &R, DvObject_var &Grad)
{
	//Assume that: DvObject_var Grad = new DvObject(0.0, DIM3);

	//Multiply the magnitude with reciprocal vectors:
	for (size_t i=0; i<3; i++) Grad->dbl(i) = sum_SCs(Vs,R,0,i, nr_satellites_);
}

void EugClusterAnalysis::computeGradMagnitude(DvObject_var &V, DvObject_var &R, DvObject_var &Grad)
{
	//First parameter is the field and second parameter contains reciprocal vectors.
	//Assume that: DvObject_var Grad = new DvObject(0.0, DIM3);
    vector<size_t>dims;
    dims.push_back(nr_satellites_);
    dims.push_back(1); // for later call
    DvObject_var Magn = new DvObject(0.0, dims);      // magnitude of vector from each SC
    DvObject_var  Vec = new DvObject(0.0, DIM3);
    int i,j;

	//1. Compute the magnitude
  for (i=0; i<nr_satellites_; i++) {
    for (j=0; j<3; j++) Vec->dbl(j) = V->dbl(0,i,j);
    Magn->dbl(0,i,0) = sqrt(computeSquareVectorNorm(Vec));
  }

	//2. Compute gradient of scalars:
	computeGradScalar(Magn, R, Grad);
}

double EugClusterAnalysis::computeSquareVectorNorm(const DvObject_var &V)
{
  double sum=0.0;
  for (int i=0; i<3; i++) sum += V->dbl(i) * V->dbl(i);
  return sum;
}

void EugClusterAnalysis::normalizeVector(const DvObject_var &vector, DvObject_var &unit_vector)
{
	//Assume that: DvObject_var vector = new DvObject(0.0, DIM3); [in]
	//Assume that: DvObject_var unit_vector = new DvObject(0.0, DIM3); [out]

	double norm = computeSquareVectorNorm(vector);
	for (size_t i = 0; i<3; i++)
		unit_vector->dbl(i) = vector->dbl(i) / norm;
}

double EugClusterAnalysis::computeEstimateDgDn(const DvObject_var &grad_p, const DvObject_var &normal,
													 const CEstimateNormalComponentFor3SpacecraftMethod &estimator,
													 size_t t,
													 size_t coord,
													 double &dgdn)
{
	//Assume that: DvObject_var grad_p = new DvObject(0.0, DIM3); [in]
	//Assume that: DvObject_var normal = new DvObject(0.0, DIM3); [in]
	//Assume that: DvObject_var vec_e  = new DvObject(0.0, DIM3); [in]

	double quality = 0.0;  //how reliable is the result. Quality should not be too small.
	dgdn = 0;
	if (estimator.get_option() == CEstimateNormalComponentFor3SpacecraftMethod::kNone) {
		if ((t==0) && (coord==0)) Status::PrintMessage(true, "No estimation for the normal component\n");
		return quality;
	}

	const double epsilon = 1e-50;
	DvObject_var estvector = estimator.qvector;

	switch (estimator.get_option()){
		case CEstimateNormalComponentFor3SpacecraftMethod::kGradientParallel: {
			//Gradient parallel to a given vector (formula 32)
			if ((t==0) && (coord==0)) Status::PrintMessage(true, "Estimating normal component using a vector parallel to gradient.\n");
			int c[]={0,1,2,0,1};  //to compute the cross product: cyclic permutations
			double prod1=0, prod2=0;
			DvObject_var qnormalized_estvector = new DvObject(0.0, DIM3);
			normalizeVector(estvector, qnormalized_estvector);
			DvObject_var prodvect = new DvObject(0.0, DIM3);
			for (unsigned int i=0; i<3; i++){  //vector components
				prod1 += qnormalized_estvector->dbl(i) * grad_p->dbl(i);
				prod2 += qnormalized_estvector->dbl(i) * normal->dbl(i);
				//vector product:
				prodvect->dbl(c[i]) = qnormalized_estvector->dbl(c[i+1]) * normal->dbl(c[i+2]) -
														qnormalized_estvector->dbl(c[i+2]) * normal->dbl(c[i+1]);
			}
			quality = computeSquareVectorNorm(prodvect);
			if (quality>epsilon)
				dgdn = (prod1 * prod2) / quality;
		}
			break;
		case CEstimateNormalComponentFor3SpacecraftMethod::kGradientPerpendicular: {
			//Gradient perpendicular to a given vector (formula 34)
			if ((t==0) && (coord==0)) Status::PrintMessage(true, "Estimating normal component using a vector perpendicular to gradient.\n");
			DvObject_var qnormalized_estvector = new DvObject(0.0,DIM3);
			normalizeVector(estvector, qnormalized_estvector);
			double prod1=0, prod2=0;
			for (unsigned int i=0; i<3; i++){  //vector components
				prod1 += qnormalized_estvector->dbl(i) * grad_p->dbl(i);
				prod2 += qnormalized_estvector->dbl(i) * normal->dbl(i);
			}
			quality = prod2;
			if (quality>epsilon)
				dgdn = - prod1 / prod2;
		}
			break;
		case CEstimateNormalComponentFor3SpacecraftMethod::kStationarity: {
			if ((t==0) && (coord==0)) Status::PrintMessage(true, "Estimating normal component using stationarity assumption.\n");
			//Here the estimator vector is the velocity of the plasma frame relative to the spacecraft array: U = U_p +U_n n
			double U_n = 0;
			for (unsigned short i=0; i<3; i++) U_n += estvector->dbl(i) * normal->dbl(i);
			quality = abs(U_n);
			DvObject_var U_p = new DvObject(0.0, DIM3);
			double UpDpg = 0;
			for (size_t i=0; i<3; i++) {
				U_p->dbl(i) = estvector->dbl(i) - U_n * normal->dbl(i);
				UpDpg += U_p->dbl(i) * grad_p->dbl(i);
			}
			dgdn = -(estimator.get_slope(t, coord) + UpDpg)/U_n;
			break;
		}
		case CEstimateNormalComponentFor3SpacecraftMethod::kNone: {
			break;
		}
	}
	return quality;
}

