/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	FFT.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS FFT wrapper
*
* PURPOSE:      Wrapper for Fast Fourier Transform.
* 
* DESCRIPTION:  Plug In module checking input parameters and
*               constructing FFT arrays.
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*              see FFT.qtpl file
*   OUT:
*
* GLOBALS:
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
* FUNCTIONS_USED:	four1.c cdft.c rdft.c
*
* ALGORITHM:  see Press et al. 'Numerical Recipes in C' (1992)
*
* HISTORY: 1.0 created  20/03/2001 Markus Fraenz 
*          1.1 size of vector output corrected 31/07/2001 wjmf
*
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"

extern "C" void four1(double data[], unsigned long nn, int isign);
extern "C" void cdft(int n, int isgn, double *a, int *ip, double *w);
extern "C" void rdft(int n, int isgn, double *a, int *ip, double *w);

extern "C" QplugReturnStatus DvFFTStart(vector <DvObject_var>&, vector <DvObject_var>&);


QplugReturnStatus DvFFTStart( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
{

// define some C variables

int	        attribute_ok,list_length;
unsigned long	Ndata,Ndata2=2,Ndata4,i,k,np,nm,Ndatah;
int             method,*ip_tak,inverseset=1,imfac;
bool            imset=false,vectorset=false;
int qar_status;

double  *data, deltat, T, *w_tak;

// define QSAS data objects

DvObject_var ts_out, ts_outi;

DvObject_var  ts_real, ts_im;



//
// ...first, unpack the input data objects
//

QplugAppendTextDisplay("FFT Plugin 1.1 31 Jul 2001\n");

// ...first item is input Time Series
DvObject_var ts_in = inputs[0];

//  check QSAS datatype

DvObject_var tt_in = ts_in->getTimeTags();
if( tt_in.is_nil() ) {
  QplugAppendTextDisplay ("FFT>> Failed to get time tags\n");
  return QPLUG_FAILURE;
}

if ( tt_in->isRegular(deltat) == false) {
	QplugAppendTextDisplay ("FFT>> ERROR - time series not evenly spaced.\n");
	return QPLUG_FAILURE; 
}

vectorset = ts_in->isThreeVector();

// ...second item is imaginary part

DvObject_var ts_ini = inputs[1];
DvObject_var tt_ini = ts_ini->getTimeTags();

imset = tt_ini.is_ok();

//  check QSAS datatype
if (imset) {  
	if( !ts_in->isJoined(*ts_ini) ) {
		QplugAppendTextDisplay ("FFT>> ERROR - Imaginary Input TS not joined to real TS.\n");
		return QPLUG_FAILURE; 
	}
 
	if(vectorset && !ts_ini->isThreeVector()) {
		QplugAppendTextDisplay ("FFT>> ERROR - Imaginary Input TS not vector.\n");
		return QPLUG_FAILURE; 
	}
}  


//	..3rd item is option 'time_interval'

DvEvent start_end = inputs[2]->getTimeRange();
 
//	..4th item is option 'Method'
method = (int) inputs[3]->asInt();


// now take the subset for the time interval chosen  

QplugAppendTextDisplay ("FFT>> Subsetting Time Series...\n");
 
 
ts_out = ts_in->subSequence(start_end);

if ( ts_out.is_nil() ){
    QplugAppendTextDisplay ("FFT>> ERROR - subset failed:\n");
    return QPLUG_FAILURE;
} 

Ndata = ts_out->seqSize();

if ( Ndata < 2  ){
    QplugAppendTextDisplay ("FFT>> Less than 2 elements in subset.\n");
    return QPLUG_FAILURE;
}


// do imaginary part
if (imset) {
	ts_outi = ts_ini->subSequence(start_end);
}


//	..extract the data from the data series object

if ( vectorset ) {
	QplugAppendTextDisplay ("FFT>> Decomposing Vector Timeseries...\n");
  	ts_real = ts_out->getVectorComponent(DV_RMAG);
	if (imset)  {
		ts_im = ts_outi->getVectorComponent(DV_RMAG);
	}
}
else {
 	ts_real = ts_out;
 	if (imset)  ts_im = ts_outi;
}


//	..extract the time tags from the data series object


//      (notice that NR routines start indices with 1!!) 
DvString str_out;
if (imset) str_out = "FFT>> complex FFT ";
else str_out = "FFT>> real FFT ";

if (method > 0) str_out += "using Ooura CDFT8G.\n";
else str_out += "using NR four1.\n";
QplugAppendTextDisplay (str_out.c_str());

if (imset || (method == 0)) imfac=2;
else imfac=1;

// zero - padding :
while (Ndata2 < Ndata) Ndata2 *= 2;
Ndata4=imfac*Ndata2;
Ndatah=Ndata2/2;

data    = new double[Ndata4]; 
for (i=0; i<Ndata4; i++) data[i] = 0.0;

for (i=0; i<Ndata; i++) data[imfac*i] =  ts_real->asDouble(i);
if (imset) for (i=0; i<Ndata; i++) data[2*i+1] =  ts_im->asDouble(i);

// four1 still uses Fortran array indices

if (method > 0) { // initialize Tak data arrays
 ip_tak = new int[(long) sqrt((float) Ndatah*imfac) +3];
 ip_tak[0]=0;
 w_tak = new double[Ndatah*imfac];
 if (imset) cdft((int) Ndata4, inverseset, data, ip_tak, w_tak);
 else rdft((int) Ndata4, inverseset, data, ip_tak, w_tak);
}
else four1(data-1, Ndata2, inverseset);


//	..create the output data series objects
// we want the output sorted by frequency, so we have to resort
// according to NRp508.

DvObject_var out_freq_series  = new DvObject(0.0, Ndatah*imfac);
DvObject_var out_remag_series = new DvObject(0.0, Ndatah*imfac);
DvObject_var out_immag_series = new DvObject(0.0, Ndatah*imfac);


T  = (double)Ndata2 * deltat;

if (imset||(method==0)){
	for (i =0; i<Ndatah; i++) {
		np=Ndatah+i;
		out_freq_series->dbl(np) =  (double)i / T;
		out_remag_series->dbl(np) = (double) data[2*i];
		out_immag_series->dbl(np) = (double) data[2*i+1];
		out_freq_series->dbl(i) = -(double)(Ndatah-i) / T; 
		out_remag_series->dbl(i) = (double) data[2*i+Ndata2];
		out_immag_series->dbl(i) = (double) data[2*i+1+Ndata2];
	}
}
else {
	for (i =0; i<Ndatah; i++) {
		out_freq_series->dbl(i) =  (double)i / T;
		out_remag_series->dbl(i) = (double) data[2*i];
		out_immag_series->dbl(i) = (double) data[2*i+1];
	}
}

DvObject_var revec_series, imvec_series;
if(vectorset) { 
	revec_series = new DvObject(0.0, DIM3, Ndatah*imfac);
	imvec_series = new DvObject(0.0, DIM3, Ndatah*imfac);
	
	for (k=0;k<3;k++) {
		for (i=0; i<Ndata4; i++) data[i] = 0.0;
		for(i=0; i<Ndata; i++)  data[imfac*i] = ts_out->asDouble(i,k,0.0);
		
		if (imset) for(i=1; i<Ndata; i++)  data[2*i+1] = ts_outi->asDouble(i,k,0.0);
 		if (method > 0) {
  			if (imset) cdft((int) Ndata4, inverseset, data, ip_tak, w_tak);
  			else rdft((int) Ndata4, inverseset, data, ip_tak, w_tak);
 		}
 		else four1(data-1, Ndata2, inverseset);
 
 		if (imset|| (method==0)) {
  			for (i =0; i<Ndatah; i++) {
   				np=Ndatah+i;
				revec_series->dbl(np,k) = (double) data[2*i];
				revec_series->dbl(i,k) = (double) data[2*i+Ndata2];
				imvec_series->dbl(np,k) = (double) data[2*i+1];
				imvec_series->dbl(i,k) = (double) data[2*i+1+Ndata2];
			}
 		}
 		else {
			for (i =0; i<Ndatah; i++) {
				revec_series->dbl(i,k) = (double) data[2*i];
				imvec_series->dbl(i,k) = (double) data[2*i+1];
			}
		}
	}

}


//	...set some attributes

 // SI_conversion
   DvString attr_ptr = ts_in->getSICInverse();
   if( !attr_ptr.empty() ){
     out_remag_series->change_xref("SI_conversion", attr_ptr);
     out_immag_series->change_xref("SI_conversion", attr_ptr);
     if(vectorset) {
    	 revec_series->change_xref("SI_conversion", attr_ptr);
    	 imvec_series->change_xref("SI_conversion", attr_ptr);
     }
   }

  str_out = "ReFFT(";
  str_out += ts_in->getXrefText("LABLAXIS");
  str_out += ")";
  out_remag_series->change_xref("LABLAXIS", str_out);
  if(vectorset) revec_series->change_xref("LABLAXIS", str_out);
  
  str_out = "ReFFT(";
  str_out += ts_in->getXrefText("FIELDNAM");
  str_out += ")";
  out_remag_series->change_xref("FIELDNAM", str_out);
  if(vectorset)   revec_series->change_xref("FIELDNAM", str_out);

  out_remag_series->copy_xref_from("UNITS", ts_in);
  if(vectorset) revec_series->copy_xref_from("UNITS", ts_in);

  out_remag_series->change_xref("DEPEND_0", out_freq_series); 
  if(vectorset)  revec_series->change_xref("DEPEND_0", out_freq_series);
   
  str_out = "ImFFT(";
  str_out += ts_in->getXrefText("LABLAXIS");
  str_out += ")";
  out_immag_series->change_xref("LABLAXIS", str_out);
  if(vectorset) imvec_series->change_xref("LABLAXIS", str_out);
  
  str_out = "ImFFT(";
  str_out += ts_in->getXrefText("FIELDNAM");
  str_out += ")";
  out_immag_series->change_xref("FIELDNAM", str_out);
  if(vectorset)   imvec_series->change_xref("FIELDNAM", str_out);

  out_immag_series->copy_xref_from("UNITS", ts_in);
  if(vectorset) imvec_series->copy_xref_from("UNITS", ts_in);
  
  out_immag_series->change_xref("DEPEND_0", out_freq_series); 
  if(vectorset) imvec_series->change_xref("DEPEND_0", out_freq_series);


  if(vectorset) {
  	DvString frame = ts_in->getFrameAttr();
    revec_series->change_xref(FRAME, frame);
    imvec_series->change_xref(FRAME, frame);
  }


  out_freq_series->change_xref("SI_conversion", "1>s^-1");
  out_freq_series->change_xref("LABLAXIS", "frequency");
  out_freq_series->change_xref("FIELDNAM", "frequency");
  out_freq_series->change_xref("UNITS", "Hz");


//	..add them to the list

  outputs.push_back(out_remag_series);     
  outputs.push_back(out_immag_series);     
  outputs.push_back(revec_series);     
  outputs.push_back(imvec_series);     


//	..some messages to the analysis interface window..

str_out = "FFT>> ";
str_out += (int) Ndata;
str_out += " data points\n" ;
QplugAppendTextDisplay (str_out);
str_out = "FFT>> ";
str_out += (int) Ndata2;
str_out += " zero padded data points\n";
QplugAppendTextDisplay (str_out);
str_out = "FFT>> ";
str_out += (int) Ndata2;
str_out += " frequencies\n";
QplugAppendTextDisplay (str_out);
QplugAppendTextDisplay ("FFT>> finished...\n");


//	..clean up and exit

delete[] data;
if (method > 0) { 
 delete ip_tak;
 delete w_tak;
}

return QPLUG_SUCCESS;


}



