// -----------------------------------------------------------------------------------------
//
// Calculate Projection matrix for constrained MVA
//
// SEH, Sep 12, 2003
// 
// Inputs :
//     B = magnetic field   (n x 3)  for Bn=0 constrain
//     V = plasma velocity  (n x 3) 
//     U = MP velocity vector  (3 comp) for (<V> - U)n = 0 constrain
//     V_alfven = alfven velocity (3 comp)
//     alpha = Walen regression slope form HT analysis
//     Evec = eigenvectormatrix from MVAB for (n perp to Evec_MVAB_max) constrain
//     nestsize = number of points in data sets
//     constrain = 1 for Bn=0, 2 for (<V> - U)n etc - see below.
//
// SEH, 12 Sep 2003
//      23 OCt 2005, tested <Valf> = <B> / sqrt(u_0 <rho>)
//      9 June 2015, AJA, migrated to DVOS
//
// -----------------------------------------------------------------------------------------


#include "qdutil.h"
#include "qplug_if.h"
#include "GenericResidue.h"
#include "DvObj.h"

DvObject_var Projection(DvObject_var &B, DvObject_var &V, 
			 DvObject_var &U, DvObject_var &V_Alfven,
			 DvObject_var &Evec, double alpha, int nestsize, int constrain, const double density) 
{  
  int i,j,k;
  double m_p= 1.67e-27;      // rho = mass dens with proton_mass m_p
  double mu_0 = 1.25664e-06;     // modified mu_0 (due to B in nT and dens in 1/ccm)
  double normfac = (double) nestsize;
  DvObject_var C =  new DvObject(0.0, DIM33);                // C = projection matrix for constrain
  DvObject_var A  = new DvObject(0.0, DIM3);
  DvObject_var ave_B  = new DvObject(0.0, DIM3);
  DvObject_var ave_V  = new DvObject(0.0, DIM3);
  DvObject_var ave_Valf  = new DvObject(0.0, DIM3);

  bool doB = B.is_ok() &&  B->seqSize() > 0 ;
  bool doV = V.is_ok() &&  V->seqSize() > 0 ;
  bool doVA = V_Alfven.is_ok() &&  V_Alfven->seqSize() > 0 ;
  
//   if (doB) printf("B size = %d\n",B->seqSize());
//   if (doV) printf("V size = %d\n",V->seqSize());
//   if (doVA) printf("VAB size = %d\n",V_Alfven->seqSize());

//   printf("in Projection...%d %d %d \n",doB, doV, doVA);


    for (k=0; k<nestsize; k++) {
      for (i=0; i<3; i++) {
		if (doB) ave_B->dbl(i) += B->dbl(k,i);
		if (doV) ave_V->dbl(i) += V->dbl(k,i);
		if (density > 0.00 && doV && doB && doVA) {
			ave_Valf->dbl(i) += alpha*1.0e-3 * B->dbl(k,i) * 1e-9 / sqrt(mu_0 * density *1e06 * m_p);
		} 
		else  if (doVA) ave_Valf->dbl(i) += alpha * V_Alfven->dbl(k,i);
      }
    }
    
    for (i=0; i<3; i++) {
      ave_B->dbl(i) /= normfac;
      if (doV) ave_V->dbl(i) /= normfac;
      if (doVA) ave_Valf->dbl(i) /= normfac;
    } 

  switch(constrain) {
  case 1:
    for (i=0; i<3; i++) A->dbl(i) = ave_B->dbl(i);
    break;
  case 2:
    for (i=0; i<3; i++) {
      A->dbl(i) = 1000.0 * ave_V->dbl(i) - U->dbl(i);
      printf("U,V = %f    %f\n", U->dbl(i), ave_V->dbl(i));
	  fflush(stdout);
    }
    break;
  case 3:
    for (i=0; i<3; i++) {
//       printf("Va, U,V = %10.3f  %10.3f    %10.3f\n",(double)(*ave_Valf)[i],(double)(*U)[i], (double)(*ave_V)[i]);
      A->dbl(i) = -1000.0 * ave_Valf->dbl(i) + 1000.0 * ave_V->dbl(i) - U->dbl(i);
    }
    break;
  case 4:
//     for (i=0; i<3; i++) (*VC1)[i] = (double)(*VC)[i] - (double)(*V_ave)[i];
//     sz2 = sqrt(vecsum2((double)(*VC1)[0],(double)(*VC1)[1],(double)(*VC1)[2]));
//     for (i=0; i<3; i++) (*VC1)[i] /= sz2;
//     for (i=0; i<3; i++) (*VC2)[i] = (double)(*B_ave)[i];
//     sz2 = sqrt(vecsum2((double)(*VC2)[0],(double)(*VC2)[1],(double)(*VC2)[2]));
//     for (i=0; i<3; i++) (*A)[i] = (double)(*VC1)[i] + (double)(*VC2)[i]/sz2;
    break;
  case 5 :
    for (i=0; i<3; i++) A->dbl(i) = Evec->dbl(0,2,i);
    break;
  default:
    for (i=0; i<3; i++) A->dbl(i) = 0.0;
    break;
  }
    
  // ---- projection matrix for constrain  ------------
  
  double sz2 = vecsum2(A);

  
  C->dbl(0,0,0) = 1.0 - (A->dbl(0) * A->dbl(0)) / sz2;
  C->dbl(0,0,1) = - (A->dbl(0) * A->dbl(1)) / sz2;
  C->dbl(0,0,2) = - (A->dbl(0) * A->dbl(2)) / sz2;
  
  C->dbl(0,1,0) = - (A->dbl(1) * A->dbl(0)) / sz2;
  C->dbl(0,1,1) = 1.0 - (A->dbl(1) * A->dbl(1)) / sz2;
  C->dbl(0,1,2) = - (A->dbl(1) * A->dbl(2)) / sz2;
  
  C->dbl(0,2,0) = - (A->dbl(2) * A->dbl(0)) / sz2;
  C->dbl(0,2,1) = - (A->dbl(2) * A->dbl(1)) / sz2;
  C->dbl(0,2,2) = 1.0 - (A->dbl(2) * A->dbl(2)) / sz2;

  return C;
}

