// -----------------------------------------------------------------------------------------
//
// Various utility functions for the generic residue plugin
// 
//
// SEH, 12 Sep 2003
//
// -----------------------------------------------------------------------------------------

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include "GenericMinVar.h"
#include "GenericResidue.h"



// -----------------------------------------------------------------------------------------
//
// Returns the weights according to Eq 44 in Sonnerup et al, JGR 2005
//
// SEH, Jul 13, 2005 - see Mail from BS, Sat 9 Jul 2005, 22:12:57
// 
// -----------------------------------------------------------------------------------------

double weight_k_calc(DvObject_var &Evalues)
{
  if (Evalues->dbl(2) == 0.0) return(1.0);
  double lower =  Evalues->dbl(0) * sqrt( Evalues->dbl(2) * Evalues->dbl(1) );
  double w_k = ( Evalues->dbl(2) - Evalues->dbl(0) ) * ( Evalues->dbl(1) - Evalues->dbl(0) )  / lower;
  return(w_k);
}



double aveTotalPressure(DvObject_var &B, DvObject_var &P, int nestsize)
{
  size_t j, k;
  double p, p_ave=0.00;
  double Mu0  = 12.566370614e-07;  

  DvObject_var B_k  = new DvObject(0.0, DIM3);

  for (k=0; k<nestsize; k++) {
    p = 0.0;
    for (j=0; j<3; j++) {
      p += (P->dbl(k,j,j))/3.0 * 1.60219E-13;                                // scalar pressure = trace(P)/3 [ kg/ms^2 ]
      B_k->dbl(j) =  B->dbl(k,j) * 1.0E-9;                                              // B field vec [T]
    }
    p_ave += p/3.0 + vecsum2(B_k)/(2.0*Mu0);
  }
  p_ave /= (double) nestsize;
  return(p_ave);
}
  


void Report(DvObject_var &Evalues, DvObject_var &Evec,  bool outstd, const char *msg, int constrain, double alpha, const char *SCc_ID, const bool changeOrder)
{
  char buf[255],extended_msg[255],constrain_msg[255];
  int order[3]={0,1,2};
  if (changeOrder) {
    order[0]=2;
    order[2]=0;
  }

  switch (constrain) {
  case 0 : 
    sprintf(buf,"\n# ------ %s for SC%s   ------\n%s_SC%s \n",msg,SCc_ID,msg,SCc_ID);
    break;
  case 3:
    sprintf(extended_msg, "%sC%d for SC%s #walenslope = %5.3f",msg,constrain,SCc_ID,alpha);
    sprintf(constrain_msg,"%sC%d_SC%s",msg,constrain,SCc_ID);
    sprintf(buf,"\n# ------ %s ------\n%s\n",extended_msg,constrain_msg);
    break;
  default:
    sprintf(extended_msg,"%sC%d",msg,constrain);
    sprintf(buf,"\n# ------ %s for SC%s   ------\n%s_SC%s \n",extended_msg,SCc_ID,extended_msg,SCc_ID);
    break;
  }

  QplugAppendTextDisplay(buf);
  if (outstd) printf("%s",buf);
  for (int i=0; i<3; i++){
    sprintf(buf,"%12.6g  [%10.6g, %10.6g, %10.6g]\n", 
	  Evalues->dbl(order[i]),
	  Evec->dbl(0,order[i],0),
	  Evec->dbl(0,order[i],1),
	  Evec->dbl(0,order[i],2) );
    QplugAppendTextDisplay(buf);
    if (outstd) printf("%s",buf);
  }
}


void ReportSiscoe(DvObject_var &q_n, DvObject_var &U,
		 DvObject_var &Bdave, DvObject_var &Vdave, DvObject_var &Vadave,
		 int nestsize, const char *msg, int constrain, double alpha, const char *SCc_ID)
{
  int i,j,k;
  DvObject_var Q = new DvObject(0.0, DIM33);
  char buf[255];
  char SC_ID_loops[255];

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          Q->dbl(0, i,j) += q_n->dbl(k,i,j) / (double) nestsize;
      }
    }
  }
  if (constrain ==3) {
    sprintf(SC_ID_loops,"%s (%d) # walenslope = %5.3f",SCc_ID,nestsize,alpha);
  } else sprintf(SC_ID_loops,"%s (%d)",SCc_ID,nestsize);

  sprintf(buf,"%s",msg);
    
  DvObject_var ProjMx = new DvObject();
  GenericMinVar minvar(0, Q, ProjMx);       // unconstrained, so ProjMx does not play any role
  DvObject_var Evec = minvar.eigenvectors();
  DvObject_var Evalues = minvar.eigenvalues();
  
  DvObject_var N_siscoe = new DvObject(0.0, DIM3);
  for (i=0; i<3; i++) N_siscoe->dbl(i) = Evec->dbl(0,2,i);
  Report(Evalues, Evec, false, buf, constrain, alpha, SC_ID_loops,true);

  //  -- return average of velocities ----

  DvObject_var U_ave = new DvObject(0.0, DIM3);

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      U_ave->dbl(i) += U->dbl(k,i) / (1000.0 * nestsize);    // to get U_ave in [km/s]
    }
  }

  double un = (U_ave->dot(N_siscoe))->dbl(0);
  double bn = (U_ave->dot(N_siscoe))->dbl(0);
  double vn = (Vdave->dot(N_siscoe))->dbl(0);
  double van = (Vadave->dot(N_siscoe))->dbl(0);
  
  sprintf(buf,"!<<B>>  = [%10.3f  %10.3f  %10.3f] (%10.3f) nT   (based on %d different nest sizes)\n", Bdave->dbl(0),Bdave->dbl(1),Bdave->dbl(2),bn,nestsize);
  QplugAppendTextDisplay(buf);
  //  printf("%s",buf);

  sprintf(buf,"!<<V>>  = [%10.3f  %10.3f  %10.3f] (%10.3f) km/s \n", Vdave->dbl(0),Vdave->dbl(1),Vdave->dbl(2),vn);
  QplugAppendTextDisplay(buf);
  //  printf("%s",buf);

  sprintf(buf,"!<<Va>> = [%10.3f  %10.3f  %10.3f] (%10.3f) km/s \n", Vadave->dbl(0),Vadave->dbl(1),Vadave->dbl(2),van);
  QplugAppendTextDisplay(buf);
  //  printf("%s",buf);

  sprintf(buf,"!<<U>>  = [%10.3f  %10.3f  %10.3f] (%10.3f) km/s \n", U_ave->dbl(0),U_ave->dbl(1),U_ave->dbl(2),un);
  QplugAppendTextDisplay(buf);

}


DvObject_var LeviCivita()
  // 
  // returns a LeviCivita 3x3x3 matrix w
  // 
  //
{
  vector <size_t> dim(3,3);
  DvObject_var LC =  new DvObject(0.0, dim);

    // leading 0 in args is record number
  LC->dbl(0,0,1,2) = 1.0;
  LC->dbl(0,2,0,1) = 1.0;
  LC->dbl(0,1,2,0) = 1.0;
  LC->dbl(0,2,1,0) = -1.0;
  LC->dbl(0,0,2,1) = -1.0;
  LC->dbl(0,1,0,2) = -1.0;

  return(LC);
}
 

void ToSpreadsheet(const char *filename, DvObject_var &Normal, DvObject_var &U, int row,  bool nl)
  // 
  // writes spreadsheet like output (CSV) for inclusion in spreadsheets
  // 
  //
{

  double un = 0.0;
  for (int i=0; i<3; i++) un += U->dbl(i) * Normal->dbl(0,row) / 1e3;

  FILE *f1 = fopen(filename,"a");
  fprintf(f1,"%g , %g , %g , ",Normal->dbl(row,0), Normal->dbl(row,1),Normal->dbl(row,2));
  fprintf(f1,"%g , %g , %g , %g ,",U->dbl(0)/1e3,U->dbl(1)/1e3,U->dbl(2)/1e3, un);
  if (nl) fprintf(f1,"\n");
  fclose(f1);
}
  

// -----------------------------------------------------------------------------------------
//
// Calculate standard deviation of a vector V dotted into a normal 
//
// SEH, 12 Jun 2003
//
// -----------------------------------------------------------------------------------------

double sigma(DvObject_var &V, DvObject_var &normal, int n)
{
 
  double vn, vnave=0.00, sumv = 0.00;
  int i,j;
  for (i=0; i<n; i++) for (j=0; j<3; j++) vnave += V->dbl(i,j) * normal->dbl(0,j);
  vnave /= (double) n;

  for (i=0; i<n; i++) {
    vn = 0.00;
    for (j=0; j<3; j++) vn += V->dbl(i,j) * normal->dbl(0,j);
    sumv += ((vn - vnave) * (vn - vnave));
  }
  sumv /= (double)(n-1);
  return(sqrt(sumv));
}
 

	       
// -----------------------------------------------------------------------------------------
//
// Calculate error according to 8.23. and  8.24 in blue book
//
// SEH, 12 Jun 2003
//
// -----------------------------------------------------------------------------------------

double err_824(DvObject_var &B_ave, DvObject_var &Evec, DvObject_var &Eigenvalues, DvObject_var &dYij,
               DvObject_var &M_MVA, double sigma, int nestsize, int modify)
{
  int i, j;
  double Bx1 = 0.00, Bx2 = 0.00, dBx3 = 0.00, l3=0.0;
  
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      dYij->dbl(0,i,j) = sqrt(Eigenvalues->dbl(0)/((double)(nestsize-1)) *
			   (Eigenvalues->dbl(i)+ Eigenvalues->dbl(j) - Eigenvalues->dbl(0)) /
			   ((Eigenvalues->dbl(i) - Eigenvalues->dbl(j)) *
			    (Eigenvalues->dbl(i) - Eigenvalues->dbl(j)) ));
    }
  }
  
  for (i=0; i<3; i++) {
    Bx2 += dYij->dbl(0,0,1) * B_ave->dbl(i) * Evec->dbl(0,1,i);
    Bx1 += dYij->dbl(0,0,2) * B_ave->dbl(i) * Evec->dbl(0,2,i);
  }

  // how to calculate the first term in 8.24 :

  switch(modify) {
  case 0:
    l3 = Eigenvalues->dbl(0) / ((double)(nestsize - 1));
    break;
  case 1:
    for (i=0; i<3; i++) for (j=0; j<3; j++) l3 += Evec->dbl(0,0,i) * M_MVA->dbl(0,i,j) * Evec->dbl(0,0,i);
     l3 /= ((double)(nestsize - 1));
    break;
  case 2:
    l3 = sigma;
    break;
  case 3:
    l3 = 0.00;
    break;
  default:
    l3 = Eigenvalues->dbl(0) / ((double)(nestsize - 1));
    break;
  }

  dBx3 = sqrt(l3  + (Bx1*Bx1) + (Bx2*Bx2));
  return(dBx3);      
}


double jumpvalue(DvObject_var &V, const int n)
{
    if (V.is_nil() || V->not_dbl()) return(-9.9e-31);

    double before=0.00, after=0.00;

    if(V->arraySize() == 3){
        int j = V->seqSize()-1;
        int m = (n >= j) ? j-1 : n;
        for (int i=0; i<m; i++) {
            before+=sqrt( V->dbl(i,0)*V->dbl(i,0) + V->dbl(i,1)*V->dbl(i,1) + V->dbl(i,2)*V->dbl(i,2));
            after +=sqrt( V->dbl(j-1,0)*V->dbl(j-1,0) + V->dbl(j-1,1)*V->dbl(j-1,1)  + V->dbl(j-1,2)*V->dbl(j-1,2));
        }
    }
    else if(V->arraySize() == 1){
        int j = V->seqSize()-1;
        int m = (n >= j) ? j-1 : n;
        for (int i=0; i<m; i++) {
            before+=sqrt( V->dbl(i));
            after +=sqrt( V->dbl(j-i));
        }
    }
    else return(-9.9e-31);
        
    return(before/after);
}


DvObject_var Q_GEN_calc(DvObject_var &TS)
{
  int i,j,k;
  DvObject_var Q  = new DvObject(0.0, DIM33);
  DvObject_var aveTS  = new DvObject(0.0, DIM3);

  int nestsize = TS->seqSize();
  double normfac = (double) nestsize;
 
  DvObject_var dTS  = new DvObject(0.0, DIM3, nestsize);

  for (k=0; k<nestsize; k++) {              // find <TS>
    for (i=0; i<3; i++) {
      aveTS->dbl(i) += TS->dbl(k,i) / normfac;;
    }
  }

  for (k=0; k<nestsize; k++) {             // find delta TS
    for (i=0; i<3; i++) {
      dTS->dbl(k,i) = TS->dbl(k,i) - aveTS->dbl(i);
    }
  }

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) Q->dbl(0,i,j) += TS->dbl(k,i) * dTS->dbl(k,j) / normfac;
    }
  }

  return(Q);
}
