// -----------------------------------------------------------------------------------------
//
// Calculate Q-matrix for Minimum Faraday Residue
//
// Inputs :
//     B = magnetic field  (n x 3)     [nT]
//     V = plasma velocity (n x 3)     [km/s]
//     N = plasma density  (n)         [1/cm^3]  (proton mass used to get mass density)
//
// For simplification, all units are converted to SI units
//
//
// SEH, Mar 15, 2003 
//
// -----------------------------------------------------------------------------------------

#include "GenericResidue.h"
#include "DvObj.h"

DvObject_var Q_MFR_calc(DvObject_var &B, DvObject_var &V, DvObject_var &U, int nestsize)
{
 int i,j,k,m;
 double d_eta2=0.00;
 double normfac = (double) nestsize;
 double m_p = 1.67e-27;                     // mass of proton [kg] )we assume ions are == protons)
 bool PO = false;                           // PrintOut to screen ? 

 DvObject_var eta        = new DvObject(0.0, DIM3, nestsize);
 DvObject_var    ave_eta = new DvObject(0.0, DIM3);
 DvObject_var q_ij    = new DvObject(0.0, DIM33, nestsize);
 DvObject_var    ave_q_ij = new DvObject(0.0, DIM33);
 DvObject_var Q   = new DvObject(0.0, DIM33);
 DvObject_var LC  = LeviCivita();


 // --- construct q_i and eta -----
    
 DvObject_var V_k = V->changeUnitsTo("1>m s^-1", "m/s");
 DvObject_var B_k = B->changeUnitsTo("1>T", "T");
 DvObject_var E_k = B_k->vec(V_k);
   
 for (k=0; k<nestsize; k++) {
   for (i=0; i<3; i++) {
     eta->dbl(k,i) = B_k->dbl(k,i);
     ave_eta->dbl(i) += eta->dbl(k,i);
     for (j=0; j<3; j++) {
       for (m=0; m<3; m++) {
           q_ij->dbl(k,i,j) += LC->dbl(i,j,m) * E_k->dbl(k,m);
       }
       ave_q_ij->dbl(0,i,j) += q_ij->dbl(k,i,j);
     }
   }
 }


  // ----- get <q_ij> and <eta>--------

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      ave_q_ij->dbl(0,i,j) /= normfac;
    }
    ave_eta->dbl(i) /= normfac;
    U->dbl(i) = 0.0;
  }

  if (PO) printf("In Q_MFR:\nq_ij matrix [km/(m s^2)] : \n");
  if (PO) for (i=0; i<3; i++) {
    printf("<q_%dj>     = [%E  %E  %E];\n",i,ave_q_ij->dbl(0,i,0), ave_q_ij->dbl(0,i,1), ave_q_ij->dbl(0,i,2));
  } 

  if (PO)
  {
	printf("\n<eta>      = [%E  %E  %E]  [kg/(m^2 s)];\n", ave_eta->dbl(0), ave_eta->dbl(1), ave_eta->dbl(2));
	fflush(stdout);
  }


  // ---   get delta_q_ij = q_ij - <q_ij> --------

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          q_ij->dbl(k,i,j) -= ave_q_ij->dbl(0,i,j);
      }
      eta->dbl(k,i) -= ave_eta->dbl(i);
      d_eta2 += eta->dbl(k,i) * eta->dbl(k,i);
    }
  }

  if (d_eta2 == 0.0) d_eta2 = 1e-99;   // avoid singularity
 
  //  printf("\n<|deta|^2> = [%E  %E  %E];\n", (double)(*d_eta2)[0], (double)(*d_eta2)[1], (double)(*d_eta2)[2]);
 if (PO)
 {
	printf("\n<|deta|^2> = %E  [kg^2/(m^2s^4)]\n", d_eta2);
	fflush(stdout);
 }



  // ---- construct U* vector ---------------

  for (k=0; k<nestsize; k++) {
    for (j=0; j<3; j++) {
      for (i=0; i<3; i++) {
          U->dbl(j) += eta->dbl(k,i) * q_ij->dbl(k,i,j);
      }
    }
  }

  for (i=0; i<3; i++) U->dbl(i) /= d_eta2;
  if (PO)
  {
	printf("\nU*         = [%E  %E  %E] [m/s];  (eq 12 in manuscript)\n\n", U->dbl(0), U->dbl(1), U->dbl(2));
	fflush(stdout);
  }


  // --- construct Q_ij matrix --------------  

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          for (m=0; m<3; m++) {
              Q->dbl(0,i,j) += (q_ij->dbl(k,m,i) - U->dbl(i)*eta->dbl(k,m)) * (q_ij->dbl(k,m,j) - U->dbl(j)*eta->dbl(k,m));
          }
      }
    }
  }

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      Q->dbl(0,i,j) /= normfac;
    }
  }

  if (PO) printf("Q_ij [km/(m s^2)] (eq 15 in manuscript)\n");
  if (PO) for (i=0; i<3; i++) {
    printf("Q_i%d       = [%E  %E  %E];\n",i, Q->dbl(0,i,0), Q->dbl(0,i,1), Q->dbl(0,i,2));
	fflush(stdout);
  } 

  return(Q);
}




