
// -----------------------------------------------------------------------------------------
//
// Calculate Q-matrix for MLMR - minimum linear momentum flux residue
//
// SEH, Sep 5, 2003
//  
// Inputs :
//     B = magnetic field  (n x 3)     [nT]
//     V = plasma velocity (n x 3)     [km/s]
//     P = pressure tensor (n x 3 x 3) [eV/cm^3]
//     N = plasma density  (n)         [1/cm^3]  (proton mass used to get mass density)
//
// For simplification, all units are converted to SI units
//
// -----------------------------------------------------------------------------------------

#include "GenericResidue.h"
#include "DvObj.h"


// -----------------------------------------------------------------------------------------
//
// Returns the Q_matrix for MLMR (Minimum Linear Momentum Residue)
//
// SEH, 07 Sep, 2001
//      19 Sep, 2001, bug fixes - U returned
//      15 Oct, 2003, Doxygen DOC
//
// Doxygen info :
/*! 
  \brief Returns the covariance matrix for the Minimum Linear Momentum Residue 
  (MLMR) method.

  \param V - vector time series with plasma velocity in units of km/s
  \param B - vector time series with the magnetic field in units of nT
  \param P - matrix time series with the full 3x3 pressure tensor
  \param N - scalar time series with the plasma density in units of 1/cm3
  \param U - vector - holds the calculated velocity in units of km/s on successful return.
  \param nestsize - number of elements in V,B,P,E,N and H.

*/ 
// -----------------------------------------------------------------------------------------
DvObject_var Q_MLMR_calc(DvObject_var &V, DvObject_var &B, DvObject_var &P,
			 DvObject_var &N, DvObject_var &U, int nestsize)
{
  int i,j,k,m;
  double m_p = 1.67e-27;                     // mass of proton [kg] )we assume ions are == protons)
  double Mu0  = 12.566370614e-07;  
  double dirac=0.00;                         // holds the dirac delta function (1 for i==j, 0 else)
  double normfac = (double) nestsize;
  double btot,d_eta2=0.00,p_ave=0.00;
  bool PO = false;                          // PrintOut to screen ?

  DvObject_var    B_k = new DvObject(0.0, DIM3);                    // holds the B(k) vector
  DvObject_var q_ij = new DvObject(0.0, DIM33, nestsize);
  DvObject_var    ave_q_ij =  new DvObject(0.0, DIM33);
  DvObject_var eta = new DvObject(0.0, DIM3, nestsize);
  DvObject_var    ave_eta = new DvObject(0.0, DIM3);
  DvObject_var    Q =  new DvObject(0.0, DIM33);

  p_ave = aveTotalPressure(B, P, nestsize);
  if (PO) printf("Q_MLMR > p_ave = %10.4g %f\n",p_ave,Mu0);
  if(PO) fflush(stdout);

  for (k=0; k<nestsize; k++) {
    //    if (k > 9) m_p = 2.0e-27;
    for (i=0; i<3; i++) B_k->dbl(i) = B->dbl(k,i);
    btot = vecsum2(B_k) * 1E-18 / (2.0 * Mu0);                         // 1E-18 -> to get SI units T^2
    if (PO)
	{
		printf("enter Q_MLMR.k = %d. \n",k);
		fflush(stdout);
	}
       
    for (i=0; i<3; i++) {                                              // loop over three components
      eta->dbl(k,i) = m_p * N->dbl(k) * V->dbl(k,i) * 1E9;                 // 1E9 -> to get SI units kg/s m^2
      ave_eta->dbl(i) += eta->dbl(k,i);

      for (j=0; j<3; j++) {
        dirac = (double)(i==j);  
        q_ij->dbl(k,i,j) =
	  m_p * N->dbl(k) * V->dbl(k,i) * V->dbl(k,j) * 1E12 -             // 1E12 -> V ion km/s, N in /cm^3
	  B->dbl(k,i) *  B->dbl(k,j) * 1E-18 / Mu0 +                     // 1E-18 -> to get B in T
	  dirac * btot +                                               // only diagonal elements herer
       	  P->dbl(k,i,j) * 1.60219E-13;                                 // 1.60219E-13 -> from eV/cm^3 to kg/ms^2
	  //	            p_ave - ( vecsum2(B_k) * 1E-18 / (2.0*Mu0));
          ave_q_ij->dbl(0,i,j) += q_ij->dbl(k,i,j);
      }
    }
  }
  
  // ----- get <q_ij> and <eta> and initialize U* --------

  if (PO)
  {
	printf("get <q_ij>.. \n");
	fflush(stdout);
  }
  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      ave_q_ij->dbl(0,i,j) /= normfac;
    }
    ave_eta->dbl(i) /= normfac;
    U->dbl(i) = 0.0;
  }

  if (PO) printf("Q_MLMR:\nq_ij matrix [km/(m s^2)] : \n");
  if (PO) for (i=0; i<3; i++) {
    printf("<q_%dj>     = [%E  %E  %E];\n",i,ave_q_ij->dbl(0,i,0), ave_q_ij->dbl(0,i,1), ave_q_ij->dbl(0,i,2));
	fflush(stdout);
  } 

  if (PO)
  {
	printf("\n<eta>      = [%E  %E  %E]  [kg/(m^2 s)];\n", ave_eta->dbl(0), ave_eta->dbl(1), ave_eta->dbl(2));
	fflush(stdout);
  }


  // ---   get delta_q_ij = q_ij - <q_ij> --------

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          q_ij->dbl(k,i,j) -= ave_q_ij->dbl(0,i,j);
      }
      eta->dbl(k,i) -= ave_eta->dbl(i);
      d_eta2 += (eta->dbl(k,i) * eta->dbl(k,i));
    }
  }

 
  //  printf("\n<|deta|^2> = [%E  %E  %E];\n", (double)(*d_eta2)[0], (double)(*d_eta2)[1], (double)(*d_eta2)[2]);
  if (PO)
  {
	printf("\n<|deta|^2> = %E  [kg^2/(m^2s^4)]\n", d_eta2);
	fflush(stdout);
  }



  // ---- construct U* vector ---------------

  for (k=0; k<nestsize; k++) {
    for (j=0; j<3; j++) {
      for (i=0; i<3; i++) {
          U->dbl(j) += eta->dbl(k,i) * q_ij->dbl(k,i,j);
      }
    }
  }

  for (i=0; i<3; i++) U->dbl(i) /= d_eta2;
  if (PO)
  {
	printf("\nU*         = [%E  %E  %E] [m/s];  (eq 12 in manuscript)\n\n", U->dbl(0), U->dbl(1), U->dbl(2) );
	fflush(stdout);
  }


  // --- construct Q_ij matrix --------------  

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          for (m=0; m<3; m++) {
              Q->dbl(0,i,j) += (q_ij->dbl(k,m,i) - U->dbl(i)*eta->dbl(k,m)) * (q_ij->dbl(k,m,j) - U->dbl(j)*eta->dbl(k,m));
          }
      }
    }
  }

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      Q->dbl(0,i,j) /= normfac;
    }
  }

  if (PO) printf("Q_ij [km/(m s^2)] (eq 15 in manuscript)\n");
  if (PO) for (i=0; i<3; i++) {
    printf("Q_i%d       = [%E  %E  %E];\n",i, Q->dbl(0,i,0), Q->dbl(0,i,1), Q->dbl(0,i,2));
	fflush(stdout);
  } 


  return (Q);

}

