// -----------------------------------------------------------------------------------------
//
// Calculate Q-matrix for Massflow
//
// Inputs :
//     B = magnetic field  (n x 3)     [nT]
//     V = plasma velocity (n x 3)     [km/s]
//     N = plasma density  (n)         [1/cm^3]  (proton mass used to get mass density)
//
// For simplification, all units are converted to SI units
//
//
// SEH, Mar 27, 2003
//      Sep 12, 2003 - modified for use with the GenericResidue plugin
//
// -----------------------------------------------------------------------------------------

#include "qdutil.h"
#include "qplug_if.h"
#include "GenericResidue.h"
#include "DvObj.h"

DvObject_var Q_MMR_calc(DvObject_var &V, DvObject_var &N, DvObject_var &U, int nestsize)
{
 int i,j,k;
 double ave_eta=0.0, d_eta2=0.00;
 double normfac = (double) nestsize;
 double m_p = 1.67e-27;                     // mass of proton [kg] )we assume ions are == protons)
 bool PO = false;                          // PrintOut to screen ? 

 DvObject_var eta =  new DvObject(0.0, nestsize);
 DvObject_var q_i = new DvObject(0.0, DIM3, nestsize);
 DvObject_var    ave_q_i = new DvObject(0.0, DIM3);
 DvObject_var Q  = new DvObject(0.0, DIM33);


 // --- construct q_i and eta -----

 for (k=0; k<nestsize; k++) {
//    m_p = 1.67e-27;
//    if (k > 10) m_p *= 2.0;   
   eta->dbl(k) = m_p * N->dbl(k) * 1E6;
   ave_eta += eta->dbl(k);
   for (i=0; i<3; i++) {
     q_i->dbl(k,i) = eta->dbl(k) * V->dbl(k,i) * 1E3;
     ave_q_i->dbl(i) += q_i->dbl(k,i);
   }
 }


  // ----- initialize U, get <q_ij> and <eta>--------

  for (i=0; i<3; i++) {
    ave_q_i->dbl(i) /= normfac;
    U->dbl(i) = 0.0;
  }
  ave_eta /= normfac;

  if (PO) printf("In Q_MMR:\nq_i vector [kg/(m^2 s)] : \n");
  if (PO) printf("<q_j>     = [%E  %E  %E]; \n",ave_q_i->dbl(0), ave_q_i->dbl(1), ave_q_i->dbl(2));
  if (PO) printf("\n<eta>      = %E [kg/m^3]\n", ave_eta);
  if (PO) fflush(stdout);

  // ---   get delta_q_i = q_i - <q_i> and delta_eta --------

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) q_i->dbl(k,i) -= ave_q_i->dbl(i);
    eta->dbl(k) -= ave_eta;
    d_eta2 += (eta->dbl(k) * eta->dbl(k));
  }

 
  if (PO)
  {
	printf("\n<|deta|^2> = %E  [kg^2/(m^2s^4)]\n", d_eta2);
	fflush(stdout);
  }

  // ---- construct U* vector ---------------

  if (d_eta2 != 0.0) {
     for (k=0; k<nestsize; k++) {
       for (i=0; i<3; i++) U->dbl(i) += eta->dbl(k) * q_i->dbl(k,i) / d_eta2;
      }  
  } else {
     printf("MMR: singularity - setting U* = <V_cis>\n");
	 fflush(stdout);
     for (k=0; k<nestsize; k++) {
       for (i=0; i<3; i++) U->dbl(i) += V->dbl(k,i) * 1E3 / normfac;
      }
  }

  if (PO)
  {
	printf("\nU*         = [%E  %E  %E];  [m/s] (eq 12 in manuscript)\n\n", U->dbl(0), U->dbl(1), U->dbl(2));
	fflush(stdout);
  }


  // --- construct Q_ij matrix --------------  

  for (k=0; k<nestsize; k++) {
    for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
          Q->dbl(0,i,j) += (q_i->dbl(k,i) - U->dbl(i)*eta->dbl(k)) * (q_i->dbl(k,j) - U->dbl(j)*eta->dbl(k));
      }
    }
  }
  

  for (i=0; i<3; i++) {
    for (j=0; j<3; j++) {
      Q->dbl(0,i,j) /= normfac;
    }
  }

  if (PO)  printf("Q_ij  [kg/(s^3)] (eq 15 in manuscript)\n");
  if (PO)  for (i=0; i<3; i++) {
    printf("Q_i%d       = [%E  %E  %E];\n",i, Q->dbl(0,i,0), Q->dbl(0,i,1), Q->dbl(0,i,2));
	fflush(stdout);
  } 
 
 return(Q);
}

