/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    OhmsLaw
* 
* MODULE:	
*
* LANGUAGE:	C++
*
* FUNCTION:
*
* PURPOSE:      Calculates first (J) and second (JxB/ne) terms of Ohms law based
*               on B field, ion and electron moments and the E-field.
* 
* DESCRIPTION:
*
* ARGUMENTS:
*    IN:
*     VECTOR_TIME_SERIES    Bfield, Vion, Velec, Efield   
*     SCALAR_TIME_SERIES    Densities
*     TIME_INTERVAL         Time                 time to calculate curl or div for
* 
*    OUT:
*     VECTOR_TIME SERIES    J, JxB/ne
*
* GLOBALS:
*
* RETURN:
*    TYPE:     QplugReturnStatus
*    VALUES:   see qplug_if.h
*
*
* ALGORITHM:   See chap ISSI Auroral Physics, Eq 3.14 or any plasma physics text book
*
* HISTORY:	SEH, 21 Sep 2016, Firts version
*
***********************************************************************/

#include "DvObj.h"
#include "qplug_if.h"
#include "qdutil.h"
#include <qdatetime.h>

extern  "C" QplugReturnStatus  DvJp_Hall_VxB  (vector <DvObject_var>&, vector <DvObject_var>&);



using namespace QSAS;

QplugReturnStatus DvJp_Hall_VxB (vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)
{
  int ninputs=0;
  int nestsize=0;
  DvString lab;

  QDateTime today=QDateTime::currentDateTime();
  DvString DateStamp = DvString(today.toString(Qt::ISODate).toStdString());

  if (inputs.size() < 6) {
    QplugAppendTextDisplay ("7 inputs required");
    return QPLUG_FAILURE; 
  } 

  // get time interval ---
  DvEvent start_end = inputs[6]->getTimeRange();
  DvObject_var tt_out = inputs[2]->getTimeTags();
 
  double deltat, avg_width;
  tt_out->isRegular(deltat); // need spacing but not regular tt
  avg_width = 2 * deltat;    // nyquist
 
  tt_out->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  tt_out->change_xref(DV_BOX_WIDTH, avg_width);
  tt_out->change_xref(DV_GAP_OPTION, DV_LINEAR);

  // -- get inputs
  DvObject_var B = inputs[0];
  if (B.is_ok()) {
    ninputs++;
    QplugAppendTextDisplay("Bfield OK");
  } else {
    QplugAppendTextDisplay("No Bfield supplied - aborting....");
    return QPLUG_FAILURE;
  }
  B->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  B->change_xref(DV_BOX_WIDTH, avg_width);
  B->change_xref(DV_GAP_OPTION, DV_LINEAR);
  B = B->Join(tt_out);
  B = B->subSequence(start_end); 
  B = B->changeUnitsTo("1.0e-9>T", "nT");
  nestsize = B->seqSize();
  DvObject_var coordsys = B->get_xref(COORDINATE_SYSTEM);

  DvObject_var E = inputs[1];
  if (E.is_ok()) {
    ninputs++;
    QplugAppendTextDisplay("Efield OK");
  } else {
    QplugAppendTextDisplay("No Efield supplied; Assuming E=0;");
    E = new DvObject(0.0, DIM3, nestsize);
  }
  E->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  E->change_xref(DV_BOX_WIDTH, avg_width);
  E->change_xref(DV_GAP_OPTION, DV_LINEAR);
  E = E->subSequence(start_end); 
  E = E->changeUnitsTo("1.0e-3>V m^-1", "mVm^-1");

  
  DvObject_var Vi = inputs[2];
  if (Vi.is_ok()) {
    ninputs++;
    QplugAppendTextDisplay("Ion velocity OK");
  } else {
    QplugAppendTextDisplay("No ion velocity supplied. Aborting...");
    return QPLUG_FAILURE;
  }
  Vi->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  Vi->change_xref(DV_BOX_WIDTH, avg_width);
  Vi->change_xref(DV_GAP_OPTION, DV_LINEAR);
  Vi = Vi->Join(tt_out);
  Vi = Vi->subSequence(start_end); 
  Vi = Vi->changeUnitsTo("1.0e3>m s^-1");

  DvObject_var Ve = inputs[3];
  if (Ve.is_ok()) {
    ninputs++;
    QplugAppendTextDisplay("Electron velocity OK");
  } else {
    QplugAppendTextDisplay("No electro velocity supplied. Aborting...");
    return QPLUG_FAILURE;
  }
  Ve->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  Ve->change_xref(DV_BOX_WIDTH, avg_width);
  Ve->change_xref(DV_GAP_OPTION, DV_LINEAR);
  Ve = Ve->Join(tt_out);
  Ve = Ve->subSequence(start_end); 
  Ve = Ve->changeUnitsTo("1.0e3>m s^-1");

  

  DvObject_var Ni = inputs[4];
  if (Ni.is_ok()) {
    ninputs++;
    QplugAppendTextDisplay("Ion density OK");
  } else {
    QplugAppendTextDisplay("No ion density supplied. Aborting...");
    return QPLUG_FAILURE;
  }
  Ni->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  Ni->change_xref(DV_BOX_WIDTH, avg_width);
  Ni->change_xref(DV_GAP_OPTION, DV_LINEAR);
  Ni = Ni->Join(tt_out);
  Ni = Ni->changeUnitsTo("1.0e6>m^-3"); // DvString understands 1.0e6 or 10^6

  DvObject_var Ne = inputs[5];
  if (Ne.is_ok()) {
    ninputs++;
    QplugAppendTextDisplay("Electron density OK");
  } else {
    QplugAppendTextDisplay("No electron density supplied. Aborting...");
    return QPLUG_FAILURE;
  }
  Ne->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  Ne->change_xref(DV_BOX_WIDTH, avg_width);
  Ne->change_xref(DV_GAP_OPTION, DV_LINEAR);
  Ne = Ne->Join(tt_out);
  Ne = Ne->changeUnitsTo("1.0e6>m^-3"); // DvString understands 1.0e6 or 10^6

  printf("Now calc Jp...for N = %d samples .......\n",nestsize);

  DvObject_var J = new DvObject(0.0, DIM3, nestsize);
  DvObject_var HallTerm = new DvObject(0.0, DIM3, nestsize);
  DvObject_var VixB = new DvObject(0.0, DIM3, nestsize);
  double qn;
  for (int i=0; i<nestsize; i++){
//    qn= 1.6021e-01*0.5*(Ne->dbl(i,0) + Ni->dbl(i,0));  // q * <N> but in nA/m2
    qn= 1.6021e-01*Ni->dbl(i,0);  // q * <N> but in nA/m2
    J->dbl(i,0) = qn * (Vi->dbl(i,0) - Ve->dbl(i,0));
    J->dbl(i,1) = qn * (Vi->dbl(i,1) - Ve->dbl(i,1));
    J->dbl(i,2) = qn * (Vi->dbl(i,2) - Ve->dbl(i,2));
  }

  double cc = 1e-3/qn;  // nA and nT 
  for (int i=0; i<nestsize; i++){
    qn= 1.6021e-01*Ni->dbl(i,0);  // q * <N> but in nA/m2
    HallTerm->dbl(i,0) = 1e-3*(J->dbl(i,1) * B->dbl(i,2)  - J->dbl(i,2) * B->dbl(i,1))/qn;
    HallTerm->dbl(i,1) = 1e-3*(J->dbl(i,2) * B->dbl(i,0)  - J->dbl(i,0) * B->dbl(i,2))/qn;
    HallTerm->dbl(i,2) = 1e-3*(J->dbl(i,0) * B->dbl(i,1)  - J->dbl(i,1) * B->dbl(i,0))/qn;
    VixB->dbl(i,0) = -1e-3*(Vi->dbl(i,1) * B->dbl(i,2)  - Vi->dbl(i,2) * B->dbl(i,1)); 
    VixB->dbl(i,1) = -1e-3*(Vi->dbl(i,2) * B->dbl(i,0)  - Vi->dbl(i,0) * B->dbl(i,2));
    VixB->dbl(i,2) = -1e-3*(Vi->dbl(i,0) * B->dbl(i,1)  - Vi->dbl(i,1) * B->dbl(i,0));
   }

  J->copy_xrefs_from(B);
  
  DvObject_var new_label_ptr = new DvObject(lab,DIM3,1);
  new_label_ptr->str(0) +="Jx_"; new_label_ptr->str(0)+=coordsys->str().trimmed();
  new_label_ptr->str(1) +="Jy_"; new_label_ptr->str(1)+=coordsys->str().trimmed();
  new_label_ptr->str(2) +="Jz_"; new_label_ptr->str(2)+=coordsys->str().trimmed();
  J->change_xref(LABL_PTR_1,new_label_ptr);
  J->change_xref(DEPEND_0, tt_out);
  J->change_xref(UNITS,"nA/m^2");
  J->change_xref(CATDESC,"Current density Jp calculated by QSAS Jp_Hall_VxB calculator");
  J->change_xref(SI_CONVERSION,"1e-9>A/m^2");
  J->change_xref("Parents","QSAS Jp_Hall_VxB");
  J->change_xref(FIELDNAM,"Current density vector Jp");
  J->change_xref("Date generated",DateStamp);

  DvObject_var e_label = new DvObject(lab,DIM3,1);
  e_label->str(0) +="Ex_"; e_label->str(0)+=coordsys->str().trimmed();
  e_label->str(1) +="Ey_"; e_label->str(1)+=coordsys->str().trimmed();
  e_label->str(2) +="Ez_"; e_label->str(2)+=coordsys->str().trimmed();
  HallTerm->copy_xrefs_from(B);
  HallTerm->change_xref(LABL_PTR_1,e_label);
  HallTerm->change_xref(UNITS,"mV/m");
  HallTerm->change_xref(DEPEND_0, tt_out);
  HallTerm->change_xref(CATDESC,"Efield from JxB/q  calculated by QSAS Jp_Hall_VxB calculator");
  HallTerm->change_xref(SI_CONVERSION,"1e-3>V/m");
  HallTerm->change_xref("Parents","QSAS Jp_Hall_VxB");
  HallTerm->change_xref(FIELDNAM,"Efield from JxB/q");
  HallTerm->change_xref("Date generated",DateStamp);

   VixB->change_xref(LABL_PTR_1,e_label);
  VixB->copy_xrefs_from(E);
  VixB->change_xref(DEPEND_0, tt_out);
  VixB->change_xref(CATDESC,"Efield from -VixB  calculated by QSAS Jp_Hall_VxB calculator");
  VixB->change_xref("Date generated",DateStamp);
  VixB->change_xref("Parents","QSAS Jp_Hall_VxB");


  outputs.push_back(J);
  outputs.push_back(HallTerm);
  outputs.push_back(VixB);



  // --- now handle electrons with their higher cadence ---
  printf("Now handling Ve x B..\n");

  tt_out = inputs[3]->getTimeTags(); // tt now from Ve
  nestsize = tt_out->seqSize();

  printf("nestsize in VexB = %d\n",nestsize);
  tt_out->isRegular(deltat); // tt from Ve;
  avg_width = 2 * deltat;    // Nyquist
 
  printf("delta t for Ve..%g\n",avg_width);
  tt_out->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  tt_out->change_xref(DV_BOX_WIDTH, avg_width);
  tt_out->change_xref(DV_GAP_OPTION, DV_LINEAR);


  B = inputs[0];
  if (B.is_ok()) {
    QplugAppendTextDisplay("Bfield OK");
  } else {
    QplugAppendTextDisplay("No Bfield supplied - aborting....");
    return QPLUG_FAILURE;
  }
  B->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  B->change_xref(DV_BOX_WIDTH, avg_width);
  B->change_xref(DV_GAP_OPTION, DV_LINEAR);
  B = B->Join(tt_out);
  B = B->subSequence(start_end); 
  B = B->changeUnitsTo("1.0e-9>T", "nT");


  Ve = inputs[3];
  if (Ve.is_ok()) {
    QplugAppendTextDisplay("Electron velocity OK");
  } else {
     QplugAppendTextDisplay("No electron velocity supplied. Aborting...");
     return QPLUG_FAILURE;
   }
  Ve->change_xref(DV_JOIN_METHOD, DV_BOXCAR);	
  Ve->change_xref(DV_BOX_WIDTH, avg_width);
  Ve->change_xref(DV_GAP_OPTION, DV_LINEAR);
  Ve = Ve->Join(tt_out);
  Ve = Ve->subSequence(start_end); 
  Ve = Ve->changeUnitsTo("1.0e3>m s^-1");
  nestsize = Ve->seqSize();


  DvObject_var VexB = new DvObject(0.0, DIM3, nestsize);
  for (int i=0; i<nestsize; i++){
    VexB->dbl(i,0) = -1e-3*(Ve->dbl(i,1) * B->dbl(i,2)  - Ve->dbl(i,2) * B->dbl(i,1)); 
    VexB->dbl(i,1) = -1e-3*(Ve->dbl(i,2) * B->dbl(i,0)  - Ve->dbl(i,0) * B->dbl(i,2));
    VexB->dbl(i,2) = -1e-3*(Ve->dbl(i,0) * B->dbl(i,1)  - Ve->dbl(i,1) * B->dbl(i,0));
  }
  VexB->copy_xrefs_from(Ve);
  VexB->change_xref(LABL_PTR_1,e_label);
  VexB->change_xref("UNITS","mV/m");
  VexB->change_xref("FIELDNAM","Efield -VexB from QSAS Jp_Hall_VxB plugin");
  VexB->change_xref(CATDESC,"Efield from -VexB  calculated by QSAS Jp_Hall_VxB calculator");
  VexB->change_xref("Date generated",DateStamp);
  VexB->change_xref("Parents","QSAS Jp_Hall_VxB");

  outputs.push_back(VexB);


  return QPLUG_SUCCESS;
}
