// ------------------------------------------------------------------
//
//  ConstrainedMinVarMFR.h
//  Qdos2 - Constrained Minimum Variance Algorithm for MFR. Based on 
//  the original MinVariance.h.
//
//  Janet Barnes
//  May 2000
//  Cloned & modified by SEH, May 2002
//  MFR version, SEH 28 May 2002
//
// ------------------------------------------------------------------

#ifndef QDOS_CONSTRAINEDMINVARMFR_H
#define QDOS_CONSTRAINEDMINVARMFR_H

#define VERBOSE_MV              1

#include "DvObj.h"


class ConstrainedMinVarMFR
{
    DvObject_var _eigvals;       // Matrix Dim(3) of eigenvalues
    DvObject_var _eigvecs;       // Matrix Dim(3,3) of eigenvectors (in rows)
    DvObject_var _eigqual;       // Matrix Dim(3) eigenvalue quality
    double _orgqual;               // orthogonality quality
    DvObject_var _covariance;    // covariance matrix
    DvObject_var P;              // projection matrix P (see ISSI book, p 195)
    DvObject_var Q;              // MFR Q matrix for which eigenvals _ eigenvecs are found
    DvObject_var temp;           // temporary storage matrix
    int i,j;
    
public:
    ConstrainedMinVarMFR(int constrained, DvObject_var &ms, DvObject_var &Q)
    {
        if(ms.is_nil())
            return;
        if(ms->nDims()!=1)
            return;
        if(ms->arraySize()!=3)
             return;
             
        int sz=ms->seqSize();
        
        // compute the variance matrix 
          
        DvObject_var ms_sum = ms->sum();
        double sum_x=ms_sum->dbl(0);
        double sum_y=ms_sum->dbl(1);
        double sum_z=ms_sum->dbl(2);
 
        _covariance = new DvObject(Q);        
   
        _covariance->dbl(0,1,0) = _covariance->dbl(0,0,1);
        _covariance->dbl(0,2,0) = _covariance->dbl(0,0,2);
        _covariance->dbl(0,2,1) = _covariance->dbl(0,1,2);
        
        // 
        // Constrained MVA : find eigenvec of [P # C # P] * n = lamda * n
        // instead of [C] * n = lambda * n. 
        //
        //   0 - standard MVA
        //   1 - constrain so that Bn == 0 (see p 209 in blue ISSI book)
        //
        // SEH, 14 May 2002
        //      28 May 2002, MFR version
        //

        if (constrained == 1) {

            if (VERBOSE_MV) printf("MonVar_MFR > Using constrain\n");

            double sz2 = sum_x*sum_x + sum_y*sum_y + sum_z*sum_z;

            P = new DvObject(0.0, DIM33);

            P->dbl(0,0,0) = 1.0 - (sum_x*sum_x) / sz2;
            P->dbl(0,0,1) = - (sum_x*sum_y) / sz2;
            P->dbl(0,0,2) = - (sum_x*sum_z) / sz2;

            P->dbl(0,1,0) = - (sum_y*sum_x) / sz2;
            P->dbl(0,1,1) = 1.0 - (sum_y*sum_y) / sz2;
            P->dbl(0,1,2) = - (sum_y*sum_z) / sz2;

            P->dbl(0,2,0) = - (sum_z*sum_x) / sz2;
            P->dbl(0,2,1) = - (sum_z*sum_y) / sz2;
            P->dbl(0,2,2) = 1.0 - (sum_z*sum_z) / sz2;

            temp = _covariance * P;
            _covariance = P*temp;

            // force symmetric covariance matrix again
            _covariance->dbl(0,1,0) = _covariance->dbl(0,0,1);
            _covariance->dbl(0,2,0) = _covariance->dbl(0,0,2);
            _covariance->dbl(0,2,1) = _covariance->dbl(0,1,2);
        }

        // get eigenvectors and values  
        DvMatrixEig eig(_covariance);
        _eigvecs=eig.eigenvectors();
        _eigvals=eig.eigenvalues();


        // If constrained : eigval 0 == 0. Swap 1st and 2nd eigenvec

        if (constrained == 1) {
            _eigvals->dbl(0) = _eigvals->dbl(1);
            _eigvals->dbl(1) = _eigvals->dbl(2);
            _eigvals->dbl(2) = 0;                                // by definition
            temp = new DvObject(_eigvecs);
            for (i=0; i<3; i++) _eigvecs->dbl(0,0,i) = temp->dbl(0,1,i);
            for (i=0; i<3; i++) _eigvecs->dbl(0,1,i) = temp->dbl(0,2,i);
            for (i=0; i<3; i++) _eigvecs->dbl(0,2,i) = temp->dbl(0,0,i);
        }

        
        // check the positive sense of the eigenvector associated with
        // the minimum eigenvalue and if necessary reverse the signs
        // of all three components i.e. vint x vmax = vector perpendicular
        // to both vint and vmax. If vmin is perpendicular to vint 
        // and vmax (i.e. system is ortogonal), then (vinx x vmax).vmin=1
        // If < 0 then not a right-handed system, so reverse signs of vmin
 
      
        // get vectors out
        DvObject_var eigV0 = _eigvecs->sliceDim(0,0);
        DvObject_var eigV1 = _eigvecs->sliceDim(0,1);
        DvObject_var eigV2 = _eigvecs->sliceDim(0,2);
 
        DvObject_var rh = eigV1->vec(eigV2);
        rh = rh->dot(eigV0);
        
        double rh_factor = rh->dbl();        
        if(rh_factor<0.) *eigV0 *= -1;
        
        _orgqual = fabs(fabs(rh_factor) - 1.0);
                
        // ensure that the sign of the average minimum variance component
        // is positive by computing average of sum of B-field data.min
        // eigenvector. If less than zero negate both vmin and vint (so
        // preserving right-handed system)
        DvObject_var avg_minvar_comp = ms->dot(eigV0);
        avg_minvar_comp = avg_minvar_comp->sum();
        if(avg_minvar_comp->dbl() < 0) 
        {
            *eigV0 *= -1.;
            *eigV1 *= -1.;
        }
        _eigqual = new DvObject(_eigvals);
        *_eigqual /= _eigvals->dbl(2);
            
        // put vectors back
        valarray <size_t> length(1);
        valarray <size_t> stride(1);
        length[0]=eigV0->totalSize();
        stride[0]=1;
		gslice ds(0, length, stride); // array_slice of all data in vector
        _eigvecs->setSliceDim(0, 0, eigV0->getDSlice(ds));
        _eigvecs->setSliceDim(0, 1, eigV1->getDSlice(ds));
        _eigvecs->setSliceDim(0, 2, eigV2->getDSlice(ds));
        
    }



    // methods for returning members and calculating variance
    DvObject_var covariance() 
    { return _covariance; }
    
    DvObject_var eigenvalues() 
    { return _eigvals;}
     
    DvObject_var eigenvectors()
    { return _eigvecs;}   
    
    DvObject_var eigenquality() 
    { return _eigqual;}
    
    double& orthogonality_quality()
    { return _orgqual;}    // orthogonality quality     

        
};
    
#endif
