// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : Plasma calculator
// 
//
// Inputs [optional]:
// 	bfield		- vector time series with magnetic field
//      density         - plasma density
//      velocity        - plasma velocity
// 	t_perp   	- scalar time series with perpendicular temp
// 	t_par	        - scalar time series with parallelle plasma temp
//
// Outputs
//      P_b             - magnetic pressure (B^2/2my0)
//      P_pperp         - plasma pressure (nkT_perp)
//      P_p             - full plasma pressure (nkT_par +2nkT_perp)/3
//      P_tot           - magnetic + plasma pressure
//      Bal             - pressure balance = P_b / P_pperp
//      A_p             - plasma anisotropy (T_par/T_perp - 1).
//      gyro            - ion gyro radius
//      inert           - ion inertial length
//      valf            - local Alfven speed
//      cs              - sound speed
//      fpe             - plasma frequency
//
// SEH, 08 Jan, 2003
//      14 Feb, 2003 - added rho as output
//      20 Nov, 2003 - added ion gyro radii and inertial length as output
//      26 Dec, 2003 - QSAS 2.0 version
//      07 Oct, 2004 - added Mach number + Alfven speed 
//      28 Jun, 2006 - added fce + cs (electron freq + sound speed
//      01 Feb, 2008 - reorganized + added temp tensor (for i.e. Themis), removed rho, mack, (1-alpha)
//      09 Nov, 2016 - output Va as a vector (not scalar any more)
//
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"


extern "C" QplugReturnStatus  DvPlasmaParams  ( vector<DvObject_var>&, vector<DvObject_var>&);

using namespace QSAS;

QplugReturnStatus  DvPlasmaParams  ( vector<DvObject_var>& inputs, vector<DvObject_var>& outputs )
{

 DvObject_var bfield_nest;
 DvObject_var velocity_nest;
 DvObject_var density_nest;
 DvObject_var Tpar_nest;
 DvObject_var Tperp_nest;
 DvObject_var  P_b;
 DvObject_var P_p;
 DvObject_var P_pfull;
 DvObject_var P_tot;
 DvObject_var Beta;
 DvObject_var A_p;
 DvObject_var gyro;
 DvObject_var inert;
 DvObject_var Valf;
 DvObject_var Mach;
 DvObject_var Alpha;
 DvObject_var OneAlpha;
 DvObject_var Entropy;
 DvObject_var ttJoined;
 bool    Bsup=false, Nsup=false, Vsup=false, Tparsup=false, Tperpsup=false, Ttenssup=false;
 long    n_vecs, nestsize;
 double  L0,theta,deltat,avg_width,Vsw,TotalEnergy=0.0,Bt;
 double  ionMass= 1.67E-6;                                         
 char    txtbuf[200];
 int nseq=0;

 // --- establish what inputs are -------

 DvObject_var B_in = inputs[0];
 if (B_in.is_ok()) {
     Bsup=true;
     nseq++;
     QplugAppendTextDisplay("B-field supplied\n");
 }

 DvObject_var V_in = inputs[1];
 if (V_in.is_ok()) {
     Vsup=true;
     nseq++;
     QplugAppendTextDisplay("Velocity supplied\n");
 }

 DvObject_var N_in = inputs[2];
 if (N_in.is_ok()) {
     QplugAppendTextDisplay("Density supplied\n");
     Nsup=true;
     nseq++;
 }


 DvObject_var Tpar_in = inputs[3];
 DvObject_var Tperp_in = inputs[4];
 
 if (Tperp_in.is_ok() && Tperp_in->asInt() != 0) {
   QplugAppendTextDisplay("T_perp supplied\n");
   nseq ++;
   Tperpsup=true;
 } 
 if (Tpar_in.is_ok() && Tpar_in->asInt() != 0) {
    if(Tpar_in->arraySize() == 1){
        QplugAppendTextDisplay("T_par supplied\n");
        nseq ++;
        Tparsup=true;
    } 
    else if(Tpar_in->arraySize() == 3){
        QplugAppendTextDisplay("Temp supplied as tensor diagonal..\n");
        nseq ++;
        Ttenssup=true;
    }
 }  
 
 // --- do we have sufficient input ---
 if (nseq == 0 || (nseq == 1 && Vsup)) {
    QplugAppendTextDisplay("To few input params\n");
    return(QPLUG_FAILURE);
 }

 // ---- check if B and V in same frame ----
 if (Vsup && Bsup){
   if ( !B_in->sameFrame(*V_in) ){
     QplugAppendTextDisplay("B field and velocity in different frames\n");
     return QPLUG_FAILURE;
   }
 }


 //  --- first we join B measurements to plasma measurements -----

DvJoinList ToJoinList;
DvJoinList JoinedList;

 if (Bsup) {
    ToJoinList.append("B", B_in);
    ttJoined = B_in->getTimeTags();   
 }
 if (Vsup) {
    ToJoinList.append("V", V_in);
    ttJoined = V_in->getTimeTags();
 }
 if (Nsup) {
    ToJoinList.append("N", N_in);
    ttJoined = N_in->getTimeTags(); 
 }
 if (Tparsup) {
    ToJoinList.append("Tpar", Tpar_in);
    ttJoined = Tpar_in->getTimeTags();
 }
 if (Tperpsup) {
    ToJoinList.append("Tperp", Tperp_in);
    ttJoined = Tperp_in->getTimeTags();
 }
 

 //  ---- get Time interval and mask ----------------

 if( inputs[5].is_nil() ) {  
   QplugAppendTextDisplay("Time interval missing\n");
   return(QPLUG_FAILURE);
 }
 DvEvent start_end = inputs[5]->getTimeRange();

 ttJoined = ttJoined->subSequence(start_end);

 if(ttJoined.is_nil() || ttJoined->seqSize() < 1) {
   QplugAppendTextDisplay("No data in interval\n");
   return(QPLUG_FAILURE);
 }
 

  ToJoinList.MultiJoin(ttJoined, JoinedList);
 
 QplugAppendTextDisplay("Join OK\n");


 // -- B field ---
 if (Bsup) {
   DvNode *Bnode = JoinedList.find("B");
   DvObject_var masked_B;
     
   if(Bnode) masked_B = Bnode->obj();
   
   if(masked_B.is_nil() || masked_B->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No B field in range ");
     return (QPLUG_FAILURE);
   }

   bfield_nest = masked_B->changeUnitsTo("1.0e-9>T", "nT"); // ensure units are in those assumed
   if(bfield_nest.is_nil() || bfield_nest->seqSize() < 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm B units are 1.0e-9>T\n");
     bfield_nest = masked_B;
   }    
   nestsize = bfield_nest->seqSize();

   printf("Nestsize B = %ld\n",nestsize);
   fflush(stdout);
 }


 // -- Velocity field ---
 if (Vsup) {
   DvNode *Vnode = JoinedList.find("V");
   DvObject_var masked_V;
     
   if(Vnode) masked_V = Vnode->obj();
   
   if(masked_V.is_nil() || masked_V->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No V in range ");
     return (QPLUG_FAILURE);
   }

   velocity_nest = masked_V->changeUnitsTo("1.0e3>m s^-1", "km/s"); // ensure units are in those assumed
   if(velocity_nest.is_nil() || velocity_nest->seqSize() < 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm V units are 1.0e3>m s^-1\n");
     velocity_nest = masked_V;
   }    
   nestsize = velocity_nest->seqSize();

   printf("Nestsize V = %ld\n",nestsize);
   fflush(stdout);
 }


 // -- Density ---
 if (Nsup) {
   DvNode *Nnode = JoinedList.find("N");
   DvObject_var masked_N;
     
   if(Nnode) masked_N = Nnode->obj();
   
   if(masked_N.is_nil() || masked_N->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No N in range ");
     return (QPLUG_FAILURE);
   }

   density_nest = masked_N->changeUnitsTo("1.0e6>m^-3", "cm^-3"); // ensure units are in those assumed
   if(density_nest.is_nil() || density_nest->seqSize() < 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm N units are 1.0e6>m^-3\n");
     density_nest = masked_N;
   }    
   nestsize = density_nest->seqSize();

   printf("Nestsize N = %ld\n",nestsize);
   fflush(stdout);
 }
 

 // -- T_par ---
 if (Tparsup ) {
   DvNode *Tparnode = JoinedList.find("Tpar");
   DvObject_var masked_Tpar;
     
   if(Tparnode) masked_Tpar = Tparnode->obj();
   
   if(masked_Tpar.is_nil() || masked_Tpar->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No Tpar in range ");
     return (QPLUG_FAILURE);
   }

   Tpar_nest = masked_Tpar->changeUnitsTo("1.0e6>K", "MK"); // ensure units are in those assumed
   if(Tpar_nest.is_nil() || Tpar_nest->seqSize() < 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm Tpar units are 1.0e6>K\n");
     Tpar_nest = masked_Tpar;
   }    
   nestsize = Tpar_nest->seqSize();

   printf("Nestsize Tpar = %ld\n",nestsize);
   fflush(stdout);
 }
 

 // -- T_perp ---
 if (Tperpsup) {
   DvNode *Tperpnode = JoinedList.find("Tperp");
   DvObject_var masked_Tperp;
     
   if(Tperpnode) masked_Tperp = Tperpnode->obj();
   
   if(masked_Tperp.is_nil() || masked_Tperp->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No Tperp in range ");
     return (QPLUG_FAILURE);
   }

   Tperp_nest = masked_Tperp->changeUnitsTo("1.0e6>K", "MK"); // ensure units are in those assumed
   if(Tpar_nest.is_nil() || Tpar_nest->seqSize() < 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm Tpar units are 1.0e6>K\n");
     Tperp_nest = masked_Tperp;
   }    
   nestsize = masked_Tperp->seqSize();

   printf("Nestsize Tperp = %ld\n",nestsize);
   fflush(stdout);
 }


 // --- Temp tensor, eventually split this into Tpar, Tperp --------
 if (Ttenssup) {
   DvNode *Tparnode = JoinedList.find("Tpar");
   DvObject_var masked_Tpar;
     
   if(Tparnode) masked_Tpar = Tparnode->obj();
   
   if(masked_Tpar.is_nil() || masked_Tpar->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No Tpar in range ");
     return (QPLUG_FAILURE);
   }

   masked_Tpar = masked_Tpar->changeUnitsTo("1.602e-19>J", "ev"); // ensure units are in those assumed
   if(masked_Tpar.is_nil() || masked_Tpar->seqSize() < 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm Tpar units are ev\n");
   }    
   nestsize = masked_Tpar->seqSize();

   printf("Nestsize Tpar = %ld\n",nestsize);
   fflush(stdout);


   Tpar_nest  = new DvObject(0.0, nestsize);
   Tperp_nest = new DvObject(0.0, nestsize);

   double inv_k_boltzmann = 1.602e-19/1.38054e-17; // [MK/ev]
   for (int i=0; i<nestsize; i++) {
     Tpar_nest->dbl(i)  = inv_k_boltzmann * masked_Tpar->dbl(i,2);
     Tperp_nest->dbl(i) = inv_k_boltzmann * 0.5 * (masked_Tpar->dbl(i,0) + masked_Tpar->dbl(i,1));
   }    
   
   Tpar_nest->change_xref(SI_CONVERSION, "1.0e6>K");
   Tperp_nest->change_xref(SI_CONVERSION, "1.0e6>K");
   Tpar_nest->change_xref(UNITS, "MK");
   Tperp_nest->change_xref(UNITS, "MK");

   Tperpsup=true;
   Tparsup=true;
   nestsize = Tperp_nest->seqSize();
   printf("Nestsize Tperp = %ld\n",nestsize);
   fflush(stdout);
 }


 QplugAppendTextDisplay("Calculating output..\n");

// ---- declare and calculate the output objects -----

 P_b     = new DvObject(0.0, nestsize);
 P_p     = new DvObject(0.0, nestsize);
 P_pfull = new DvObject(0.0, nestsize);
 P_tot   = new DvObject(0.0, nestsize);
 Beta    = new DvObject(0.0, nestsize);
 Alpha   = new DvObject(0.0, nestsize);
 Entropy = new DvObject(0.0, nestsize);
 Valf    = new DvObject(0.0, DIM3, nestsize);
 A_p     = new DvObject(0.0, nestsize);
 gyro    = new DvObject(0.0, nestsize);
 inert   = new DvObject(0.0, nestsize);
 DvObject_var cs  = new DvObject(0.0, nestsize);
 DvObject_var fpe  = new DvObject(0.0, nestsize);

 float pi = 3.14159;        // good enough here....
 float Mu_0 = 1.25664E-06;
 float k = 1.38054e-23;    // Boltzmann constant
 float e = 1.609e-19;      //  elementary charge C (Vs)
 float sfac = 1.0E21;      //  1E6 * 1E6 * 1E9 due to n in 1/ccm and T in units of 1E6K -> nPa
 float mp= 1.67e-27;       //  proton mass
 float me= 9.1091e-31;     //  electron mass
 float vtot2=0.00;

 DvObject_var btot2;
 if (Bsup) {
    btot2 = bfield_nest->dot(bfield_nest);
    *btot2 *= 1.0e-18; // in T^2
 }
   
 for(int i=0; i<nestsize; i++)  {
    vtot2 = 0.00;
    
    if (Bsup) {
      P_b->dbl(i) = btot2->dbl(i) / (2.0 * Mu_0) * 1.0E9;                                    // Pb in nPa;
    }
    
    if (Nsup) {
      inert->dbl(i) = 227.0 / sqrt(density_nest->dbl(i));
      fpe->dbl(i) = 8968.0 * sqrt(density_nest->dbl(i));
      if (Bsup) {
	Valf->dbl(i,0) = 21.83 * bfield_nest->dbl(i,0)/sqrt(density_nest->dbl(i)); // SEH Nov 2016 X; assuming ions==protons
 	Valf->dbl(i,1) = 21.83 * bfield_nest->dbl(i,1)/sqrt(density_nest->dbl(i)); // SEH Nov 2016 Y; assuming ions==protons
	Valf->dbl(i,2) = 21.83 * bfield_nest->dbl(i,2)/sqrt(density_nest->dbl(i)); // SEH Nov 2016 Z; assuming ions==protons
      }
    }
    if (Nsup && Tperpsup) {
      P_p->dbl(i) = density_nest->dbl(i) * k * Tperp_nest->dbl(i) * sfac;    // Pp in nPa;
    }
    if (Nsup && Tperpsup && Bsup) {
      	Beta->dbl(i) = P_p->dbl(i) / P_b->dbl(i);                             // Plasma beta
    }
    if (Tperpsup && Bsup) {
      gyro->dbl(i) = 1300.0 * sqrt(Tperp_nest->dbl(i)) / (sqrt(btot2->dbl(i)) * 1e09) ;
    }
    if (Nsup && Tperpsup && Tparsup) {
      P_pfull->dbl(i) = density_nest->dbl(i) * k * (2.0 * Tperp_nest->dbl(i) + Tpar_nest->dbl(i))/3.0 * sfac; // Ppfull in nPa;
      A_p->dbl(i) = Tpar_nest->dbl(i)/ Tperp_nest->dbl(i) - 1.00;            // temp anisotropy
      if (Bsup) {
        P_tot->dbl(i) = P_p->dbl(i) + P_b->dbl(i);
      }
    }
    if (Vsup) {
      for (size_t j=0; j<3; j++) {
        vtot2 += velocity_nest->dbl(i,j) * velocity_nest->dbl(i,j);
      }  
    }
    if (Nsup && Tparsup && Tperpsup && Bsup) {
      Alpha->dbl(i) = density_nest->dbl(i) * k * sfac * (Tpar_nest->dbl(i) - Tperp_nest->dbl(i)) /  (btot2->dbl(i) / (Mu_0) * 1.0E9);       
      Entropy->dbl(i) = P_pfull->dbl(i) / pow(density_nest->dbl(i),(5.0/3.0));   
      cs->dbl(i) = 989.0 * sqrt(P_pfull->dbl(i)/density_nest->dbl(i));
    }    
 }    

 // printf("Completed calculations..\n");

 // --- send some output to QSAS window ---
 sprintf(txtbuf,"Nest size (# of data points) = %ld",nestsize);
 QplugAppendTextDisplay(txtbuf);


 // --- insert timetags + attributes and push to WL ---


 DvObject_var empty;//,empty2,empty3,empty4,empty5,empty6,empty7,empty8,empty9,empty10,empty11,empty12,empty13;

 if (Bsup) {
   P_b->change_xref(DEPEND_0, ttJoined);
   P_b->change_xref("SI_conversion","1e-9 Pa");
   P_b->change_xref("UNITS","nPa");
   P_b->change_xref("LABLAXIS","Mag_press");
   P_b->change_xref("CATDESC","Magnetic pressure");
   outputs.push_back(P_b);
 } else  outputs.push_back(empty);

 if (Nsup && (Tperpsup)) {
   P_p->change_xref(DEPEND_0, ttJoined);
   P_p->change_xref("SI_conversion","1E-9 Pa");
   P_p->change_xref("UNITS","nPa");
   P_p->change_xref("LABLAXIS","Plasma_press");
   P_p->change_xref("CATDESC","Plasma pressure P = nkT_perp");
   outputs.push_back(P_p);
 }  else  outputs.push_back(empty);

 if (Nsup && Tperpsup && Tparsup) {
   P_pfull->change_xref(DEPEND_0, ttJoined);
   P_pfull->change_xref("SI_conversion","1E-9 Pa");
   P_pfull->change_xref("UNITS","nPa");
   P_pfull->change_xref("LABLAXIS","Plasma_press");
   P_pfull->change_xref("CATDESC","Plasma pressure P = nkT");
   outputs.push_back(P_pfull);
 } else  outputs.push_back(empty);

 if (Bsup && Nsup && Tparsup && Tperpsup) {
   P_tot->change_xref(DEPEND_0, ttJoined);
   P_tot->change_xref("SI_conversion","1E-9 Pa");
   P_tot->change_xref("UNITS","nPa");
   P_tot->change_xref("LABLAXIS","Pp_tot");
   P_tot->change_xref("CATDESC","Total plasma pressure (Pp + Pp_perp");
   outputs.push_back(P_tot);
 } else  outputs.push_back(empty);

 if (Bsup && Nsup && Tperpsup) {
   Beta->change_xref(DEPEND_0, ttJoined);
   Beta->change_xref("LABLAXIS","Beta");
   Beta->change_xref("CATDESC","Plasma Beta");
   outputs.push_back(Beta);
   QplugAppendTextDisplay("4 - Bsup Nsup & Tperp-> Beta");
 } else  outputs.push_back(empty);

 if (Tparsup && Tperpsup) {
   A_p->change_xref(DEPEND_0, ttJoined);
   A_p->change_xref("LABLAXIS","A_p");
   A_p->change_xref("CATDESC","Temperature anisotropy");
   outputs.push_back(A_p);
 } else  outputs.push_back(empty);

 if (Nsup && Tparsup && Tperpsup && Bsup) {
   Alpha->change_xref(DEPEND_0, ttJoined);
   Alpha->change_xref("LABLAXIS","Alpha");
   Alpha->change_xref("CATDESC","Alpha");
   Alpha->change_xref("UNITS","NA");
   outputs.push_back(Alpha);

   Entropy->change_xref(DEPEND_0, ttJoined);
   Entropy->change_xref("LABLAXIS","Entropy");
   Entropy->change_xref("CATDESC","Entropy");
   outputs.push_back(Entropy);
   
   cs->change_xref(DEPEND_0, ttJoined);
   cs->change_xref("LABLAXIS","CS");
   cs->change_xref("Frame","scalar>na");
   cs->change_xref("UNITS","m s^-1");
   cs->change_xref("CATDESC","Sound speed");
   outputs.push_back(cs);
 } else {
   outputs.push_back(empty);
   outputs.push_back(empty);
   outputs.push_back(empty);
 } 
  
 if (Nsup) {
   fpe->change_xref(DEPEND_0, ttJoined);
   fpe->change_xref("LABLAXIS","fpe");
   fpe->change_xref("Frame","scalar>na");
   fpe->change_xref("UNITS","s^-1");
   fpe->change_xref("CATDESC","Plasma frequency");
   outputs.push_back(fpe);

   inert->change_xref(DEPEND_0, ttJoined);
   inert->change_xref("LABLAXIS","Ion inertial length");
   inert->change_xref("Frame","scalar>na");
   inert->change_xref("UNITS","km");
   inert->change_xref("CATDESC","Ion inertial length");
   outputs.push_back(inert);
 } else {
   outputs.push_back(empty);
   outputs.push_back(empty);
 }   

 if (Tperpsup && Bsup) {
   gyro->change_xref(DEPEND_0, ttJoined);
   gyro->change_xref("LABLAXIS","GyroRadius");
   gyro->change_xref("Frame","scalar>na");
   gyro->change_xref("UNITS","km");
   gyro->change_xref("CATDESC","Ion gyro radius");
   outputs.push_back(gyro);
 }else {
   outputs.push_back(empty);
 }
 // --- SEH 9 Nov 2016 set frame, coordsys for vector V_A --
 if (Nsup && Bsup) {
   DvString frame_B = B_in->getXrefText("Frame");
   DvString csys_B = B_in->getXrefText("COORDINATE_SYSTEM");
   Valf->change_xref(DEPEND_0, ttJoined);
   Valf->change_xref("LABLAXIS","Alfven Velocity");
   Valf->change_xref("TENSOR_ORDER","1");
   Valf->change_xref("Frame",frame_B);
   Valf->change_xref("COORDINATE_SYSTEM",csys_B);
   Valf->change_xref("UNITS","km/s");
   Valf->change_xref("CATDESC","Alfven velocity");
   outputs.push_back(Valf);
 }else {
   outputs.push_back(empty);
 }
 

 return QPLUG_SUCCESS;

}




