/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	ShockNormalCoords.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Shock Normal Coordinates Plugin wrapper
*
* PURPOSE:      Calculate Rotation Matrix to transform into Shock
*               Normal Coordinates lmn. n is normal to a model 
*               bow shock, the l direction is the projection of the
*               upstream field onto the surface, and m completes the
*               right handed lmn system..
* 
* DESCRIPTION:  .
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*
*             pos_in      Constant vector or Vector TS
*             time_c      Time for crossing
*             Vsw         Approx solar wind speed to calculate aberration
*             upstream_vec Upstream magnetic field vector or time series
*             time_int_up Upstream time interval to average upstream_vec
*             imodel      Model number from Ch10 of ISSI Book
*             frame_out   GSE->lmn (0) or GSM->lmn (1)
*
*   OUT:
*             frametolmn  Rotation Matrix R. 
*                          R dot vector(frame_out) = vector(lmn)		
*
* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
* FUNCTIONS USED: built on qrtn and qbowshock modules
*
* HISTORY: July 2001 Steve Schwartz 
*
***********************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"

#include "qdutil.h"
#include "qrtn.h"
#include "DvObj.h"
#include "qbowshock.h"

const double Re = 6.3712e6; //Earth radius [m]  - a global	

using namespace QSAS;  

// local prototypes

QplugReturnStatus QSNPosition(RtnVector *vec_gsep,
								DvObject_var pos_in,
								DvObject_var time_c);

QplugReturnStatus QSNMakeGSERe (RtnVector *vecp);

QplugReturnStatus QSNmake_upstream_vec(RtnVector *vec, DvObject_var vec_obj, 
									DvEvent time_int_up);

void QrVector_diag_print(RtnVector *vecp);
  
/**********************************************************************/
extern "C" QplugReturnStatus DvShockNormalCoords( vector<DvObject_var>& inputs, vector<DvObject_var>& outputs ){
/**********************************************************************/

// define some C variables

long  check_rtn;
float Vsw, r_vec[3], n_gse[3];
QBSMP_models   imodel;
int   i, qbs_rtn;

// call-list arguments
 
DvObject_var    pos_in;        // 0
DvObject_var    time_c;        // 1
DvObject_var    Vsw_var;       // 2
DvObject_var    ups_vec_obj;   // 3
DvEvent         time_int_up;   // 4
DvObject_var    imodel_var;    // 5
DvObject_var    frame_out_var; // 6
DvObject_var    frametolmn;    // 7


RtnVector  vec_gse, vec_up, l, m, n;

QplugReturnStatus ret_status;
QplugAppendTextDisplay("\n");

// ...first 2 items are input position vector/vector TS and time interval

pos_in = inputs[0];
if (pos_in.is_nil()){
  QplugAppendTextDisplay("SNCoords>> Error getting pos_in object\n");
  return QPLUG_FAILURE;
}

time_c = inputs[1];
if( time_c.is_nil() ) {
  QplugAppendTextDisplay ("SNCoords>> Time object invalid.\n");
  return QPLUG_FAILURE;
}

vec_gse.IsPositionVector = 0;     // force so QrDoRotation will accept

ret_status = QSNPosition(&vec_gse, pos_in, time_c);
if (ret_status != QPLUG_SUCCESS){
  QplugAppendTextDisplay("SNCoords>> QBNPosition failed\n");
  return QPLUG_FAILURE;
}

// vec_gse now has a sequence with 1 element, the gse_xyz position
// of the crossing in Re with one timetag. 

Vsw_var = inputs[2];
if (Vsw_var.is_nil()){
   QplugAppendTextDisplay("SNCoords>> Input arg Vsw not Scalar\n");
   return QPLUG_FAILURE;
}

imodel_var = inputs[5];
if(imodel_var.is_nil()){
   QplugAppendTextDisplay("SNCoords>> Input arg imodel not Scalar\n");
   return QPLUG_FAILURE;
}


// unpack inputs for QBOWSHOCK routines and find gse normal;
Vsw = (float) Vsw_var->asDouble();

imodel = (QBSMP_models)imodel_var->asInt();

for (i=0; i<3; i++){ r_vec[i] = (float)vec_gse.ptr->dbl(0,i); }

qbs_rtn = QBSNormalAtSurfaceGSE(r_vec, Vsw, imodel, n_gse);
if (qbs_rtn != QBS_OK){
   QplugAppendTextDisplay("SNCoords>> QBSNormal returned error\n");
   return QPLUG_FAILURE;
}


// put normal in n

n          = vec_gse;   // copy over n_recs = 1, timetag, ...
n.in_frame = GSE;
n.ptr      = new DvObject(0.0, DIM3);
for(size_t i=0; i<3; i++) n.ptr->dbl(i) = n_gse[i];

n.ptr->setFrameAttr( "vector>gse_xyz"); // needed for new qrtn

// get upstream field vector from inputs; average over time interval

ups_vec_obj = inputs[3];
if (ups_vec_obj.is_nil()){
  QplugAppendTextDisplay("SNCoords>> Error getting upstream vector object\n");
  return QPLUG_FAILURE;
}

time_int_up = inputs[4]->getTimeRange();

ret_status = QSNmake_upstream_vec(&vec_up, ups_vec_obj, time_int_up);
if (ret_status != QPLUG_SUCCESS){
  QplugAppendTextDisplay ("SNCoords>>make_upstream_vec failed\n");
  return QPLUG_FAILURE;
}

frame_out_var = inputs[6];
if(frame_out_var.is_nil()){
   QplugAppendTextDisplay("SNCoords>> Frame_out not a scalar\n");
   return QPLUG_FAILURE;
}

frametolmn = new DvObject(0.0, DIM33);

switch ( frame_out_var->asInt() ){
   case 1:      // gsm
      n.out_frame = GSM;
      vec_up.out_frame = GSM;
      frametolmn->setFrameAttr("rot>gsm_xyz");
      break;
   case 0: default:    // gse
      n.out_frame = GSE;
      vec_up.out_frame = GSE;
      frametolmn->setFrameAttr("rot>gse_xyz");
      break;
}

check_rtn = n.QrDoRotation();  // put n in correct frame
if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
   QplugAppendTextDisplay( "SNCoords>> Failed to put n into out frame\n");
   return QPLUG_FAILURE;
}

n.in_frame = n.out_frame;      // make in_frame consistent with holding
check_rtn = vec_up.QrDoRotation();  // put vec_up in correct frame
if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
   QplugAppendTextDisplay( "SNCoords>> Failed to put vec_up into out frame\n");
   return QPLUG_FAILURE;
}
vec_up.in_frame = vec_up.out_frame; // make in_frame consistent with holding

// find n x B_up 

m.ptr = n.ptr->vec(vec_up.ptr);
if (m.ptr.is_nil()) {
    QplugAppendTextDisplay ("SNCoords>> n cross B_u failed\n");
    return QPLUG_FAILURE;
}

DvObject_var modm = m.ptr->abs();

if (modm.is_nil()){
    QplugAppendTextDisplay ("SNCoords>> |m| returned invalid number\n");
    return QPLUG_FAILURE;
}

m.ptr =  m.ptr / modm;

if ((m.ptr).is_nil()) {
    QplugAppendTextDisplay ("SNCoords>> m/|m| failed\n");
    return QPLUG_FAILURE;
}

l.ptr = m.ptr->vec(n.ptr);
if (l.ptr.is_nil()){
    QplugAppendTextDisplay ("SNCoords>> l = m x n failed\n");
    return QPLUG_FAILURE;
}

for (i=0; i<3; i++){
    frametolmn->dbl(0,0,i) = l.ptr->dbl(i);
    frametolmn->dbl(0,1,i) = m.ptr->dbl(i);
    frametolmn->dbl(0,2,i) = n.ptr->dbl(i);
}


// set the attributes for the output rotation matrix

frametolmn->change_xref(FIELDNAM, "RMatrix_to_lmn");
frametolmn->change_xref(SI_CONVERSION, "1>()");
frametolmn->change_xref("UNITS", " ");
frametolmn->change_xref("ToFrame", "vector>lmn_xyz");
char model_name[200];
sprintf(model_name,"Model %d: %s",imodel,BSMP_name[imodel]);
frametolmn->change_xref("Boundary Model", model_name);

char crossing_str[200];
sprintf(crossing_str,"( %9.3g, %9.3g, %9.3g )GSE Re", 
    vec_gse.ptr->dbl(0),vec_gse.ptr->dbl(1),vec_gse.ptr->dbl(2) );
frametolmn->change_xref("Crossing",crossing_str);
frametolmn->change_xref("Crossing Time", vec_gse.t->time().getISOstring().c_str());
char B_up_str[200];
sprintf(B_up_str,"B_up = (%9.3g, %9.3g, %9.3g)", 
  vec_up.ptr->dbl(0), vec_up.ptr->dbl(1), vec_up.ptr->dbl(2) );
frametolmn->change_xref("B_upstream", B_up_str);
frametolmn->change_xref("Upstream Interval", time_int_up.getISOstring().c_str());

DvObject_var thetabn_ptr = vec_up.ptr->dot(n.ptr);
DvObject_var vec_up_norm = vec_up.ptr->abs();
DvObject_var n_norm = n.ptr->abs();
double thetabn = thetabn_ptr->dbl() / (vec_up_norm->dbl() * n_norm->dbl() );

thetabn = (45.0/atan(1.0))*acos(thetabn);
char thetabn_str[200];
sprintf(thetabn_str,"Theta_bnu = %11.5g deg", thetabn);
frametolmn->change_xref("Theta_bn",thetabn_str);

outputs.push_back(frametolmn);     

// put some results to plug-in text window
char outstr[200];
sprintf(outstr,"SNCoords>> Crossing %s at %s\n", crossing_str, vec_gse.t->time().getISOstring().c_str());
QplugAppendTextDisplay(outstr);
sprintf(outstr,"SNCoords>> Normal is (%9.3g, %9.3g, %9.3g) Frame = %s\n",
   n.ptr->dbl(0), n.ptr->dbl(1), n.ptr->dbl(2),
   frametolmn->getFrameAttr().c_str() );
QplugAppendTextDisplay(outstr);
QplugAppendTextDisplay(B_up_str);QplugAppendTextDisplay("\n");
QplugAppendTextDisplay("Upstream Time Interval: ");
QplugAppendTextDisplay(time_int_up.getISOstring().c_str());QplugAppendTextDisplay("\n");
QplugAppendTextDisplay(thetabn_str);QplugAppendTextDisplay("\n");
QplugAppendTextDisplay("SNCoords>> finished successfully\n");

return QPLUG_SUCCESS;

}

/**********************************************************************/
QplugReturnStatus QSNPosition(RtnVector *vec_gsep,
                              DvObject_var pos_in,
                              DvObject_var time_c){
/**********************************************************************/

long  check_rtn;
QplugReturnStatus ret_status;

// sort out position of crossing. Input can either be a time series or
// a constant vector, so this a bit complicated. Separate routines
// return in vec_gse a sequence of 1 cartesian vector whose timetag
// is the time  time_c, and whose frame is
// that of pos_in. This is then rotated into gse. [NB qrtn routines at
// present (July 2001) assume vectors are _xyz]

if (!pos_in->isThreeVector()){
   QplugAppendTextDisplay("SNCoords>> Const pos_in not a 3Vector\n");
   return QPLUG_FAILURE;
}

if( pos_in->seqSize() == 1) {
    vec_gsep->ptr = new DvObject(pos_in);
} 
else {
    vec_gsep->ptr = pos_in->interpAt(time_c->time(0));
}

vec_gsep->ptr->ensureVectorXYZ();

vec_gsep->n_recs = 1;
vec_gsep->t = time_c;

vec_gsep->in_frame = vec_gsep->QrGetFrameFromAttr(pos_in);
if (vec_gsep->in_frame == UNKNOWN){
    QplugAppendTextDisplay("SNCoords>>  pos_in unknown frame\n");
    return QPLUG_FAILURE;
}
    
ret_status = QSNMakeGSERe(vec_gsep);
if (ret_status != QPLUG_SUCCESS){
    QplugAppendTextDisplay("SNCoords>> QSNMakeGSERe failed\n");
    return QPLUG_FAILURE;
}

//QrVector_diag_print(vec_gsep);
vec_gsep->out_frame = GSE;
check_rtn = vec_gsep->QrDoRotation();  // make sure vec_gse in gse
if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
   QplugAppendTextDisplay( "SNCoords>> Failed to put pos into gse\n");
   return QPLUG_FAILURE;
}

//QrVector_diag_print(vec_gsep);
vec_gsep->in_frame = GSE;        // advertise in GSE so don't double rotate

return QPLUG_SUCCESS;
}


/**********************************************************************/
QplugReturnStatus QSNMakeGSERe (RtnVector *vecp){
/**********************************************************************/

// rotate vecp to GSE and scale to Re

int check_rtn;
double factor, SIfactor;

// Rotate

vecp->out_frame = GSE;
//printf("QBNMakeGSEre received:\n");QrVector_diag_print(vecp);
check_rtn = vecp->QrDoRotation();
if (check_rtn != QRTN_OK && check_rtn != SAME_FRAME){
   QplugAppendTextDisplay("SNCoords>> QSNMakeGSERe rotate to gse failed\n");
   printf("QSNMakeGSERe QrDoRotation returned %d\n",check_rtn);
   fflush(stdout);
   return QPLUG_FAILURE;
}
vecp->in_frame = GSE;    // if don't do this, another call to QrDoRot will
                         // rotate again!
// Scale
// Algorithm:
//    Take SI_conversion, verify base unit = m, and use SI_conv/Re as factor
//    If SI_conversion exists but isn't -> m or not positive, assume factor=1.
//    If no SI_conversion, try to make sense of UNITS; default to factor = 1

factor = 1.0;  // assume in Re if all else fails; actually this is done
               // explicitly below, but this is here for safety.

//check whether SI_conversion exists
DvString attr = vecp->ptr->getXrefText("SI_conversion");
if(attr.empty()){                // no SI_conversion
   QplugAppendTextDisplay (
    "SNCoords>>WARNING:Input Vector SI_conversion not defined, trying UNITS.\n");
   attr = vecp->ptr->getXrefText("UNITS");
   if(attr.empty()) {
     QplugAppendTextDisplay (
     "SNCoords>>WARNING: UNITS not defined, assuming [Re].\n");
     factor = 1.0;
   } 
   else {                           // yes, we have UNITS
     if(attr.contains("RE")) factor=1.0;
     else if(attr.contains("KM") ) factor=1000.0/Re;
     else {
       QplugAppendTextDisplay (
         "SNCoords>>WARNING: UNITS not recognized, assuming [Re].\n");
       factor = 1.0;
     }
   }   
} 
else {      // Found SI_conversion.
  // check whether it is position vector, but don't change vecp->IsPos...
  // so can rotate it
  if(vecp->QrIsPositionVector(vecp->ptr)!= 1){  // checks base SI == m
      QplugAppendTextDisplay("SNCoords>> Input Vector not position vector. ");
      QplugAppendTextDisplay("SNCoords>> WARNING: Re will be assumed.\n");
      factor = 1.0;
  } 
  else {
      SIfactor = vecp->ptr->convFactor();  // get SI_conversion
      if(SIfactor <= 0.0) {
        QplugAppendTextDisplay("SNCoords>> Input Vec SI_conversion invalid.\n");
        QplugAppendTextDisplay ("SNCoords>> WARNING: Re will be assumed.\n");
        factor = 1.0;
      } 
      else {
        factor = SIfactor/Re;
      }
  }
}

*(vecp->ptr) *= factor;

//printf("QBNMakeGSEre returned:\n");QrVector_diag_print(vecp);

return QPLUG_SUCCESS;
}


/*********************************************************************/
QplugReturnStatus QSNmake_upstream_vec(RtnVector *vec, DvObject_var vec_obj, 
                            DvEvent time_int){
/*********************************************************************/

// If ups_vec_obj is a constant, return it in QrVector *vec
// If it is a time series, average it over time_int

QplugReturnStatus ret_status;

// be sure vec_obj in cartesians; returns same obj if already xyz
DvObject_var vecxyz_obj = new DvObject(vec_obj);

vecxyz_obj->ensureVectorXYZ();

// prepare vec
vec->n_recs = 1;
vec->in_frame =  vec->QrGetFrameFromAttr(vecxyz_obj);
if (vec->in_frame == UNKNOWN){
  QplugAppendTextDisplay("SNCoords>> Upstream vec in unknown frame\n");
  return QPLUG_FAILURE;
}


// constant vector used as is; sequence averaged

if ( vecxyz_obj->seqSize() == 1 ){  // constant - just use it!
   QplugAppendTextDisplay("SNCoords>>Const Upstr Vec; time ignored\n");
   vec->ptr= vecxyz_obj; 
} 
else  {  
    // sequence - average it
  
    vecxyz_obj = vecxyz_obj->subSequence(time_int);
    DvString comp = 0;
    comp += ",";
    comp =  comp + (int) (vecxyz_obj->seqSize()-1);

    vec->ptr = vecxyz_obj->getTimeAve(comp);
} 


vec->t = vecxyz_obj->getTimeTags();


return QPLUG_SUCCESS;
}



/**************************************************************************/
void QrVector_diag_print(RtnVector *vecp){
  printf("IsPositionVector = %d, in_frame = %d, out_frame = %d, n_recs = %d\n",
        vecp->IsPositionVector,vecp->in_frame,vecp->out_frame, vecp->n_recs);
  printf("First Vector: (%9.3g, %9.3g, %9.3g) at %s\n",
     vecp->ptr->dbl(0), vecp->ptr->dbl(1), vecp->ptr->dbl(2),
     vecp->t->time().getISOstring().c_str());
	fflush(stdout);
  return;
}



