//  qslot.cc
//
// slot Utility Widgets
//
// Aug 2007: Tony Allen
//

/// \file qgu.cc
///  Implementation file for the main GUI classes that are used within QSAS which are: \n
///  \arg \c QguListItem \n
///  \arg \c QguListView \n
///  \arg \c QguWLinput  \n
///  \arg \c QguWindow   \n \n
///  Those classes specialise respectively the QT3 built-in classes as listed below\n
///  \arg \c Q_LIST_VIEW_ITEM \n
///  \arg \c QListView     \n
///  \arg \c QLineEdit     \n
///  \arg \c QMainWindow   \n 
///  \author Tony Allen    \n
///  \version 2.1         \n
///  \date    2007

#include <qcolor.h>
#include <qpalette.h>
#include <qglobal.h>
#include <qpaintdevice.h>
#include <qfont.h>

#include "qslot.h"
#include "qwui.h"
#include "qprofile.h"

#include <stdio.h>

extern qwuiWindow* qsasWindow;


QColor col_White("#ffffff");
QColor col_Light_Grey("#eeeeee");
QColor col_Grey("#8c8c8c");
QColor col_Black("#000000");
QColor col_Green("#228b22");
QColor col_Red("#cc0000");
QColor col_DarkRed("#990000");
QColor col_Yellow("#ffff00");
QColor col_Amber("#ddbb33");
QColor col_LightYellow("#ffffaa");
QColor col_Blue("#000099");



// QguWLinput class
//----------------------------

/// \fn QguWLinput::QguWLinput(QWidget *parent, const char *name)
/// \brief Constructor of QguWLinput class
/// \param     parent  of type QWidget      
/// \param     name    of type const char *  
/// \exception None
/// \return    None
///
QguWLinput::QguWLinput(QWidget *parent, const char *name, int width)
	:  QWidget(parent)
{
	graphicsParent = NULL;
	widgetParent = parent;
	QHBoxLayout * layout=new QHBoxLayout(this);

	this->setLayout(layout);

	WLI_palette = new QPalette(qApp->palette());
	QFont slot_font = QFont(qApp->font());
	QFontMetrics slot_font_mt(slot_font);
	int font_width = slot_font_mt.width("n");
	int font_height = slot_font_mt.height();
	this->setFont(slot_font);

	active = true;       // turn this slot on
	ObjectName_nil = QString("");
	ObjectName = ObjectName_nil;
	
    if(QSAS_profiles->current()->slotLR == 1)
	{
		dropDownArray = new QToolButton(this);
		dropDownArray->setArrowType(Qt::DownArrow);
		dropDownArray->setFixedSize(24,20);
		dropDownArray->setPopupMode(QToolButton::InstantPopup);
		layout->addWidget(dropDownArray);
		
		lineBox = new QguLineEdit(this, "InputLineBox");
		lineBox->WLinput = this;
		layout->addWidget(lineBox);
	}
	else
	{
		lineBox = new QguLineEdit(this, "InputLineBox");
		lineBox->WLinput = this;
		layout->addWidget(lineBox);
 
		dropDownArray = new QToolButton(this);
		dropDownArray->setArrowType(Qt::DownArrow);
		dropDownArray->setFixedSize(24,20);
		dropDownArray->setPopupMode(QToolButton::InstantPopup);
		layout->addWidget(dropDownArray);
	  
	}
	connect(dropDownArray, SIGNAL(clicked(bool)), this, SLOT(dropdownClicked(bool)));

  // initialise the dropDown menu possibilities
    
	defaultPopup = 0;
	TrangePopup = 0;
	TimePopup = 0;
	TimeFromELPopup = 0;
	TivlFromEvtTablePopup = 0;
	TivlSlicePopup = 0;
	vectorPopup = 0;
	arrayPopup = 0;
  
	layout->setMargin(0);
	layout->setSpacing(0);
  
	dropDownArray->hide();
	setColour(COL_OK);
	connect(lineBox, SIGNAL(textChanged(const QString &)), this, SLOT(resetCache()));
	connect(lineBox, SIGNAL(hadDropEvent()), this, SIGNAL(hadDropEvent()));
	connect(lineBox, SIGNAL(textEdited(const QString &)), this, SLOT(validateInput(const QString &)));
	connect(qsasWindow->wlview, SIGNAL(objChanged(const QString&)), this, SLOT(validateObjChanged(const QString&)));
	connect(qsasWindow->wlview, SIGNAL(deleted(const QString&)), this, SLOT(validateDeletion(const QString&)));
	connect(qsasWindow->wlview, SIGNAL(pathChanged(const QString&, const QString&)), this, SLOT(validatePathChanged(const QString&, const QString&)));
  
	this->resize(font_width*width,(int)(font_height*1.5)); // large enough for a time string
	this->setMinimumWidth(font_width*width);
	takesAlias=false;
}


/// \fn QguWLinput::~QguWLinput()
/// \brief Destructor of QguWLinput class
/// \exception None
/// \return    None
///
QguWLinput::~QguWLinput()
{
	delete lineBox;
	if(WLI_palette) delete WLI_palette;
	if(defaultPopup) delete defaultPopup;
	if(TrangePopup) delete TrangePopup;
	if(TimePopup) delete TimePopup;
	if(TivlFromEvtTablePopup) delete TivlFromEvtTablePopup;
	if(TivlSlicePopup) delete TivlSlicePopup;
	if(vectorPopup) delete vectorPopup;
	if(arrayPopup) delete arrayPopup;

	delete dropDownArray;
  
}

// --------------------------
void QguWLinput::resetCache(){
	DVOS_cache = nilDVObject;
}

// --------------------------
void QguWLinput::set3DSubsetDef(const char *setting, bool sliceTime){
	rtpDefault = string(setting);
	if(sliceTime) rtpDefault.append("0");
	
	// used to override defaults for subsample of particle data with preferred options
}

// --------------------------
QString QguWLinput::get3DSubsetDef(DvObject_var &dobj){
	// used to override defaults for subsample of particle data with preferred options

	QString compTxt = "";
	if(dobj.is_nil()) return compTxt;
	
	
	size_t nDims=dobj->nDims();
		
	if(nDims != 3) return compTxt;
	
	if(rtpDefault.empty()) return compTxt;

		compTxt += " [ ";
	bool getR = (rtpDefault.find("r") != string::npos);
	bool getPhi = (rtpDefault.find("p") != string::npos);
	bool getTheta = (rtpDefault.find("t") != string::npos);
	
	char strDepend[30];

	for(int i=1; i<=3; i++) {
		DvString dep("Depend_");
		dep.append(i);
		
		if(dobj->xref_exists(dep)){
		// find what this dimension is
		DvObject_var dep_i = dobj->get_xref(dep);
		if(dep_i.is_nil()) continue;

		if( !dep_i->xref_exists(SI_CONVERSION) ) return QString(""); // can't do isDeg etc
		
		bool isDeg = dep_i->isDeg();
		bool isAngle = dep_i->isAngle();
		bool isR = !isAngle;

		double minVal;
		double fillVal = -1.;
		int kStart = 0;
		int kEnd = 0;
		bool isPhi = dep_i->isPhiAngle(isDeg, i, minVal, kStart, kEnd);
		bool isTheta = !isPhi;
			
			// handle each component, skips bins with fills
			if(isR){
				if(getR){
					compTxt +=  QString(" (%1,%2) ").arg(kStart).arg(kEnd);
				}
				else{
					compTxt +=  QString(" Sum(%1,%2) ").arg(kStart).arg(kEnd);
				}
			}
			else if(isPhi){
				if(getPhi){
					compTxt +=  QString(" (%1,%2) ").arg(kStart).arg(kEnd);
				}
				else{
					compTxt +=  QString(" Sum(%1,%2) ").arg(kStart).arg(kEnd);
				}
			}
			else if(isTheta){
				if(getTheta){
					compTxt +=  QString(" (%1,%2) ").arg(kStart).arg(kEnd);
				}
				else{
					compTxt +=  QString(" Sum(%1,%2) ").arg(kStart).arg(kEnd);
				}
			}
		}
	}
	compTxt += " ]";

	if(rtpDefault.find("0") != string::npos) compTxt += " <0>";
	
	
	return compTxt;
	
}

   
// --------------------------
QMenu * QguWLinput::getDefaultPopup(int seqChoice, bool event)
{
	if(this->defaultPopup == 0) {
 	   defaultPopup = new QMenu(this->dropDownArray);
	}
	this->defaultPopup->clear();

	defaultPopup->addAction("This Object", this, SLOT(setDropdownText_dobj()));
	defaultPopup->addSeparator();
	if(event){
		defaultPopup->addAction("Time Subsample... ", this, SLOT(openSliceS()));
	}
	else{
		if(seqChoice == QS_SEQ ) defaultPopup->addAction("Time Subsample... ", this, SLOT(openSliceS_seq()));
		else if(seqChoice == QS_NOT_SEQ ) defaultPopup->addAction("Time Subsample... ", this, SLOT(openSliceS_not_seq()));
		else if(seqChoice == QS_ANY) defaultPopup->addAction("Time Subsample... ", this, SLOT(openSliceS()));
	}
	defaultPopup->addSeparator();
	defaultPopup->addAction(" Time and Component Slice ", this, SLOT(openSliceV()));
	defaultPopup->addAction(" Time and Array Slice ", this, SLOT(openArrayReduce()));

	defaultPopup->addSeparator();
	defaultPopup->addAction("Metadata", this, SLOT(openSelectMetadata()));

	this->defaultPopup->hide();
	return this->defaultPopup;
}

// --------------------------
QMenu * QguWLinput::getTrangePopup()
{
	if(this->TrangePopup == 0){

		TrangePopup = new QMenu(this->dropDownArray);

		TrangePopup->addAction("From Object", this, SLOT(setDropdownText_dobj()));
		TrangePopup->addSeparator();
		TrangePopup->addAction("Subset...", this, SLOT(openSliceT_seq()));
		TrangePopup->addSeparator();
		TrangePopup->addAction("Metadata", this, SLOT(openSelectMetadata()));
	}
	this->TrangePopup->hide();
	return this->TrangePopup;
}

// --------------------------
QMenu * QguWLinput::getTimePopup()
{
	if(this->TimePopup == 0){
		TimePopup = new QMenu(this->dropDownArray);
		TimePopup->addAction("Default", this, SLOT(setTimeDropdownDefault()));
		TimePopup->addSeparator();
		TimePopup->addAction("Start Time", this, SLOT(setDropdownTstart()));
		TimePopup->addAction("Centre Time", this, SLOT(setDropdownTcentre()));
		TimePopup->addAction("End Time", this, SLOT(setDropdownTend()));
		TimePopup->addSeparator();
		TimePopup->addAction("Time at Record", this, SLOT(openSliceT_not_seq()));
		TimePopup->addSeparator();
		TimePopup->addAction("Metadata", this, SLOT(openSelectMetadata()));	
	}

	this->TimePopup->hide();
	return this->TimePopup;
}

// --------------------------
QMenu * QguWLinput::getTimeFromELPopup()
{
	if(this->TimeFromELPopup == 0){
		TimeFromELPopup = new QMenu(this->dropDownArray);
		TimeFromELPopup->addAction("Default", this, SLOT(setETDropdownDefault()));
		TimeFromELPopup->addSeparator();
		TimeFromELPopup->addAction("Select Event and time", this, SLOT(selectEventAndTime()));
		TimeFromELPopup->addSeparator();
		TimeFromELPopup->addAction("Metadata", this, SLOT(openSelectMetadata()));	
	}

	this->TimeFromELPopup->hide();
	return this->TimeFromELPopup;
}

// --------------------------

QMenu * QguWLinput::getTivlFromEvtTablePopup()
{
	if(this->TivlFromEvtTablePopup == 0){
		TivlFromEvtTablePopup = new QMenu(this->dropDownArray);
		TivlFromEvtTablePopup->addAction("Default Event", this, SLOT(setDropdownText_event()));
		TivlFromEvtTablePopup->addAction("Select Event", this, SLOT(selectTivlEvent()));
	}
	this->TivlFromEvtTablePopup->hide();
	return this->TivlFromEvtTablePopup;
}

// --------------------------

QMenu * QguWLinput::getTivlSeqSlicePopup()
{
	if(this->TivlSlicePopup == 0){
		TivlSlicePopup = new QMenu(this->dropDownArray);
		TivlSlicePopup->addAction("This Object", this, SLOT(setDropdownText_dobj()));
		TivlSlicePopup->addSeparator();
		TivlSlicePopup->addAction("Select Subset", this, SLOT(selectTivlSeqSubset()));
		TivlSlicePopup->addAction("Select Event", this, SLOT(selectTivlEvent()));

	}
	this->TivlSlicePopup->hide();
	return this->TivlSlicePopup;
}

 // --------------------------
 
QMenu * QguWLinput::getVectorPopup(DvObject_var dobj)
{
	if(this->vectorPopup == 0){
		vectorPopup = new QMenu(this->dropDownArray);

		vectorPopup->addAction("Default", this, SLOT(setVectorDropdownDefault()));
		vectorPopup->addAction(DV_RMAG, this, SLOT(setDropdownText_r()));
		vectorPopup->addAction(DV_X, this, SLOT(setDropdownText_x()));
		vectorPopup->addAction(DV_Y, this, SLOT(setDropdownText_y()));
		vectorPopup->addAction(DV_Z, this, SLOT(setDropdownText_z()));
		vectorPopup->addAction(DV_LAT_DEG, this, SLOT(setDropdownText_ld()));
		vectorPopup->addAction(DV_THETA_DEG, this, SLOT(setDropdownText_td()));
		vectorPopup->addAction(DV_PHI_DEG, this, SLOT(setDropdownText_pd()));
		vectorPopup->addAction(DV_LAT_RAD, this, SLOT(setDropdownText_lr()));
		vectorPopup->addAction(DV_THETA_RAD, this, SLOT(setDropdownText_tr()));
		vectorPopup->addAction(DV_PHI_RAD, this, SLOT(setDropdownText_pr()));
		if(dobj->getFrame() == "gse") vectorPopup->addAction(DV_LOCAL_T, this, SLOT(setDropdownText_lt()));
		vectorPopup->addSeparator();
		vectorPopup->addAction(" Time and Component Slice ", this, SLOT(openSliceV()));
		vectorPopup->addAction(" Time and Array Slice ", this, SLOT(openArrayReduce()));
		vectorPopup->addSeparator();
		vectorPopup->addAction("Metadata", this, SLOT(openSelectMetadata()));
	}
	this->vectorPopup->hide();
	return this->vectorPopup;
}

 // --------------------------
 
QMenu * QguWLinput::getArrayPopup()
{
	if(this->arrayPopup == 0){
		arrayPopup = new QMenu(this->dropDownArray);

		arrayPopup->addAction("Default", this, SLOT(setArrayDropdownDefault()));
		arrayPopup->addSeparator();
		arrayPopup->addAction("Time and Array Slice", this, SLOT(openArrayReduce()));
		arrayPopup->addSeparator();
		arrayPopup->addAction("Metadata", this, SLOT(openSelectMetadata()));
	}
	this->arrayPopup->hide();
	return this->arrayPopup;
}

 // --------------------------
 
bool QguWLinput::hasFocus(){
	return lineBox->hasFocus();
}

 // --------------------------
 
void QguWLinput::clearFocus(){
	return lineBox->clearFocus();
}

bool QguWLinput::hasComponent(){

	DvString StrCmp = this->text().toLatin1().data();

	if( StrCmp.contains("[") ) return true;
	if( StrCmp.contains("<") ) return true;
	
	return false;
}

// --------------------------
QguComp_e QguWLinput::getComponent(){

	QguComp_e component;
	QString StrCmp = QString( this->text().toLatin1().data() );

	int compEndAt = StrCmp.indexOf(" [");
	int sliceEndAt = StrCmp.indexOf(" <");
	
	if (StrCmp.contains(T_STR_0_START)) component = T_0_START_COMP;
	else if(compEndAt != -1)
	{
		StrCmp.remove(0,compEndAt); // get " [ x ]" etc
		if (StrCmp.contains( "mag"))       component = R_COMP;
		else if (StrCmp.contains( "x"))  component = X_COMP;
		else if (StrCmp.contains( "y"))  component = Y_COMP;
		else if (StrCmp.contains( "z"))  component = Z_COMP;
		else if (StrCmp.contains( "Lat") && StrCmp.contains( "deg")) component = LAT_DEG_COMP;
		else if (StrCmp.contains( "Lat") && StrCmp.contains( "rad")) component = LAT_RAD_COMP;
		else if (StrCmp.contains( "theta") && StrCmp.contains( "deg")) component = THETA_DEG_COMP;
		else if (StrCmp.contains( "theta") && StrCmp.contains( "rad")) component = THETA_RAD_COMP;
		else if (StrCmp.contains( "phi") && StrCmp.contains( "deg")) component = PHI_DEG_COMP;
		else if (StrCmp.contains( "phi") && StrCmp.contains( "rad")) component = PHI_RAD_COMP;
		else if (StrCmp.contains( "Local") ) component = LOCAL_T;
		else if (StrCmp.contains("Start")) component = T_START_COMP;
		else if (StrCmp.contains("Centre")) component = T_CENTRE_COMP;
		else if (StrCmp.contains( "End"))   component = T_END_COMP;
		else if (StrCmp.contains("[=>") ) component = XREF_COMP;
		else component = ARRAY_COMP;
	}
	else if(sliceEndAt !=-1)
		component=SLICE_COMP;
	else
	{
		component = WL_OBJ;
	}

	return component;
}


void QguWLinput::openSliceT_seq(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceT "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "T", this, "tSliceWindow", true, false);
	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());
	
	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceT_not_seq(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceT "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "T", this, "tSliceWindow", false, true);
	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());

	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceS_seq(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceS_seq "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "S", this, "tSliceWindow", true, false);
	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());

	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceS_not_seq(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceS "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "S", this, "tSliceWindow", false, true);
	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());

	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceS(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceS "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "S", this, "tSliceWindow", true, true);
	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());

	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}

void QguWLinput::openSliceV(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openSliceV "  << endl;

	QString objName = this->textObject();

	TimeSliceWindow * tSliceWindow;
	tSliceWindow = new TimeSliceWindow(objName, "V", this, "tSliceWindow", true, true);

	
	tSliceWindow->setPalette(qApp->palette());

	tSliceWindow->parentSlot = this;
	tSliceWindow->move(QCursor::pos());

	tSliceWindow->show();
	tSliceWindow->setToLastChoice();
  
}



void QguWLinput::setDropdownText_event(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_event " << EVENT_0_STR << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(EVENT_0_STR));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	this->reValidate(true);
}

void QguWLinput::setDropdownText_r(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_r " << R_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(R_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(R_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_x(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_x " << X_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(X_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(X_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_y(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_y " << Y_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Y_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Y_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_z(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_z " << Z_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Z_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Z_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_ld(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_ld " << Ld_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Ld_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Ld_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_td(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_td " << Td_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Td_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Td_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_pd(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_pd " << Pd_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Pd_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Pd_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_lr(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_lr " << Lr_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Lr_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Lr_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_tr(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_tr " << Tr_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Tr_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Tr_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_pr(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_pr " << Pr_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Pr_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Pr_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_lt(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_lt " << Lt_STR_CMP << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(Lt_STR_CMP));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->componentSelected(Lt_STR_CMP, "VECTOR");
	this->reValidate(true);
}

void QguWLinput::setDropdownText_dobj(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownText_dobj " << "this obj" << endl;
		
	qApp->processEvents(); // to avoid calculator crash

	this->lineBox->setText(QString(ObjectName.toLatin1().data()));
    this->lineBox->end(false);
	this->lineBox->setFocus();
	
	emit this->componentSelected("", "VECTOR");
	this->reValidate(true);
	
	qApp->processEvents(); // to avoid calculator crash

}

void QguWLinput::setDropdownTstart(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownTstart " << T_STR_START << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(T_STR_START));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->TimeChosen();
	emit this->componentSelected(T_STR_START, "TIME");
	this->reValidate(true);
}

void QguWLinput::setDropdownTcentre(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownTcentre " << T_STR_CENTRE << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(T_STR_CENTRE));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->TimeChosen();
	emit this->componentSelected(T_STR_CENTRE, "TIME");
	this->reValidate(true);
}
void QguWLinput::setDropdownTend(){
  
	if(QuSlotDebug())
		std::cout << "QguWLinput::setDropdownTend " << T_STR_END << endl;

	this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(T_STR_END));
    this->lineBox->end(false);
	this->lineBox->setFocus();

	emit this->TimeChosen();
	emit this->componentSelected(T_STR_END, "TIME");
	this->reValidate(true);
}
void QguWLinput::setTimeDropdownDefault(){
	
	if(QuSlotDebug()) std::cout << "QguWLinput::setTimeDropdownDefault "  << endl;

	DvObject_var dobj = getWLObject(ObjectName.toLatin1().data());
	
	if(dobj->is_time() && dobj->seqSize() == 1)
	{
		this->lineBox->setText(ObjectName);
		this->setColour(COL_OK);
		emit this->TimeChosen();
		this->reValidate(true);
	}
	else
	{
		// set start time
		this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(T_STR_START));
        this->lineBox->end(false);
		this->lineBox->setFocus();

		emit this->TimeChosen();
		emit this->componentSelected(T_STR_START, "TIME");
		this->reValidate(true);
	}
}

void QguWLinput::setETDropdownDefault(){
	
	if(QuSlotDebug()) std::cout << "QguWLinput::setETDropdownDefault "  << endl;

	DvObject_var dobj = getWLObject(ObjectName.toLatin1().data());
	
	if(dobj->is_time() && dobj->seqSize() == 1)
	{
		this->lineBox->setText(ObjectName);
		this->setColour(COL_OK);
		emit this->TimeChosen();
		this->reValidate(true);
	}
	else
	{
		// set start time
		this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(T_STR_0_START));
        this->lineBox->end(false);
		this->lineBox->setFocus();

		emit this->TimeChosen();
		emit this->componentSelected(T_STR_0_START, "TIME");
		this->reValidate(true);
	}
}

void QguWLinput::setVectorDropdownDefault(){
	
	if(QuSlotDebug()) std::cout << "QguWLinput::setVectorDropdownDefault "  << endl;
		
	bool takesVector = false;
	for(int i=0; i<acceptConditions.size(); i++)
	{
		if(acceptConditions[i].condDim == QS_VECTOR
		|| acceptConditions[i].condData == QS_TI
		|| acceptConditions[i].condObj == QS_WL_OBJ) takesVector = true;
	}
	
	if(takesVector)
	{
		setDropdownText_dobj();		
	}
	else
	{
		setDropdownText_r();
	}
	// reValidate done in setDropdown calls
}

void QguWLinput::setArrayDropdownDefault(){
	
	if(QuSlotDebug()) std::cout << "QguWLinput::setArrayDropdownDefault "  << endl;
		
	int targetDim = 0;
	int targetSeq = QS_SEQ;
	int cDim = 3;
	for(int i=0; i<acceptConditions.size(); i++)
	{
		// take largest acceptable dim, otherwise it may be impossible to get to base object
		if(acceptConditions[i].condDim == QS_0D) cDim = 0;
		else if(acceptConditions[i].condDim == QS_1D) cDim = 1;
		else if(acceptConditions[i].condDim == QS_2D) cDim = 2;
		else if(acceptConditions[i].condDim == QS_ARRAY) cDim = 3;
		else if(acceptConditions[i].condDim == QS_ANY) cDim = 3;
		if(targetDim < cDim) 
		{
			targetDim = cDim;
			targetSeq = acceptConditions[i].condSeq;
		}
		
	}
	
	QString defText = getDefaultComp(targetDim, targetSeq);
	this->setText(defText);
	reValidate(true);
}


void QguWLinput::openSelectMetadata(){

	QString objName = this->textObject();

	DvObject_var dobj = getWLObject(objName.toLatin1().data());
	if( dobj.is_ok() )
	{
		selectMetadataWindow * metaWindow =	new selectMetadataWindow(dobj, this);
		metaWindow->parentSlot = this;
		metaWindow->move(QCursor::pos());
		metaWindow->show();
	}
	else setColour(COL_NOT_OK);
}


void QguWLinput::openArrayReduce(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::openArrayReduce () " << this << endl;

	QString objName = this->textObject();
	vector <size_t> mdim;	
	
	DvObject_var dobj = getWLObject(objName.toLatin1().data());

	if(dobj.is_nil() ){
		setColour(COL_OK);
		return;
	}
		
	if (dobj->arraySize() == 1 ){
			
		cout << "Cannot resolve matrix object" << endl;
		setColour(COL_NOT_OK);
		return;
	}
	mdim = dobj->Dims();
	
	ArrayReduceWindow * arrWindow = new ArrayReduceWindow(objName, mdim, this, "ArrayWindow", false);
	arrWindow->setPalette(qApp->palette());

	arrWindow->parentSlot = this;
	arrWindow->move(QCursor::pos());
	arrWindow->show();
	arrWindow->setToLastChoice();

}

void QguWLinput::setEnabled(bool enabled){
	lineBox->setEnabled(enabled);
	if( enabled) setColour(COL_OK);
	else setColour(COL_INACTIVE);
}
 
 
/// \fn QguWLinput::setReadOnly()
/// \brief passes flag to setReadOnly of contained QLineEdit
/// \exception None
/// \return    None
///  
void QguWLinput::setReadOnly(bool readOnly)
{
	lineBox->setReadOnly(readOnly);
}

/// \fn QguWLinput::clear()
/// \brief passes clear to contained QLineEdit
/// \exception None
/// \return    None
///
void QguWLinput::clear()
{
	lineBox->clear();
}


void QguWLinput::validateObjChanged(const QString& context)
{
	if(this->text().startsWith(context))
	{
		DVOS_cache = new DvObject();
	}
    reValidate(true);

}

void QguWLinput::validateDeletion(const QString& path)
{
	const QString txt = this->text();
	if(txt.startsWith(path))
	{
		DVOS_cache = new DvObject();
		reValidate();
	}
}

void QguWLinput::validatePathChanged(const QString& from, const QString& to)
{
	const QString txt = this->text();
	if(txt.startsWith(from) || txt.startsWith(to))
	{
		DVOS_cache = new DvObject();
		reValidate(true);
	}
}

void QguWLinput::resetColour(){

	this->setPalette(*WLI_palette);
	this->lineBox->setPalette(*WLI_palette);
	this->dropDownArray->setPalette(*WLI_palette);
}

void QguWLinput::selectTivlEvent(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::selectTivlEvent "  << endl;

	QString objName = this->textObject();

	SelectTivlWindow * selectWindow = new SelectTivlWindow(objName, this, "tSliceWindow", false);
	selectWindow->setPalette(qApp->palette());

	selectWindow->parentSlot = this;
	selectWindow->move(QCursor::pos());

	selectWindow->exec();
}

void QguWLinput::selectEventAndTime(){

	if(QuSlotDebug())
		std::cout << "QguWLinput::selectEventAndTime "  << endl;

	QString objName = this->textObject();

	SelectETWindow * selectWindow = new SelectETWindow(objName, this, "etWindow", false);
	selectWindow->setPalette(qApp->palette());

	selectWindow->parentSlot = this;
	selectWindow->move(QCursor::pos());

	selectWindow->exec();
}



void QguWLinput::selectTivlSeqSubset(){
	if(QuSlotDebug())
		std::cout << "QguWLinput::selectTivlSeqSubset "  << endl;

	QString objName = this->textObject();

	SelectTivlSubsetWindow * selectWindow = new SelectTivlSubsetWindow(objName, this, "tSliceWindow", false);
	selectWindow->setPalette(qApp->palette());

	selectWindow->parentSlot = this;
	selectWindow->move(QCursor::pos());

	selectWindow->show();
}

/// \fn void QguWLinput::setColour(int validity)
/// \brief Method that sets the colour of the line edit object on the basis of the validity state of the content of this object
/// \param     validity of type const QString
/// \exception None
/// \return    None
///
void QguWLinput::setColour(int validity){

	WLI_palette->setBrush(QPalette::Active, QPalette::Button, WLI_palette->light());
	WLI_palette->setBrush(QPalette::Inactive, QPalette::Button, WLI_palette->light());
	WLI_palette->setBrush(QPalette::Inactive, QPalette::Text, WLI_palette->mid());
	WLI_palette->setBrush(QPalette::Inactive, QPalette::Base, WLI_palette->midlight());

	_validity=validity;

// palettes need to be changed BEFORE signals are emitted, to guarantee that
// any fields subsequently changed by slots connected to these signals,
// are not undone on return to this function

	if( validity == COL_OK){
		WLI_palette->setColor(QPalette::Active, QPalette::Base, col_White);
		WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Black);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_White);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Black);

		this->setPalette(*WLI_palette);
		this->lineBox->setPalette(*WLI_palette);
		this->dropDownArray->setPalette(*WLI_palette);
	}
	else if( validity == COL_INACTIVE){
		WLI_palette->setColor(QPalette::Active, QPalette::Base, col_White);
		WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Grey);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_White);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Grey);

		this->setPalette(*WLI_palette);
		this->lineBox->setPalette(*WLI_palette);
		this->dropDownArray->setPalette(*WLI_palette);
	}
	else if( validity == COL_NOT_OK){
		WLI_palette->setColor(QPalette::Active, QPalette::Base, col_Red);
		WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Light_Grey);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_DarkRed);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Light_Grey);

		this->setPalette(*WLI_palette);
		this->lineBox->setPalette(*WLI_palette);
		this->dropDownArray->setPalette(*WLI_palette);
	}
	else if( validity == COL_INCOMPLETE){
		WLI_palette->setColor(QPalette::Active, QPalette::Base, col_Amber);
		WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Black);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_Yellow);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Black);

		this->setPalette(*WLI_palette);
		this->lineBox->setPalette(*WLI_palette);
		this->dropDownArray->setPalette(*WLI_palette);
	}
	else if( validity == COL_COMP){
		WLI_palette->setColor(QPalette::Active, QPalette::Base, col_Yellow);
		WLI_palette->setColor(QPalette::Active, QPalette::Text, col_Blue);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Base, col_LightYellow);
		WLI_palette->setColor(QPalette::Inactive, QPalette::Text, col_Blue);

		this->setPalette(*WLI_palette);
		this->lineBox->setPalette(*WLI_palette);
		this->dropDownArray->setPalette(*WLI_palette);

	}
	return;
}

bool QguWLinput::isValid() const
{
	return (_validity==COL_OK || _validity==COL_COMP);
}

// -----------------------------------------------------------------

bool QguWLinput::reValidate(bool propagate)
{
	if(QuSlotDebug())
	{
		printf("QguWLinput::reValidate %d\n", propagate);
		fflush(stdout);
	}
	
	if(takesAlias) {
		emit tryAlias(propagate);
		return false;
	}
	
	bool OK;
	// This method allows the calling method to force slot validation

	if(!(this->active))
	{
		// grey out if inactive
		ObjectName = ObjectName_nil;
		setColour(COL_INACTIVE);
		return false;
	}

	QString txt = this->text();
	
	// if slot is empty, only validate once to set colour and don't propagate
	if(txt.isEmpty() && acceptConditions.size() > 0){
		OK = validateCondition(txt, acceptConditions[0]);
		return OK;	
	}
	
	// non-propagating validation when program sets text
	for(int i=0; i<acceptConditions.size(); i++)
	{
		OK = validateCondition(txt, acceptConditions[i]);
		if( OK ) 
		{
			if(propagate) emit edited_OK();
			// else no signal is emitted, just provides visual feedback
			return OK;
		}
	}
	
	return false;
}
// -----------------------------------------------------------------

void QguWLinput::validateInput(const QString& txt)
{
	if(QuSlotDebug())
	{
		printf("QguWLinput::validateInput %s\n", txt.toLatin1().data());
		fflush(stdout);
	}
	
	resetCache();
	
	if(takesAlias) {
		emit tryAlias(false);
		return;
	}
	

	// if slot is empty, only validate once to set colour
	if(txt.isEmpty() && acceptConditions.size() > 0){
		validateCondition(txt, acceptConditions[0]);
		return;
	}
	
	for(int i=0; i<acceptConditions.size(); i++)
	{
		if( validateCondition(txt, acceptConditions[i]) )  {
			emit edited_OK();
			return;
		}
	}
	
	return;
}


bool QguWLinput::validateCondition(const QString &txt, Qcondition &cond)
{

	if(QuSlotDebug())
	{
		printf("QguWLinput::validateCondition  '%s'\n", txt.toLatin1().data());
		fflush(stdout);
	}
	
	QString noWhite = txt.trimmed();
	if(noWhite.isEmpty())
	{
		emit slotEmpty();
		ObjectName = ObjectName_nil;
		setColour(COL_OK);
		if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
		return false; // will not emit signal edited_OK
	}
		
	// Does this resolve to a folder
	WLItem *WLitem = qsasWindow->wlview->itemFromPath(ObjectName);
	if(WLitem != NULL && WLitem->isFolder() && cond.condData != QS_TEXT)
	{
		ObjectName = ObjectName_nil;
		setColour(COL_INCOMPLETE);
		if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
		return false;
	}

	ObjectName = QString(baseObjectText(txt.toLatin1().data()).c_str()); // deep copy
	DvObject_var dobj = getWLObject(ObjectName.toLatin1().data());

	// QS_WL_OBJ
	// ----------
	if(cond.condObj == QS_WL_OBJ)
	{
		if(dobj.is_nil())
		{
			if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
			setColour(COL_NOT_OK);
			return false;
		}

//		if( dobj->seqSize() == 1 )
//		{
//			setDropdown(getDefaultPopup(QS_NOT_SEQ));
//		}
//		else
        if( dobj->arraySize() > 1 )
		{
			if( dobj->isThreeVector() ) 
			{
				setDropdown(getVectorPopup(dobj));
			}
			else
			{
				setDropdown(getArrayPopup());
			}
		}
		else if( dobj->is_event()){
			if(dobj->seqSize() > 1) setDropdown(getTivlFromEvtTablePopup());
			else setDropdown(getTimePopup());
		}
		else
		{
			setDropdown(getDefaultPopup(QS_ANY));
		}
		 
		
		if(!this->hasComponent()) {
			emit validated_WL_OBJ(ObjectName);
			setColour(COL_OK);
		}
		else setColour(COL_COMP);

		return true;
	}
	
	// QS_NARROW
	// ----------
	else if(cond.condObj == QS_NARROW)
	{
		if(dobj.is_nil())
		{
			if ( dropDownArray->isVisibleTo(this)) dropDownArray->hide();
			
			// only value typed in slot possible
			if(cond.condTT == QS_TT ||
			   cond.condSeq == QS_SEQ ||
			   (cond.condDim != QS_0D && cond.condDim != QS_ANY) )
			{
				setColour(COL_NOT_OK);
				return false;
			}
			
			if(cond.condData == QS_DBL)
			{
				bool OK;
				txt.toDouble(&OK); 
				if(OK) 
				{
					setColour(COL_OK);
					return true;
				}
				else
				{
					setColour(COL_NOT_OK);
					return false;
				}
			}
			else if(cond.condData == QS_INT)
			{
				bool OK;
				txt.toInt(&OK); 
				if(OK) 
				{
					setColour(COL_OK);
					return true;
				}
				else
				{
					setColour(COL_NOT_OK);
					return false;
				}
			}
			else if(cond.condData == QS_TIME)
			{
				DvTime timeValue;
				if(timeValue.setFromISOstring(txt.toLatin1().data()))
				{
					setColour(COL_OK);
					return true;
				}
				else
				{
					setColour(COL_INCOMPLETE);
					return false;
				}
			}
			else if(cond.condData == QS_TI)
			{
				DvEvent TIvalue;
				if(TIvalue.setFromISOstring(txt.toLatin1().data()))
				{
					setColour(COL_OK);
					return true;
				}
				else
				{
					setColour(COL_INCOMPLETE);
					return false;
				}
			}
			else
			{
				// always accept  QS_ANY or QS_TEXT
				setColour(COL_OK);
				return true;
			}
		}
		else
		{
			// object exists, can we narrow or subsample to required condition
			
			bool isComp = this->hasComponent();
			if(isComp) {
				setColour(COL_COMP);
			}
			else 
			{
				 setColour(COL_OK);
			}

			// sequence conditions
			int seqLen = dobj->seqSize();
			if( seqLen < 1 && cond.condSeq == QS_SEQ    ) 
			{
				setColour(COL_NOT_OK);
				return true;
			}
			
			// TT conditions
			DvObject_var tt = dobj->getTimeTags(); // object or its DEPEND_0
			if( tt.is_nil()  && cond.condTT == QS_TT)  
			{
				setColour(COL_NOT_OK);
				return false;
			}
			
			if(cond.condData != QS_TIME && cond.condData != QS_TI && dobj->is_event()){
				setColour(COL_NOT_OK);
				return false;
			}
			
			// data type conditions
			if(cond.condData == QS_TIME)
			{
				if(!dobj->is_time() && !dobj->is_event() && tt.is_nil() )
				{
					setColour(COL_NOT_OK);
					return false;
				}
				
				if(cond.condSeq == QS_NOT_SEQ)
				{
					if( dobj->is_time() && dobj->seqSize() == 1)
					{
						setColour(COL_OK);
					}
					else if(dobj->is_event()) setDropdown(getTimeFromELPopup());
					else 					  setDropdown(getTimePopup());
				}
				else
				{
					if( !dobj->is_time() && tt.is_nil() )
					{
						setColour(COL_NOT_OK);
						return false;
					}
					setDropdown(getDefaultPopup(cond.condSeq));
				}
						
				setColour(COL_COMP);
				return true;
			}
	
			if(cond.condData == QS_TI)
			{
				if(!dobj->is_event() && tt.is_nil() )
				{
					setColour(COL_NOT_OK);
					return false;
				}
				
				if(!dobj->is_event()){
					setDropdown(getDefaultPopup(cond.condSeq, true));
				}
				else{
					if(cond.condSeq == QS_NOT_SEQ)
					{
						if(dobj->seqSize() > 1)
						{
							setDropdown(getTivlFromEvtTablePopup());
							if(!isComp )
							{
								this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(EVENT_0_STR));
								setColour(COL_COMP);
							}
						}
						else
						{
							setDropdown(getTrangePopup());
						}
					}
					else
					{
						setDropdown(getTivlSeqSlicePopup());
					}
				}
				setColour(COL_COMP);
				return true;
				
			}
			
			if(cond.condData == QS_TEXT)
			{
				setDropdown(getDefaultPopup(cond.condSeq));
				return true;
			}
			
			if(cond.condData == QS_DBL || cond.condData == QS_INT)
			{

				if( !dobj->is_dbl() && !dobj->is_int()  )
				{
					setColour(COL_NOT_OK);
					setDropdown(getDefaultPopup(cond.condSeq));
					return false;
				}
				
				if( dobj->arraySize() == 1 )
				{
					setDropdown(getDefaultPopup(cond.condSeq));
					return true;
				}
				else 
				{
				if( dobj->isThreeVector() ) 
					{
						setDropdown(getVectorPopup(dobj));
						if(!isComp && cond.condDim == QS_0D) 
						{
							this->lineBox->setText(QString(ObjectName.toLatin1().data())+QString(R_STR_CMP));
							setColour(COL_COMP);
						}
						return true;
					}
					else
					{
						setDropdown(getArrayPopup());
						if( !isComp )
						{
							QString defComp = getDefaultComp(cond.condDim, cond.condSeq);

							this->lineBox->setText(defComp);
							setColour(COL_COMP);
						}
						return true;
					}
				}
			}
			
			if(cond.condData == QS_ANY )
			{
				// just do dropdown for now 
				// could handle each data type differently
				setDropdown(getDefaultPopup(cond.condSeq));
				return true;
			}
			
		}
		
	}
		
	return false;
}

void  QguWLinput::setDropdown(QMenu *dropdownChoice)
{
//	dropDownArray->setMenu(dropdownChoice);
	dropDownArray->hide();
	if ( !(dropDownArray->isVisibleTo(this)))
	{
		dropDownArray->show();
	}
	activeMenu = dropdownChoice;
	return;
}


// --------------------------------

void QguWLinput::setDefaultComp(int targetDim, int targetIsSeq, QString &obj_name)
{
	this->setText(obj_name);
	QString compText = this->getDefaultComp(targetDim, targetIsSeq);
	this->setText(compText);
}

// --------------------------------

QString QguWLinput::getDefaultComp(int targetDim, int targetIsSeq)
{
	
	if(QuSlotDebug()) cout << "QguWLinput::getDefaultComp:  "  << targetDim <<  " " << targetIsSeq << endl;
	
	QString outText = QString(baseObjectText(this->text().toLatin1().data()).c_str());
	
	if(targetDim >= 3) return outText;
	
	if(outText.isEmpty())  return this->text();

	 QString compDef;
	 if(targetDim == 0 ) compDef = QSAS_profiles->current()->Array3Dto0D;
	 else if(targetDim == 1 ) compDef = QSAS_profiles->current()->Array3Dto1D;
	 else if(targetDim == 2 ) compDef = QSAS_profiles->current()->Array3Dto2D;

	if(compDef.isEmpty()) return outText;
	
	DvObject_var dobj = getWLObject(outText.toLatin1().data());
	
	QString guiChoice = get3DSubsetDef(dobj);
	
	if( !guiChoice.isEmpty() ) {
		outText += guiChoice;
		return outText;
	}

	QString compTxt = " [";
		
	int n = (int)dobj->nDims();
	if( n <= targetDim) return outText;
	vector <size_t> dims = dobj->Dims();
	
	if(compDef.contains("[") || compDef.contains("<"))
	{
		compTxt = QString(" ")+compDef;
	}
	else 
	{
		for (int i=0; i<compDef.length(); i++) {
			if( n <= i) return outText;
		
			if(compDef[i] == QChar('i'))
			{
				int di = dims[i]; 
				compTxt +=  QString("(1,%1) ").arg(di);
			}
			else if(compDef[i] == QChar('S'))
			{
				int di = dims[i];
				compTxt +=  QString("Sum(1,%1) ").arg(di);
			}
			else if(compDef[i] == QChar('A'))
			{
				int di = dims[i];
				compTxt +=  QString("Ave(1,%1) ").arg(di);
			}

			else return outText;
		}
		if(targetIsSeq == QS_NOT_SEQ  && dobj->seqSize() > 1) compTxt +=  "] < 0 >";
		else compTxt +=  "]";
	}
	
	outText += QString(compTxt);
	
	return outText;

}

bool QguWLinput::validObject()
{
	DvString ObjName = baseObjectText(this->text().toLatin1().data());
	return validObjName(ObjName.c_str());
}



DvObject_var QguWLinput::getObject()
{
	// return local cached object
	if(DVOS_cache.is_ok()) return DVOS_cache;
	
	// always returns a DvObject_var pointer.
	// Returns pointer to empty object on failure (test on obj.is_nil() or is_ok() ) 

	if ( QuSlotDebug() )
	{
		printf( " QguWLinput::getObject for '%s' ObjectName= '%s' \n", text().toLatin1().data(), ObjectName.toLatin1().data());
		fflush(stdout);
	}

	// get slot text
	DvString txt = this->text().toLatin1().data();
	
	txt.trimmed(); // remove leading and trailing white
	
	bool isNumber = txt.isNumber();
	
	// give up now if slot empty
	if(txt.empty()) {
		return  DVOS_cache;
	}

	// try to resolve to one of the conditions in the order they are set
	for(int i=0; i<acceptConditions.size(); i++)
	{
		Qcondition cond = acceptConditions[i];


		// QS_WL_OBJ
		// ----------
	
		if(cond.condObj == QS_WL_OBJ)
		{ 	
			DvObject_var baseObj = getWLObject(txt);

 			if(baseObj.is_ok()) {
			
				DvString compStr = txt.betweenIncl("[", "]");
				DvString recStr = txt.betweenIncl("<", ">");
				compStr += " ";
				compStr += recStr;
	
				DVOS_cache = baseObj->getObjComponent(compStr);

				if( DVOS_cache.is_ok() ) return  DVOS_cache;
			}
		}
	
		// QS_NARROW
		// ----------
	
		else if(cond.condObj == QS_NARROW)
		{
			DvObject_var baseObj = getWLObject(txt); 
			if(baseObj.is_nil())
			{
				// only value typed in slot possible
				if(cond.condTT == QS_TT || cond.condSeq == QS_SEQ || (cond.condDim != QS_0D && cond.condDim != QS_ANY) )  
				{
					return  DVOS_cache;
				}	
				
				if(cond.condData == QS_INT && isNumber)
				{					
					int value = txt.toInt(); 
					DVOS_cache = new DvObject(value);
					return DVOS_cache;
					
				}
                
				if(cond.condData == QS_DBL && isNumber)
				{				
					double value = txt.toDouble(); 
					DVOS_cache = new DvObject(value);
					return DVOS_cache;
					
				}
			
				else if(cond.condData == QS_TIME)
				{
					DvTime timeValue;
					if(timeValue.setFromISOstring(txt))
					{
						DVOS_cache = new DvObject(timeValue);
						return DVOS_cache;
					}
				}

				else if(cond.condData == QS_TI)
				{
					DvEvent TIvalue;
					if(TIvalue.setFromISOstring(txt))
					{
						DVOS_cache = new DvObject(TIvalue);
						return DVOS_cache;
					}
				}

				else if(cond.condData == QS_TEXT)
				{
					DVOS_cache = new DvObject(txt);
					return DVOS_cache;

				}

			}
			else
			{
				// base object exists, can we narrow or subsample to required condition

				// sequence conditions
				int seqLen = baseObj->seqSize();
				if( seqLen <= 1 && cond.condSeq == QS_SEQ ) continue; // try next condition
			
				// TT conditions
				if( !baseObj->xref_exists(DEPEND_0)  && cond.condTT == QS_TT)  continue; // try next condition

				// data type conditions
				if(cond.condData == QS_TIME)
				{
					DVOS_cache = baseObj->getTimeComponent(txt);
				}
				
				else if(cond.condData == QS_TI)
				{
					DVOS_cache = baseObj->getIntervalComponent(txt);
				}
			
				else if(cond.condData == QS_TEXT)
				{
					DVOS_cache = baseObj->getObjComponent(txt);
				}
			
				else if(cond.condData == QS_DBL || cond.condData == QS_INT)
				{
					DVOS_cache = baseObj->getObjComponent(txt);
				}
			
				else if(cond.condData == QS_ANY )
				{
					DVOS_cache = baseObj->getObjComponent(txt);
				}
				
				if( DVOS_cache.is_ok() ) {
					return DVOS_cache;
				}
			}
		}
  
	} // end for on condition list
	
	return DVOS_cache; //  is_nil if we get here
}


double QguWLinput::value()
{
	bool acceptNumber=false;
	bool acceptTime=false;
	bool acceptTI=false;
	for (int i=0; i<acceptConditions.size(); i++)
	{
		Qcondition cond = acceptConditions[i];
		if( cond.condDim == QS_0D && cond.condSeq == QS_NOT_SEQ)
		{
			if( cond.condData == QS_DBL || cond.condData == QS_INT ) acceptNumber = true;
			if( cond.condData == QS_TIME ) acceptTime = true;
			if( cond.condData == QS_TI ) acceptTI = true;
		}
	}
	
	if (acceptNumber)
	{
		bool ok;
		return this->lineBox->text().toDouble(&ok);
	}
	else if (acceptTime)
	{
		DvTime slotTime;
		slotTime.setFromISOstring(this->lineBox->text().toLatin1().data());
		
		return  slotTime.getEpoch2000(); // used for slots needing to get/set from double
	}
	else if (acceptTI)
	{
		DvEvent slotTI;
		slotTI.setFromISOstring(this->lineBox->text().toLatin1().data());
		
		return  slotTI.duration(); // used for slots needing to get/set from double
	}
	else return 0;
}

/// \fn void QguWLinput::setDouble()
/// \brief Method that sets the slot text as time string from double if acceptTime is true 
/// \param     None
/// \exception None
/// \return    none
///
void QguWLinput::setDouble(double value)
{

	bool acceptNumber=false;
	bool acceptTime=false;
	bool acceptTI=false;
	for (int i=0; i<acceptConditions.size(); i++)
	{
		Qcondition cond = acceptConditions[i];
		if( cond.condDim == QS_0D && cond.condSeq == QS_NOT_SEQ)
		{
			if( cond.condData == QS_DBL || cond.condData == QS_INT ) acceptNumber = true;
			if( cond.condData == QS_TIME ) acceptTime = true;
			if( cond.condData == QS_TI ) acceptTI = true;
		}
	}

	if (acceptTime)
	{
		DvTime slotTime(value);
		this->lineBox->setText(slotTime.getISOstring(0).c_str());
	}
	if(acceptNumber)
	{
        DvString strVal = "";
        strVal += value;
		this->lineBox->setText(strVal.c_str());
	}
	return;
}
 
// -------------------------------------------



/// \fn void QguWLinput::dropEvent(QDropEvent *e)
/// \brief This event handler is called when the drag is dropped on this widget. It has been re-implemented from QWidget
///
/// Sets and validates dropped text
/// \param     e of type QDropEvent *
/// \exception None
/// \return    None
///
void QguWLinput::dropEvent(QDropEvent *e)
{
	if(QuSlotDebug())
	{
		printf("QguWLinput::dropEvent \n");
		fflush(stdout);
	}

	QApplication::setOverrideCursor( QCursor(Qt::WaitCursor) );

	qApp->processEvents();


	// use valid ObjectName in feedback

	if (e->mimeData()->hasText())
	{
		string newText = e->mimeData()->text().toLatin1().data();
		cleanWLtext(newText);
		setTextKeepComp(QString(newText.c_str()));
			
	}
	
	QApplication::restoreOverrideCursor();

	qApp->processEvents();
	lineBox->setFocus();
	emit hadDropEvent();
	reValidate(true); // causes edited_OK to be emitted on success
}




void QguWLinput::setText(QString txt)
{
	this->lineBox->setText(txt);
	return;

}


void QguWLinput::setText(DvString txt)
{
	this->lineBox->setText(txt.c_str());
	return;

}

void QguWLinput::setText(const char * txt)
{
	this->lineBox->setText(txt);
	return;

}

void QguWLinput::setTextKeepComp(const QString &txt){
	
	if ( QuSlotDebug() )
	{
		printf( " QguWLinput::setTextKeepComp for  '%s'\n",  txt.toLatin1().data());
		fflush(stdout);
	}

	bool compOK = true;
	DvString oldText(this->text().toLatin1().data()); 

	DvString comp = oldText.betweenIncl(" [", "]");
	DvString tcomp = oldText.betweenIncl(" <", ">");

	DvString newText(txt.toLatin1().data()); //  copy of dropped name
	newText = newText.front("[");
	newText = newText.front("<");

	
	// process modifiers found in oldText (of the form ?str1=str2?)
	// used by mult-trace templates
	
	newText = newText.replaceTokens(oldText.c_str());
	
	DvObject_var dobj = getWLObject(newText);

	if(dobj.is_nil()) {
		this->lineBox->setText(newText.c_str());
		return;
	}
	
	ObjectName = newText.c_str();
	
	if ( !comp.empty() ){

		compOK = compatibleComp(dobj, comp);


		// now replace with components if still valid
		if( compOK ) 
		{
			newText +=  comp;
			
		}
	}
	if ( !tcomp.empty() ){
		int idx;
		DvString recs = tcomp.between("<", ">");
		if(recs.contains(",")){
			// two values, take second
			idx = recs.after(",").toInt();
		}
		else idx = recs.toInt();

		// now replace with components if still valid
		if( idx < dobj->seqSize() ) newText +=  tcomp;
			
	}

	this->lineBox->setText(newText.c_str());

	// Empty slot or no component and subset.
	// Set new object and return
	bool acceptNumber=false;
	bool acceptTime=false;
	for (int i=0; i<acceptConditions.size(); i++)
	{
		Qcondition cond = acceptConditions[i];
		if( cond.condDim == QS_0D && cond.condSeq == QS_NOT_SEQ)
		{
			if( cond.condData == QS_DBL || cond.condData == QS_INT ) acceptNumber = true;
			if( cond.condData == QS_TIME ) acceptTime = true;
		}
	}
	

	// now correct for special requests
	
	if( acceptTime && dobj->is_event() ) setETDropdownDefault();
	else if(acceptTime) setTimeDropdownDefault();
	else if(acceptNumber) 
	{
		if(dobj->isVectorXYZ()) setVectorDropdownDefault();
		else setArrayDropdownDefault();
	}

	this->reValidate(); // needed to set objectName
	
	emit dropDetected(txt.toLatin1().data());	
	return;

}


// ------------------------------------------------------
 
bool QguWLinput::compatibleComp(DvObject_var &dobj,  DvString comp){
	
	bool vectorComp=false;

	if(	comp.contains(DV_RMAG) ||  
		comp.contains(DV_X) ||
		comp.contains(DV_Y) ||
		comp.contains(DV_Z) ||
		comp.contains(DV_LAT_DEG) ||
		comp.contains(DV_LAT_RAD) ||
		comp.contains(DV_THETA_DEG) ||
		comp.contains(DV_THETA_RAD) ||
		comp.contains(DV_PHI_DEG) ||
		comp.contains(DV_PHI_RAD) ||
		comp.contains(DV_LOCAL_T) ) vectorComp = true;

	if( dobj->isVectorXYZ() || dobj->arraySize() == 3)
	{
		// look only for Vector comp types
		if(vectorComp) return true;
		return false;
	}
	else return false;

	// most things take time comp
	if(comp.contains(DV_T_START) ||
	   comp.contains(DV_T_CENTRE)||
	   comp.contains(DV_T_END)  ) return true;

	// check array dimensions
	

	for(int i=0; i<dobj->nDims(); i++)
	{
		DvString next = comp.between("(", ")");
		
		if(next.empty()) return false;
		
		if(next.contains(",")){
			// two values, take second
			next = next.after(",");
		}
		int idx = next.toInt();

		if(idx > dobj->Dims()[i]) return false;
		
		// move to next comp
		comp = comp.after(")");
		
	}

	return true;
	
}


// ------------------------------------------------------



QString QguWLinput::text()
{

	return this->lineBox->text();

}


QString QguWLinput::textObject()
{
	return QString(baseObjectText(this->lineBox->text().toLatin1().data()).c_str());
}


void QguWLinput::setCursorPosition(int posn)
{
	this->lineBox->setCursorPosition(posn);
	return;

}


/// \fn void QguWLinput::setAlignment(int flag)
/// \brief Method that sets the text alignment in the underlying lineEdit
/// \param     flag of type int
/// \exception None
/// \return    None

void QguWLinput::setAlignment(Qt::AlignmentFlag flag)
{
   // Takes int flag of form QT:AlignLeft of QT:AlignAuto
	this->lineBox->setAlignment(flag);

	return;

}


/// \fn void QguWLinput::setMaxLength(int len)
/// \brief Method that sets the max text length in the underlying lineEdit
/// \param     len of type int
/// \exception None
/// \return    None
void QguWLinput::setMaxLength(int len)
{

	this->lineBox->setMaxLength(len);

	return;

}

void QguWLinput::dropdownClicked(bool checked){
	this->raise();
	activeMenu->popup(dropDownArray->mapToGlobal(QPoint(0,0)));
	if(widgetParent == NULL && graphicsParent != NULL) graphicsParent->setSelected(true); // to get menu to top
	activeMenu->activateWindow();
}




dimTripleStack QguWLinput::parseArrayReduce()
{
	if ( QuSlotDebug() )
		printf( "QguWLinput::parseArrayReduce this = %s \n", this->text().toLatin1().data() );

	dimTripleStack arrayTriples;

	QString compTxt = this->lineBox->text().toLatin1().data();

	int ptr = compTxt.indexOf(" [");
	if(ptr == -1) return arrayTriples;  // no component section found, arrayTriples is empty vector

	return getDimTripleStack(compTxt.toLatin1().data());

}

/// \fn vector <dimTriple *> QguWLinput::parseTimeSlice()
/// \brief Method that parses text in QguLineEdit and finds the array reduction requested
/// \param     None
/// \exception None
/// \return    vector <dimTriple *>

dimTriple QguWLinput::parseTimeSlice()
{
	if ( QuSlotDebug() )
		printf( "QguWLinput::parseTimeSlice this = %s \n", this->text().toLatin1().data() );

	DvString cmpTxt = DvString(this->lineBox->text().toLatin1().data());
	return getSliceTriple(cmpTxt);
}

// QguLIneEdit Class

/// \fn void QguLineEdit::QguLineEdit(QWidget *parent, const char *name)
/// \brief QguLineEdit constructor
///
/// \param    parent of type QWidget*
/// \param    name of type const char*
/// \exception None
/// \return    None
///
QguLineEdit::QguLineEdit(QWidget *parent, const char *name)
	:QLineEdit(parent)
{
      // simple class to allow us to overload drop
	active=true;
	activeColor=QApplication::palette().color(QPalette::Active, QPalette::Light);
	inactiveColor=QApplication::palette().color(QPalette::Active, QPalette::Midlight);
	activeText=QApplication::palette().color(QPalette::Active, QPalette::Text);
	inactiveText=QApplication::palette().color(QPalette::Inactive, QPalette::Text);

	setAcceptDrops(true);
	setDragEnabled(true);

}

QguLineEdit::~QguLineEdit(){

}

void QguLineEdit::setActive(bool active)
{

	if(this->active!=active)
	{
		if(active)
		{
			QPalette palette(this->palette());
 			palette.setColor(this->backgroundRole(), activeColor);
 			palette.setColor(this->foregroundRole(), activeText);
 			this->setPalette(palette);
		}
		else
		{
			QPalette palette(this->palette());
 			palette.setColor(this->backgroundRole(), inactiveColor);
 			palette.setColor(this->foregroundRole(), inactiveText);
 			this->setPalette(palette);
		}
		this->active=active;
	}

}
/*
void QguLineEdit::focusInEvent(QFocusEvent* e)
{
	QLineEdit::focusInEvent(e);
	this->setFocus();
}
 */
/// \fn void QguLineEdit::dropEvent(QDropEvent *e)
/// \brief This event handler is called when the drag is dropped on this widget. It has been re-implemented from QWidget
///
/// Sets and validates dropped text
/// \param     e of type QDropEvent *
/// \exception None
/// \return    None
///
void QguLineEdit::dropEvent(QDropEvent *e)
{
	if(QuSlotDebug())
	{
		printf("QguLineEdit::dropEvent  \n");
		fflush(stdout);
	}

	if( !(WLinput->active) ) return;
	
	WLinput->dropEvent(e);

}



// ------------------------------------------------------
void QguLineEdit::mouseDoubleClickEvent ( QMouseEvent * event )  
{
	this->selectAll();
}



/*
QMimeData* QguLineEdit::mimeData ( const QString slotText ) const 
{
	QMimeData *mimeData = new QMimeData();

	mimeData->setData("application/slotdata", slotText.toAscii());
	mimeData->setText(slotText);
	return mimeData;

}
*/
QStringList QguLineEdit::mimeTypes() const
{
	QStringList types;
	types << "application/slotdata";
	return types;
}

void QguLineEdit::focusInEvent(QFocusEvent *e)
{
	QLineEdit::focusInEvent(e);
	emit gotFocus();
}


