// qpui_frame.cc
//
// Plot Interface Implementation - Frame Object
//
//  Tony Allen
//  Jan 2008, Jan 2014

#include "qpui.h"

extern QPalette QPUI_normal_palette;
extern int qsas_restore_release;

//     FrameItem CLASS
//     ----------------------

FrameItem::FrameItem(const QString& name, PageItem* parent, QpFrameType frame_type, int position) : QpuiItem(name, parent)
{
	if(parent==NULL)
	{
		std::cerr << "FrameItem::FrameItem: creating FrameItem with NULL parent" << std::endl;
		return;
	}
	restoring = false;
	
	tabWidget=parent->tabWidget;

	I_Type = QP_FRAME;
	_frameType = frame_type;

	if(name.isEmpty())
	{
		setNextName();
	}
	
	
	// initialise options tab widgets
	PropertiesTab = new QpPropertiesTab(tabWidget, "PropertiesTab", this);
	PropertiesTab->hide();
	
	DisplayTab = new QpDisplayTab(tabWidget, "DisplayTab", this);
	DisplayTab->hide();
	
	TitleTab = new QpTitleTab(tabWidget, "TitleTab", this);
	TitleTab->hide();

	tAxisTab = new QpTAxisTab(tabWidget,  "tAxisTab", this);
	tAxisTab->hide();


	xAxisTab = new QpXAxisTab(tabWidget, "xAxisTab", this);
	xAxisTab->hide();
	

	if(frame_type == QP_TS) xtAxisTab = tAxisTab;
	else xtAxisTab = xAxisTab;
	
	setIcon(QP_FRAME);

}
//-------------

FrameItem::FrameItem(FrameItem* frame, PageItem* parent, int position) : QpuiItem("", parent)
{
	if(parent==NULL)
	{
		std::cerr << "FrameItem::FrameItem: creating FrameItem with NULL parent" << std::endl;
		return;
	}
	restoring = false;

	
	setName(frame->getName());
	
	_title_colour=frame->_title_colour;
	_line_colour=frame->_line_colour;
	
	tabWidget=parent->tabWidget;

	I_Type = QP_FRAME;
	_frameType = frame->_frameType;
	
	// initialise options tab widgets
	PropertiesTab = new QpPropertiesTab(frame->PropertiesTab, this);
	PropertiesTab->hide();
	
	DisplayTab = new QpDisplayTab(frame->DisplayTab, this);
	DisplayTab->hide();
	
	TitleTab = new QpTitleTab(frame->TitleTab, this);
	TitleTab->hide();
	
	tAxisTab = new QpTAxisTab(frame->tAxisTab, this);
	tAxisTab->hide();
	
	xAxisTab = new QpXAxisTab(frame->xAxisTab, this);
	xAxisTab->hide();

	if(_frameType == QP_TS) xtAxisTab = tAxisTab;
	else xtAxisTab = xAxisTab;
	
	for(int i=0; i<frame->childCount(); i++)
	{
		PanelItem* panelItem=dynamic_cast<PanelItem*>(frame->child(i));
		if(panelItem!=NULL)
		{
			new PanelItem(panelItem, this);
		}
	}
	
	// set some user selections
	if(frame->DisplayTab){
		DisplayTab->_ticks_char_size_fld->setValue(frame->DisplayTab->_ticks_char_size_fld->value());

	}
	
	setIcon(QP_FRAME);

}
//-------------

FrameItem::~FrameItem()
{

	delete PropertiesTab;
	delete DisplayTab;
	delete TitleTab;
	delete tAxisTab;
	delete xAxisTab;
}
//-------------

bool FrameItem::panelBlank(int position){
	if(position < 0) return true;
	if(position >= childCount()) return true;

	PanelItem * panel = dynamic_cast<PanelItem*> (child(position));
	if(panel && panel->isBlank()) return true;
	
	return false;
	
}

// ----------------------------------

void FrameItem::setDefaultOptions()
{
	TitleTab->_title_fld->setText("");
	TitleTab->_subtitle_fld->setText("");
	TitleTab->_autotitle_fld->setChecked(true);
	TitleTab->_title_char_size_fld->setValue(QSAS_profiles->current()->BaseFontSize+QPUI_TITLE_CHAR_SIZE_ADD);
	TitleTab->_title_disp_fld->setValue(QPUI_TITLE_DISP);
	TitleTab->_subtitle_char_size_fld->setValue(QSAS_profiles->current()->BaseFontSize+QPUI_SUBTITLE_CHAR_SIZE_ADD);
	TitleTab->_subtitle_disp_fld->setValue(QPUI_SUBTITLE_DISP);
	TitleTab->_title_just_fld->setCurrentIndex(QPUI_HCENTRE);
	TitleTab->_title_font_fld->setFontParams(PL_FCI_SANS, false, false);
	
	
	QPalette line_palette = QPalette(QPUI_normal_palette);
	line_palette.setColor(QPalette::Active ,QPalette::ButtonText, Qt::black);
	line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, Qt::black);
	TitleTab->_title_colour_fld->setPalette(line_palette);
    _title_colour = Qt::black;
	
	DisplayTab->_line_width_fld->setValue(1.);
	DisplayTab->_ticks_length_fld->setValue(QPUI_TICKS_LENGTH);
	
	DisplayTab->_line_colour_fld->setPalette(line_palette);
    _line_colour = Qt::black;
	
	DisplayTab->_show_top_axis_fld->setChecked(true);
	DisplayTab->_show_bottom_axis_fld->setChecked(true);
	DisplayTab->_show_left_axis_fld->setChecked(true);
	DisplayTab->_show_right_axis_fld->setChecked(true);
	DisplayTab->_show_top_ticks_labels_fld->setChecked(false);
	DisplayTab->_show_bottom_ticks_labels_fld->setChecked(true);
	DisplayTab->_show_left_ticks_labels_fld->setChecked(true);
	DisplayTab->_show_right_ticks_labels_fld->setChecked(false);
	DisplayTab->_show_top_labels_fld->setChecked(false);
	DisplayTab->_show_bottom_labels_fld->setChecked(true);
	DisplayTab->_show_left_labels_fld->setChecked(true);
	DisplayTab->_show_right_labels_fld->setChecked(false);
	DisplayTab->_show_xiszero_fld->setChecked(false);
	DisplayTab->_show_yiszero_fld->setChecked(false);

	xAxisTab->_axis_linlog_fld->setCurrentIndex(QP_LINEAR);
	

	xtAxisTab->_label_fld->setText("");
		
	xtAxisTab->_axis_autoscale_fld->setCurrentIndex(QP_FROM_META);
	xtAxisTab->_axis_major_ticks_num_fld->setValue(0);
	xtAxisTab->_axis_minor_ticks_num_fld->setValue(0);
	xtAxisTab->_axis_major_tickmarks_fld->setCurrentIndex(QPUI_IN);
	xtAxisTab->_axis_minor_tickmarks_fld->setCurrentIndex(QPUI_IN);
	xtAxisTab->_axis_autolabel_fld->setChecked(true);
	xtAxisTab->_label_char_size_fld->setValue(QSAS_profiles->current()->BaseFontSize+QPUI_LABELS_CHAR_SIZE_ADD);
	xtAxisTab->_label_disp_fld->setValue(QPUI_LABELS_DISP);
	xtAxisTab->_label_rotation_fld->setValue(0);
	xtAxisTab->_label_font_fld->setFontParams(PL_FCI_SANS, false, false);
	xtAxisTab->_label_colour_fld->setPalette(line_palette);
    xtAxisTab->_label_colour = Qt::black;


}

// ------------------------------------

void FrameItem::updateAxes()
{
	if(QuSlotDebug()) cout << "FrameItem::updateAxes " << endl;
	
	// Note range done first to ensure _xdo exists for other calls
	
	xtAxisTab->setAxisMinMax(); // always set axis, if user set will just ensure yobj is there
	if( xtAxisTab->_axis_autolabel_fld->isChecked()) xtAxisTab->setAutoLabelText();
	
	// with x or t axis set we must update the y axis
	PanelItem* panel;
	for(int i=0; i<this->childCount(); ++i)
	{
		panel=dynamic_cast <PanelItem*>(this->child(i));
		panel->updateAxes();
	}

	// set Title
	if( TitleTab->_autotitle_fld->isChecked())
	{
		string title = getTitle();
		TitleTab->_title_fld->clearFocus(); // ensure field doesn't try to turn autotitle off
		TitleTab->_title_fld->setText(title.c_str());
	}

}

// -------------------------------------------

void FrameItem::showTabs()
{
	// ensure all tabs are hidden
	int Ntabs = tabWidget->count();
	for(int i = Ntabs-1; i >= 0; i--) 
	{
		if(tabWidget->widget(i))
		{
			tabWidget->widget(i)->hide();
			tabWidget->removeTab(i);
		}
	}

	if (frameType() == QP_TS) 
	{
		tabWidget->addTab(tAxisTab, "T Axis");
		xtAxisTab=tAxisTab;
	}
	else 
	{
		tabWidget->addTab(xAxisTab, "X Axis");
		xtAxisTab=xAxisTab;
	}
	tabWidget->addTab(PropertiesTab, "Position");
	
	tabWidget->addTab(DisplayTab, "Frame");
	
	tabWidget->addTab(TitleTab, "Title");
	
	// value return is always shown
	tabWidget->addTab(page()->ValueReturnTab, "Value Return");
	
 	if(this->Tree()->I_LastTab == QP_POSITION_TAB)  tabWidget->setCurrentIndex(tabWidget->indexOf(PropertiesTab));
 	else if(this->Tree()->I_LastTab == QP_FRAME_TAB)  tabWidget->setCurrentIndex(tabWidget->indexOf(DisplayTab));
 	else if(this->Tree()->I_LastTab == QP_TITLE_TAB)  tabWidget->setCurrentIndex(tabWidget->indexOf(TitleTab));
 	else tabWidget->setCurrentIndex(tabWidget->indexOf(xtAxisTab));

	
}

// -------------------------------------------

int FrameItem::numberInStack()
{
	// returns number of stack sub panels in this frame
	return this->childCount();
}

// -------------------------------------------------

string FrameItem::getTitle()
{
	if ( QuSlotDebug() )
	{
		printf( "QpFrameItem::getTitle \n" );
		fflush(stdout);
	}

	string title="";
	// get frame title from panel(s)
	
	PanelItem* panelItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		panelItem=dynamic_cast<PanelItem*>(this->child(i));
		string item_title = panelItem->getTitle();
		
		if(title != item_title) 
		{
			if(title != "" && title != " " ) title += " / ";
			title += item_title;
		}
	}
	
	title = SafePGPString(title.c_str()); // truncate to be safe length for pgplot
	return title;
}

void FrameItem::updateIsotropic(bool resize)
{
	if(QuSlotDebug()) cout << "FrameItem::updateIsotropic " << resize << endl;
	if (this->restoring) resize = false;
	
	if(PropertiesTab->_aspect_ratio_fld->currentIndex()==QPUI_ISOTROPIC)
	{
		double aspectRatio = PropertiesTab->setIsotropicRatio();
		if(!resize) return;
		
		double newWidth; //= PropertiesTab->_height_fld->value()/aspectRatio;
		double newHeight; //= PropertiesTab->_width_fld->value() * aspectRatio;
		double l_offset = PropertiesTab->_loffset_fld->value();
		double t_offset = PropertiesTab->_toffset_fld->value();
		/*
		if (newWidth + l_offset > 0.99)
		{
			newWidth = 0.99 - l_offset;
			double newHeight = newWidth * aspectRatio;
            PropertiesTab->_height_fld->setValue(newHeight);
		}
		*/
		// fit to width
		newWidth = 0.98 - l_offset;
		newHeight = newWidth * aspectRatio;
		// ensure height is not too large
		if(newHeight + t_offset > 0.96){
			newHeight = 0.96 - t_offset;
			newWidth = newHeight / aspectRatio;
		}
		
		PropertiesTab->_height_fld->setValue(newHeight);
        PropertiesTab->_width_fld->setValue(newWidth);

	}
	qApp->processEvents();

}

// -------------------------------------------------

bool FrameItem::getXmetaRange(double &from, double &to)
{
    
    double xmin = DBL_MAX;
	double xmax = -DBL_MAX;

    bool rangeFound = false;
	int isLog=0;
	
	// get union of ranges of panels
	PanelItem* panelItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		panelItem=dynamic_cast<PanelItem*>(this->child(i));
		double  item_from,  item_to;
		// if any object has scalemin/max, use it
		if(panelItem->getXmetaRange( item_from,  item_to) ) 
		{
			rangeFound = true;
			xmin = min(xmin, item_from);
			xmax = max(xmax, item_to);
		}
		// get Scaletype accumulation after range to ensure objects exist
		if(isLog == 0) isLog = panelItem->getXmetaIsLog(); // only update until one plot is log
	}
	
	if(rangeFound)
	{
		from = xmin;
		to = xmax;
	}
	
	if(frameType() == QP_XY && !xAxisTab->userSetLinLog)
	{	
		if(isLog == 1 ) xAxisTab->_axis_linlog_fld->setCurrentIndex(QP_LOG);
		else  xAxisTab->_axis_linlog_fld->setCurrentIndex(QP_LINEAR);
	}
	
	return rangeFound;
}

// -------------------------------------------------

bool FrameItem::getXRange(double &from, double &to)
{
	double xmin = DBL_MAX;
	double xmax = -DBL_MAX;
	
	bool rangeFound = false;
	// get union of ranges of panels
	PanelItem* panelItem;
	
	for(int i=0; i<this->childCount(); ++i)
	{
		panelItem=dynamic_cast<PanelItem*>(this->child(i));
		
		double  item_from,  item_to;
		if(panelItem->getXRange( item_from,  item_to) )
		{
			rangeFound = true;
			xmin = min(xmin, item_from);
			xmax = max(xmax, item_to);
		}
	}
	
	from = xmin;
	to = xmax;
	return rangeFound;
}
// -------------------------------------------------

void FrameItem::getYObjects()
{
	
	// tell all panels
	PanelItem* panelItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		panelItem=dynamic_cast<PanelItem*>(this->child(i));
		panelItem->getYObjects();
	}
	
	return;
}
// ------------------------------------

void FrameItem::setPanelProperties(double loffset, double width, double toffset, double height, int num_rows, int num_cols)
{
	if(QuSlotDebug())
	{
		printf("QpFrameItem::setPanelProperties L=%g, W=%g, T=%g, H=%g\n", loffset, width, toffset, height);
		fflush(stdout);
	}

    if(PropertiesTab->_aspect_ratio_fld->currentIndex()==QPUI_ISOTROPIC ||
       PropertiesTab->_aspect_ratio_fld->currentIndex()==QPUI_SQUARE_ASPECT){
        
        double aspectRatio = PropertiesTab->_aspect_ratio;
        
        this->PropertiesTab->_loffset_fld->setValue(loffset);
        this->PropertiesTab->_toffset_fld->setValue(toffset);
        
        if(num_rows == 1) this->PropertiesTab->_width_fld->setValue(width);
        if(num_cols == 1) this->PropertiesTab->_height_fld->setValue(height); 
        if(num_rows > 1 && num_cols > 1) {
            this->PropertiesTab->_width_fld->setValue(width); 
            this->PropertiesTab->_height_fld->setValue(height);
        }
    }
    else{
        this->PropertiesTab->_loffset_fld->setValue(loffset);
        this->PropertiesTab->_width_fld->setValue(width);
        this->PropertiesTab->_toffset_fld->setValue(toffset);
        this->PropertiesTab->_height_fld->setValue(height);
    }
    
    // cascade number of cols to set colourBar width
    PanelItem* panelItem;
    for(int i=0; i<this->childCount(); ++i)
    {
        panelItem=dynamic_cast<PanelItem*>(this->child(i));
        PlotItem* plotItem;
        
        for(int j=0; j<panelItem->childCount(); ++j)
        {
            plotItem=dynamic_cast<PlotItem*>(panelItem->child(j));
            SpectrumItem * spectrumItem = dynamic_cast <SpectrumItem*>(plotItem);
            if(spectrumItem) spectrumItem->ColourscaleTab->_colourscale_colour_bar_width_fld->setValue(2.5/num_cols);
            
            TraceItem * traceItem = dynamic_cast <TraceItem*>(plotItem);
            if(traceItem) traceItem->TraceTab->_colour_bar_width->setValue(2.5/num_cols);
        }
    }
}
// ------------------------------------

bool FrameItem::containsSpectrum()
{
	PanelItem* panelItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		panelItem=dynamic_cast<PanelItem*>(this->child(i));
		PlotItem* plotItem;
		
		for(int j=0; j<panelItem->childCount(); ++j)
		{
			plotItem=dynamic_cast<PlotItem*>(panelItem->child(j));
			SpectrumItem * spectrumItem = dynamic_cast <SpectrumItem*>(plotItem);
			if(spectrumItem) return true;
			
			TraceItem * traceItem = dynamic_cast <TraceItem*>(plotItem);
			if(traceItem) {
				// X-Y-Z trace can have colour bar
				if( traceItem->TraceTab->_show_colour_bar->isChecked() )return true;
			}
		}
	}
	
	return false;
}

// ------------------------------------

QpuiReturnStatus FrameItem::plot(PlotWindow* win)
{
	// catch signal processing out opf order, ensure this is the correct window to plot in
	win->setPLFocus();
	
	// plots contents of each panel in frame
	bool hasSpectrum = page()->containsSpectrum();
	
	// count number of panels in frame
	int stackNum = numberInStack();
	int num = 0;
	double isotropic_offset=0;
	
	PanelItem* panelItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		panelItem=dynamic_cast<PanelItem*>(this->child(i));
		if(PropertiesTab->_aspect_ratio_fld->currentIndex() == QPUI_ISOTROPIC)
		{
			// special handling for isotropic panels which can have different heights
			double height  = PropertiesTab->_height_fld->value() * panelItem->yAxisTab->_axis_xy_fld->getLength() / PropertiesTab->_isotropic_y_range;
			double toffset = PropertiesTab->_toffset_fld->value() + isotropic_offset;
			isotropic_offset += height;
			double loffset = PropertiesTab->_loffset_fld->value();
			double width   = PropertiesTab->_width_fld->value();
			if(panelItem->yAxisTab->_axis_autolabel_fld->isChecked())
			{
				DvString _label="";
				panelItem->getLabel(Y_AXIS, _label);
				panelItem->yAxisTab->_label_fld->setText(_label.c_str());
			}
			panelItem->plot(height,  width, toffset, loffset, win);
		}
		else{
			double height  = PropertiesTab->_height_fld->value() / stackNum;
			double toffset = PropertiesTab->_toffset_fld->value() + num * height;
			double loffset = PropertiesTab->_loffset_fld->value();
			double width   = PropertiesTab->_width_fld->value();
			if(hasSpectrum && PropertiesTab->_aspect_ratio_fld->currentIndex()!=QPUI_ISOTROPIC) width *= 0.9;  // allow space for colour scale
			if(panelItem->yAxisTab->_axis_autolabel_fld->isChecked())
			{
				DvString _label="";
				panelItem->getLabel(Y_AXIS, _label);
				panelItem->yAxisTab->_label_fld->setText(_label.c_str());
			}
			panelItem->plot(height,  width, toffset, loffset, win);
		
		}num++;
	}

	return QPUI_SUCCESS;
}

// ---------------------------

int FrameItem::panelNum(PanelItem *panelItem)
{
	int num=0;
	PanelItem* item;
	for(int i=0; i<this->childCount(); ++i)
	{
		item=dynamic_cast<PanelItem*>(this->child(i));
		if(item == panelItem) return num;
		num++;
	}
	
	return 0; // just return first panel to be safe
}

// ---------------------------

void FrameItem::changeFrame(int frameType)
{
	// Tells panels to reset data tabs for appropriate inputs
	// NOTE: X and t axis tabs are changed by updateOnSelection()
	
	this->_frameType = (QpFrameType)frameType;
	
	// Now tell all contained plots they are now of new type TS or XY
	
	PanelItem* panelItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		panelItem=dynamic_cast<PanelItem*>(this->child(i));
		PlotItem* item;
		for(int j=0; j<panelItem->childCount(); ++j)
		{
			item=dynamic_cast<PlotItem*>(panelItem->child(j));
			item->changeFrameType(frameType);
		}
	}
	
	if (frameType == QP_TS) this->xtAxisTab=tAxisTab;
	else xtAxisTab=xAxisTab;
	
	// may need to reset axis ranges
	this->updateAxes();
}  
	
QDomElement FrameItem::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Frame_Item");
    xml.appendChild(createTextXMLNode(doc, "Name", getName()));
    xml.appendChild(createTextXMLNode(doc, "Frame_Type", _frameType));
    xml.appendChild(createTextXMLNode(doc, "Line_Colour_Red", _line_colour.red()));
    xml.appendChild(createTextXMLNode(doc, "Line_Colour_Green", _line_colour.green()));
    xml.appendChild(createTextXMLNode(doc, "Line_Colour_Blue", _line_colour.blue()));
    xml.appendChild(createTextXMLNode(doc, "Title_Colour_Red", _title_colour.red()));
    xml.appendChild(createTextXMLNode(doc, "Title_Colour_Green", _title_colour.green()));
    xml.appendChild(createTextXMLNode(doc, "Title_Colour_Blue", _title_colour.blue()));
    QpAxisTab *axisTab;
    if(_frameType == QP_TS){
        axisTab = tAxisTab;
        xml.appendChild(axisTab->toXMLElement(doc)); // do bits common to all axes
        xml.appendChild(createTextXMLNode(doc, "Time_String", axisTab->_t_data_fld->text()));
    }
    else
    {
        axisTab = xAxisTab;
        xml.appendChild(axisTab->toXMLElement(doc)); // do bits common to all axes
        xml.appendChild(createTextXMLNode(doc, "Lin_Log", axisTab->_axis_linlog_fld->currentIndex()));
    }

    xml.appendChild(PropertiesTab->toXMLElement(doc));
    xml.appendChild(DisplayTab->toXMLElement(doc));
    xml.appendChild(TitleTab->toXMLElement(doc));

    int cnt = childCount();//numberInStack();
    xml.setAttribute("size", childCount());
    
    PanelItem* panelItem;
    for(int i=0; i<cnt; ++i)
    {
        panelItem=dynamic_cast<PanelItem*>(child(i));
        xml.appendChild(panelItem->toXMLElement(doc));
    }

    return xml;
}

void FrameItem::fromXMLElement(QDomElement& el)
{
	this->restoring = true;
    QString s;
    if(getXMLNodeText(el, "Name", s))
    {
        setName(s);
    }
    if(getXMLNodeText(el, "Frame_Type", s))
    {
        _frameType=(QpFrameType)s.toInt();
    }
    int r(0), g(0), b(0);
    if(getXMLNodeText(el, "Line_Colour_Red", s))
    {
        r=s.toInt();
    }
    if(getXMLNodeText(el, "Line_Colour_Green", s))
    {
        g=s.toInt();
    }
    if(getXMLNodeText(el, "Line_Colour_Blue", s))
    {
        b=s.toInt();
    }
    _line_colour=QColor(r, g, b);
	QPalette line_palette = QPalette(QPUI_normal_palette);
    line_palette.setColor(QPalette::Active ,QPalette::ButtonText, _line_colour);
    line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, _line_colour);
    DisplayTab->_line_colour_fld->setPalette(line_palette);


    r=0;
    g=0;
    b=0;
    if(getXMLNodeText(el, "Title_Colour_Red", s))
    {
        r=s.toInt();
    }
    if(getXMLNodeText(el, "Title_Colour_Green", s))
    {
        g=s.toInt();
    }
    if(getXMLNodeText(el, "Title_Colour_Blue", s))
    {
        b=s.toInt();
    }
    _title_colour=QColor(r, g, b);
    QPalette title_palette = QPalette(QPUI_normal_palette);
    title_palette.setColor(QPalette::Active ,QPalette::ButtonText, _title_colour);
    title_palette.setColor(QPalette::Inactive, QPalette::ButtonText, _title_colour);
    TitleTab->_title_colour_fld->setPalette(title_palette);

    QDomElement axisEl=el.firstChildElement("Axis_Tab");
    if(!axisEl.isNull())
    {
        QpAxisTab *axisTab;
        
        if(_frameType == QP_TS){
            axisTab = tAxisTab;
            
            axisTab->fromXMLElement(axisEl); // do bits common to all axes

            if(getXMLNodeText(el, "Time_String", s))
            {
                tAxisTab->_t_data_fld->setText(s);
            }
            xtAxisTab=tAxisTab;
        }
        else
        {
            axisTab = xAxisTab;
            
            axisTab->fromXMLElement(axisEl); // do bits common to all axes

            if(getXMLNodeText(el, "Lin_Log", s))
            {
                xAxisTab->_axis_linlog_fld->setCurrentIndex(s.toInt());
            }
            xtAxisTab=xAxisTab;
        }
    }

    QDomElement propertiesElement=el.firstChildElement("Properties_Tab");
	if(!propertiesElement.isNull())
	{
		PropertiesTab->fromXMLElement(propertiesElement);
	}

    QDomElement displayElement=el.firstChildElement("Display_Tab");
	if(!displayElement.isNull())
	{
		DisplayTab->fromXMLElement(displayElement);
	}
    QDomElement titleElement=el.firstChildElement("Title_Tab");
	if(!titleElement.isNull())
	{
		TitleTab->fromXMLElement(titleElement);
	}

    QDomElement panelElement=el.firstChildElement("Panel_Item");
    while(!panelElement.isNull())
    {
        PanelItem *panelItem = new PanelItem("", this);
        panelItem->fromXMLElement(panelElement);

        panelElement=panelElement.nextSiblingElement("Panel_Item");
    }
	qApp->processEvents();
	this->updateAxes(); 
	qApp->processEvents();
	
	this->restoring = false;

}

