// qpui_panel.cc
//
// Plot Interface Implementation - Panel Object
//
// Janet Barnes
// July 2002
//
//  Tony Allen
//  Feb 2008
//  June 2013 (Label Panels)

#include "qpui.h"
#include "qdutil.h"
#include "qsas_plplot_time_helpers.h"

extern QPalette QPUI_normal_palette;
extern int qsas_restore_release;

//     PanelItem CLASS
//     ----------------------

PanelItem::PanelItem(const QString& name, FrameItem* parent, int position) : QpuiItem(name, parent)
{

	if(parent==NULL)
	{
		std::cerr << "PanelItem::PanelItem: creating PanelItem with NULL parent" << std::endl;
		return;
	}

		
	tabWidget=parent->tabWidget;
	I_Type = QP_PANEL;

	if(name.isEmpty())
	{
		setNextName();
	}
	
	// initialise options tab widgets
	LegendTab = new QpLegendTab(tabWidget, "LegendTab", this);
	LegendTab->hide();

	yAxisTab = new QpYAxisTab(tabWidget, "yAxisTab", this);
	yAxisTab->hide();
	
	setIcon(QP_PANEL);

}

PanelItem::PanelItem(PanelItem* panel, FrameItem* parent, int position) : QpuiItem("", parent)
{
	if(parent==NULL)
	{
		std::cerr << "Error: PanelItem::PanelItem: creating PanelItem with NULL parent" << std::endl;
		return;
	}
	setName(panel->getName());
	
	tabWidget=parent->tabWidget;
	I_Type = QP_PANEL;
	
	// initialise options tab widgets
	LegendTab = new QpLegendTab(panel->LegendTab, this);
	LegendTab->hide();
	
	yAxisTab = new QpYAxisTab(panel->yAxisTab, this);
	yAxisTab->hide();
	
	for(int i=0; i<panel->childCount(); ++i)
	{
		QTreeWidgetItem *item = panel->child(i);
		TraceItem* traceItem=dynamic_cast<TraceItem*>(item);
		EventItem* eventItem=dynamic_cast<EventItem*>(item);
		SpectrumItem* spectrumItem=dynamic_cast<SpectrumItem*>(item);
		TLabItem* labItem=dynamic_cast<TLabItem*>(item);
		
		if(traceItem!=NULL)
		{
			TraceItem* newTrace = new TraceItem(traceItem, this);
			newTrace->setDataYObj(traceItem->DataTab->_y_data_fld->text());
			newTrace->setDataXObj(traceItem->DataTab->_x_data_fld->text());
		}
		if(eventItem!=NULL)
		{
			EventItem*  newEvent = new EventItem(eventItem, this);
			newEvent->setDataYObj(eventItem->eventTab->_y_data_fld->text());
			newEvent->setDataXObj(eventItem->eventTab->_x_data_fld->text());
		}
		if(spectrumItem!=NULL)
		{
			SpectrumItem* newSpectrum = new SpectrumItem(spectrumItem, this);
			newSpectrum->setDataZObj(spectrumItem->DataTab->_z_data_fld->text());
			newSpectrum->setDataYObj(spectrumItem->DataTab->_y_start_data_fld->text());
			newSpectrum->setDataXObj(spectrumItem->DataTab->_x_start_data_fld->text());
		}
		if(labItem!=NULL)
		{
			TLabItem*  newLab = new TLabItem(labItem, this);
			newLab->TLabDataTab->_x_data_fld->setText(labItem->TLabDataTab->_x_data_fld->text());
			newLab->TLabDataTab->_x_data_fld->reValidate(true);
		}
		
	}
	setIcon(QP_PANEL);

}

PanelItem::~PanelItem()
{
	delete LegendTab;
	delete yAxisTab;
}

// ----------------------------------

void PanelItem::setDefaultOptions()
{

	yAxisTab->_axis_linlog_fld->setCurrentIndex(QP_LINEAR);
	
	yAxisTab->_label_fld->setText("");
		
	yAxisTab->handleAutoScale(QP_FROM_META);
	yAxisTab->_axis_major_ticks_num_fld->setValue(0);
	yAxisTab->_axis_minor_ticks_num_fld->setValue(0);
	yAxisTab->_axis_major_tickmarks_fld->setCurrentIndex(QPUI_IN);
	yAxisTab->_axis_minor_tickmarks_fld->setCurrentIndex(QPUI_IN);
	yAxisTab->_axis_autolabel_fld->setChecked(true);
	yAxisTab->_label_char_size_fld->setValue(QSAS_profiles->current()->BaseFontSize+QPUI_LABELS_CHAR_SIZE_ADD);
	yAxisTab->_label_disp_fld->setValue(QPUI_LABELS_DISP);
	yAxisTab->_label_rotation_fld->setValue(0);
	yAxisTab->_label_font_fld->setSansFont();
	yAxisTab->_label_font_fld->setItalic(false);
	yAxisTab->_label_font_fld->setBold(false);

	QPalette line_palette = QPalette(QPUI_normal_palette);
	line_palette.setColor(QPalette::Active ,QPalette::ButtonText, Qt::black);
	line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, Qt::black);
	yAxisTab->_label_colour_fld->setPalette(line_palette);
    yAxisTab->_label_colour = Qt::black;
	
	LegendTab->_show_legend_fld->setChecked(false);
	LegendTab->_legend_border_fld->setCurrentIndex(QPUI_LINE);
	LegendTab->_legend_background_fld->setCurrentIndex(QPUI_TRANSPARENT);
	LegendTab->_legend_num_rows_fld->setValue(1);
	LegendTab->_legend_num_cols_fld->setValue(1);
	LegendTab->_legend_width_fld->setValue(QPUI_LEGEND_WIDTH);
	LegendTab->_legend_height_fld->setValue(QPUI_LEGEND_HEIGHT);
	LegendTab->_legend_max_char_length_fld->setValue(15);
	LegendTab->_legend_line_length_fld->setValue(0.2);
	LegendTab->_legend_char_size_fld->setValue(0.5);
	LegendTab->_legend_just_fld->setCurrentIndex(QPUI_VCENTRE);
	LegendTab->_legend_font_fld->setSansFont();
	LegendTab->_legend_font_fld->setItalic(false);
	LegendTab->_legend_font_fld->setBold(false);
	LegendTab->_legend_align_fld->setCurrentIndex(QPUI_INSIDE_TOP_RIGHT);
	LegendTab->_legend_loffset_fld->setValue(0.75);
	LegendTab->_legend_toffset_fld->setValue(0.);
}

// -------------------------------------------

void PanelItem::showTabs()
{	

	// ensure all tabs are hidden
	int Ntabs = tabWidget->count();
	for(int i = Ntabs-1; i >= 0; i--) 
	{
		if(tabWidget->widget(i))
		{
			tabWidget->widget(i)->hide();
			tabWidget->removeTab(i);
		}
	}

	if(!isBlank()){
		tabWidget->addTab(yAxisTab, "Y Axis");
		tabWidget->addTab(LegendTab, "Legend");
	}	
		
	// value return is always shown
	tabWidget->addTab(page()->ValueReturnTab, "Value Return");
	if( isBlank() ) tabWidget->setCurrentIndex(tabWidget->indexOf(page()->ValueReturnTab));
 	else if(this->Tree()->I_LastTab == QP_LEGEND_TAB)  tabWidget->setCurrentIndex(tabWidget->indexOf(LegendTab));
 	else tabWidget->setCurrentIndex(tabWidget->indexOf(yAxisTab));
}

// -------------------------------------------

int PanelItem::getStackNumber()
{
	// get position in frame stack, 1 = first
	
	return childNumber()+1;

}

// -------------------------------------------

int PanelItem::numberTraces()
{
	// get number of plot items that are traces in this panel
	
	int cnt = 0;
	TraceItem* item;
	for(int i=0; i<this->childCount(); ++i)
	{
		item=dynamic_cast<TraceItem*>(this->child(i));
		if(item && item->DataTab->_show_legend_fld->isChecked()) cnt++;
	}
	return cnt;

}

// -------------------------------------------

int PanelItem::numberSpectra()
{
	// get number of plot items that are traces in this panel
	
	int cnt = 0;
	SpectrumItem* item;
	for(int i=0; i<this->childCount(); ++i)
	{
		item=dynamic_cast<SpectrumItem*>(this->child(i));
		if(item) cnt++;
	}
	return cnt;

}

int PanelItem::numberThisSpectrum(SpectrumItem *thisItem)
{
	// get number of plot items that are traces in this panel
	
	SpectrumItem* item;
	for(int i=0; i<this->childCount(); ++i)
	{
		item=dynamic_cast<SpectrumItem*>(this->child(i));
		if(item == thisItem) return i;
	}
	return 0;

}

bool PanelItem::isInterlaced()
{
	SpectrumItem* item;
	for(int i=0; i<this->childCount(); ++i)
	{
		item=dynamic_cast<SpectrumItem*>(this->child(i));
		if(item && item->DataTab->interlace_toggle->isChecked()) return true;
	}
	return false;

}

DvString PanelItem::getTitle()
{
	if(QuSlotDebug())
	{
		printf("QpPanelItem::getTitle\n");
		fflush(stdout);
	}
	qApp->processEvents();
	
	DvString title="";
	// get panel title from plot elements
	
	for(int i=0; i<this->childCount(); ++i)
	{
		QTreeWidgetItem *item = this->child(i);
		DvString item_title;
		TraceItem* traceItem = dynamic_cast<TraceItem*>(item);
		SpectrumItem* spectrumItem = dynamic_cast<SpectrumItem*>(item);      

		if(traceItem) item_title = traceItem->getTitle();
		else if(spectrumItem) item_title = spectrumItem->getTitle();
		
		if( title != item_title) 
		{
			if( title == "" ) title = item_title;
			else {
				DvString stem = title.common(item_title);
				if( stem.endsIn(" ") ) stem = stem.beforeR(' ');
				if( stem.endsIn("_") ) stem = stem.beforeR('_');
				if( stem.endsIn("[") ) stem = stem.beforeR('[');
				if( stem != "" ) {
					title = stem;
				}
				else{
					title += " / ";
					title += item_title;
				}
			}
		}
	}

	title = SafePGPString(title.c_str());
	return title;
}

// ---------------------------------------------

void PanelItem::prePlot()
{
	// sets panel parameters (i.e. called BEFORE plot component is plotted)
		
	plwid(1); // force reset

	PLFLT xmin, xmax;
	if(frameType() == QP_TS)
	{
		xmin = 0.0; // scale window to range
		xmax = frame()->tAxisTab->_axis_time_fld->getDuration();
	}
	else
	{
		xmin = frame()->xAxisTab->_axis_xy_fld->getStart();
		xmax = frame()->xAxisTab->_axis_xy_fld->getEnd();
	}
	
	if(frame()->xtAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
	{
		xmin = SafeLog10(xmin);
		xmax = SafeLog10(xmax);
	}
	
	PLFLT ymin = yAxisTab->_axis_xy_fld->getStart();
	PLFLT ymax = yAxisTab->_axis_xy_fld->getEnd();
	if(yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
	{
		ymin = (PLFLT)SafeLog10(ymin);
		ymax = (PLFLT)SafeLog10(ymax);
	}
	
	plwind(xmin, xmax, ymin, ymax);
	
	if(LegendTab->_show_legend_fld->isChecked())
	{
		// calculate width and height of legend entry
// 		string legend_text = SafePGPString(yAxisTab->_label_fld->text().toLatin1().data());
		plschr(QPUI_CHAR_HEIGHT, LegendTab->_legend_char_size_fld->value());
// 		PLFLT xch, ych;
// 		plgchr(&xch, &ych);
		_legend_entry = 0;
// 		PLFLT loffset, roffset, toffset, boffset;
// 		PLFLT x0vp_save, x1vp_save, y0vp_save, y1vp_save;
// 		plgvpd(&x0vp_save, &x1vp_save, &y0vp_save, &y1vp_save);

	}

	plschr(QPUI_CHAR_HEIGHT, frame()->DisplayTab->_ticks_length_fld->value());
}

// ------------------------------------------------

void PanelItem::postPlot(bool show_top_axis, bool show_bottom_axis, bool show_title, bool show_T_lab)
{
	PLINT family, style, weight;
	
	double ymin = yAxisTab->_axis_xy_fld->getStart();
	double ymax = yAxisTab->_axis_xy_fld->getEnd();
// 	if(ymin >= ymax) return;
	
	// draw frame

		// set colours, font, character size, line style and line width
	PLINT  line_colour_index = getPGPColourIndex(frame()->_line_colour);
	plcol0(line_colour_index);
	PLFLT _ticks_length = frame()->DisplayTab->_ticks_length_fld->value();
	PLFLT _ticks_label_size = frame()->DisplayTab->_ticks_char_size_fld->value();
// 		plschr(40., _ticks_length);
	pllsty(QP_SOLID);
	plwid(QtWidth2PlWidth(frame()->DisplayTab->_line_width_fld->value()));
		
	string xopt, yopt;
	string x_subopt, y_subopt;
	if(frame()->DisplayTab->_show_bottom_axis_fld->isChecked())
	{
		xopt+="b";
		x_subopt+="b";
	}
	if(frame()->DisplayTab->_show_left_axis_fld->isChecked())
	{
		yopt+="b";
		y_subopt+="b";
	}
	if(frame()->DisplayTab->_show_top_axis_fld->isChecked())
	{
		xopt+="c";
		x_subopt+="c";
	}
	if(frame()->DisplayTab->_show_right_axis_fld->isChecked())
	{
		yopt+="c";
		y_subopt+="c";
	}
	if(frame()->DisplayTab->_show_lr_alternating_axis_fld->isChecked())
	{
		if(getStackNumber()%2==1)
		{
			yopt+="b";
			y_subopt+="b";
		}
		else
		{
			yopt+="c";
			y_subopt+="c";
		}
	}
	if(frame()->DisplayTab->_show_rl_alternating_axis_fld->isChecked())
	{
		if(getStackNumber()%2==0)
		{
			yopt+="b";
			y_subopt+="b";
		}
		else
		{
			yopt+="c";
			y_subopt+="c";
		}
	}
				
	if(frame()->DisplayTab->_show_bottom_axis_fld->isChecked() || frame()->DisplayTab->_show_top_axis_fld->isChecked() || frame()->DisplayTab->_show_yiszero_fld->isChecked() )
	{
		frame()->xtAxisTab->_label_font_fld->getFontParams(family, style, weight);
		plsfont(family, style, weight);

		switch(frame()->xtAxisTab->_axis_minor_tickmarks_fld->currentIndex())
		{
			case QPUI_IN:
				x_subopt+="st";
				break;
			case QPUI_OUT: default:
				x_subopt+="ist";
				break;
			case QPUI_NOTICKS:
				break;
		}
		switch(frame()->xtAxisTab->_axis_major_tickmarks_fld->currentIndex())
		{
			case QPUI_IN: default:
				xopt += "t";
				break;
			case QPUI_OUT:
				xopt += "it";
				break;
			case QPUI_GRID:
				xopt+="gt";
				break;
			case QPUI_NOTICKS:
				break;
		}

		if(show_bottom_axis && frame()->DisplayTab->_show_bottom_ticks_labels_fld->isChecked()) xopt+="n";
			
		if(show_top_axis && frame()->DisplayTab->_show_top_ticks_labels_fld->isChecked()) xopt+="m";
			
		PLFLT xmin, xmax;
		PLINT nsubx = frame()->xtAxisTab->_axis_minor_ticks_num_fld->value();
		PLFLT xtick = 0.;
		int nx=frame()->xtAxisTab->_axis_major_ticks_num_fld->value();
		char fmt_ticks[50], fmt_label[50];

        PLLabelDefaults x_label_opt; // do nothing if QP_TS. Else used to move exponential label
        
		if(frameType() == QP_TS)
		{
			PLINT auto_nsubx;
			PLFLT auto_xtick;
			char labelS[50];	
			char labelE[50];	
			DvTime start=frame()->tAxisTab->_axis_time_fld->getStart();
			DvTime end=frame()->tAxisTab->_axis_time_fld->getEnd();
// 			struct tm tm_start, tm_end;
				
			xopt += "d";
			x_subopt += "d";
				
// 			start=*(frame()->tAxisTab->_axis_time_fld->getStartTime());
// 			end=*(frame()->tAxisTab->_axis_time_fld->getEndTime());

// 			getXminXmaxTS(start, end, &xmin, &xmax, &tm_start, &tm_end);
			xmin = start.getEpoch1970();
			xmax = end.getEpoch1970();
				
			int separator=(QSAS_profiles->current()->TimeRep=="Colon"?1:0);
			computeFormat(fabs(xmax-xmin), fmt_ticks, fmt_label, &auto_xtick, &auto_nsubx, separator);
				
			ymin = yAxisTab->_axis_xy_fld->getStart();
			ymax = yAxisTab->_axis_xy_fld->getEnd();
			if(yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
			{
				ymin = SafeLog10(ymin);
				ymax = (PLFLT)SafeLog10(ymax);
			}

			plwind(xmin, xmax, (PLFLT)ymin, (PLFLT)ymax);
				
			if(nx!=0)
			{
				xtick=(xmax-xmin)/(PLFLT)nx;
			}
			else
			{
				xtick=auto_xtick;
			}
				
			if(nsubx==0) nsubx=auto_nsubx;
				
			if(frame()->tAxisTab->tick_time_fmt_auto_fld->isChecked())
			{
				pltimefmt(fmt_ticks);
				frame()->tAxisTab->tick_time_fmt_fld->setText(QString(fmt_ticks));
			}
			else
			{
				pltimefmt(frame()->tAxisTab->tick_time_fmt_fld->text().toLatin1().data());
			}
				
			// print labels
			bool showStart = true;
			bool showEnd = true;
			DvString format;
			
			frame()->tAxisTab->_label_font_fld->getFontParams(family, style, weight);
			plsfont(family, style, weight);

			plschr(0., _ticks_label_size);
				
			if(frame()->tAxisTab->time_labels_fmt_auto_fld->isChecked())
			{
				format = fmt_label;
				start.strfMJD(labelS, 49, fmt_label);
				end.strfMJD(labelE, 49, fmt_label);
				frame()->tAxisTab->time_labels_fmt_fld->setText(QString(fmt_label));
			}
			else
			{
				format = frame()->tAxisTab->time_labels_fmt_fld->text().toLatin1().data();
				start.strfMJD(labelS, 49, format.c_str());
				end.strfMJD(labelE, 49, format.c_str());
			}
			
			this->handleBoundary(format, start, end, showStart, showEnd);

			if(show_bottom_axis && frame()->DisplayTab->_show_bottom_ticks_labels_fld->isChecked()){
				if(showStart) Q_plmtex("b",3.1,0.0,0.3,labelS);
				if(showEnd) Q_plmtex("b",3.1,1.0,0.7,labelE);
				
			}
			if(show_top_axis && frame()->DisplayTab->_show_top_ticks_labels_fld->isChecked()){
				if(showStart) Q_plmtex("t",3.1,0.0,0.3,labelS);
				if(showEnd) Q_plmtex("t",3.1,1.0,0.7,labelE);
			}
		}
		else
		{
            x_label_opt.exp_label_disp=3.2+frame()->xAxisTab->_exp_label_offset_y_fld->value();
            x_label_opt.exp_label_pos=1.02+frame()->xAxisTab->_exp_label_offset_x_fld->value()/5.;
            x_label_opt.exp_label_just=.5;
            
			xmin = frame()->xAxisTab->_axis_xy_fld->getStart();
			xmax = frame()->xAxisTab->_axis_xy_fld->getEnd();
			if(nx>0)
			{
				xtick=(xmax-xmin)/(PLFLT)nx;
			}
		}
			
		if(frame()->xtAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG){
			xopt+="l";
			x_subopt+="l";
			xmin = SafeLog10(xmin);
			xmax = SafeLog10(xmax);
		}

        if(frameType() == QP_XY)
        {
            plslabelfunc(NULL, &x_label_opt);
        }
		// fix for plplot putting grid and major ticks in different place (suppress major tick)		
		plsmaj(0., _ticks_length);
		plschr(0., _ticks_label_size);
		plbox(xopt.c_str(), xtick, nsubx, "", 0, 0);
		plsmaj(0., 0.);
		plsmin(0., _ticks_length*0.6);
		plbox(x_subopt.c_str(), xtick, nsubx, "", 0, 0);
        plslabelfunc(NULL, NULL);
			
		// draw y=0
		if(frame()->DisplayTab->_show_yiszero_fld->isChecked() )
		{
			PLFLT yzero_x[2] = {xmin, xmax};
			PLFLT yzero_y[2]={0., 0.};
			plline(2,yzero_x,yzero_y);
		}
	}

	yAxisTab->_label_font_fld->getFontParams(family, style, weight);
	plsfont(family, style, weight);


	switch(yAxisTab->_axis_minor_tickmarks_fld->currentIndex())
	{
		case QPUI_IN:
		default:
			y_subopt += "st";
			break;
		case QPUI_OUT:
			y_subopt += "ist";
			break;
		case QPUI_NOTICKS:
			break;
	}
		
	if(yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
	{
		yopt += "l";
		y_subopt+="l";
		ymin = SafeLog10(ymin);
		ymax = SafeLog10(ymax);
	}
		
	switch(yAxisTab->_axis_major_tickmarks_fld->currentIndex())
	{
		case QPUI_IN: default:
			yopt += "t";
			break;
		case QPUI_OUT:
			yopt += "it";
			break;
		case QPUI_GRID:
			yopt += "gt";
			break;
		case QPUI_NOTICKS:
			break;
	}

    bool drawLabelsLeftAxis=false;
    bool drawLabelsRightAxis=false;
    
	if(frame()->DisplayTab->_show_left_ticks_labels_fld->isChecked())
	{
		yopt += "nv";
        drawLabelsLeftAxis=true;
	}
	if(frame()->DisplayTab->_show_right_ticks_labels_fld->isChecked())
	{
		yopt += "mv";
        drawLabelsRightAxis=true;
	}
	if(frame()->DisplayTab->_show_lr_alternating_ticks_labels_fld->isChecked())
	{
		if(getStackNumber()%2==1)
		{
			yopt += "nv";
            drawLabelsLeftAxis=true;
		}
		else
		{
			yopt += "mv";
            drawLabelsRightAxis=true;
		}
	}
	if(frame()->DisplayTab->_show_rl_alternating_ticks_labels_fld->isChecked())
	{
		if(getStackNumber()%2==0)
		{
			yopt += "nv";
            drawLabelsLeftAxis=true;
		}
		else
		{
			yopt += "mv";
            drawLabelsRightAxis=true;
		}
	}
		
	PLINT nsuby = yAxisTab->_axis_minor_ticks_num_fld->value();
	int ny=yAxisTab->_axis_major_ticks_num_fld->value();
	PLFLT ytick;
	if(ny!=0)
	{
		ytick=(ymax-ymin)/(PLFLT)ny;
	}
	else ytick=0.;

    if(drawLabelsRightAxis && drawLabelsLeftAxis)
    {
        // Draw axes in 2 times because the exponential label offsets differ for each axis
        std::string bkp=yopt;
        size_t position;
        PLLabelDefaults y_label_opt;
        y_label_opt.exp_label_disp = yAxisTab->_exp_label_offset_y_fld->value();
        // draw left axis
        // remove "mv" commands
        do
        {
            position=yopt.find("mv");
            if(position!=std::string::npos) yopt.replace(position, 2, "");
        }
        while(position!=std::string::npos);
		
        y_label_opt.exp_label_pos=-0.02-yAxisTab->_exp_label_offset_x_fld->value()/5.;
        y_label_opt.exp_label_just=1.;
        plslabelfunc(NULL, &y_label_opt);
		// fix for plplot putting grid and major ticks in different place (suppress major tick)
        plsmaj(0., _ticks_length);
        plschr(0., _ticks_label_size);
        plbox("", 0., 0, yopt.c_str(), ytick, nsuby);
        plsmaj(0., 0.);
        plsmin(0., _ticks_length*0.6);
        plbox("", 0., 0, y_subopt.c_str(), ytick, nsuby);
        plsmaj(0., _ticks_length);
        plslabelfunc(NULL, NULL);

        yopt=bkp;
        // draw right axis
        // remove "nv" commands
        do
        {
            position=yopt.find("nv");
            if(position!=std::string::npos) yopt.replace(position, 2, "");
        }
        while(position!=std::string::npos);
        y_label_opt.exp_label_pos=1.02+yAxisTab->_exp_label_offset_x_fld->value()/5.;
        y_label_opt.exp_label_just=0.;
        plslabelfunc(NULL, &y_label_opt);
        plsmaj(0., _ticks_length);
        plschr(0., _ticks_label_size);
        plbox("", 0., 0, yopt.c_str(), ytick, nsuby);
        plsmaj(0., 0.);
        plsmin(0., _ticks_length*0.6);
        plbox("", 0., 0, y_subopt.c_str(), ytick, nsuby);
        plsmaj(0., _ticks_length);
        plslabelfunc(NULL, NULL);
    }
    else
    {
        PLLabelDefaults y_label_opt;
        y_label_opt.exp_label_disp = yAxisTab->_exp_label_offset_y_fld->value();
        if(drawLabelsLeftAxis)
        {
            y_label_opt.exp_label_pos=-0.02+yAxisTab->_exp_label_offset_x_fld->value()/5.;
            y_label_opt.exp_label_just=1.;
        }
        else
        {
            y_label_opt.exp_label_pos=1.02+yAxisTab->_exp_label_offset_x_fld->value()/5.;
            y_label_opt.exp_label_just=0.;
        }
        plslabelfunc(NULL, &y_label_opt);
        plsmaj(0., _ticks_length);
        plschr(0., _ticks_label_size);
        plbox("", 0., 0, yopt.c_str(), ytick, nsuby);
        plsmaj(0., 0.);
        plsmin(0., _ticks_length*0.6);
        plbox("", 0., 0, y_subopt.c_str(), ytick, nsuby);
        plsmaj(0., _ticks_length);
        plslabelfunc(NULL, NULL);
    }
		
	//draw x=0
	if(frame()->DisplayTab->_show_xiszero_fld->isChecked())
	{
		PLFLT xzero_x[2] = {0.,0.};
		PLFLT xzero_y[2] = {ymin, ymax};
		plline(2, xzero_x, xzero_y);
	}
		
	// draw labels - text line width needs to be adjustable
	int _text_line_width = 1;
	plwid(QtWidth2PlWidth(_text_line_width));
				
	PLFLT x0win_save, x1win_save, y0win_save, y1win_save;
	PLFLT x0vp_save, x1vp_save, y0vp_save, y1vp_save;
			
	plgvpw(&x0win_save, &x1win_save, &y0win_save, &y1win_save);
	plgvpd(&x0vp_save, &x1vp_save, &y0vp_save, &y1vp_save);
				
	float xch_ndc = yAxisTab->_label_char_size_fld->value()/40.;
	float ych_ndc = frame()->xtAxisTab->_label_char_size_fld->value()/40.;
			
	plvpor(0., 1., 0., 1.);
	plwind(0., 1., 0., 1.);
				
	if(show_top_axis || show_bottom_axis)
	{
		frame()->xtAxisTab->_label_font_fld->getFontParams(family, style, weight);
		plsfont(family, style, weight);

		plschr(QPUI_CHAR_HEIGHT, frame()->xtAxisTab->_label_char_size_fld->value());
		double p_def, p_ht;
		plgchr (&p_def, &p_ht);	

		PLINT labels_colour_index = getPGPColourIndex(frame()->xtAxisTab->_label_colour);
		plcol0(labels_colour_index);
			
		if(frame()->DisplayTab->_show_top_labels_fld->isChecked() && show_top_axis)
		{
			PLFLT dx, dy;
			dx=cos(frame()->xtAxisTab->_label_rotation_fld->value()*M_PI/180.);
			dy=sin(frame()->xtAxisTab->_label_rotation_fld->value()*M_PI/180.);
			plptex((x0vp_save+x1vp_save)/2.,
					y1vp_save+getCharHeight()*(1.+frame()->xtAxisTab->_label_disp_fld->value()),	
					dx, dy, 0.5,SafePGPString(frame()->xtAxisTab->_label_fld->text().toLatin1().data()).c_str());
		}
		
		if(frame()->DisplayTab->_show_bottom_labels_fld->isChecked()  && show_T_lab)
		{
			PLFLT dx, dy;
			dx=cos(frame()->xtAxisTab->_label_rotation_fld->value()*M_PI/180.);
			dy=sin(frame()->xtAxisTab->_label_rotation_fld->value()*M_PI/180.);
			plptex((x0vp_save+x1vp_save)/2.,
					y0vp_save-getCharHeight()*(1.+frame()->xtAxisTab->_label_disp_fld->value()),
					 dx, dy
					,0.5,
					SafePGPString(frame()->xtAxisTab->_label_fld->text().toLatin1().data()).c_str());
		}
	}
				
	if(frame()->DisplayTab->_show_right_labels_fld->isChecked() || frame()->DisplayTab->_show_left_labels_fld->isChecked() || frame()->DisplayTab->_show_lr_alternating_labels_fld->isChecked() || frame()->DisplayTab->_show_rl_alternating_labels_fld->isChecked())
	{
		float y_label_just = 13./(1. + yAxisTab->_label_char_size_fld->value() * yAxisTab->_label_fld->text().length());
		if(y_label_just > 1.) y_label_just = 1.;
		
		PLINT family, style, weight;

		yAxisTab->_label_font_fld->getFontParams(family, style, weight);
		plsfont(family, style, weight);

		plschr(QPUI_CHAR_HEIGHT, yAxisTab->_label_char_size_fld->value());
				
		PLINT labels_colour_index = getPGPColourIndex(yAxisTab->_label_colour);
		plcol0(labels_colour_index);
				
		if(frame()->DisplayTab->_show_right_labels_fld->isChecked() ||
		(frame()->DisplayTab->_show_lr_alternating_labels_fld->isChecked() && getStackNumber()%2==0) ||
		(frame()->DisplayTab->_show_rl_alternating_labels_fld->isChecked() && getStackNumber()%2==1))
		{
			PLFLT dx, dy;
			dx=cos(yAxisTab->_label_rotation_fld->value()*M_PI/180.);
			dy=sin(yAxisTab->_label_rotation_fld->value()*M_PI/180.);
			plptex(x1vp_save+getCharHeight()*(1.+yAxisTab->_label_disp_fld->value()),
					(y0vp_save+y1vp_save)/2.,
					dx, dy,
					0.5,
					SafePGPString(yAxisTab->_label_fld->text().toLatin1().data()).c_str());
		}
		if(frame()->DisplayTab->_show_left_labels_fld->isChecked() ||
		(frame()->DisplayTab->_show_lr_alternating_labels_fld->isChecked() && getStackNumber()%2==1) ||
		(frame()->DisplayTab->_show_rl_alternating_labels_fld->isChecked() && getStackNumber()%2==0))
		{
			PLFLT dx, dy;
			dx=cos(yAxisTab->_label_rotation_fld->value()*M_PI/180.);
			dy=sin(yAxisTab->_label_rotation_fld->value()*M_PI/180.);
			plptex(x0vp_save-getCharHeight()*(1.+yAxisTab->_label_disp_fld->value()),
					(y0vp_save+y1vp_save)/2.,
					dx, dy,
					0.5,
				SafePGPString(yAxisTab->_label_fld->text().toLatin1().data()).c_str());
		}
	}
				
	plvpor(x0vp_save, x1vp_save, y0vp_save, y1vp_save);
	plwind(x0win_save, x1win_save, y0win_save, y1win_save);
			
	if(show_title)
	{
		frame()->TitleTab->_title_font_fld->getFontParams(family, style, weight);
		plsfont(family, style, weight);

		plschr(QPUI_CHAR_HEIGHT, frame()->TitleTab->_title_char_size_fld->value());
		PLINT title_colour_index = getPGPColourIndex(frame()->_title_colour);
		plcol0(title_colour_index);
		PLFLT fjust;
		PLFLT coord;
		switch(frame()->TitleTab->_title_just_fld->currentIndex())
		{
			case QPUI_HCENTRE:
			default:
				coord = 0.5;
				fjust = 0.5;
				break;
			case QPUI_LEFT:
				coord = 0.0;
				fjust = 0.0;
				break;
			case QPUI_RIGHT:
				coord = 1.0;
				fjust = 1.0;
				break;
		}
				
		Q_plmtex("t", frame()->TitleTab->_title_disp_fld->value(), coord, fjust, SafePGPString(frame()->TitleTab->_title_fld->text().toLatin1().data()).c_str());
				
		plschr(QPUI_CHAR_HEIGHT, frame()->TitleTab->_subtitle_char_size_fld->value());
				
		Q_plmtex("t", frame()->TitleTab->_subtitle_disp_fld->value(), coord, fjust, SafePGPString(frame()->TitleTab->_subtitle_fld->text().toLatin1().data()).c_str());
	}
	
	// draw legend if show_legend set
	if(LegendTab->_show_legend_fld->isChecked() && LegendTab->_legend_border_fld->currentIndex()  != QPUI_NO_BORDER && numberTraces() > 0)
	{
		PLFLT x0win_save, x1win_save, y0win_save, y1win_save;
		PLFLT x0vp_save, x1vp_save, y0vp_save, y1vp_save;
	
		plgvpw(&x0win_save, &x1win_save, &y0win_save, &y1win_save);
		plgvpd(&x0vp_save, &x1vp_save, &y0vp_save, &y1vp_save);
	
	//         QPlot_window->save();
	
		if(LegendTab->_legend_border_fld->currentIndex() == QPUI_LINE) pllsty(QP_SOLID);
		else pllsty(QP_NOLINE);
			
		plschr(QPUI_CHAR_HEIGHT, LegendTab->_legend_char_size_fld->value());
		PLFLT xch, ych;
		plgchr(&xch, &ych);
		PLFLT x0_vp = x0vp_save + LegendTab->_legend_loffset_fld->value() * (x1vp_save - x0vp_save);
		PLFLT x1_vp = x0_vp + LegendTab->_legend_width_fld->value() * (x1vp_save - x0vp_save) ;
		PLFLT y1_vp = y1vp_save - LegendTab->_legend_toffset_fld->value() * (y1vp_save - y0vp_save);
		PLFLT y0_vp = y1_vp - LegendTab->_legend_height_fld->value() * (y1vp_save - y0vp_save);
		if(x1_vp > 0.999) x1_vp = 0.999;

		plvpor(x0_vp, x1_vp, y0_vp, y1_vp);
		plwind(0., 1., 0., 1.);

		plwid(QtWidth2PlWidth(0.6*frame()->DisplayTab->_line_width_fld->value()));
		plbox("bc", 0., 0, "bc", 0., 0);
	
		plscol0(0, 255, 255, 255);
		plscol0(1, 0, 0, 0);
		plcol0(1);
		//         QPlot_window->unsave();
		plvpor(x0vp_save, x1vp_save, y0vp_save, y1vp_save);
		plwind(x0win_save, x1win_save, y0win_save, y1win_save);
	}
}

// ------------------------------------

void PanelItem::handleBoundary(DvString format, DvTime start, DvTime end, bool &showStart, bool &showEnd)
{
	showStart = true;
	showEnd = false;

	if(end < start) return handleBoundary(format, end, start, showEnd, showStart);
	
	// finds if time axis clicks over boundary so time label needed both ends
	if(QuSlotDebug()) cout << "QpPanelItem::handleBoundary " << endl;
	
	// find length of axis
	double length = end - start;
	
	// find start components
	int year, month, day, hour, min;
	double second;
	
	start.breakDownMJD(&year, &month, &day, &hour, &min, &second);
	
	// locate last unit in format string
	format = format.afterR('%');
	
	// if click over is near start leave off start
	// else if click over is near end leave off end
	
	if(format[0] == 'Y'){
		// Year,  next unit is month
		
		int startYear = year;
		end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
		
		if( year != startYear){
			
			// clickover too near start?
			end = start + 0.1 * length;
			end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
			
			if(year != startYear) {
				showStart = false;
				showEnd = true;
				return;
			}
			
			// clickover too near end?
			end = start + 0.9 * length;
			end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
			
			if(year == startYear) {
				showEnd = false;
				showStart = true;
			}
			else{
				showEnd = true;
				showStart = true;
			}
			
			return;
		}
	}
	else if(format[0] == 'm'){
		// month,  next unit is day
		
		int startMonth = month;
		end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
		
		if( month != startMonth){
			
			// clickover too near start?
			end = start + 0.1 * length;
			end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
			
			if(month != startMonth) { // year may click over too, then month decreases
				showStart = false;
				showEnd = true;
				return;
			}
			
			// clickover too near end?
			end = start + 0.9 * length;
			end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
			
			if(month == startMonth) {
				showEnd = false;
				showStart = true;
			}
			else{
				showEnd = true;
				showStart = true;
			}
			
			return;
		}
	}
	else if(format[0] == 'd'){
		// day,  next unit is hour
		
		int startDay = day;
		end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
		
		if( day != startDay){
			
			// clickover too near start?
			end = start + 0.1 * length;
			end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
			
			if(day != startDay) {
				showStart = false;
				showEnd = true;
				return;
			}
			
			// clickover too near end?
			end = start + 0.9 * length;
			end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
			
			if(day == startDay) {
				showEnd = false;
				showStart = true;
			}
			else{
				showEnd = true;
				showStart = true;
			}
			
			return;
		}
	}
	else if(format[0] == 'H'){
		// Hour,  next unit is min
		int startHour = hour;
		end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
		
		if( hour != startHour){
			
			// clickover too near start?
			end = start + 0.1 * length;
			end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
			
			if(hour != startHour) {
				showStart = false;
				showEnd = true;
				return;
			}
			
			// clickover too near end?
			end = start + 0.9 * length;
			end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
			
			if(hour == startHour) {
				showEnd = false;
				showStart = true;
			}
			else{
				showEnd = true;
				showStart = true;
			}
			
			return;
		}
	}
	else if(format[0] == 'M'){
		// Minute,  next unit is second
		int startMin = min;
		end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
		
		if( min != startMin){
			
			// clickover too near start?
			end = start + 0.1 * length;
			end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
			
			if(min != startMin) {
				showStart = false;
				showEnd = true;
				return;
			}
			
			// clickover too near end?
			end = start + 0.9 * length;
			end.breakDownMJD(&year, &month, &day, &hour, &min, &second);
			
			if(min == startMin) {
				showEnd = false;
				showStart = true;
			}
			else{
				showEnd = true;
				showStart = true;
			}
			
			return;
		}
	}
	
}

// ------------------------------------

void PanelItem::updateAxes()
{
	if(QuSlotDebug()) cout << "QpPanelItem::updateAxes " << endl;
	// this is called from QpFrame::updateAxes to ensure x or t axis set first
	// Note range done first to ensure _ydo exists for other calls
	
	if( yAxisTab->_axis_autoscale_fld->currentIndex() != QP_FROM_USER)  yAxisTab->setAxisMinMax();

	if( yAxisTab->_axis_autolabel_fld->isChecked())  yAxisTab->setAutoLabelText();
	
	LegendTab->updateNumTraces();

}

// ---------------------------------------

void PanelItem::getLabel(const QpAxes& axis, DvString& label)
{
	// get label for axis

	label = "-";
	DvString itemLabel;
	
	for(int i=0; i<this->childCount(); i++)
	{
		QTreeWidgetItem *item = this->child(i);
		TraceItem* traceItem=dynamic_cast<TraceItem*>(item);
		SpectrumItem* spectrumItem=dynamic_cast<SpectrumItem*>(item);

		// ignore const trace types
		if(traceItem && traceItem->I_Type == QP_TRACE)
		{
			traceItem->getLabel(axis, itemLabel);
			if(label == "-") label = itemLabel;
			else{
				label = label.common(itemLabel);
				if( label.endsIn("_") ) label = label.beforeR('_');
			}
		}
		else if(spectrumItem)
		{
			// only allowed one spectrum in a panel
			spectrumItem->getLabel(axis, label);
			break;
		}
	}

	label = label.replace("theta","#gh");
	label = label.replace("phi","#gf");

}

// ---------------------------------------

int PanelItem::getXmetaIsLog()
{	
	int isLog=0;
	
	PlotItem* plotItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->child(i));
		if(isLog == 0 && plotItem) isLog = plotItem->getXmetaIsLog(); // update until log is set 
	}
	return isLog;
}

// ---------------------------------------

bool PanelItem::getXmetaRange(double& panel_xmin, double& panel_xmax)
{
	// get range of values for x axis

    double xmin = DBL_MAX;
	double xmax = -DBL_MAX;
	
	
	double item_xmin, item_xmax;
	bool rangeFound=false;
	
	// descend thru' panel hierarchy
	
	PlotItem* plotItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->child(i));
		if(plotItem) plotItem->get_dobjs();

        if(plotItem && plotItem->getXmetaRange(item_xmin, item_xmax))
		{
			xmin = min(xmin, item_xmin);
			xmax = max(xmax, item_xmax);
			rangeFound = true;
		}
	}
	
	if(rangeFound)
	{
		panel_xmin = xmin;
		panel_xmax = xmax;
	}

	return rangeFound;
}


// ---------------------------------------

void PanelItem::getYObjects()
{
	// Get data objects so get Y range has them.
	// Called when XT axis is user set.
	
	// descend thru' panel hierarchy
	
	PlotItem* plotItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->child(i));
		if(plotItem) plotItem->get_dobjs();
	}
	
	return;
}

// ---------------------------------------

bool PanelItem::getXRange(double& panel_xmin, double& panel_xmax)
{
	// get range of values for x axis
	double xmin = DBL_MAX;
	double xmax = -DBL_MAX;
	
	double item_xmin, item_xmax;
	bool rangeFound=false;
	
	// descend thru' panel hierarchy
	
	PlotItem *plotItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->child(i));
		if(plotItem) plotItem->get_dobjs();
		if(plotItem && plotItem->getXRange(item_xmin, item_xmax))
		{
			xmin = min(xmin, item_xmin);
			xmax = max(xmax, item_xmax);
			rangeFound = true;
		}
	}
	if(rangeFound)
	{
		panel_xmin = xmin;
		panel_xmax = xmax;
	}

	return rangeFound;
}

// ---------------------------------------

bool PanelItem::getYRange(double& panel_ymin, double& panel_ymax)
{
	// get range of values for y axis
	if ( QuSlotDebug() )
	{
		printf( "QpPanelItem::getYRange  \n" );
		fflush(stdout);
	}
	double ymin = DBL_MAX;
	double ymax =- DBL_MAX;
	
	double item_ymin, item_ymax;
	bool rangeFound = false;
    int isLog = 0;

	for(int i=0; i<this->childCount(); ++i)
	{
		QTreeWidgetItem * item = this->child(i);
		TraceItem* traceItem=dynamic_cast<TraceItem*>(item);
		SpectrumItem* spectrumItem=dynamic_cast<SpectrumItem*>(item);
		if(traceItem)
		{
			if(traceItem->getYRange(item_ymin,item_ymax))
			{
				ymin = min(ymin, item_ymin);
				ymax = max(ymax, item_ymax);
				rangeFound = true;
			}
            if(isLog == 0) isLog = traceItem->getYmetaIsLog();

		}
		else if(spectrumItem)
		{
			if(spectrumItem->getYRange( item_ymin, item_ymax))
			{
				ymin = min(ymin, item_ymin);
				ymax = max(ymax, item_ymax);
				rangeFound = true;
			}
            if(isLog == 0) isLog = spectrumItem->getYmetaIsLog();

		}
	}

    if( !yAxisTab->userSetLinLog ){
        if(isLog == 1) yAxisTab->_axis_linlog_fld->setCurrentIndex(QP_LOG);
        else yAxisTab->_axis_linlog_fld->setCurrentIndex(QP_LINEAR);
    }

	if(!rangeFound) return rangeFound;
	
	if(yAxisTab->_axis_y_nudge_fld->isChecked())
	{
		if(yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			double nudgeUp = pow(fabs(ymax / (ymin +1e-29)),   0.015);
			double nudgeDown = pow(fabs(ymax / (ymin +1e-29)), 0.03);
			ymax *= nudgeUp;
			ymin /= nudgeDown;
		}
		else
		{
			double nudge = fabs(ymax - ymin)*0.03;
			ymax += nudge;
			ymin -= nudge;
		}
	}
	panel_ymin = ymin;
	panel_ymax = ymax;
	return rangeFound;
}

// ---------------------------------------

bool PanelItem::getYmetaRange(double& panel_ymin, double& panel_ymax)
{
	// get range of values for y axis
	double ymin = DBL_MAX;
	double ymax =- DBL_MAX;
	
	double item_ymin, item_ymax;
	bool rangeFound = false;
    int isLog = 0;
	
	PlotItem* plotItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->child(i));
		if(plotItem && plotItem->getYmetaRange(item_ymin, item_ymax))
		{
			ymin = min(ymin, item_ymin);
			ymax = max(ymax, item_ymax);
			rangeFound = true;
		}

		if(plotItem && isLog == 0) isLog = plotItem->getYmetaIsLog();
	}

    if( !yAxisTab->userSetLinLog ){
        if(isLog == 1) yAxisTab->_axis_linlog_fld->setCurrentIndex(QP_LOG);
        else yAxisTab->_axis_linlog_fld->setCurrentIndex(QP_LINEAR);
    }
    
	if(yAxisTab->_axis_y_nudge_fld->isChecked())
	{
		if(yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG)
		{
			double nudgeUp = pow(fabs(ymax / (ymin +1e-29)), 0.015);
			double nudgeDown = pow(fabs(ymax / (ymin +1e-29)), 0.03);
			ymax *= nudgeUp;
			ymin /= nudgeDown;
		}
		else
		{
			double nudge = fabs(ymax - ymin)*0.03;
			ymax += nudge;
			ymin -= nudge;
		}
	}
	
	if(rangeFound)
	{
		panel_ymin = ymin;
		panel_ymax = ymax;
	}
	
	
	return rangeFound;
}

int PanelItem::numTLabel(TLabItem *item){
	// number of TLabItems above item
	int n=0;
	for(int i=0; i<item->childNumber(); ++i)
	{
		TLabItem *labItem=dynamic_cast<TLabItem*>(this->child(i));
		if(labItem) n++;
	}
	return n;
}

// ---------------------------------------

QpuiReturnStatus PanelItem::plot(double _height, double _width, double _toffset, double _loffset, PlotWindow* win)
{
	// defend against signals out of order, set set focus window, ensure plot in right window
	win->setPLFocus();
	
	// set viewport
    PLFLT x0, x1, y0, y1;
    x0 = _loffset;
    x1 = x0 + _width;
    y1 = 1 - _toffset;
    y0 = y1 - _height;

    plvpor(x0, x1, y0, y1);
    
    // preplot
    if(!isBlank()) this->prePlot();
    
    // now handle each plot object within this panel

	PlotItem* plotItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->child(i));
		if(plotItem){
        	win->setSmooth(false);
	    	if(QPUI_FAILURE == plotItem->plot())
	    	{
		    	cout << "Problem plotting plot item " << (plotItem->getName().toLatin1().data()) << endl;
	    	}
	    	win->setSmooth(true);
		}
	}
	
	TLabItem* labItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		labItem=dynamic_cast<TLabItem*>(this->child(i));
		if(labItem){
        	win->setSmooth(false);
	    	if(QPUI_FAILURE == labItem->plot())
	    	{
		    	cout << "Problem plotting T Lab Item " << (labItem->getName().toLatin1().data()) << endl;
	    	}
	    	win->setSmooth(true);
		}
	}

	for(int i=0; i<this->childCount(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->child(i));
		if(plotItem){
			TraceItem* traceItem=dynamic_cast<TraceItem*>(plotItem);
			if(traceItem)
        	{
				try{
					traceItem->plotLegend();
				}
				catch(...){;}
        	}
		}
	}

    bool show_top_axis = false;
    bool show_bottom_axis = false;
    bool show_title = false;
    bool show_T_lab = false;
    int stackNumber = getStackNumber();
    if(stackNumber == 1) show_title = true;
    if(stackNumber == frame()->numberInStack() ) show_T_lab = true;
	
	if( frame()->panelBlank(childNumber()-1) ) show_top_axis = true;
	if( frame()->panelBlank(childNumber()+1) ) show_bottom_axis = true;
	if( stackNumber == frame()->numberInStack()-1 &&  show_bottom_axis ) show_T_lab = true;

    if(!isBlank()) this->postPlot(show_top_axis, show_bottom_axis, show_title, show_T_lab);

    return QPUI_SUCCESS;
}

void PanelItem::updateYZRange()
{
  	if ( QuSlotDebug() ) cout << "QpPanelItem::updateYZRange" << endl;

	// called when X or T axis range is changed
	PlotItem* plotItem;
	for(int i=0; i<this->childCount(); ++i)
	{
		plotItem=dynamic_cast<PlotItem*>(this->child(i));
		if(plotItem) plotItem->clearObj_id();
	}
	this->yAxisTab->setAxisMinMax();
	
	// update colourscale in spectra
	for(int i=0; i<this->childCount(); ++i)
	{		
		SpectrumItem* spectrumItem = dynamic_cast<SpectrumItem*>(this->child(i));
		if(spectrumItem) spectrumItem->ColourscaleTab->setColourscaleMinMax();
	}

}

bool PanelItem::isBlank(){
	// blank if it has no traces or spectra (TLabelItems leave it blank)
	for(int i=0; i<this->childCount(); ++i)
	{
		PlotItem *plotItem=dynamic_cast<PlotItem*>(this->child(i));
		if(plotItem) return false;
	}
	return true;
}

bool PanelItem::hasTLabel(){
	// true if it has TLabelItem
	for(int i=0; i<this->childCount(); ++i)
	{
		TLabItem *tlabItem=dynamic_cast<TLabItem*>(this->child(i));
		if(tlabItem) return true;
	}
	return false;
}



// ---------------------------------------

QDomElement PanelItem::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Panel_Item");
    xml.appendChild(createTextXMLNode(doc, "Name", getName()));

    xml.appendChild(yAxisTab->toXMLElement(doc));// do bits common to all axes
    xml.appendChild(createTextXMLNode(doc, "Lin_Log", yAxisTab->_axis_linlog_fld->currentIndex()));

    xml.appendChild(LegendTab->toXMLElement(doc));
    
    int cnt = childCount();
    xml.setAttribute("size", QString::number(cnt));
   
    PlotItem* plotItem;
    TLabItem* labItem;
    for(int i=0; i<cnt; ++i)
    {
        plotItem=dynamic_cast<PlotItem*>(child(i));
        labItem=dynamic_cast<TLabItem*>(child(i));
        if(plotItem && plotItem->I_Type == QP_TRACE)
        {
            TraceItem *traceItem = dynamic_cast<TraceItem*>(plotItem);
            xml.appendChild(traceItem->toXMLElement(doc));
        }
        else if(plotItem && plotItem->I_Type == QP_SPECTROGRAM)
        {
            SpectrumItem *spectrumItem = dynamic_cast<SpectrumItem*>(plotItem);
            xml.appendChild(spectrumItem->toXMLElement(doc));
        }
        else if(plotItem && plotItem->I_Type == QP_CONST_Y)
        {
            TraceItem *constYItem = dynamic_cast<TraceItem*>(plotItem);
            xml.appendChild(constYItem->toXMLElement(doc));
        }
        else if(plotItem && plotItem->I_Type == QP_CONST_XT)
        {
            TraceItem *constXTItem = dynamic_cast<TraceItem*>(plotItem);
            xml.appendChild(constXTItem->toXMLElement(doc));
        }
        else if(plotItem && plotItem->I_Type == QP_REG_LINE)
        {
            TraceItem *regItem = dynamic_cast<TraceItem*>(plotItem);
            xml.appendChild(regItem->toXMLElement(doc));
        }
        else if(plotItem && plotItem->I_Type == QP_EVENT)
        {
            EventItem *eventItem = dynamic_cast<EventItem*>(plotItem);
            xml.appendChild(eventItem->toXMLElement(doc));
        }
        else if(labItem)
        {
            xml.appendChild(labItem->toXMLElement(doc));
        }
    }
    
    return xml;
}

void PanelItem::fromXMLElement(QDomElement& el)
{
    QString s;
    
    if(getXMLNodeText(el, "Name", s))
    {
        setName(s);
    }

    QDomElement axisElement=el.firstChildElement("Axis_Tab");
    if(!axisElement.isNull())
    {
        yAxisTab->fromXMLElement(axisElement);
    }
    if(getXMLNodeText(el, "Lin_Log", s))
    {
        yAxisTab->_axis_linlog_fld->setCurrentIndex((bool)s.toInt());
    }

    QDomElement legendElement=el.firstChildElement("Legend_Tab");
    if(!legendElement.isNull())
    {
        LegendTab->fromXMLElement(legendElement);
    }

    QDomElement plotElement=el.firstChildElement("Plot_Item");

    while(!plotElement.isNull())
    {
        QString type=plotElement.attribute("Type", "");
        if(type=="Trace")
        {
            TraceItem *traceItem = new TraceItem("", this, "", QP_TRACE);
            traceItem->fromXMLElement(plotElement);
        }
        else if(type=="Spectrogram")
        {
            SpectrumItem *spectrumItem = new SpectrumItem("", this);
            spectrumItem->fromXMLElement(plotElement);
        }
        else if(type=="Const_Y")
        {
            TraceItem *constYItem = new TraceItem("", this, "", QP_CONST_Y);
            constYItem->fromXMLElement(plotElement);
        }
        else if(type=="Const_XT")
        {
            TraceItem *constXTItem = new TraceItem("", this, "", QP_CONST_XT);
            constXTItem->fromXMLElement(plotElement);
        }
        else if(type=="Reg_Line")
        {
            TraceItem *regLine = new TraceItem("", this, "", QP_REG_LINE);
            regLine->fromXMLElement(plotElement);
        }
        else if(type=="Event")
        {
            EventItem *eventItem = new EventItem("", this);
            eventItem->fromXMLElement(plotElement);
        }
        
        plotElement=plotElement.nextSiblingElement("Plot_Item");
    }
    
	QDomElement labElement=el.firstChildElement("T_Lab_Item");

	while(!labElement.isNull())
    {
	 	TLabItem *labItem = new TLabItem("", this);
		labItem->fromXMLElement(labElement);
		labElement=labElement.nextSiblingElement("T_Lab_Item");

	}
}

// ------------------------------- TLabItem Class -----------------
//
// Belongs here as really a panel property

TLabItem::TLabItem(const QString& name, PanelItem* parent, const QString& WLcontext, int position) : QpuiItem(name, parent)
{
	if(parent==NULL)
	{
		std::cerr << "TLabItem::TLabItem: creating PlotItem with NULL parent" << std::endl;
		return;
	}
	
	I_Type = QP_T_LAB;
		
	if(name.isEmpty())
	{
		setNextName();
	}
	
	tabWidget=parent->tabWidget;

	if(WLcontext.isEmpty())
	{
		
		if(validObjName(name.toLatin1().data())) this->WLcontext=name;
		else this->WLcontext="";
	}
	else
	{
		this->WLcontext=WLcontext;
	}
	
	TLabDataTab = new QpTLabDataTab(tabWidget, "LabDataTab", this, QP_T_LAB);
	TLabDataTab->hide();

	setIcon(QP_T_LAB);

}

TLabItem::TLabItem(TLabItem* plot, PanelItem* parent, int position) : QpuiItem("", parent)
{
	if(parent==NULL)
	{
		std::cerr << "TLabItem::TLabItem: creating PlotItem with NULL parent" << std::endl;
		return;
	}

	I_Type = QP_T_LAB;

	setName(plot->getName());

	tabWidget=parent->tabWidget;

	WLcontext=plot->WLcontext;

	TLabDataTab = new QpTLabDataTab(tabWidget, "LabDataTab", this, QP_T_LAB);
	TLabDataTab->hide();
	
	setIcon(QP_T_LAB);

}



QpFrameType TLabItem::frameType()
{
	if(parent()) return parent()->frameType();
	return QP_TS;
}

TLabItem::~TLabItem()
{
	clear();
	if(TLabDataTab) delete TLabDataTab;
	
}

void TLabItem::showTabs()
{
	if(QuSlotDebug()) cout << "TLabItem::showTabs '" << this->getName().toLatin1().data() << "'" << endl;
	
	// ensure all tabs are hidden
	int Ntabs = tabWidget->count();
	for(int i = Ntabs-1; i >= 0; i--)
	{
		if(tabWidget->widget(i))
		{
			tabWidget->widget(i)->hide();
			tabWidget->removeTab(i);
		}
	}


	tabWidget->addTab(TLabDataTab, "Time Axis Data Label");

	// value return is always shown
	tabWidget->addTab(page()->ValueReturnTab, "Value Return");
 	tabWidget->setCurrentIndex(tabWidget->indexOf(TLabDataTab));
	
}


//--------------------------

QpuiReturnStatus TLabItem::plot()
{
	// SJS code
	if(QuSlotDebug())
	{
		printf("TLabItem::plot \n");
		fflush(stdout);
	}	
	
	int nx=frame()->xtAxisTab->_axis_major_ticks_num_fld->value();
			

	DvObject_var dobj = TLabDataTab->_x_data_fld->getObject();
	
	if(dobj.is_nil()) return QPUI_FAILURE;
	
	if(dobj->not_dbl()) return QPUI_FAILURE;

	DvTime tmin = frame()->tAxisTab->_axis_time_fld->getStart();
	DvTime tmax = frame()->tAxisTab->_axis_time_fld->getEnd();
	double tstart = tmin.getEpoch1970();
	double tend = tmax.getEpoch1970();
	PLFLT ttick1=0;
	if(nx!=0)
	{
		ttick1=(tmax-tmin)/(PLFLT)nx;
	}	
	else{
		// most of this is ignored
		int separator=1;
		PLINT auto_nsubx;
		char fmt_ticks[50], fmt_label[50];
		computeFormat(fabs(tmax-tmin), fmt_ticks, fmt_label, &ttick1, &auto_nsubx, separator);
	}
	
	QString units = TLabDataTab->unitText->text();
	QString label = TLabDataTab->metaText->text();
	
	int n = panel()->numTLabel(this);
	double y = TLabDataTab->_offset_y_fld->value();
	PLFLT voffset = (PLFLT) y - n*2.0;
		
	tickLabels((PLFLT)tstart, (PLFLT)tend, ttick1, -voffset, 13, dobj, label.toLatin1().data(), units.toLatin1().data());

	plscol0(0, 255, 255, 255);
	plscol0(1, 0, 0, 0);
	plcol0(1);

	return QPUI_SUCCESS;
}

// T Label Utilities

// sjs ticklabeller
//
// Label time axes with the values of a time series of data.
// Assumes viewport and window are already set up, so just makes
// calls to plmtex to add extra labels. This is done manually, but
// uses (private copies of) the plplot pldtik and pldtfac to determine
// the tick locations.
//
// Inputs:
// tstart, tend : times (in the plplot sense) corresponding to the time axis limits
// ttick1:  major tick spacing (in seconds). Set to zero for automatic
// disp:  displacement of the labels outward from the time axis
// length: max length string for the label. The string passed to qlabel
//         should be at least this long.
// data: a pointer to the time series values. This can be cast to be a pointer
//       of any type. It is only dereferenced in the data_interpolate routine to
//       return a PLFLT containing the value interpolated to the tick time.
// linelabel:  String that is written to the LEFT of the labels indicating what
//       it is, e.g., "X(GSE)"
// units:  String that is written to the RIGHT of the labels indicating the
//         units, e.g., "Re". This is prepended with an exponential label if the
//         data are scaled by a power of ten, e.g. "x10^3 Re".
// Helper routines:
//    tickscaler: samples data at several (currently 10) times between tstart
//        and tend to determine magnitude and range of values. Pickes an
//        appropriate power of ten to scale values before writing labels. This
//        does NOT use the plplot algorithms, which make assumptions about
//        monotonicity and get much more involved. My version is crude, but robust.
//
//    data_interpolate: interpolates the value of data at a given time.
//
//    qlabel:  creates the string coorresponding to the label. Calls data_interpolate
//
//    qpldtik and qpldfac: private versions of the corresponding plplot routines.
//        Unchanged apart from the names (including internal calls from one to the
//        other. Necessary because these aren't part of the public plplot API, so
//        aren't found by the linker. This may not be the case in the qsas
//        implementation.
//
void TLabItem::tickLabels(PLFLT tstart, PLFLT tend, PLFLT ttick1, PLFLT disp, PLINT length, DvObject_var &data, const char *linelabel, const char *units)
{
	PLFLT factor, tstart1;
	PLFLT tn, tp;
	PLFLT pos, just, dt;
	PLINT nxsub1, powten;
	char ticklabel[128], unitstr[128];
	PLFLT vpwxmi, vpwxma;
	PLINT family, style, weight;
	
	bool labVertical = TLabDataTab->_label_V_fld->isChecked();
	bool unitsVertical = TLabDataTab->_unit_V_fld->isChecked();

	TLabDataTab->_label_font_fld->getFontParams(family, style, weight);
	plsfont(family, style, weight);
	plschr(0, TLabDataTab->_label_char_size_fld->value());

	PLINT labels_colour_index = getPGPColourIndex(TLabDataTab->_label_colour);
	plcol0(labels_colour_index);

	if (length>127) length=127; 
	if (length < 13) length = 13;  // min for %g specifier
	
	dt=tend-tstart;
	
	if (dt>0.0)  //plplot routines expect vpwxmi < vpwxma
	{
		vpwxmi = tstart - 1.e-5*dt;  //plplot does this in plcore routines.
		vpwxma = tend + 1.e-5*dt;
	}
	else
	{
		vpwxmi = tend + 1.e-5*dt;
		vpwxma = tstart - 1.e-5*dt;
	}
	
	qpldtik( vpwxmi, vpwxma ,&ttick1,&nxsub1, TRUE); 
	qpldtfac(vpwxmi, vpwxma, &factor, &tstart1);
	tp = ttick1*(floor((vpwxmi-tstart1 )/ttick1))+tstart1;
//	printf("ttick1 = %0.2g tstart1 = %0.2g tp = %0.2g\n",ttick1,tstart1,tp);
//	powten = tickscaler(tstart,tend,data);

	powten = (PLINT) TLabDataTab->powText->value();
	
	bool isHr=false;
	DvString unitPtr(units);
	if(unitPtr == "h:m") {
		isHr=true;
		powten=0;
	}
	for(;;)
	{
		tn = tp+ttick1;
		pos = (tn-tstart)/dt;
		just = 0.5; 
		if ( pos > 1.0) break;
		if(pos > 0.0){
			qlabel(tn,length,powten,ticklabel,data,isHr);
			Q_plmtex("b",disp, pos,just,ticklabel);
		}
		tp=tn;
	}
	if(labVertical) Q_plmtex("bv",disp, -0.065,0.5,linelabel);
	else Q_plmtex("b",disp, -0.03,1,linelabel);
	
	if(unitsVertical){
		if (powten)
		{
			snprintf(unitstr, 127, "x10#u%d#d %s",powten,unitPtr.c_str());
			Q_plmtex("bv",disp, 1.065, 0.5, unitstr);
		}
		else 
		{
			Q_plmtex("bv",disp, 1.03,0,unitPtr.c_str());
		}
	}
	else{
		if (powten)
		{
			snprintf(unitstr, 127, "x10#u%d#d %s",powten,unitPtr.c_str());
			Q_plmtex("b",disp, 1.03, 0, unitstr);
		}
		else 
		{
			Q_plmtex("b",disp, 1.03,0,unitPtr.c_str());
		}
	}
}	


// ----------------------------------

QDomElement TLabItem::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("T_Lab_Item");
    xml.appendChild(createTextXMLNode(doc, "Name", getName()));

    xml.appendChild(TLabDataTab->toXMLElement(doc, QP_T_LAB));

    return xml;
}

void TLabItem::fromXMLElement(QDomElement& el)
{
    QString s;
	
 	if(QuSlotDebug()) cout <<  "TLabItem restore from XML (>>)" << endl;
   
    if(getXMLNodeText(el, "Name", s))
    {
        setName(s);
    }

    QDomElement dataElement=el.firstChildElement("T_Lab_Data_Tab");
    if(!dataElement.isNull())
    {
        TLabDataTab->fromXMLElement(dataElement);
    }


}
// -------------------------------------------

void TLabItem::replaceXData(QString obj_name)
{
	if(QuSlotDebug()) cout << "TLabItem::replaceXData" << endl;
	if(obj_name.isEmpty()) return;

	TLabDataTab->_x_data_fld->setTextKeepComp(obj_name);  // set data slot
	// setTextKeepComp can change text for comp or substitute string
	TLabDataTab->_x_data_fld->reValidate(true);
	
/*	if(  !(_x_data_fld->text().isEmpty())) {
		I_LabItem->setName(_x_data_fld->text());
	}*/

}

// -------------------------------------------

void TLabItem::replaceXComponent(QString comp)
{
	if(QuSlotDebug()) cout << "TLabItem::replaceXComponent" << endl;
	if(comp.isEmpty()) return;
	
	DvString newName = baseObjectText(TLabDataTab->_x_data_fld->text().toLatin1().data());
	newName += comp.toLatin1().data();
	TLabDataTab->_x_data_fld->setText(newName.c_str());  // set data slot
	
}


// QpTLabDataTab class
//------------------


QpTLabDataTab::QpTLabDataTab(QTabWidget *tabWidget,  const QString& name, TLabItem * _item, QpItemType plotType):
		QpTab(tabWidget, name, QP_LAB_DATA_TAB)
{
	// Constructor for the Data Tab for the Trace Options
  
	// keep track of item which owns this tab
	I_LabItem = _item;
	I_PlotType = plotType;

	create();
}

QpTLabDataTab::QpTLabDataTab(QpTLabDataTab* tab, TLabItem* _item):
		QpTab((QTabWidget*)(tab->parent()), "LabDataTab", tab->TabType)
{
	I_LabItem = _item;
	I_PlotType = tab->I_PlotType;

	create();
}

void QpTLabDataTab::create()
{
	old_BaseXName = "";
	old_XComp = "";
	
	UserUnit=false;
	UserText=false;
	
	_x_data_fld=NULL;
	
	QVBoxLayout *data_top_layout = new QVBoxLayout(this);
	data_top_layout->setSpacing(4);

	QGroupBox *data_plot_groupbox = new QGroupBox(this);
	
	QGridLayout *data_plot_layout=new QGridLayout;
	data_plot_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	data_plot_layout->setMargin(QSAS_profiles->current()->GUImargins);


	_x_label_fld = new QLabel("Data: ");
	data_plot_layout->addWidget(_x_label_fld,0,0,Qt::AlignRight);
	_x_data_fld = new QguWLinput(data_plot_groupbox,"x_data");
	_x_data_fld->setMinimumWidth(500);
	data_plot_layout->addWidget(_x_data_fld,0,1,1,2,Qt::AlignLeft);
		
		
	data_plot_groupbox->setLayout(data_plot_layout);
        
	data_top_layout->addWidget(data_plot_groupbox);
		
		
		// info box 
	QGroupBox *meta_groupbox=new QGroupBox(this);
	meta_groupbox->setTitle("Metadata");
		
	QGridLayout *meta_layout=new QGridLayout;
	meta_layout->setSpacing((int)(QSAS_profiles->current()->GUImargins*2.5));
	meta_layout->setMargin((int)(QSAS_profiles->current()->GUImargins*2.5));		
	
	QLabel *metaLabel = new QLabel("Label: ");
	meta_layout->addWidget(metaLabel,0,1,Qt::AlignRight);
	metaText = new QLineEdit();
	metaText->setMinimumWidth(400);
	meta_layout->addWidget(metaText,0,2,Qt::AlignLeft);
		
	QLabel *unitLabel = new QLabel("Unit Text: ");
	meta_layout->addWidget(unitLabel,1,1,Qt::AlignRight);
	unitText = new QLineEdit();
	unitText->setMinimumWidth(400);
	meta_layout->addWidget(unitText,1,2,Qt::AlignLeft);
	unitText->setToolTip("Unit string for labels - does not rescale data");
		
	QLabel *powLabel = new QLabel("Power of 10: ");
	meta_layout->addWidget(powLabel,2,1,Qt::AlignRight);
	powText = new FloatSpinBox(-20., 20., 0, 1.0);
	powText->setMinimumWidth(40);
	powText->setValue(0.0);
	meta_layout->addWidget(powText,2,2,Qt::AlignLeft);
	powText->setToolTip("Power of 10 for tick values - ignored for units of 'h:m' used in Local Time ");
		
	meta_groupbox->setLayout(meta_layout);

	data_top_layout->addWidget(meta_groupbox);		

	// Position box
	QGroupBox *posn_groupbox=new QGroupBox(this);
	posn_groupbox->setTitle("Position");
		
	QGridLayout *posn_layout=new QGridLayout;
	posn_layout->setSpacing((int)(QSAS_profiles->current()->GUImargins*2.5));
	posn_layout->setMargin((int)(QSAS_profiles->current()->GUImargins*2.5));

    posn_layout->addWidget(new QLabel("Vertical offset:"),0,0, Qt::AlignRight);
    posn_layout->addWidget(new QLabel("(in char height from bottom):"),0,2, Qt::AlignLeft);
    _offset_y_fld=new FloatSpinBox(-40., 40., 2, 0.05);
    posn_layout->addWidget(_offset_y_fld,0,1, Qt::AlignCenter);
	_offset_y_fld->setValue(1.5);

	_label_V_fld = new QCheckBox("Rotate Label");
    posn_layout->addWidget(_label_V_fld,1,2, Qt::AlignLeft);
			
	_unit_V_fld = new QCheckBox("Rotate Units");
    posn_layout->addWidget(_unit_V_fld,1,3, Qt::AlignLeft);

	posn_groupbox->setLayout(posn_layout);
	data_top_layout->addWidget(posn_groupbox);		

	QLabel* label_char_size_label = new QLabel("Character size: ");
	posn_layout->addWidget(label_char_size_label, 1, 0, Qt::AlignRight);
	_label_char_size_fld = new FloatSpinBox(0.01, 40.,1, .1, posn_groupbox, "label_char_size");
	_label_char_size_fld->setFixedWidth(100);
	posn_layout->addWidget(_label_char_size_fld, 1, 1, Qt::AlignLeft);
	_label_char_size_fld->setValue(QSAS_profiles->current()->BaseFontSize+QPUI_LABELS_CHAR_SIZE_ADD);

	QLabel* font_label = new QLabel("Font: ");
	posn_layout->addWidget(font_label, 2, 0, Qt::AlignRight);
	_label_font_fld = new FontButton;
	posn_layout->addWidget(_label_font_fld, 2, 1, Qt::AlignLeft);

	QLabel* colour_label = new QLabel("Colour: ");
	posn_layout->addWidget(colour_label, 2, 2, Qt::AlignRight);
	_label_colour_fld = new QPushButton("Colour...");
	_label_colour_fld->setFixedWidth(100);
	_label_colour_fld->setFixedHeight(30);
	QFont f = _label_colour_fld->font();
	f.setBold(TRUE);
	_label_colour_fld->setFont(f);
	posn_layout->addWidget(_label_colour_fld, 2, 3, Qt::AlignLeft);

    posn_layout->addWidget(new QLabel("Note: Panel itself will not be drawn if otherwise empty"),3,0,1,4, Qt::AlignLeft);

	QString PlotTitle = "Additional Time Axis Data Label";

	_x_label_fld->setVisible(true);
	_x_data_fld->setVisible(true);
	_x_data_fld->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_0D, QS_SEQ, QS_TT));
    
    data_plot_groupbox->setTitle(PlotTitle);
	
	data_top_layout->addSpacing(4);
	data_top_layout->addStretch(10);

	connect(_x_data_fld, SIGNAL(edited_OK()), this, SLOT(setXData()));
	connect(metaText, SIGNAL(textEdited( const QString &)), this, SLOT(setMeta( const QString &)));
	connect(unitText, SIGNAL(textEdited( const QString &)), this, SLOT(setUnit( const QString &)));
	connect(powText, SIGNAL(valueChanged( double)), this, SLOT(setPow( double)));
	connect(_offset_y_fld, SIGNAL(valueChanged( double)), this, SLOT(setOffset( double)));
	connect(_label_V_fld, SIGNAL(toggled( bool)), this, SLOT(setLabVert( bool)));
	connect(_unit_V_fld, SIGNAL(toggled( bool)), this, SLOT(setUnitVert( bool)));
	connect(_label_char_size_fld, SIGNAL(valueChanged( double)), this, SLOT(setCharSize( double)));
	connect(_label_font_fld, SIGNAL(familyChanged( int)), this, SLOT(setFontFamily( int)));
	connect(_label_font_fld, SIGNAL(styleChanged( bool)), this, SLOT(setFontStyle( bool)));
	connect(_label_font_fld, SIGNAL(weightChanged( bool)), this, SLOT(setFontWeight( bool)));
	connect(_label_colour_fld, SIGNAL(clicked()), this, SLOT(setLabelColour()));

}

void QpTLabDataTab::setUnit(const QString &text){

	if ( QuSlotDebug() ) printf( "QpTLabDataTab::setUnit \n" );
	
	if(text == "h:m"){
		powText->setEnabled(false);
		powText->setValue(0.0);
	}
	else{
		powText->setEnabled(true);
	}

	if(text.isEmpty()) UserUnit = false;
	else UserUnit = true;
	if(this->isVisible() ) {

		// propagate to selected items 

		I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_UNITS, 0, 0.0, text);
	}
}

void QpTLabDataTab::setPow(double d){

	if ( QuSlotDebug() ) printf( "QpTLabDataTab::setPow \n" );

	if(this->isVisible() ) {

		// propagate to selected items 

		I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_POWER, 0, d);
	}
}

//  ---------------------

void QpTLabDataTab::setLabelColour()
{	
	bool changed = false;
	// ask the user for colour choice
	QColor c = QuGetColour(_label_colour, &changed, this);
	if(changed)
	{
        QPalette line_palette = QPalette(QPUI_normal_palette);
        line_palette.setColor(QPalette::Active ,QPalette::ButtonText, c);
        line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, c);
        this->_label_colour_fld->setPalette(line_palette);
        _label_colour = c;
	}

		// propagate to selected items 

		I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_COLOUR, 0, 0., "", false, c);
   
}

void QpTLabDataTab::setOffset(double d){
	
	if ( QuSlotDebug() ) printf( "QpTLabDataTab::setOffset \n" );

	if(this->isVisible() ) {

		// propagate to selected items if not hard linked

		I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_OFFSET, 0, d);
	}
}

void QpTLabDataTab::setCharSize(double d){
	
	if ( QuSlotDebug() ) printf( "QpTLabDataTab::setCharSize \n" );

	if(this->isVisible() ) {

		// propagate to selected items if not hard linked

		I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_CHAR, 0, d);
	}
}

void QpTLabDataTab::setFontFamily(int i){
	
	if ( QuSlotDebug() ) printf( "QpTLabDataTab::setFontFamily \n" );
	
	if(this->isVisible() ) {

		// propagate to selected items if not hard linked

		I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_FONT_FAMILY, i);
	}
}

void QpTLabDataTab::setFontStyle(bool b){
	
	if ( QuSlotDebug() ) printf( "QpTLabDataTab::setFontStyle \n" );

	if(this->isVisible() ) {

		// propagate to selected items if not hard linked

		I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_FONT_STYLE, 0, 0.0, "", b);
	}
}

void QpTLabDataTab::setFontWeight(bool b){
	
	if ( QuSlotDebug() ) printf( "QpTLabDataTab::setFontWeight \n" );
	
	if(this->isVisible() ) {

		// propagate to selected items if not hard linked

		I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_FONT_WEIGHT, 0, 0.0, "", b);
	}
}

void QpTLabDataTab::setLabVert(bool v){
	
	if ( QuSlotDebug() ) printf( "QpTLabDataTab::setLabVert \n" );

	if(this->isVisible() ) {

		// propagate to selected items if not hard linked

		I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_LAB_VERT, 0, 0.0, "", v);
	}
}
void QpTLabDataTab::setUnitVert(bool v){
	
	if ( QuSlotDebug() ) printf( "QpTLabDataTab::setUnitVert \n" );

	if(this->isVisible() ) {

		// propagate to selected items if not hard linked

		I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_UNIT_VERT, 0, 0.0, "", v);
	}
}

void QpTLabDataTab::setMeta(const QString &text){

	if ( QuSlotDebug() ) printf( "QpTLabDataTab::setMeta \n" );

	if(text.isEmpty()) UserText = false;
	else UserText = true;
	
	if(this->isVisible() ) {

		// propagate to selected items if not hard linked

		I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_META, 0, 0.0, text);
	}

}

void QpTLabDataTab::setXData()
{
	if( QuSlotDebug() ) cout << " QpDataTab::setXData  " << endl;

	if( !I_LabItem) return;


	// is change to base object or component...
	DvString BaseName = baseObjectText( _x_data_fld->text().toLatin1().data() ); // new value
	DvString CompText = compText( _x_data_fld->text().toLatin1().data() ); // new value
	
	bool compOnly = false;
	if(old_BaseXName == BaseName && old_XComp != CompText) compOnly = true;
	
	old_BaseXName = BaseName;
	old_XComp = CompText;

	
	if(this->isVisible() ) {	

		// propagate to selected items if not hard linked

		if(compOnly) I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_XCOMP, 0, 0.0, CompText.c_str());
		else I_LabItem->Tree()->setTLabDataOption(T_LAB_DATA_XDATA, 0, 0.0, BaseName.c_str());
	}
	
	DvObject_var dobj = _x_data_fld->getObject();
	
	if(dobj.is_nil()) return;
	
	DvTime tmin = I_LabItem->frame()->tAxisTab->_axis_time_fld->getStart();
	DvTime tmax = I_LabItem->frame()->tAxisTab->_axis_time_fld->getEnd();
	double tstart = tmin.getEpoch1970();
	double tend = tmax.getEpoch1970();
	double powten = (double) tickscaler(tstart, tend, dobj);

	powText->setValue(powten);

	if(!UserUnit){
		if(dobj->xref_exists(UNITS)) {
			DvString units = dobj->getXrefText(UNITS );
			if(units == "hr"){
				powText->setEnabled(false);
				powText->setValue(0.0);
				units="h:m";
			}
			else{
				powText->setEnabled(true);
			}
			unitText->setText(units.c_str());
		}
	}
	if(!UserText){
		if(dobj->xref_exists(LABLAXIS)) {
			DvString text = dobj->getXrefText(LABLAXIS);
			metaText->setText(text.c_str());
		}
	}
	

	if(  !(_x_data_fld->text().isEmpty())) {
		I_LabItem->setName(_x_data_fld->text());
	}

}


QDomElement QpTLabDataTab::toXMLElement(QDomDocument& doc, QpItemType I_Type) const
{
    QDomElement xml=doc.createElement("T_Lab_Data_Tab");

	xml.appendChild(createTextXMLNode(doc, "X_Data", _x_data_fld->text()));    
	xml.appendChild(createTextXMLNode(doc, "Label_Data", metaText->text()));    
	xml.appendChild(createTextXMLNode(doc, "Unit_Data", unitText->text()));    
	xml.appendChild(createTextXMLNode(doc, "Offset", _offset_y_fld->value()));    
	xml.appendChild(createTextXMLNode(doc, "CharSize", _label_char_size_fld->value()));    
	xml.appendChild(createTextXMLNode(doc, "LabVert", _label_V_fld->isChecked()));    
	xml.appendChild(createTextXMLNode(doc, "UnitVert", _unit_V_fld->isChecked()));    

    PLINT family, style, weight;
    _label_font_fld->getFontParams(family, style, weight);
    xml.appendChild(createTextXMLNode(doc, "Label_Font_Family", family));
    xml.appendChild(createTextXMLNode(doc, "Label_Font_Style", style));
    xml.appendChild(createTextXMLNode(doc, "Label_Font_Weight", weight));

    xml.appendChild(createTextXMLNode(doc, "Label_Red", _label_colour.red()));
    xml.appendChild(createTextXMLNode(doc, "Label_Green", _label_colour.green()));
    xml.appendChild(createTextXMLNode(doc, "Label_Blue", _label_colour.blue()));

	xml.appendChild(createTextXMLNode(doc, "powerOfTen", powText->value()));    

    return xml;
}

void QpTLabDataTab::fromXMLElement(QDomElement& el)
{
    QString s;

	if(getXMLNodeText(el, "X_Data", s))
	{
		_x_data_fld->setText(s);
	}
	if(getXMLNodeText(el, "Label_Data", s))
	{
		metaText->setText(s);
	}
	if(getXMLNodeText(el, "Unit_Data", s))
	{
		unitText->setText(s);
	}
	if(getXMLNodeText(el, "Offset", s))
	{
		_offset_y_fld->setValue(s.toDouble());
	}
	if(getXMLNodeText(el, "CharSize", s))
	{
		_label_char_size_fld->setValue(s.toDouble());
	}
	if(getXMLNodeText(el, "LabVert", s))
	{
		_label_V_fld->setChecked(s.toInt());
	}
	if(getXMLNodeText(el, "UnitVert", s))
	{
		_unit_V_fld->setChecked(s.toInt());
	}
    PLINT family(PL_FCI_SANS), style(0), weight(0);
    if(getXMLNodeText(el, "Label_Font_Family", s))
    {
        family=s.toInt();
    }
    if(getXMLNodeText(el, "Label_Font_Style", s))
    {
        style=s.toInt();
    }
    if(getXMLNodeText(el, "Label_Font_Weight", s))
    {
        weight=s.toInt();
    }
    switch(family)
    {
        case PL_FCI_SERIF: _label_font_fld->setSerifFont(); break;
        case PL_FCI_MONO: _label_font_fld->setMonotypeFont(); break;
        case PL_FCI_SCRIPT: _label_font_fld->setDecorativeFont(); break;
        default: _label_font_fld->setSansFont(); break;
    }
    _label_font_fld->setItalic(style==PL_FCI_ITALIC);
    _label_font_fld->setBold(weight==PL_FCI_BOLD);

    int r(0), g(0), b(0);
    if(getXMLNodeText(el, "Label_Red", s))
    {
        r=s.toInt();
    }
    if(getXMLNodeText(el, "Label_Green", s))
    {
        g=s.toInt();
    }
    if(getXMLNodeText(el, "Label_Blue", s))
    {
        b=s.toInt();
    }
    _label_colour = QColor(r,g,b);
    QPalette line_palette = QPalette(QPUI_normal_palette);
    line_palette.setColor(QPalette::Active ,QPalette::ButtonText, _label_colour);
    line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, _label_colour);
    _label_colour_fld->setPalette(line_palette);

	if(getXMLNodeText(el, "powerOfTen", s))
	{
		powText->setValue(s.toDouble());
	}

	_x_data_fld->reValidate(false);
			
			
}


// Determine whether values for ticks should be scaled by a power
// of ten. Simplified version of plplot algorithms, due to non-monotonic
// nature of data and other irrelevant things.
//
// Returns power of ten to use to scale data for labels, and to
// annotate line with 10^xx accordingly.
// Return value of zero means no scaling
PLINT tickscaler(PLFLT tstart, PLFLT tend, DvObject_var &data)
{
	PLINT i, N=10; 
	PLFLT t, dt, x, xmaxabs,xminabs;
	PLINT xmaxfloor;

// Sample data at N times, and find extremal values of |x|
	dt = (tend-tstart)/(PLFLT)N;
	xmaxabs = ABS(data_interpolate(tstart,data)); // initialise
	xminabs = xmaxabs;
	for (i=0,t=tstart+dt; i<N; i++, t+=dt)
	{
		x = ABS(data_interpolate(t,data));
		if (x>xmaxabs) xmaxabs=x;
		if (x<xminabs) xminabs=x;
	}
// Try to optimise label format. Use %.3g to ensure 3 sig figs always.
// Alt is to give user some control, but dangerous.
//
// Logic:
// 1 If only zeroes, don't scale and return.
// 2 If xmin is zero, can't take log so just scale based on xmax
// 3 Calculate floor of log10(xmax), i.e. xmax = a.bb x 10^floor
// 4 If xmax/xmin > 1000, nudge floor by 2 places so xmin and xmax both
//   well represented, e.g., as 0.123 and 456.
// 5 Else if xmax/xmin > 100 nudge by 1 place
// 6 Else if xmax < 1000 don't scale
// 6 Else use floor
//
//  This algorithm doesn't guarantee width of label, nor possibility
//  that you get a line like  1.234e+003  ... x10^1, but it's not too bad in
//  examples I've tried.
	if (xmaxabs == 0.0) return 0;
	// calculate floor of xmax
	xmaxfloor = (PLINT) floor (log10 (xmaxabs));
	if (xminabs > 0.0)
	{
		if (xmaxabs/xminabs > 100.) return (PLINT)( xmaxfloor-2);
		else if (xmaxabs/xminabs > 10.) return (PLINT) (xmaxfloor -1);
		else return  xmaxfloor;		
	}
	else return  xmaxfloor ;
}

// call Interpolation routine. 
PLFLT data_interpolate(PLFLT t, DvObject_var &data)
{
	DvTime target;
	target.setFromEpoch1970(t);
	
	// We know this is a scalar object, but call is generic
	DvObject_var interpVal = data->interpAt(target);
	return (PLFLT) interpVal->asDouble(0); 
}

// Routine to take t and return label string based on data(t)
// Values are scaled by 10^powten
void qlabel(PLFLT t, PLINT length, PLINT powten, char* label, DvObject_var &data, bool isHr)
{
	PLFLT val = (PLFLT) data_interpolate(t, data);		

	if(isHr){
		PLINT hr = (PLINT) val;
		val -= (PLFLT) hr;
		
		PLINT min = (PLINT)(val * 60);
		snprintf ( label, length, "%2.2d:%2.2d", hr, min);
	}
	else{
		if (powten)
		{
			val /= pow(10., powten);
		}	
		// print label. This should be robust, but could lead to
		// strange-looking labels, e.g.:  1.2e+001 .... x10^6
		// due to simplistic tickscaler
		snprintf ( label, length, "%.3g", val);
	}
}

//
// Following routines are unchanged apart from names from plplot versions,
// which are not accessible via the standard plplot libraries. It is possible
// that they are in the qsas implementation of plplot. SJS

//--------------------------------------------------------------------------
// void pldtik()
//
// Determine tick spacing: works out a "nice" interval (if tick == 0) such
// that there are between 3 and 7.5 major tick intervals in the input
// range vmin to vmax.  The recommended number of subticks is returned in
// "nsubt" unless the routine is entered with a non-zero value of "nsubt".
// n.b. big change: now returns only positive values of tick and nsubt
//--------------------------------------------------------------------------

void
qpldtik( PLFLT vmin, PLFLT vmax, PLFLT *tick, PLINT *nsubt, PLBOOL ld )
{
    PLFLT t1, t2, tick_reasonable;
    PLINT np, ns;
    PLFLT factor;


    if ( ld )
    {
        // Check suitable units for tick spacing
        qpldtfac( vmin, vmax, &factor, NULL );

        *tick = *tick / factor;
        vmin  = vmin / factor;
        vmax  = vmax / factor;
    }

// Magnitude of min/max difference to get tick spacing

    t1 = (PLFLT) log10( ABS( vmax - vmin ) );
    np = (PLINT) floor( t1 );
    t1 = t1 - np;

// Get tick spacing.

    if ( t1 > 0.7781512503 )
    {
        t2 = 2.0;
        ns = 4;
    }
    else if ( t1 > 0.4771212549 )
    {
        t2 = 1.0;
        ns = 5;
    }
    else if ( t1 > 0.1760912591 )
    {
        t2 = 5.0;
        ns = 5;
        np = np - 1;
    }
    else
    {
        t2 = 2.0;
        ns = 4;
        np = np - 1;
    }

// Now compute reasonable tick spacing

    tick_reasonable = t2 * pow( 10.0, (double) np );
    if ( *tick == 0 )
    {
        *tick = t2 * pow( 10.0, (double) np );
    }
    else
    {
        *tick = ABS( *tick );
        if ( *tick < 1.e-4 * tick_reasonable )
        {
            plexit( "pldtik: magnitude of specified tick spacing is much too small" );
            return;
        }
    }
    if ( *nsubt == 0 )
        *nsubt = ns;

    *nsubt = ABS( *nsubt );

    if ( ld )
    {
        *tick = *tick * factor;
    }
}

//--------------------------------------------------------------------------
// PLFLT pldtfac()
//
// Calculate factor to convert a date/time interval in seconds
// into a more natural units (minutes, hours, days, week, years).
// Also optionally calculate the sensible start time for counting ticks
// from (e.g. beginning of day, beginning of year).
// Used to calculate sensible tick and label spacings.
//--------------------------------------------------------------------------
void
qpldtfac( PLFLT vmin, PLFLT vmax, PLFLT *factor, PLFLT *start )
{
    PLFLT diff;
    PLINT year, month, day, hour, min;
    PLFLT sec;

    diff = vmax - vmin;

    if ( start != NULL )
    {
        plbtime( &year, &month, &day, &hour, &min, &sec, vmin );
    }

    if ( diff < 3.0 * 60.0 )
    {
        // Seconds
        *factor = 1.0;
        if ( start != NULL )
        {
            sec = 0.;
            plctime( year, month, day, hour, min, sec, start );
        }
    }
    else if ( diff < 3.0 * 60.0 * 60.0 )
    {
        // Minutes
        *factor = 60.0;
        if ( start != NULL )
        {
            sec = 0.;
            min = 0;
            plctime( year, month, day, hour, min, sec, start );
        }
    }
    else if ( diff < 3.0 * 60.0 * 60.0 * 24.0 )
    {
        // Hours
        *factor = 60.0 * 60.0;
        if ( start != NULL )
        {
            sec  = 0.;
            min  = 0;
            hour = 0;
            plctime( year, month, day, hour, min, sec, start );
        }
    }
    else if ( diff < 3.0 * 60.0 * 60.0 * 24.0 * 7.0 )
    {
        // Days
        *factor = 60.0 * 60.0 * 24.0;
        if ( start != NULL )
        {
            sec  = 0.;
            min  = 0;
            hour = 0;
            plctime( year, month, day, hour, min, sec, start );
        }
    }
    else if ( diff < 3.0 * 60.0 * 60.0 * 24.0 * 365 )
    {
        // Weeks
        *factor = 60.0 * 60.0 * 24.0 * 7.0;
        if ( start != NULL )
        {
            sec  = 0.;
            min  = 0;
            hour = 0;
            plctime( year, month, day, hour, min, sec, start );
        }
    }
    else
    {
        // Years
        *factor = 60.0 * 60.0 * 24.0 * 365.25;
        if ( start != NULL )
        {
            sec   = 0.;
            min   = 0;
            hour  = 0;
            day   = 0;
            month = 0;
            plctime( year, month, day, hour, min, sec, start );
        }
    }
}

