// qpolar_window.cc
//

// SJS 1 Dec 2015:
// I've edited the two sets of calls to cont_store, one in the cartesian and
// one in the polar for the case of colourLog->isChecked. The original code
// gave spurious plots in both cases when the user adjusted the X/R axis limits.
// The cartesian one displaced and distorted the contours in the X direction. 
// The polar one displaced them to the other side of the plot, i.e., a pitch
// angle distribution with theta in (0,180) found itself with contours in the
// (180-360) range.
//
// There is still a bug in the sense that for restricted ranges you find fewer
// contour levels. Perhaps zmin and zmax should be found over the restricted
// to populate the contour levels based on the restricted range of points.
//
// There are also serious issues with attempting to change the plot size
// (custom, portrait, etc.). Most don't work, and changing landscape->protrait
// leaves it stuck in portrait even if you try to switch back. There's a lot#
// of code in this section. I'd be inclined to set up an aspect ratio that was
// tuned to the plot (removing a lot of wasted white space); users can always
// print onto any size paper, and that makes the plots suitable for use without
// trimming. I also hit a crash in playing with all this, but didn't try to pin
// it down at all.

#include "qsurface.h"
#include "qwui.h"
#include "qdutil.h"


extern qwuiWindow* qsasWindow;
QsurfaceWindow *QSurface_widget=static_cast<QsurfaceWindow*>(NULL);

void plfill3(polygon*);
void plpoly3(polygon*);

void QSurfaceMakeWindow( QWidget *parent,const char* Title )
{
	if(QSurface_widget==NULL)
	{
		QSurface_widget = new QsurfaceWindow(parent);
		QSize qsasGUISize = qsasWindow->frameSize();
		QSurface_widget->adjustSize();  
		QSurface_widget->move((int)(qsasGUISize.width()), 25);

		QSurface_widget->connect(parent, SIGNAL(saveSessionRequest(QString)), QSurface_widget, SLOT(saveSession(QString)) );
		QSurface_widget->connect(parent, SIGNAL(restoreSessionRequest(QString)), QSurface_widget, SLOT(restoreSession(QString)) );

		QSurface_widget->setWindowTitle(Title);
		QSurface_widget->_filename+=QString("/saveSet");
		QSurface_widget->SAVE_EXTENSION=QString(".qx_surface_plot");
		QSurface_widget->_filename+=QSurface_widget->SAVE_EXTENSION;
		
	}

}

void QsurfaceShow()
{
    QSurface_widget->show();
}


//     QsurfaceWindow CLASS
//     ----------------

QsurfaceWindow::QsurfaceWindow( QWidget *parent ):QMainWindow(0)
{
	// Constructor
	if(QuSlotDebug())
	{
		printf("QsurfaceWindow::QsurfaceWindow \n");
		fflush(stdout);
	}
	// used later by plot win

    startSlider = 0;
    startElevation = 70;
    startAzimuth = 30;
    elevation = NULL;
    azimuth = NULL;
    atSlider = NULL;
    delay = NULL;
    runBtn = NULL;
    animateBtn = NULL;
    
	// initialise defaults as they may get used before setPage() is called
	DPI = 72;
	conv = (double) DPI;
	escPushed = false;
	
	// set up colour group palettes
	QPalette QSurface_normal_palette = QPalette(qApp->palette());
	this->setPalette(QSurface_normal_palette);

	setAcceptDrops(true);
	setFocusPolicy(Qt::WheelFocus);

	QSplitter *_top_splitter=new QSplitter(Qt::Vertical, this);
	setCentralWidget( _top_splitter);

	tab_widget = new QTabWidget( _top_splitter);
	int margins = QSAS_profiles->current()->GUImargins;
	tab_widget->setContentsMargins(margins, margins, margins, margins);

	// data tab
	// ---------
	
	QWidget *dataTab=new QWidget();
	QGridLayout *dataTabLayout=new QGridLayout(dataTab);
	dataTabLayout->setSpacing(QSAS_profiles->current()->GUImargins);
	
	dataTabLayout->setMargin(QSAS_profiles->current()->GUImargins);	
	dataTabLayout->setColumnStretch(0, 1);
	dataTabLayout->setColumnStretch(1, 3);
	dataTabLayout->setColumnStretch(2, 3);
	dataTabLayout->setColumnStretch(3, 3);
	dataTabLayout->setColumnStretch(4, 3);
	dataTabLayout->setColumnStretch(5, 10);
	dataTabLayout->setColumnStretch(6, 10);
	
	// data input
	QLabel * dataLabel = new QLabel("2D data: ");
	dataTabLayout->addWidget(dataLabel, 0, 0, Qt::AlignRight);

	dataSlot = new QguWLinput(dataTab, "data");
	dataSlot->setMinimumWidth(400);
	dataSlot->setToolTip("Accepts a 2D array or sequence of 2D arrays - use pulldown to subsample");
	dataSlot->setAcceptable(Qcondition(QS_NARROW, QS_DBL, QS_2D, QS_ANY, QS_ANY));
	connect(dataSlot, SIGNAL(edited_OK()), this, SLOT(setDataObj()) );
	dataTabLayout->addWidget(dataSlot, 0, 1, 1, 6);
	
	//time selection
	atSlider = new  RecordSlider("At ", dataTab, "atSlider");
	atSlider->setToolTip("Select time (or record) at which to plot if a sequence of 2D records");
	dataTabLayout->addWidget(atSlider, 1, 0, 1, 7);
	connect(atSlider, SIGNAL(recordSliderChanged()), this, SLOT(setRecord()) );


	dataTab->adjustSize();
	tab_widget->addTab(dataTab, "Data");
	

	// Options Tab
	// ----------
	QWidget *optionsTab=new QWidget();
	QGridLayout *optionsTabLayout=new QGridLayout(optionsTab);
	optionsTabLayout->setSpacing(QSAS_profiles->current()->GUImargins);
	optionsTabLayout->setMargin(QSAS_profiles->current()->GUImargins);	
	
	QLabel * titleLab = new QLabel("Plot Title: ");
	optionsTabLayout->addWidget(titleLab, 0, 0, Qt::AlignRight);
	
	titleSlot = new QLineEdit();
	titleSlot->setMinimumWidth(200);
	connect(titleSlot, SIGNAL(textEdited(const QString& )), this, SLOT(setUserTitle()));
	optionsTabLayout->addWidget(titleSlot, 0, 1, 1, 3);

    autoTitle = new QCheckBox("Auto-Title");
    optionsTabLayout->addWidget(autoTitle, 0, 4, Qt::AlignLeft);
    autoTitle->setChecked(true);
    connect(autoTitle, SIGNAL(toggled(bool)), this, SLOT(setTitle()) );

    swapR = new QCheckBox("Swap R-theta");
    swapR->setToolTip("Force other dimension to be used for radial data (useful when metadata is missing and plot guesses wrongly)");
    optionsTabLayout->addWidget(swapR, 0, 5, Qt::AlignLeft);
    connect(swapR, SIGNAL(toggled(bool)), this, SLOT(setSwapRTheta()) );
    
	RaxisLab = new QLabel("R axis Text: ");
	optionsTabLayout->addWidget(RaxisLab, 1, 2, Qt::AlignRight);
	
	RXLabSlot = new QLineEdit();
	RXLabSlot->setMinimumWidth(160);
	connect(RXLabSlot, SIGNAL(textEdited(const QString& )), this, SLOT(setUserRXLab()));
	optionsTabLayout->addWidget(RXLabSlot, 1, 3, 1, 1);

	ThAxisLab = new QLabel("Angle axis Text: ");
	optionsTabLayout->addWidget(ThAxisLab, 1, 4, Qt::AlignRight);
	
	ThYLabSlot = new QLineEdit();
	ThYLabSlot->setMinimumWidth(160);
	connect(ThYLabSlot, SIGNAL(textEdited(const QString& )), this, SLOT(setUserThYLab()));
	optionsTabLayout->addWidget(ThYLabSlot, 1, 5, 1, 1);

	QLabel * typeLab = new QLabel("Plot Type: ");
	optionsTabLayout->addWidget(typeLab, 1, 0, Qt::AlignRight);
	plotType = new QComboBox();
	plotType->insertItem(0, "Mesh");
	plotType->insertItem(1, "Surface Filled");
	plotType->insertItem(2, "Surface Fill & Mesh");
	plotType->insertItem(3, "Contour");
	plotType->insertItem(4, "Histogram");
	plotType->insertItem(5, "Histogram (no edges)");
	plotType->setToolTip("Surface-Filled: uses blocks of colour on mesh segments.\nContour stacks coloured contours\n Histogram uses filled blocks");
	optionsTabLayout->addWidget(plotType, 1, 1, Qt::AlignLeft);
	connect(plotType, SIGNAL(currentIndexChanged(int)), this, SLOT(setPlotType(int)));

    QLabel* axis_width_label=new QLabel("Axis width: ");
    optionsTabLayout->addWidget(axis_width_label,2,0,Qt::AlignRight);
    axis_width_fld = new FloatSpinBox(0.1,20.,1,0.1, NULL, "axis_width");
    axis_width_fld->setValue(QSAS_profiles->current()->LineWidth);
    axis_width_fld->setFixedWidth(50);
    optionsTabLayout->addWidget(axis_width_fld,2,1,Qt::AlignLeft);
   
	QLabel * axisLab = new QLabel("Axis Type: ");
	optionsTabLayout->addWidget(axisLab, 2, 2, Qt::AlignRight);
	axisType = new QComboBox();
	axisType->insertItem(0, "Auto detect");
	axisType->insertItem(1, "Polar");
	axisType->insertItem(2, "Cartesian");
	axisType->setToolTip("R-Theta or X-Y axes for base dimensions");
	optionsTabLayout->addWidget(axisType, 2, 3, Qt::AlignLeft);    
	connect(axisType, SIGNAL(currentIndexChanged(int)), this, SLOT(setAxisOpt()));
	
	QLabel * alphaLab = new QLabel("Surface Opacity: ");
	optionsTabLayout->addWidget(alphaLab, 2, 4, Qt::AlignRight);
	alphaSpin = new FloatSpinBox(-2.0, 1.0, 1, 0.1, NULL, "alpha");
	alphaSpin->setValue(0.9);
	alphaSpin->setToolTip("Alpha value for mesh etc, 0.0 (transparent) to  1.0 (opaque)\n Colour runs from 1.0 at top to alpha at bottom.\nNegative alpha causes transparency to occur before bottom of colour bar");
	optionsTabLayout->addWidget(alphaSpin, 2, 5, Qt::AlignLeft);

	showCont = new QCheckBox("Base Contours");
	optionsTabLayout->addWidget(showCont, 3, 0, Qt::AlignLeft);
	connect(showCont, SIGNAL(toggled(bool)), this, SLOT(setContours(bool)));

	showContLabel = new QCheckBox("Contour Labels");
	optionsTabLayout->addWidget(showContLabel, 3, 1, Qt::AlignLeft);
	showContLabel->setEnabled(false);
	
	QLabel * contNumLab = new QLabel("Contour interval: ");
	optionsTabLayout->addWidget(contNumLab, 3, 2, Qt::AlignRight);
	contSpace=new QLineEdit();
	contSpace->setText("Default");
	contSpace->setToolTip("contour interval as a numeric value, (increment in power of 10 if log colour scale)");
	optionsTabLayout->addWidget(contSpace, 3, 3, Qt::AlignLeft);
	
	QLabel * zOffsetLab = new QLabel("Displacement: ");
	optionsTabLayout->addWidget(zOffsetLab, 3, 4, Qt::AlignRight);
	zOffset = new FloatSpinBox(0.0, 1.0, 1, 0.1, NULL, "zOffset");
	zOffset->setValue(0.2);
	zOffset->setToolTip("Vertical offset to allow base contours to be seen");
	optionsTabLayout->addWidget(zOffset, 3, 5, Qt::AlignLeft);	
	zOffset->setEnabled(false);

	RXLog = new QCheckBox("Log Radius ");
	optionsTabLayout->addWidget(RXLog, 4, 0, Qt::AlignLeft);
	connect(RXLog, SIGNAL(toggled(bool)), this, SLOT(setLogRX()) );

	RXAuto = new QCheckBox("Auto radius");
	RXAuto->setChecked(true);
	connect(RXAuto, SIGNAL(toggled(bool)), this, SLOT(setAutoRX()));
	optionsTabLayout->addWidget(RXAuto, 4, 1, Qt::AlignRight);

	RXMinLab = new QLabel("Min Radius: ");
	optionsTabLayout->addWidget(RXMinLab, 4, 2, Qt::AlignRight);
	RXMin = new QLineEdit();
	RXMin->setMinimumWidth(30);
	RXMin->setMaximumWidth(100);
	RXMin->setToolTip("Minimum radial value (log(0) not allowed)");
	connect(RXMin, SIGNAL(textEdited(const QString& )), this, SLOT(changeRX()));
	optionsTabLayout->addWidget(RXMin, 4, 3, Qt::AlignLeft);
	
	RXMaxLab = new QLabel("Max Radius: ");
	optionsTabLayout->addWidget(RXMaxLab, 4, 4, Qt::AlignRight);
	RXMax = new QLineEdit();
	RXMax->setMinimumWidth(30);
	RXMax->setMaximumWidth(100);
	RXMax->setToolTip("Maximum radial value ");
	connect(RXMax, SIGNAL(textEdited(const QString& )), this, SLOT(changeRX()));
	optionsTabLayout->addWidget(RXMax, 4, 5, Qt::AlignLeft);
	
	
	YLog = new QCheckBox("Log Y axis");
	optionsTabLayout->addWidget(YLog, 5, 0, Qt::AlignLeft);
	connect(YLog, SIGNAL(toggled(bool)), this, SLOT(setLogY()) );
	YLog->setEnabled(false);
	
	ThYAuto = new QCheckBox("Auto angle");
	ThYAuto->setChecked(true);
	connect(ThYAuto, SIGNAL(toggled(bool)), this, SLOT(setAutoThY()));
	optionsTabLayout->addWidget(ThYAuto, 5, 1, Qt::AlignRight);

	ThYMinLab = new QLabel("Min angle: ");
	optionsTabLayout->addWidget(ThYMinLab, 5, 2, Qt::AlignRight);
	ThYMin = new QLineEdit();
	ThYMin->setMinimumWidth(30);
	ThYMin->setMaximumWidth(100);
	ThYMin->setToolTip("Minimum angle");
	connect(ThYMin, SIGNAL(textEdited(const QString& )), this, SLOT(changeThY()));
	optionsTabLayout->addWidget(ThYMin, 5, 3, Qt::AlignLeft);
	
	ThYMaxLab = new QLabel("Max angle: ");
	optionsTabLayout->addWidget(ThYMaxLab, 5, 4, Qt::AlignRight);
	ThYMax = new QLineEdit();
	ThYMax->setMinimumWidth(30);
	ThYMax->setMaximumWidth(100);
	ThYMax->setToolTip("Maximum angle for polar axis");
	connect(ThYMax, SIGNAL(textEdited(const QString& )), this, SLOT(changeThY()));
	optionsTabLayout->addWidget(ThYMax, 5, 5, Qt::AlignLeft);
		
	colourLog = new QCheckBox("Log ColourScale");
	colourLog->setChecked(true);
	connect(colourLog, SIGNAL(toggled(bool)), this, SLOT(setLogCol()));
	optionsTabLayout->addWidget(colourLog, 6, 0, Qt::AlignLeft);

	colourAuto = new QCheckBox("Auto ColourScale");
	colourAuto->setChecked(true);
	connect(colourAuto, SIGNAL(toggled(bool)), this, SLOT(autosetCol()));
	optionsTabLayout->addWidget(colourAuto, 6, 1, Qt::AlignRight);

	QLabel * colMinLab = new QLabel("Min value: ");
	optionsTabLayout->addWidget(colMinLab, 6, 2, Qt::AlignRight);
	colMin = new QLineEdit();
	colMin->setMinimumWidth(50);
	colMin->setMaximumWidth(100);
	colMin->setText("0.0");
	colMin->setToolTip("Minimum data value (log(0) not allowed)");
	connect(colMin, SIGNAL(textEdited(const QString & )), this, SLOT(setColUser()));
	optionsTabLayout->addWidget(colMin, 6, 3, Qt::AlignLeft);

	QLabel * colMaxLab = new QLabel("Max value: ");
	optionsTabLayout->addWidget(colMaxLab, 6, 4, Qt::AlignRight);
	colMax = new QLineEdit();
	colMax->setMinimumWidth(50);
	colMax->setMaximumWidth(100);
	colMax->setToolTip("Maximum data value (log(0) not allowed)");
	colMax->setText("1.0e6");
	connect(colMax, SIGNAL(textEdited(const QString & )), this, SLOT(setColUser()));
	optionsTabLayout->addWidget(colMax, 6, 5, Qt::AlignLeft);

	QLabel* colour_table_label = new QLabel("Colour Table: ");
	optionsTabLayout->addWidget(colour_table_label, 7, 0, Qt::AlignRight);
    colourTable = new ColourTableButton;
    optionsTabLayout->addWidget(colourTable, 7, 1, Qt::AlignLeft);

	QLabel * colLabLab = new QLabel("Colour bar label: ");
	optionsTabLayout->addWidget(colLabLab, 7, 2, Qt::AlignRight);
	
	colLabSlot = new QLineEdit();
	colLabSlot->setMinimumWidth(200);
	connect(colLabSlot, SIGNAL(textEdited(const QString& )), this, SLOT(setUserColLab()));
	optionsTabLayout->addWidget(colLabSlot, 7, 3, 1, 3);

    QLabel* line_width_label=new QLabel("Line widths: ");
    optionsTabLayout->addWidget(line_width_label,8,0,Qt::AlignRight);
    line_width_fld = new FloatSpinBox(0.1,20.,1,0.1, NULL, "line_width");
    line_width_fld->setValue(QSAS_profiles->current()->LineWidth);
    line_width_fld->setFixedWidth(50);
    optionsTabLayout->addWidget(line_width_fld,8,1,Qt::AlignLeft);
    
	QLabel* title_size_label = new QLabel("Title font size: ");
	optionsTabLayout->addWidget(title_size_label, 8, 2, Qt::AlignRight);
	title_size_fld = new FloatSpinBox(0.01, 40.,1, .1, NULL, "title_char_size");
	title_size_fld->setFixedWidth(50);
	optionsTabLayout->addWidget(title_size_fld, 8, 3, Qt::AlignLeft);
	title_size_fld->setValue(QSAS_profiles->current()->BaseFontSize+0.3);

	QLabel* text_size_label = new QLabel("Text font size: ");
	optionsTabLayout->addWidget(text_size_label, 8, 4, Qt::AlignRight);
	text_size_fld = new FloatSpinBox(0.01, 40.,1, .1, NULL, "text_char_size");
	text_size_fld->setFixedWidth(50);
	optionsTabLayout->addWidget(text_size_fld, 8, 5, Qt::AlignLeft);
	text_size_fld->setValue(QSAS_profiles->current()->BaseFontSize+0.2);



	tab_widget->addTab(optionsTab, "Plot Options");
	
	// Page Tab
	// ----------
	QWidget *pageTab=new QWidget();

	QVBoxLayout *page_properties_top_layout=new QVBoxLayout(pageTab);
	page_properties_top_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	page_properties_top_layout->setMargin(QSAS_profiles->current()->GUImargins);

	QGroupBox *page_size_groupbox = new QGroupBox(pageTab);
	page_size_groupbox->setTitle("Size");

	QGridLayout *page_size_layout = new QGridLayout;
	page_size_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	page_size_layout->setMargin(QSAS_profiles->current()->GUImargins);

	QLabel* page_size_label = new QLabel("Page size: ");
	page_size_layout->addWidget(page_size_label, 1, 0, Qt::AlignRight);
	page_size = new QComboBox();
	page_size->insertItem(QPUI_A4, "A4");
	page_size->insertItem(QPUI_CUSTOM, "Custom");
	page_size->insertItem(QPUI_EXECUTIVE, "Executive");
	page_size->insertItem(QPUI_LEGAL, "Legal");
	page_size->insertItem(QPUI_LETTER, "Letter");
	page_size->insertItem(QPUI_DEFAULT, "Default");

	QFontMetrics page_size_fm(page_size->font());
	int page_size_w = page_size_fm.width("Executive");
	int page_size_h = page_size_fm.height();
	page_size->setMinimumSize(page_size_w, page_size_h);

	page_size->setToolTip("Select page size");
	page_size_layout->addWidget(page_size, 1, 1, Qt::AlignLeft);
	
	if(QSAS_profiles->current()->PageSize=="A4")
		page_size->setCurrentIndex(QPUI_A4);
	else if(QSAS_profiles->current()->PageSize=="Executive")
		page_size->setCurrentIndex(QPUI_EXECUTIVE);
	else if(QSAS_profiles->current()->PageSize=="Legal")
		page_size->setCurrentIndex(QPUI_LEGAL);
	else if(QSAS_profiles->current()->PageSize=="Letter")
		page_size->setCurrentIndex(QPUI_LETTER);
	else if(QSAS_profiles->current()->PageSize=="Default")
		page_size->setCurrentIndex(QPUI_DEFAULT);
	else page_size->setCurrentIndex(QPUI_CUSTOM);
	
	QLabel* units_label = new QLabel("Units: ");
	page_size_layout->addWidget(units_label, 1, 2, Qt::AlignRight);
	pageUnits = new QComboBox();
	pageUnits->insertItem(QPUI_PIXELS, "pixels");
	pageUnits->insertItem(QPUI_INCHES, "inches");
	pageUnits->insertItem(QPUI_CM, "cm");
	pageUnits->insertItem(QPUI_MM, "mm");
	pageUnits->insertItem(QPUI_PSPOINTS, "PS points");
	QFontMetrics units_fm(pageUnits->font());
	int units_w = units_fm.width("PS points");
	int units_h = page_size_fm.height();
	pageUnits->setMinimumSize(units_w, units_h);

	pageUnits->setToolTip("Select units for page size");
	page_size_layout->addWidget(pageUnits, 1, 3, Qt::AlignLeft);

	
	oldUnits = pageUnits->currentIndex();
	
	QLabel* dpi_label = new QLabel("Resolution: ");
	page_size_layout->addWidget(dpi_label, 1, 4, Qt::AlignRight);
	dpiPage = new QComboBox();
	dpiPage->insertItem(QPUI_72_DPI, "72 dpi");
	dpiPage->insertItem(QPUI_150_DPI, "150 dpi");
	dpiPage->insertItem(QPUI_300_DPI, "300 dpi");
	dpiPage->insertItem(QPUI_600_DPI, "600 dpi");
	dpiPage->insertItem(QPUI_1200_DPI, "1200 dpi");
	switch(QSAS_profiles->current()->PageDPI)
	{
		case 72: default:
			dpiPage->setCurrentIndex(QPUI_72_DPI);
			break;
		case 150:
			dpiPage->setCurrentIndex(QPUI_150_DPI);
			break;
		case 300:
			dpiPage->setCurrentIndex(QPUI_300_DPI);
			break;
		case 600:
			dpiPage->setCurrentIndex(QPUI_600_DPI);
			break;
		case 1200:
			dpiPage->setCurrentIndex(QPUI_1200_DPI);
			break;
	}

	page_size_layout->addWidget(dpiPage, 1, 5, Qt::AlignLeft);
	
	QLabel* width_label = new QLabel("Width: ");
	page_size_layout->addWidget(width_label, 3, 0, Qt::AlignRight);
	pageWidth = new QLineEdit;
	pageWidth->setFixedWidth(100);
	pageWidth->setValidator(new QDoubleValidator(pageWidth));

	page_size_layout->addWidget(pageWidth, 3, 1, Qt::AlignLeft);

	QLabel* height_label = new QLabel("Height: ");
	page_size_layout->addWidget(height_label, 3, 2, Qt::AlignRight);
	pageHeight = new QLineEdit;
	pageHeight->setFixedWidth(100);
	pageHeight->setValidator(new QDoubleValidator(pageHeight));
	page_size_layout->addWidget(pageHeight, 3, 3, Qt::AlignLeft);
	
	
	// set appropriate text in width and height fields
	if(QSAS_profiles->current()->PageSize=="Custom")
	{
		width = QSAS_profiles->current()->PageWidth;
		height = QSAS_profiles->current()->PageHeight;
		pageWidth->setText(QString().setNum(width));
		pageHeight->setText(QString().setNum(height));
	}

	QLabel* orientation_label = new QLabel("Orientation: ");
	page_size_layout->addWidget(orientation_label, 5, 0, Qt::AlignRight);
	pageOrientation = new QComboBox();
	pageOrientation->insertItem(QPUI_PORTRAIT, "Portrait");
	pageOrientation->insertItem(QPUI_LANDSCAPE, "Landscape");
	QFontMetrics orientation_fm(pageOrientation->font());
	int orientation_w = orientation_fm.width("Landscape");
	int orientation_h = orientation_fm.height();
	pageOrientation->setMinimumSize(orientation_w, orientation_h);

	pageOrientation->setToolTip("Select portrait or landscape");
	page_size_layout->addWidget(pageOrientation, 5, 1, Qt::AlignLeft);
	
	if(QSAS_profiles->current()->PageOrientation=="Landscape")
		pageOrientation->setCurrentIndex(QPUI_LANDSCAPE);
	else pageOrientation->setCurrentIndex(QPUI_PORTRAIT);


	
	page_size_groupbox->setLayout(page_size_layout);

	page_properties_top_layout->addWidget(page_size_groupbox);

	QGroupBox *page_output_groupbox = new QGroupBox(this);

	page_output_groupbox->setTitle("Plot Window");

	QGridLayout *page_output_layout = new QGridLayout;
	page_output_layout->setSpacing(QSAS_profiles->current()->GUImargins);
	page_output_layout->setMargin(QSAS_profiles->current()->GUImargins);


    newWindow = new QCheckBox("New window for each plot");
    page_output_layout->addWidget(newWindow,0,0,Qt::AlignLeft);
    
    plotWin_id = new PlotID("Surface Plot");
    page_output_layout->addWidget(plotWin_id, 0, 3, Qt::AlignLeft);
    
	
	page_output_groupbox->setLayout(page_output_layout);

	page_properties_top_layout->addWidget(page_output_groupbox);


	tab_widget->addTab(pageTab, "Page");

	tab_widget->adjustSize();

	// feedback text area
	textArea = new QTextEdit(_top_splitter);
	textArea->setReadOnly(true);


	// set up menu bar
	QMenu* file_menu = 	menuBar()->addMenu("&File");
	menuBar()->addSeparator();
	QMenu* help_menu = 	menuBar()->addMenu("&Help" );


	file_menu->clear();
    file_menu->addAction("Save As XML...", this, SLOT(saveAsXML()));
	file_menu->addSeparator();
    file_menu->addAction("Restore XML...", this, SLOT(restoreXML()));

	help_menu->clear();
	help_menu->addAction("QSAS &Manual", this, SLOT( launchManual() ), Qt::CTRL+Qt::Key_M );
	help_menu->addSeparator();
	help_menu->addAction("Surface Plot Help", this, SLOT( launchSurfacePlotHelp() ));
	help_menu->addSeparator();
	help_menu->addAction("Module Version", this, SLOT( launchVersion() ));

	_status_bar = statusBar();
	_status_bar->showMessage(" ");

	QWidget * buttonsWidget = new QWidget(_top_splitter);
	QHBoxLayout *buttons_layout = new QHBoxLayout(buttonsWidget);

	plotBtn = new QPushButton("Plot",this);
	plotBtn->setMinimumSize(plotBtn->sizeHint());
	plotBtn->setFocusPolicy(Qt::StrongFocus);
	connect(plotBtn, SIGNAL(clicked()), this, SLOT(doPlot()));
	buttons_layout->addWidget(plotBtn);
	plotBtn->setDefault(true);
	
	dismissBtn = new QPushButton( "Dismiss", this);
	dismissBtn->setMinimumSize(dismissBtn->sizeHint());
	dismissBtn->setFocusPolicy(Qt::StrongFocus);
	connect( dismissBtn, SIGNAL(clicked()), this, SLOT(close()) );
	buttons_layout->addWidget(dismissBtn);
	
	this->setSizePolicy(QSizePolicy(QSizePolicy::Expanding,QSizePolicy::Expanding));


	_top_splitter->setStretchFactor(0, 10);
	_top_splitter->setStretchFactor(1, 10);
	_top_splitter->setStretchFactor(1, 1);
	
	adjustSize();
	userThLab = false;
	userRLab = false;
	userColLab = false;
	
	// ensure pointers are NULL for trap on free()
	z = 0;
	zLog = 0;
	nr = 0;
	ntheta = 0;
	cgrid.nx = 0;
	cgrid.ny = 0;
	r = 0;
	r_plus = 0;
	r_minus = 0;
	theta = 0;
	theta_plus = 0;
	theta_minus = 0;

	connect(dpiPage, SIGNAL(currentIndexChanged(int)), this, SLOT(setPage()));
	connect(pageUnits, SIGNAL(currentIndexChanged(int)), this, SLOT(setPage()));
	connect(page_size, SIGNAL(currentIndexChanged(int)), this, SLOT(setPage()));
	setPage();
	
	setGUItoPolar(true);
}

// ------------------------

void QsurfaceWindow::setGUItoPolar(bool isPolar)
{
	if(isPolar){
		RXLog->setText(QString("Log Radius"));
		RXLog->setChecked(true);
		RXAuto->setText(QString("Auto radius"));
		ThYAuto->setText(QString("Auto angle"));
		YLog->setEnabled(false);
		RXMinLab->setText("Min Radius: ");
		RXMaxLab->setText("Max Radius: ");
		ThYMinLab->setText("Min angle: ");
		ThYMaxLab->setText("Max angle: ");
		RaxisLab->setText("R axis Text: ");
		ThAxisLab->setText("Angle axis Text: ");
		swapR->setText("Swap R-theta");
		dataSlot->set3DSubsetDef("rp");
	}
	else{
		RXLog->setText(QString("Log X"));
		RXLog->setChecked(false);
		RXAuto->setText(QString("Auto X range"));
		ThYAuto->setText(QString("Auto Y range"));
		YLog->setEnabled(true);
		RXMinLab->setText("Min X: ");
		RXMaxLab->setText("Max X: ");
		ThYMinLab->setText("Min Y: ");
		ThYMaxLab->setText("Max Y: ");
		RaxisLab->setText("X axis Text: ");
		ThAxisLab->setText("Y axis Text: ");
		swapR->setText("Swap X-Y");
		dataSlot->set3DSubsetDef("tp");
	}
}

// ------------------------

QsurfaceWindow::~QsurfaceWindow()
{
	if (z != 0) plFree2dGrid(z, nr, ntheta+1); // z is always r,theta even if dobj is theta,r
		
	if(nr != 0 && ntheta != 0) {
		plFree2dGrid(cgrid.xg, nr, ntheta+1);
		plFree2dGrid(cgrid.yg, nr, ntheta+1);
	}
	
	if (r != 0) delete[] r;
	if (r_plus != 0) delete[] r_plus;
	if (r_minus != 0) delete[] r_minus;
	if (theta != 0) delete[] theta;
	if (theta_plus != 0) delete[] theta_plus;
	if (theta_minus != 0) delete[] theta_minus;

}

// ------------------------

void QsurfaceWindow::setPlotType(int type)
{
	if(type == 3) {
		showCont->setChecked(false);
		showCont->setEnabled(false);
	}
	else showCont->setEnabled(true);
}
// ------------------------

void QsurfaceWindow::setAxisOpt()
{
	if(dobj.is_nil() ) {
		// handle case of changing toggle before data present
		bool polar = true;
		if(axisType->currentIndex() == 2) polar = false;
		setGUItoPolar(polar);
		return;
	}
	
	if( !initData() ) { // this will call setGUItoPolar()
		QsurfaceAppendTextDisplay("Data object needs to be at least 2D with Depend_1 and Depend_2");
		return;
	}

	setData();
	setTitle();
	this->autosetCol();
	this->autosetRX();
	this->autosetThY();
}

// ------------------------

void QsurfaceWindow::setDataObj()
{
	if( !initData() ) { // this will call setGUItoPolar()
		QsurfaceAppendTextDisplay("Data object needs to be at least 2D with Depend_1 and Depend_2");
		return;
	}
	setData();
	setTitle();
	this->autosetCol();
	this->autosetRX();
	this->autosetThY();
}


// ------------------------

void QsurfaceWindow::setRecord(bool andPlot)
{
	setData();
	setTitle();
	this->autosetCol();
	this->autosetRX();
	this->autosetThY();
	if(andPlot) doPlot(true);
}


//--------------

void QsurfaceWindow::setSwapRTheta()
{
	userThLab = false;
	userRLab = false;
	if( !initData() ) { // this will call setGUItoPolar()
		QsurfaceAppendTextDisplay("Data object needs to be at least 2D with Depend_1 and Depend_2");
		return;
	}
	setData();
	autosetCol();
	autosetRX();
	autosetThY();
	setTitle();
	
}

// ------------------------

void QsurfaceWindow::autosetRX()
{

	if(dobj.is_nil()) return;
	
	if(RXAuto->isChecked())
	{	
		if(RXLog->isChecked()) RXMax->setText( QString("%1").arg(pow(10., rmax)*userRmin) );
		else RXMax->setText( QString("%1").arg(rmax) );
		
		if(RXLog->isChecked()) RXMin->setText(QString("%1").arg(rminNonZero));
		else RXMin->setText(QString("%1").arg(rmin));	
	}
}

// ------------------------

void QsurfaceWindow::autosetThY()
{

	if(dobj.is_nil()) return;
	
	if(ThYAuto->isChecked())
	{
		if(YLog->isChecked()) ThYMax->setText( QString("%1").arg((thetamax + thetamin)*rad2deg) );
		else ThYMax->setText( QString("%1").arg(thetamax*rad2deg ) );

		if(YLog->isChecked()) ThYMin->setText(QString("%1").arg(thetaminNonZero*rad2deg));
		else ThYMin->setText(QString("%1").arg(thetamin*rad2deg));
	}
}

// ------------------------

void QsurfaceWindow::autosetCol()
{

	if(dobj.is_nil()) return;
	
		
	if(colourAuto->isChecked())
	{
		// set min/max
		//plMinMax2dGrid(z, nr, ntheta, &zmax, &zmin);
				
		if(colourLog->isChecked()){
			nonZeroMinMax();
		}
		else getMinMax();

		colMax->setText(QString("%1").arg(zmax));
		colMin->setText(QString("%1").arg(zmin));
	}
}

// ------------------------

void QsurfaceWindow::changeView()
{
	// re-render viewed from new angles
	doPlot(true);

}

void QsurfaceWindow::nonZeroMinMax()
{
	// used for log colour scale
	if(z == 0) return;
	
	zmin = PLFLT_MAX; // initialise to high value
	zmax = PLFLT_MIN; // initialise to safe value
	
	
	PLFLT val;
	for(int i=0; i<nr; i++){
		for(int j=0; j<ntheta; j++){
			val = z[i][j];
			if( !isfinite(val) ) continue;
			if(val < zmin && val > PLFLT_MIN) zmin = val;
			if(val > zmax ) zmax = val;
		}
	}
	
	if(zmax - zmin < PLFLT_MIN) zmin = 0.5*zmax;

}
// ------------------------

void QsurfaceWindow::getMinMax()
{
	if(z == 0) return;

	// used for linear colour scale
	zmin = PLFLT_MAX; // initialise to high value
	zmax = PLFLT_MIN; // initialise to safe value
	
	PLFLT val;
	for(int i=0; i<nr; i++){
		for(int j=0; j<ntheta; j++){
			val = z[i][j];
			if( !isfinite(val) ) continue;
			if(val < zmin ) zmin = val;
			if(val > zmax ) zmax = val;
		}
	}
	if(zmax - zmin < PLFLT_MIN) zmax += 1.;

}
// ------------------------

void QsurfaceWindow::setColUser()
{
	
	colourAuto->setChecked(false);

}

// ------------------------

void QsurfaceWindow::setContours(bool on)
{
	if(on){
		zOffset->setEnabled(true);
		showContLabel->setEnabled(true);
	}
	else{
		zOffset->setEnabled(false);
		showContLabel->setEnabled(false);
	}

}

// ------------------------

void QsurfaceWindow::setLogRX()
{
	// must handle RTh or XY case
	
	if(rmin <= 0) RXAuto->setChecked(true); // ensure negative values get reset accordingly
	
	setData();
	setTitle();
	this->autosetCol();
	this->autosetRX();
	this->autosetThY();
}

// ------------------------

void QsurfaceWindow::setLogY()
{
	// must handle RTh or XY case
	
	if(thetamin <= 0) ThYAuto->setChecked(true); // ensure negative values get reset accordingly
	
	setData();
	setTitle();
	this->autosetCol();
	this->autosetRX();
	this->autosetThY();
}

// ------------------------

void QsurfaceWindow::setLogCol()
{
	bool OK;
	double min = colMin->text().toDouble(&OK);
	if( ( !OK || min <= 1.0e-30 ) && colourLog->isChecked() ) 
	{
		colourAuto->setChecked(true); // ensure negative values get reset accordingly
	
		setData();
		setTitle();
		this->autosetCol();
	}
		
	contSpace->setText(QString("Default"));
		
}

// ------------------------

void QsurfaceWindow::setAutoRX()
{	
	setData();
	setTitle();
	this->autosetCol();
	this->autosetRX();
	this->autosetThY();
}

// ------------------------

void QsurfaceWindow::setAutoThY()
{	
	setData();
	setTitle();
	this->autosetCol();
	this->autosetRX();
	this->autosetThY();
}

// ------------------------

void QsurfaceWindow::changeRX()
{
	
	RXAuto->setChecked(false);
	setData();
	setTitle();
	this->autosetCol();

}


// ------------------------

void QsurfaceWindow::changeThY()
{
	
	ThYAuto->setChecked(false);
	setData();
	setTitle();
	this->autosetCol();

}




void QsurfaceWindow::setTitle()
{
	if(dobj.is_nil()) return;
	
	setColLab(); // also set default col label
	setAxisLabels();
	
	if(autoTitle->isChecked()) 
	{

		string title = dataSlot->text().toLatin1().data();
		int posn = title.rfind("/");
		if (posn != string::npos) title = title.substr(posn+1);
		titleSlot->setText(title.c_str());

	}
	timeText = "";
	if( atSlider->record <= atSlider->maxRecord){
		timeText = atSlider->ttags->asStr(atSlider->record);
	}
}

void QsurfaceWindow::setUserTitle()
{
	autoTitle->setChecked(false);

}


void QsurfaceWindow::setUserRXLab()
{
	userRLab = true;

}


void QsurfaceWindow::setUserThYLab()
{
	userThLab = true;

}


void QsurfaceWindow::setAxisLabels()
{
	if(thetaDep.is_nil()) return;

	if(!userThLab){
		string thetalabel;
		if (thetaDep->xref_exists(LABLAXIS)) {
			thetalabel = string(thetaDep->getXrefText(LABLAXIS));
		} 
		else{
			if(axisType->currentIndex() == 1) thetalabel = string("Angle");
			else if(axisType->currentIndex() == 2) thetalabel = string("Y");
			else thetalabel = string("");
		}
		
		if (thetaDep->xref_exists(UNITS)) {
			thetalabel += " (";
			thetalabel += string(thetaDep->getXrefText(UNITS));
			thetalabel += ")";
		}

		ThYLabSlot->setText(thetalabel.c_str());

	}
	
	if(!userRLab)
	{	
		string rlabel;

		if (rDep->xref_exists(LABLAXIS)) {
			rlabel = string(rDep->getXrefText(LABLAXIS));
		}
		else {
			if(axisType->currentIndex() == 1) rlabel = string("r");
			else if(axisType->currentIndex() == 2) rlabel = string("X");
			else rlabel = string("");
		}
		 
		
		if (rDep->xref_exists(UNITS)) {
			rlabel += " (";
			rlabel += string(rDep->getXrefText(UNITS));
			rlabel += ")";
		}

		RXLabSlot->setText(rlabel.c_str());

	}

	
}

void QsurfaceWindow::setColLab()
{
	if(dobj.is_nil()) return;
	
	if(userColLab) return;

	string zlabel;
	
	if (dobj->xref_exists(LABLAXIS)) {
		zlabel = string(dobj->getXrefText(LABLAXIS));
	}
	else zlabel = string("z");
	
	if (dobj->xref_exists(UNITS)){
		zlabel += " (";
		zlabel += string(dobj->getXrefText(UNITS));
		zlabel += ")";
	}

	colLabSlot->setText(zlabel.c_str());


}

void QsurfaceWindow::setUserColLab()
{
	userColLab = true;

}

void QsurfaceWindow::keyPressEvent(QKeyEvent * e )
{  
	if(QuSlotDebug())
	{
		printf("QsurfaceWindow::keyPressEvent \n");
		fflush(stdout);
	}

	// use escape key to toggle default buttons
	if(e->key() == Qt::Key_Escape){
		if(   plotBtn->isDefault()) {
			plotBtn->setDefault(false);
			dismissBtn->setDefault(true);
		}
		else if(dismissBtn->isDefault())  {
			dismissBtn->setDefault(false);
			plotBtn->setDefault(true);
		}
		escPushed = true;

		return;
	}
	// use return or enter key to perform default
	if(e->key() == Qt::Key_Return ||  e->key() == Qt::Key_Enter) {
		if(plotBtn->isDefault()) doPlot();
		else if (dismissBtn->isDefault()) close();
		return;
	}

	// else pass on to list view 
	e->ignore(); // let selection change

}




// --------------------------------------

void QsurfaceWindow::launchSurfacePlotHelp()
{
	// slot invoked from Plotting Help menu sub item
	// launches browser showing Qsas Plot help docuumentation

	if(QuSlotDebug())
	{
		printf("QsurfaceWindow::launchSurfacePlotHelp \n");
		fflush(stdout);
	}

	qsasWindow->help_window->manEntry("QSAS_surface.html");
}

void QsurfaceWindow::launchManual()
{
	// slot invoked from Qsas Manual Help menu sub item
	// launches browser showing Qsas help index

	if(QuSlotDebug())
	{
		printf("QsurfaceWindow::launchManual \n");
		fflush(stdout);
	}

	qsasWindow->help_window->manEntry("QSAS_help_index.html");
}

void QsurfaceWindow::launchVersion()
{
	// slot invoked from Module Version Help menu sub item
	// launches browser showing Qsas version numbers

	if(QuSlotDebug())
	{
		printf("QsurfaceWindow::launchVersion \n");
		fflush(stdout);
	}

	QString text;
	text  = QString("Version Numbers...\n");
	text += QString(QSAS_VERSION_ID);
	text += "\n Surface Plot ";
	text += QString(QSURFACE_VERSION);
	QMessageBox::information( this, "QSAS information", text.toLatin1().data(), "OK" );
}


QDomElement QsurfaceWindow::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Tabs");
    
    xml.appendChild(createTextXMLNode(doc, "autoTitle", autoTitle->isChecked()));
    xml.appendChild(createTextXMLNode(doc, "userRLab", userRLab));
    xml.appendChild(createTextXMLNode(doc, "userThLab", userThLab));
    xml.appendChild(createTextXMLNode(doc, "userColLab", userColLab));
    xml.appendChild(createTextXMLNode(doc, "userRmax", userRmax));
    xml.appendChild(createTextXMLNode(doc, "userRmin", userRmin));
    xml.appendChild(createTextXMLNode(doc, "userThmax", userThmax));
    xml.appendChild(createTextXMLNode(doc, "userThmin", userThmin));


    xml.appendChild(createTextXMLNode(doc, "dataSlot", dataSlot->text()));
	xml.appendChild(createTextXMLNode(doc, "atSlider", atSlider->record));
    xml.appendChild(createTextXMLNode(doc, "swapR", swapR->isChecked()));
    if(elevation) xml.appendChild(createTextXMLNode(doc, "elevation", elevation->value()));
    if(azimuth) xml.appendChild(createTextXMLNode(doc, "azimuth", azimuth->value()));
    if(delay) xml.appendChild(createTextXMLNode(doc, "delay", delay->value()));

    xml.appendChild(createTextXMLNode(doc, "titleSlot", titleSlot->text()));
    xml.appendChild(createTextXMLNode(doc, "RXLabSlot", RXLabSlot->text()));
    xml.appendChild(createTextXMLNode(doc, "ThYLabSlot", ThYLabSlot->text()));
	
    xml.appendChild(createTextXMLNode(doc, "plotType", plotType->currentIndex()));
    xml.appendChild(createTextXMLNode(doc, "axisType", axisType->currentIndex()));
    xml.appendChild(createTextXMLNode(doc, "alphaSpin", alphaSpin->value()));

    xml.appendChild(createTextXMLNode(doc, "showCont", showCont->isChecked()));
    xml.appendChild(createTextXMLNode(doc, "showContLabel", showContLabel->isChecked()));
    xml.appendChild(createTextXMLNode(doc, "contSpace", contSpace->text()));
	xml.appendChild(createTextXMLNode(doc, "zOffset", zOffset->value()));

    xml.appendChild(createTextXMLNode(doc, "RXLog", RXLog->isChecked()));
    xml.appendChild(createTextXMLNode(doc, "RXAuto", RXAuto->isChecked()));
    xml.appendChild(createTextXMLNode(doc, "RXMin", RXMin->text()));
    xml.appendChild(createTextXMLNode(doc, "RXMax", RXMax->text()));
	
    xml.appendChild(createTextXMLNode(doc, "YLog", YLog->isChecked()));
    xml.appendChild(createTextXMLNode(doc, "ThYAuto", ThYAuto->isChecked()));
    xml.appendChild(createTextXMLNode(doc, "ThYMin", ThYMin->text()));
    xml.appendChild(createTextXMLNode(doc, "ThYMax", ThYMax->text()));

	xml.appendChild(createTextXMLNode(doc, "RXMinLab", RXMinLab->isEnabled()));
	xml.appendChild(createTextXMLNode(doc, "RXMaxLab", RXMaxLab->isEnabled()));
	xml.appendChild(createTextXMLNode(doc, "ThYMinLab", ThYMinLab->isEnabled()));
	xml.appendChild(createTextXMLNode(doc, "ThYMaxLab", ThYMaxLab->isEnabled()));
	xml.appendChild(createTextXMLNode(doc, "RaxisLab", RaxisLab->isEnabled()));
	xml.appendChild(createTextXMLNode(doc, "ThAxisLab", ThAxisLab->isEnabled()));


    xml.appendChild(createTextXMLNode(doc, "colourLog", colourLog->isChecked()));
    xml.appendChild(createTextXMLNode(doc, "colourAuto", colourAuto->isChecked()));
    xml.appendChild(createTextXMLNode(doc, "colMin", colMin->text()));
    xml.appendChild(createTextXMLNode(doc, "colMax", colMax->text()));


    ColourTable ct=colourTable->getCurrentTable();

    int _colour_table_type;
    int _colour_table_id;
    if(stdPalette.contains(ct))
    {
        _colour_table_type=0;
        _colour_table_id=stdPalette.indexOf(ct);
    }
    else
    {
        _colour_table_type=1;
        _colour_table_id=savedPalette.indexOf(ct);
    }
    xml.appendChild(createTextXMLNode(doc, "Colour_Table_Type", _colour_table_type));
    xml.appendChild(createTextXMLNode(doc, "Colour_Table_ID", _colour_table_id));


    xml.appendChild(createTextXMLNode(doc, "colLabSlot", colLabSlot->text()));

    xml.appendChild(createTextXMLNode(doc, "page_size", page_size->currentIndex()));
    xml.appendChild(createTextXMLNode(doc, "pageUnits", pageUnits->currentIndex()));
    xml.appendChild(createTextXMLNode(doc, "dpiPage", dpiPage->currentIndex()));
	
    xml.appendChild(createTextXMLNode(doc, "pageWidth", pageWidth->text()));
    xml.appendChild(createTextXMLNode(doc, "pageHeight", pageHeight->text()));
	
    xml.appendChild(createTextXMLNode(doc, "pageOrientation", pageOrientation->currentIndex()));

    return xml;
}


void QsurfaceWindow::fromXMLElement(QDomElement& el)
{
    QString s;

	if(getXMLNodeText(el, "autoTitle", s))
    {
		autoTitle->setChecked(s.toInt());
    }
	if(getXMLNodeText(el, "userRLab", s))
    {
		userRLab = s.toInt();
    }
	if(getXMLNodeText(el, "userThLab", s))
    {
		userThLab = s.toInt();
    }
	if(getXMLNodeText(el, "userColLab", s))
    {
		userColLab = s.toInt();
    }

	if(getXMLNodeText(el, "userRmax", s))
    {
		userRmax = s.toDouble();
    }
	if(getXMLNodeText(el, "userRmin", s))
    {
		userRmin = s.toDouble();
    }

	if(getXMLNodeText(el, "userThmax", s))
    {
		userThmax = s.toDouble();
    }
	if(getXMLNodeText(el, "userThmin", s))
    {
		userThmin = s.toDouble();
    }

	if(getXMLNodeText(el, "swapR", s))
    {
        swapR->setChecked(s.toInt());
    }

	if(getXMLNodeText(el, "titleSlot", s))
    {
        titleSlot->setText(s);
    }
	if(getXMLNodeText(el, "RXLabSlot", s))
    {
        RXLabSlot->setText(s);
    }
	if(getXMLNodeText(el, "ThYLabSlot", s))
    {
        ThYLabSlot->setText(s);
    }


	if(getXMLNodeText(el, "plotType", s))
    {
        plotType->setCurrentIndex(s.toInt());
    }
	if(getXMLNodeText(el, "axisType", s))
    {
        axisType->setCurrentIndex(s.toInt());
    }
	if(getXMLNodeText(el, "alphaSpin", s))
    {
        alphaSpin->setValue(s.toDouble());
    }

	if(getXMLNodeText(el, "elevation", s))
    {
        if(elevation){
            elevation->blockSignals(true);
            elevation->setValue(s.toInt());
            elevation->blockSignals(false);
        }
        else startElevation = s.toInt();
    }

	if(getXMLNodeText(el, "azimuth", s))
    {
        if(azimuth){
            azimuth->blockSignals(true);
            azimuth->setValue(s.toInt());
            azimuth->blockSignals(false);
        }
        else startAzimuth = s.toInt();
    }

	if(getXMLNodeText(el, "delay", s))
    {
        if(delay){
            delay->setValue(s.toInt());
        }
    }


	if(getXMLNodeText(el, "showCont", s))
    {
        showCont->setChecked(s.toInt());
    }
	if(getXMLNodeText(el, "contSpace", s))
    {
        contSpace->setText(s);
    }
	if(getXMLNodeText(el, "showContLabel", s))
    {
        showContLabel->setChecked(s.toInt());
    }
	if(getXMLNodeText(el, "zOffset", s))
    {
        zOffset->setValue(s.toDouble());
    }
	

	if(getXMLNodeText(el, "RXLog", s))
    {
        RXLog->setChecked(s.toInt());
    }
	if(getXMLNodeText(el, "RXAuto", s))
    {
        RXAuto->setChecked(s.toInt());
    }
	if(getXMLNodeText(el, "RXMin", s))
    {
        RXMin->setText(s);
    }
	if(getXMLNodeText(el, "RXMax", s))
    {
        RXMax->setText(s);
    }
	
	if(getXMLNodeText(el, "YLog", s))
    {
        YLog->setChecked(s.toInt());
    }
	if(getXMLNodeText(el, "ThYAuto", s))
    {
        ThYAuto->setChecked(s.toInt());
    }
	if(getXMLNodeText(el, "ThYMin", s))
    {
        ThYMin->setText(s);
    }
	if(getXMLNodeText(el, "ThYMax", s))
    {
        ThYMax->setText(s);
    }
	
	if(getXMLNodeText(el, "RXMinLab", s))
    {
        RXMinLab->setEnabled(s.toInt());
    }
	if(getXMLNodeText(el, "RXMaxLab", s))
    {
        RXMaxLab->setEnabled(s.toInt());
    }
	if(getXMLNodeText(el, "ThYMinLab", s))
    {
        ThYMinLab->setEnabled(s.toInt());
    }
	if(getXMLNodeText(el, "ThYMaxLab", s))
    {
        ThYMaxLab->setEnabled(s.toInt());
    }
	if(getXMLNodeText(el, "RaxisLab", s))
    {
        RaxisLab->setEnabled(s.toInt());
    }
	if(getXMLNodeText(el, "ThAxisLab", s))
    {
        ThAxisLab->setEnabled(s.toInt());
    }


	if(getXMLNodeText(el, "colourLog", s))
    {
        colourLog->setChecked(s.toInt());
    }
	if(getXMLNodeText(el, "colourAuto", s))
    {
        colourAuto->setChecked(s.toInt());
    }
	if(getXMLNodeText(el, "colMin", s))
    {
        colMin->setText(s);
    }
	if(getXMLNodeText(el, "colMax", s))
    {
        colMax->setText(s);
    }


    int _colour_table_type(0);
    int _colour_table(0);
    if(getXMLNodeText(el, "Colour_Table_ID", s))
    {
        _colour_table=s.toInt();
    }
    if(getXMLNodeText(el, "Colour_Table_Type", s))
    {
        _colour_table_type=(s.toInt());
    }
    if(_colour_table_type!=0)
    {
        colourTable->setSessionPalette(_colour_table+importedPaletteOffset);
    }
    else
    {
        colourTable->setStdPalette(_colour_table);
    }


 	if(getXMLNodeText(el, "colLabSlot", s))
    {
        colLabSlot->setText(s);
    }
	
	if(getXMLNodeText(el, "page_size", s))
    {
        page_size->setCurrentIndex(s.toInt());
    }
	if(getXMLNodeText(el, "pageUnits", s))
    {
        pageUnits->setCurrentIndex(s.toInt());
    }
	if(getXMLNodeText(el, "dpiPage", s))
    {
        dpiPage->setCurrentIndex(s.toInt());
    }	
		
	if(getXMLNodeText(el, "pageWidth", s))
    {
        pageWidth->setText(s);
    }
	if(getXMLNodeText(el, "pageHeight", s))
    {
        pageHeight->setText(s);
    }

		if(getXMLNodeText(el, "pageOrientation", s))
    {
        pageOrientation->setCurrentIndex(s.toInt());
    }


	if(getXMLNodeText(el, "dataSlot", s))
    {
        dataSlot->setText(s);
		dataSlot->reValidate(true);
    }

	if(getXMLNodeText(el, "atSlider", s))
    {
        if(atSlider){
            atSlider->blockSignals(true);
            atSlider->adjustRec(s.toInt());
            atSlider->blockSignals(false);
        }
        else startSlider = s.toInt();
    }
    

}

void QsurfaceWindow::saveAsXML()
{
	QString path=QFileDialog::getSaveFileName(this, "Save as...", QString("saveSet"), "QSAS surface plot save (*.qx_surface_plot);; XML files (*.xml)");
    if(!path.isEmpty())
    {
		if(!path.contains(".qx_surface_plot") && !path.contains(".xml")){
			path += ".qx_surface_plot";
		}
		saveAsXML(path);
	}
}

void QsurfaceWindow::saveAsXML(QString path)
{
 
        ofstream file;
        file.open(path.toLatin1().data(),ios::out);

        if(file.is_open())
        {
            _filename = path;
            
            QDomDocument doc("QSAS_surface_plot_save");

            QDomElement root=doc.createElement("QSAS_Surface_Plot");
            doc.appendChild(root);
            string guiID;
            guiID=QSAS_VERSION_ID;

            QDomElement version=createTextXMLNode(doc, "QSAS_Version", guiID);
            root.appendChild(version);

            root.appendChild(this->toXMLElement(doc));
            
            QString xml = doc.toString();
            file << xml.toLatin1().data();
            file.close();
        }
}

void QsurfaceWindow::restoreXML()
{
    QString path=QFileDialog::getOpenFileName(this, "Open...", QString(), "QSAS surface plot save (*.qx_surface_plot);; XML files (*.xml);; All files (*)");
	restoreXML(path);
}

void QsurfaceWindow::restoreXML(QString path, bool replace)
{
	// restore
	
	if(!path.isEmpty())
    {
        _filename = path;
        
        QDomDocument doc;
        QFile myFile(path);
        doc.setContent(&myFile);
        myFile.close();

        QDomElement docElem = doc.documentElement();
        if(!docElem.isNull())
        {

            string guiID="0.0.00";

            QDomElement version=docElem.firstChildElement("QSAS_Version");
            if(!version.isNull())
            {
                guiID=version.firstChild().nodeValue().toLatin1().data();
            }
            setRestoreVersion(guiID.c_str());

            QDomElement tabElement=docElem.firstChildElement("Tabs");

            if(!tabElement.isNull())
            {
               this->fromXMLElement(tabElement);
            }
			return;
        }
    }
	return;
}


// ------------------------

void QsurfaceWindow::QsurfaceAppendTextDisplay(const char *txt)
{
	textArea->append(QString(txt));
	textArea->update();
	
}


void QsurfaceWindow::clearTextArea()
{
	textArea->clear();
}

// ------------------------

bool QsurfaceWindow::getDimensions(int &nx, int &ny)
{
	if( dobj.is_nil() || dobj->nDims() != 2) {
		QsurfaceAppendTextDisplay("matrix not 2D");
		return false;
	}
	
	nx = dobj->Dims()[0];
	ny = dobj->Dims()[1];
	
	return true;

}
// ------------------------

bool QsurfaceWindow::initData()
{
	clearTextArea();
	
	isPolar = false; // default to cartesian, but reset if angle found

	if( !dobj.is_nil()) {
		// data has been set successfully previously
		// free memory used and allocate with dimensions of new data object
		dobj = new DvObject();
				
		if (z != 0) plFree2dGrid(z, nr, ntheta); // z is always r,theta even if mobj is theta,r
		if (zLog != 0) plFree2dGrid(zLog, nr, ntheta); // z is always r,theta even if mobj is theta,r
		if(nr != 0 && ntheta != 0) {
			plFree2dGrid(cgrid.xg, nr, ntheta+1);
			plFree2dGrid(cgrid.yg, nr, ntheta+1);
		}

		z = 0;
		zLog = 0;
		nr = 0;
		ntheta = 0;
		cgrid.nx = 0;
		cgrid.ny = 0;
		
		if (r != 0) delete[] r;
		if (r_plus != 0) delete[] r_plus;
		if (r_minus != 0) delete[] r_minus;
		if (theta != 0) delete[] theta;
		if (theta_plus != 0) delete[] theta_plus;
		if (theta_minus != 0) delete[] theta_minus;
		r = 0;
		r_plus = 0;
		r_minus = 0;
		theta = 0;
		theta_plus = 0;
		theta_minus = 0;
		
		
	}
	
	dobj = dataSlot->getObject();
	
	if(dobj.is_nil()) return false;
		
	// Initialising Data
	
	// set time slider
	
	atSlider->ttags = dobj->getDep0(); // may be nil
	if(dobj->seqSize() == 1)
	{
		atSlider->recBar->setRange(0, 0);
		atSlider->setEnabled(false);
		atSlider->maxRecord = 0;
	}
	else{
		atSlider->recBar->setRange(0, atSlider->ttags->seqSize()-1);
		atSlider->setEnabled(true);
		atSlider->maxRecord = atSlider->ttags->seqSize()-1;
		
	}
	atSlider->recBar->setValue(startSlider);
	
	// Obtaining r and theta from xref
	int nx, ny;
	if( !getDimensions(nx, ny) ) {
		QsurfaceAppendTextDisplay("Input object must be 2D matrix or sequence of 2D matrices");
		return false;
	}
	
	
	//Make sure that depend 1 and 2 exist. If they don't, add our own arbitary depend.
	if (!(dobj->xref_exists(DEPEND_1))) {
		vector <size_t> dimx(1);
		dimx[0] = nx;
		DvObject_var bin1 = new DvObject(0.0, dimx);

		for (int i = 0; i < nx; i++)
			bin1->dbl(i) = i+1;

		DvString lablaxis("(index)");
		bin1->change_xref(LABLAXIS, lablaxis);
		dobj->change_xref(DEPEND_1, bin1);

	}
	if (!(dobj->xref_exists(DEPEND_2))) {
		//Add our own depend according to the bin number.
		vector <size_t> dimy(1);
		dimy[0] = ny;
		DvObject_var bin2 = new DvObject(0.0, dimy);

		for (int i = 0; i < ny; i++)
			bin2->dbl(i) = i+1;

		DvString lablaxis("(index)");
		bin2->change_xref(LABLAXIS, lablaxis);
		dobj->change_xref(DEPEND_2, bin2);
	}


	// find which dimension is theta and get fillvals and units
	rfillVal 		= NAN;
	thetafillVal 	= NAN;
	zfillVal 		= NAN;
	
	DvObject_var aObj; // local ptr used for attrs
	
	if(dobj->xref_exists(FILL_VALUE)) 
	{
		aObj = dobj->get_xref(FILL_VALUE);
		if (aObj.is_ok() && aObj->is_dbl()) zfillVal = aObj->dbl();
	}

	// find depend that is an angle variable and set R and Th depends
	DvObject_var dep1 = dobj->get_xref(DEPEND_1);
	DvObject_var dep2 = dobj->get_xref(DEPEND_2);
	
	if(dep1.is_nil() || dep2.is_nil()) return false;

	orderIsRTh = true;
	
	if(dep1->isAngle()) 
	{
		orderIsRTh = false;
	
		isPolar = true;
	}
	else if(dep2->isAngle()) isPolar = true;
	
	
	// Allow user to force other choice (useful for angle vs angle)
	if(swapR->isChecked()) 
	{
		orderIsRTh = !orderIsRTh;
	}

	if(orderIsRTh){
		rDep = dep1;
		thetaDep = dep2;
		nr = nx;
		ntheta = ny;
	}
	else{
		rDep = dep2;
		thetaDep = dep1;
		nr = ny;
		ntheta = nx;
	}
	
	// find fillval for Theta
	if(thetaDep->xref_exists(FILL_VALUE)) 
	{
		aObj = thetaDep->get_xref(FILL_VALUE);
		if (aObj.is_ok() && aObj->is_dbl()) thetafillVal = aObj->dbl();
	}
	
	// find fillval for R
	if(rDep->xref_exists(FILL_VALUE)) 
	{
		aObj = rDep->get_xref(FILL_VALUE);
		if (aObj.is_ok() && aObj->is_dbl()) rfillVal = aObj->dbl();
	}


	r = new PLFLT[nr];
	r_plus = new PLFLT[nr];
	r_minus = new PLFLT[nr];

	theta = new PLFLT[ntheta];
	theta_plus = new PLFLT[ntheta];
	theta_minus = new PLFLT[ntheta];

	plAlloc2dGrid(&z, nr, ntheta+1);
	plAlloc2dGrid(&zLog, nr, ntheta+1);

	// allocate plplot x,y grid
	
	plAlloc2dGrid(&cgrid.xg, nr, ntheta+1);
	plAlloc2dGrid(&cgrid.yg, nr, ntheta+1);

	// use explicit axis choice
	if(axisType->currentIndex() == 1) isPolar = true;
	else if(axisType->currentIndex() == 2) isPolar = false;
	
	setGUItoPolar(isPolar);
	
	return true;
}


// ------------------------

void QsurfaceWindow::setData()
{
	if(dobj.is_nil()) return;
	if(thetaDep.is_nil()) return;
	
	zLogEmpty = true;
	
	// ensure we don't free cgrid arrays if they are not allocated (done later)
	cgrid.nx = 0;
	cgrid.ny = 0;
	

	// Load dimension data
	
		// theta dimension
	
	if(isPolar && thetaDep->isDeg() ) 
	{
		ang2rad = M_PI / 180.;
		rad2deg = 1. / ang2rad;
	}
	else 
	{
		// data in radians or cartesian plots
		ang2rad = 1.0;
		rad2deg = 1.0;
	}
	
			
	// allow for record varying depends
	int thetaAtRec = 0;
	if(thetaDep->seqSize() == dobj->seqSize()) thetaAtRec = atSlider->record;
			
	if( ntheta != thetaDep->Dims()[0]){
		QsurfaceAppendTextDisplay("data dimension does not match depend for theta");
		return;
	}

	for (int j = 0; j < ntheta; j++) {
		double val = thetaDep->dbl( thetaAtRec * ntheta + j);
		if (val == thetafillVal || !isfinite(val)) theta[j] = 0.;
		else theta[j] = val * ang2rad;
	}
	
	
	// r dimension
	
	// allow for record varying depends
	int rAtRec = 0;
	if(rDep->seqSize() == dobj->seqSize()) rAtRec = atSlider->record;
	
	nr = rDep->Dims()[0]; // reset here as it changes due to removal of fills or r below RadMin
	
	// n is total number of data points
	n = nr * ntheta;
	
	for (int i = 0; i < nr; i++) {
		double val = rDep->dbl( rAtRec * nr + i);
		if (val == rfillVal || !isfinite(val) ) r[i] = 0;
		else r[i] = val;
	}

	//  Get delta plus and minus for theta
	// if deltas don't exist set bin edge to same as point
	
	dThetaSet = setDeltaTheta(theta_plus, DELTA_PLUS);
	dThetaSet = (setDeltaTheta(theta_minus, DELTA_MINUS) || dThetaSet) ;
	
	//  Get delta plus and minus for r
	// if deltas don't exist set bin edge to same as point

	dRSet = setDeltaRad(r_plus, DELTA_PLUS);
	dRSet = (setDeltaRad(r_minus, DELTA_MINUS) || dRSet) ;
	
	
	full = false;
	if (   abs(sin((theta_plus[0] - theta_minus[ntheta - 1]) / 2)) < 0.01
		|| abs(sin((theta_plus[ntheta - 1] - theta_minus[0]) / 2)) < 0.01) {
		
		if(isPolar) full = true;
				
	} 

	// fill the z data array

	// get Matrix object, take record if dobj is a seq
	int atRec = atSlider->record;

	// fill z[][]

	for (int i = 0; i < nr; i++) {
				
		if(orderIsRTh)
		{
			for (int j = 0; j < ntheta; j++) {
				z[i][j] = dobj->dbl(atRec, i, j);
				if (z[i][j] == zfillVal) z[i][j] = NAN;
			}
		} 
		else {
			for (int j = 0; j < ntheta; j++) {
				z[i][j] = dobj->dbl(atRec, j, i);
				if (z[i][j] == zfillVal) z[i][j] = NAN;
			}
		}
		z[i][ntheta] = z[i][0];
	}
	
	
	// set up ordered arrays	

	 sort_r_theta(); // sorts r/x and theta/y in ascending order and adjusts z 
	 				 // also applies new order to deltas and initialises rmin and rmax
					 
	
	// remove -ve values if log axis
	if(rmin <= 0 && RXLog->isChecked() ) {
		rmin = rminNonZero;
		
		QString txt = QString("Negative or zero r values: min R set to %1").arg(rmin);
		QsurfaceAppendTextDisplay(txt.toLatin1().data());

		for (int i = 0; i < nr; i++) {
			if(r[i] < rmin) r[i] = rmin;
			if(r_plus[i] < rmin) r_plus[i] = 1.1*rmin;
			if(r_minus[i] < rmin) r_minus[i] = rmin;
		}
	}
	
	if( !isPolar && thetamin <= 0 && YLog->isChecked()) {
		thetamin = thetaminNonZero;
		
		QString txt = QString("Negative or 0 Y values: min set to %1").arg(thetamin);
		QsurfaceAppendTextDisplay(txt.toLatin1().data());

		for (int i = 0; i < ntheta; i++) {
			if(theta[i] < thetamin) theta[i] = thetamin;
			if(theta_plus[i] < thetamin) theta_plus[i] = 1.01*thetamin;
			if(theta_plus[i] < thetamin) theta_minus[i] = thetamin;
		}
		
	}
	
	
	// set user R Max/Min to use in plots

	if(RXAuto->isChecked()){
		userRmax = rmax;
		if(RXLog->isChecked()) userRmin = rminNonZero;
		else userRmin = rmin;
	}	
	else{
		bool OK;
		double userValue = RXMin->text().toDouble(&OK);
		if(OK) userRmin = (PLFLT) userValue;
		else userRmin = rmin;

		userValue = RXMax->text().toDouble(&OK);
		if(OK) userRmax = (PLFLT) userValue;
		else userRmax = rmax;
	}

	// set user Theta Max/Min to use in plots

	if(ThYAuto->isChecked()){
		userThmax = thetamax;
		if(YLog->isChecked()) userThmin = thetaminNonZero;
		else userThmin = thetamin;
	}	
	else{
		bool OK;
		double userValue = ThYMin->text().toDouble(&OK);
		if(OK) userThmin = (PLFLT) userValue * ang2rad;
		else userThmin = thetamin;

		userValue = ThYMax->text().toDouble(&OK);
		if(OK) userThmax = (PLFLT) userValue * ang2rad;
		else userThmax = thetamax;
	}

	// set start and stop indices for R/X
	
	// move nr_start to above userRmin 
	nr_start = 0;
	for(int nthis=0; nthis<nr-1; nthis++)
	{
		if(r[nthis] >= userRmin ) {
			nr_start = nthis;
			break;
		}
	}
	
	nr_end = nr;
	// move nr_end to below userRmax, but never lower than 2
	for(int nthis=nr; nthis>1; nthis--)
	{
		if(r[nthis-1] <= userRmax ) {
			nr_end = nthis;
			break;
		}
	}

	for (int i = 0; i < nr; i++) {
		// plot starts when r_plus above userRmin, so catch small r and r_minus
		if(r_minus[i] < userRmin) r_minus[i] = userRmin;
	}	
	
	for (int i = 0; i < nr; i++) {
		// plot ends when r_minus above userRmax, so catch large r and r_plus
		if(r_plus[i] >= userRmax) r_plus[i] = userRmax;
	}	
	
	// set start and stop indices for Th/Y

	nth_start = 0;
	// move nth_start to above userThmin 
	for(int nthis=0; nthis<ntheta-1; nthis++)
	{
		if(theta[nthis] >= userThmin ) {
			nth_start = nthis;
			break;
		}
	}
	
	nth_end = ntheta;
	// move nth_end to below userThmax, but never lower than 2
	for(int nthis=ntheta; nthis>1; nthis--)
	{
		if(theta[nthis-1] <= userThmax ) {
			nth_end = nthis;
			break;
		}
	}
	
	for (int i = 0; i < ntheta; i++) {
		// plot starts when theta_plus above userThmin, so catch small theta and theta_minus
		if(theta_minus[i] < userThmin) theta_minus[i] = userThmin;
	}	
	
	for (int i = 0; i < ntheta; i++) {
		// plot ends when theta_minus above userThmax, so catch large theta and theta_plus
		if(theta_plus[i] > userThmax) theta_plus[i] = userThmax;
	}	

	QString txt;
	if(isPolar){
	
		if (RXLog->isChecked()) {
			//log graph

			if (userRmin < 1.e-30) {
				txt = QString("Log Radial Min: %1 too small").arg(userRmin);
				QsurfaceAppendTextDisplay(txt.toLatin1().data());
				userRmin = rminNonZero;
				txt = QString("Reset to: %1").arg(userRmin);
				QsurfaceAppendTextDisplay(txt.toLatin1().data());
			}
		
			for (int i = 0; i < nr; i++) {
				r[i] = SaferLog10(r[i] / userRmin);
				r_plus[i] = SaferLog10(r_plus[i] / userRmin);
				r_minus[i] = SaferLog10(r_minus[i] / userRmin);
			}
			rmax = SaferLog10(rmax / userRmin);
			userRmax = SaferLog10(userRmax / userRmin);

		}
		else {
			txt = QString("Radial Origin reset to: %1 ").arg(userRmin);
			QsurfaceAppendTextDisplay(txt.toLatin1().data());
			for (int i = 0; i < nr; i++) {
				r[i] -= userRmin;
				r_plus[i] -= userRmin;
				r_minus[i] -= userRmin;
			}
			userRmax -= userRmin;
			rmax -= userRmin;
		
		}
	}
	else{
		// Cartesian plot
		
		if (RXLog->isChecked()) {
			//log graph

			if (userRmin < 1.e-30) {
				txt = QString("Log Radial Min: %1 too small").arg(userRmin);
				QsurfaceAppendTextDisplay(txt.toLatin1().data());
				userRmin = rminNonZero;
				txt = QString("Reset to: %1").arg(userRmin);
				QsurfaceAppendTextDisplay(txt.toLatin1().data());
			}
		
			for (int i = 0; i < nr; i++) {
				r[i] = SaferLog10(r[i]);
				r_plus[i] = SaferLog10(r_plus[i]);
				r_minus[i] = SaferLog10(r_minus[i]);
			}
			rmax = SaferLog10(rmax);
			userRmax = SaferLog10(userRmax);
			userRmin = SaferLog10(userRmin);

		}
		
		if (YLog->isChecked()) {
			//log graph

			if (userThmin < 1.e-30) {
				txt = QString("Log Y Min: %1 too small").arg(userThmin);
				QsurfaceAppendTextDisplay(txt.toLatin1().data());
				userThmin = thetaminNonZero;
				txt = QString("Reset to: %1").arg(userThmin);
				QsurfaceAppendTextDisplay(txt.toLatin1().data());
			}
		
			for (int i = 0; i < ntheta; i++) {
				theta[i] = SaferLog10(theta[i]);
				theta_plus[i] = SaferLog10(theta_plus[i]);
				theta_minus[i] = SaferLog10(theta_minus[i]);
			}
			thetamax = SaferLog10(thetamax);
			userThmax = SaferLog10(userThmax);
			userThmin = SaferLog10(userThmin);

		}	
	}
	
	cgrid.nx = nr;
	if(full) cgrid.ny = ntheta + 1;
	else cgrid.ny = ntheta;	


}

// --------------------------------------

bool QsurfaceWindow::setDeltaTheta(PLFLT *result, const char *delta)
{
	if(thetaDep.is_nil()) return false;

	double sign = 1;
	if(strcmp(delta, DELTA_MINUS) == 0) sign = -1;
		
	bool Set = false; // reset to true if either delta exists 

	if(isPolar)
	{
		DvObject_var obj_delta = thetaDep->get_xref(delta);
		if (obj_delta.is_ok() && obj_delta->is_dbl()) {
		
			int atRec = 0;
			if(obj_delta->seqSize() > atSlider->record) atRec = atSlider->record;

			if (obj_delta->arraySize() == 1) {						//Sometimes the delta_plus is a scalar instead of an array.
				double val = obj_delta->dbl(atRec);

				if ( val != thetafillVal){
					Set = true;
					for (int j = 0; j < ntheta; j++)  result[j] = theta[j] + (val * sign ) * ang2rad;
				} 
				else{
					QsurfaceAppendTextDisplay("No suitable delta found for Theta");
					for (int j = 0; j < ntheta; j++)  result[j] = theta[j];			
				}
			}
			else {
				Set = true;
				for (int j = 0; j < ntheta; j++) {
					double val = obj_delta->dbl(atRec * ntheta + j);
					if (val != thetafillVal){
						result[j] = theta[j] + (val * sign) * ang2rad;
					}
					else
						result[j] = theta[j];
				}
			}
		}
		else{
			QsurfaceAppendTextDisplay("No delta_plus not found for theta");
			for (int j = 0; j < ntheta; j++){
				result[j] = theta[j];
			}
		}
	}
	else{

		DvObject_var obj_delta = thetaDep->get_xref(delta);
		if (obj_delta.is_ok() && obj_delta->is_dbl()) {
				
			int atRec = 0;
			if(obj_delta->seqSize() > atSlider->record) atRec = atSlider->record;

			if (obj_delta->arraySize() == 1) {
				double val = obj_delta->dbl(atRec);
				// Sometimes delta+ is a scalar instead of an array.
				if (val != thetafillVal) { 
					Set = true;
					for (int j = 0; j < ntheta; j++)  result[j] = theta[j] + val * sign ;
				} 
				else{
					QsurfaceAppendTextDisplay("No suitable delta_plus not found for y");
					for (int j = 0; j < ntheta; j++){
						result[j] = theta[j];
					}
				}
			} 
			else {
				Set = true;
				for (int j = 0; j < ntheta; j++) {
					double val = obj_delta->dbl(atRec * ntheta + j);
					if (val != thetafillVal){
						result[j] = theta[j] + (val * sign);
					}
					else
						result[j] = theta[j];
				}
			}
		}
		else{
			QsurfaceAppendTextDisplay("No delta_plus not found for Y");
			for (int j = 0; j < ntheta; j++){
				result[j] = theta[j];
			}
		}
	}
	return Set;
}

// --------------------------------------

bool QsurfaceWindow::setDeltaRad(PLFLT *result, const char *delta)
{

	bool Set = false; // reset to true if either delta exists 

	double sign = 1.;
	if(strcmp(delta, DELTA_MINUS) == 0) sign = -1.;

	DvObject_var obj_delta = rDep->get_xref(delta);
	if (obj_delta.is_ok() && obj_delta->is_dbl()) {
				
		// if sequence get matrix for this record
		int atRec = 0;
		if(obj_delta->seqSize() > atSlider->record) atRec = atSlider->record;

		if (obj_delta->arraySize() == 1) {
			// Sometimes delta+ is a scalar instead of an array.
			double val = obj_delta->dbl(atRec);
			if (val != rfillVal) { 
				Set = true;
				for (int i = 0; i < nr; i++){
					result[i] = val * sign + r[i];
				}
			} 
			else{
				QsurfaceAppendTextDisplay("No suitable delta_plus not found for r");
				for (int i = 0; i < nr; i++){
					result[i] = r[i];
				}
			}
		} 
		else {
			Set = true;
			for (int i = 0; i < nr; i++) {
				double val = obj_delta->dbl(atRec * nr + i);
				if (val != rfillVal)
					result[i] = val * sign + r[i];
				else
					result[i] = r[i];
			}
		}
	}
	else{
		QsurfaceAppendTextDisplay("No delta_plus not found for r");
		for (int i = 0; i < nr; i++){
			result[i] = r[i];
		}
	}

	return Set;
}
	
void QsurfaceWindow::stopAnimate()
{
	if(animateBtn->text() == QString(" || ") ) doAnimate(); // call button toggle if user did not already pause it
	
}
	
void QsurfaceWindow::doAnimate()
{
	static bool animationRunning=false;

    doPlot(true); // ensure window exists
    
	if(animateBtn->text() == QString(" || ") ){
		qApp->processEvents(); // get rid of mutiple clicks
		QsurfaceAppendTextDisplay("\n View animation paused");
		animateBtn->setText("  Animate View  ");
		animationRunning = false;
	}
	else{
		
		// set time out until interrupt possible
     	QTimer::singleShot(30000, this, SLOT(stopAnimate()));
		
		QsurfaceAppendTextDisplay("\n Run view animation");
		animateBtn->setText(" || ");
		animationRunning = true;
		
		
		if(delay->value() != 0){
			while(animationRunning){
				QTime wait;
				wait.start();
				while(wait.elapsed() < delay->value()) ; // do nothing
				stepAnimate();
				qApp->processEvents();
			}
		}
		else while(animationRunning) {
			stepAnimate();
			qApp->processEvents();
		}
	}
	animationRunning=false;
	animateBtn->setText("  Animate View  ");

}
void QsurfaceWindow::stepAnimate()
{

	static int daz=1;
	static int del=1;
	int azNow = azimuth->value();
	int elNow = elevation->value();
	
	qApp->processEvents();
	// turn slowly at ends of motion
	if(azNow > 85  && daz > 0 ) daz = 0;
	else if(azNow > 85 && daz == 0) daz = -1;
	else if(azNow < -85 && daz < 0) daz = 0;
	else if(azNow < -85 && daz == 0) daz = 1;

	if(elNow > 85  && del > 0) del = 0;
	else if(elNow > 85 && del == 0) del = -1;
	else if(elNow < 5 && del < 0) del = 0;
	else if(elNow < 5 && del == 0) del = 1;

	azNow += daz;
	elNow += del;
	elevation->setValue(elNow);
	azimuth->setValue(azNow);
					
	doPlot(true);
}

// --------------------------------------

void QsurfaceWindow::doRun()
{
	escPushed = false;
	static bool Running=false;


	if(runBtn->text()  == QString(" || ")){
		qApp->processEvents(); // get rid of multiple clicks now
		QsurfaceAppendTextDisplay("\n Time animation paused");
		runBtn->setText("  Animate Time  ");
		Running = false;
	}
	else{
		bool colAutoSet = colourAuto->isChecked();
		
		// find max colour
		if(colAutoSet){

			// force fixed col max/min
			QsurfaceAppendTextDisplay("\n Finding Colour range...");
            atSlider->blockSignals(true);
			atSlider->adjustRec(0);
			setRecord(false);
			double colMinVal = zmin;
			double colMaxVal = zmax;
			if(zmin < 1.e-30) zmin = 1.;
			
			for(int nt=1; nt<=atSlider->maxRecord; nt++){
				atSlider->adjustRec(nt);
				setRecord(false);
				if(colMinVal > zmin && zmin > 1.e-30 ) colMinVal = zmin;
				if(colMaxVal < zmax) colMaxVal = zmax;
				qApp->processEvents();
				if(escPushed) break;
			}
            atSlider->blockSignals(false);

			colMin->setText(QString("%1").arg(colMinVal));
			colMax->setText(QString("%1").arg(colMaxVal));
			colourAuto->setChecked(false);
		}
		
		// start at beginning if at end
		if(atSlider->record >= atSlider->maxRecord ) atSlider->adjustRec(0);
		
		QsurfaceAppendTextDisplay("\n Run time animation");
		runBtn->setText(" || ");
		Running = true;

		if(delay->value() != 0){
			while(Running  ){
				QTime wait;
				wait.start();
				while(wait.elapsed() < delay->value()) ; // do nothing
				
				doPlot(true);
				if(atSlider->record >= atSlider->maxRecord ) {
					atSlider->adjustRec(0);
					colourAuto->setChecked(colAutoSet); // reset auto col range
					QsurfaceAppendTextDisplay("\n Animation finished");
					break;
				}
				atSlider->incRec();
				qApp->processEvents();
			}
		}
		else while(Running && !atSlider->ttags.is_nil() ){
			doPlot(true);
			if(atSlider->record >= atSlider->maxRecord ) {
				atSlider->adjustRec(0);
				colourAuto->setChecked(colAutoSet); // reset auto col range
				QsurfaceAppendTextDisplay("\n Animation finished");
				break;
			}
			atSlider->incRec();
			qApp->processEvents();
		}
		Running=false;
		runBtn->setText("  Animate Time  ");
	}
}


// --------------------------------------

void QsurfaceWindow::doPlot(bool forceSameWindow)
{
    zroom = 0.0; //how much deeper, compared to zmax-zmin, should he box extend? (such that the contours are visible)

	// slot invoked when Plot push button clicked
  	if(QuSlotDebug())
	{
		printf("QsurfaceWindow::doPlot \n");
		fflush(stdout);
	}
	
	if(dobj.is_nil()) return;
	
	// check that data has not been deleted (only edited_OK() signal resets objects
	if(	dataSlot->reValidate(false /* don't emit edited_OK */) == false) 
	{
		QsurfaceAppendTextDisplay("Data no longer valid\n");
		return;
	}

	if(nr <= 0 || ntheta <= 0){
		QsurfaceAppendTextDisplay("Zero bin dimension\n");
		return;
	}

	// User settings
		
	bool OK;			
	double _zmin = colMin->text().toDouble(&OK);
	if(OK) zmin = (PLFLT) _zmin;
	double _zmax = colMax->text().toDouble(&OK);
	if(OK) zmax = (PLFLT) _zmax;
	
	alpha = (PLFLT) alphaSpin->value();
	
	// Initialize PLplot.
	
	setwindow(forceSameWindow);
    
	// To make sure the new width will be validated:
	// plwid only stores changes in width
	// If no change from the previous plot, the default
	// width is not stored in the device buffer.
	// Thus we force a change, to store it.
	plwid(100);
	pladv(0);

	plscmap0n(16);
	plscol0a (0, 0 , 0 , 0 , 1.0 );
	plscol0a (15, 255 , 255 , 255 , alpha );
	plcol0(0);
	plschr(QPUI_CHAR_HEIGHT, 1.2*text_size_fld->value());

	plwid(4);
	
	if (colourLog->isChecked()) {
		zmin = SaferLog10(zmin);
		zmax = SaferLog10(zmax);
	
		for (int i = nr_start; i < nr_end; i++) {  // only take logs used
			for (int j = 0; j <= ntheta; j++) { // include last value (= first in case full is set)
				zLog[i][j] = SaferLog10(z[i][j]);
			}
		}	
	}

	
	// convert int viewing angles to PLFLT
	
	az = (PLFLT) azimuth->value();
	el = (PLFLT) elevation->value();
	
	// convert deg to rad
	el *= M_PI/180.;
	az = atan2(sin(az*M_PI/180.), cos(az*M_PI/180.));	//This forces az to be between -pi to pi radians.

	// make extra space in z if base contours used
	if(showCont->isChecked()) zroom = zOffset->value();
	
	boxzmin = zmin - (zmax-zmin)*zroom ;

	select_vpor(0);
	pllab("", "", timeText.c_str());

	// call axis specific plot 
	
	if(isPolar) {
		doPlotPolar();
	}
	else {
		doPlotCartesian();
	}

#ifdef Q_OS_MACX
// fix Mac OS problem with plot not showing until focus change
	
	if(forceSameWindow) return; // avoid flicker

	win->show();
#endif

}

void QsurfaceWindow::doPlotCartesian()
{
	// slot invoked when Plot push button clicked
  	if(QuSlotDebug())
	{
		printf("QsurfaceWindow::doPlotCartesian \n");
		fflush(stdout);
	}
	
	select_vpor(-1);

	plcol0(0);
	plschr(QPUI_CHAR_HEIGHT, 1.0*title_size_fld->value());

	pllab("", "", titleSlot->text().toLatin1().data());

	plschr(QPUI_CHAR_HEIGHT, 1.0*text_size_fld->value());

	select_vpor(0);

	// set colour table
	colourTable->getCurrentTable().apply(alphaSpin->value());

	if(plotType->currentIndex() == 4 || plotType->currentIndex() == 5){
		room = 1.4; //how much larger (factor) than the maximum radius of the plot should the 3d box be?
		bin_histogram_cartesian();
	}
	else {
		room = 1.2; //how much larger (factor) than the maximum radius of the plot should the 3d box be?
		surf_cartesian();
	}

	drawcolscale();


}

// --------------------------------------

void QsurfaceWindow::bin_histogram_cartesian()
{
	int n = (nr_end - nr_start) * (nth_end - nth_start);
	
	PLFLT **zp = z;
	if(colourLog->isChecked()) zp = zLog;

	PLFLT xtop[4], ytop[4], ztop[4];
	vector <polygon*> top(n);

    plwid(QtWidth2PlWidth(axis_width_fld->value()));

	int next=0;
	for(int i=nr_start; i<nr_end; i++){
		for(int j=nth_start; j<nth_end; j++){
		
			if( isnan(zp[i][j]) ) continue; // skip fill val data

			xtop[0] = r_minus[i];
			xtop[1] = r_plus[i];
			xtop[2] = r_plus[i];
			xtop[3] = r_minus[i];

			ytop[0] = theta_minus[j];
			ytop[1] = theta_minus[j];
			ytop[2] = theta_plus[j];
			ytop[3] = theta_plus[j];

			ztop[0] = ztop[1] = ztop[2] = ztop[3] = zp[i][j];
		
			top[next] = new polygon(xtop, ytop, ztop, 4, el, az);
			next++;
		}
	}
	if(next>2) quicksortpoly(top, 0, next-1);


	PLFLT half_boxdia = sqrt(3)/1.8;
    plwind( -half_boxdia, half_boxdia, -half_boxdia + 0.5*cos(el), half_boxdia + 0.5*cos(el));
    plcol0( 0 );

    plw3d( 1., 1., 1., userRmin, userRmax, userThmin, userThmax, boxzmin, zmax, el*180./M_PI, az*180./M_PI);

    char xopt[50], yopt[50], zopt[50];

    if(RXLog->isChecked()) sprintf(xopt, "bnlstu");
    else sprintf(xopt, "bnstu");
    if(YLog->isChecked()) sprintf(yopt, "bnlstu");
    else sprintf(yopt, "bnstu");
    if(colourLog->isChecked()) sprintf(zopt, "lbcdmnstuv");
    else sprintf(zopt, "bcdmnstuv");

	plschr(0., 0.8*text_size_fld->value());
	plbox3(xopt, RXLabSlot->text().toLatin1().data(), 0.0, 0, yopt, ThYLabSlot->text().toLatin1().data(), 0.0, 0, zopt, "", 0.0, 0);

	if(showCont->isChecked() )  draw_cont_cartesian();
	
	bool withEdges = false;
	if(plotType->currentIndex() == 4) withEdges = true;

	plot_hist_bar3(top, withEdges, next);

	// clean up
	for(int k=0; k<top.size(); k++){
		delete top[k];
	}
}

// -------------------------------------

//Draws the surfaces.

void QsurfaceWindow::surf_cartesian() {
	int i, j, k;
	int n = 2*(nr_end - nr_start - 1)*(nth_end - nth_start - 1 );

    if(nr_end == nr_start || nth_end == nth_start) n=0;

	vector <triangle*> tri(n);	

    plwid(QtWidth2PlWidth(axis_width_fld->value()));
    
	PLFLT **zp = z;
	if(colourLog->isChecked()) zp = zLog;	

	PLFLT xtri[4], ytri[4], ztri[4];
	threevector a, b, c, d;
	k = 0;
	for (i = nr_start; i < nr_end - 1; i++) {
		for (j = nth_start; j < nth_end - 1; j++) {

			a.setx( r[i] );
			a.sety( theta[j] );
			a.setz( zp[i][j] );

			b.setx( r[i + 1] );
			b.sety( theta[j] );
			b.setz( zp[i+1][j] );

			c.setx( r[i + 1] );
			c.sety( theta[j + 1] );
			c.setz( zp[i+1][j+1] );

			d.setx( r[i] );
			d.sety( theta[j + 1] );
			d.setz( zp[i][j+1] );

			tri[k] = new triangle(a, b, c, el, az);
			tri[k + 1] = new triangle(c, d, a, el, az);

			if(i==nr_start)
				tri[k]->setedge(-1);
			else if(i==nr_end-2)
				tri[k + 1]->setedge(1);

			k += 2;
		}
	}

	PLFLT half_boxdia = sqrt(3.)/2.;
    plwind( -half_boxdia, half_boxdia, -half_boxdia + 0.4*cos(el), half_boxdia + 0.4*cos(el));
    plcol0( 0 );

    plw3d( 1., 1., 1., userRmin - (room-1.)*(userRmax-userRmin)/2., userRmax + (room-1.)*(userRmax-userRmin)/2.,
    		userThmin - (room-1.)*(userThmax-userThmin)/2., userThmax + (room-1.)*(userThmax-userThmin)/2.,
    		boxzmin, zmax, el*180./M_PI, az*180./M_PI );

    char xopt[50], yopt[50], zopt[50];

    if(RXLog->isChecked()) sprintf(xopt, "bnlstu");
    else sprintf(xopt, "bnstu");
    if(YLog->isChecked()) sprintf(yopt, "bnlstu");
    else sprintf(yopt, "bnstu");
    if(colourLog->isChecked()) sprintf(zopt, "lbcdmnstuv");
    else sprintf(zopt, "bcdmnstuv");

	plschr(0., 0.8*text_size_fld->value());
	plbox3(xopt, RXLabSlot->text().toLatin1().data(), 0.0, 0, yopt, ThYLabSlot->text().toLatin1().data(), 0.0, 0, zopt, "", 0.0, 0);


	// base contours or surface contours
    if(showCont->isChecked() || plotType->currentIndex() == 3)  
		draw_cont_cartesian();

    plwid(QtWidth2PlWidth(line_width_fld->value()));

 	// do surface
    if(plotType->currentIndex() != 3) {

		if(n>2) quicksortpoly(tri, 0, n - 1);

		PLBOOL draw[3] = { true, true, true };
		PLBOOL drawskirt1[3] = { false, true, false };
		PLBOOL drawskirt2[3] = { false, true, false };

		bool outofrange;
		for (i = n - 1; i > -1; i--) {

			xtri[0] = tri[i]->get0().getx();
			ytri[0] = tri[i]->get0().gety();
			ztri[0] = tri[i]->get0().getz();

			xtri[1] = tri[i]->get1().getx();
			ytri[1] = tri[i]->get1().gety();
			ztri[1] = tri[i]->get1().getz();

			xtri[2] = tri[i]->get2().getx();
			ytri[2] = tri[i]->get2().gety();
			ztri[2] = tri[i]->get2().getz();

			xtri[3] = tri[i]->get0().getx();
			ytri[3] = tri[i]->get0().gety();
			ztri[3] = tri[i]->get0().getz();

    		if(ztri[0] < zmin || ztri[1] < zmin || ztri[2] < zmin ||
    				ztri[0] > zmax || ztri[1] > zmax || ztri[2] > zmax ||
    				isnan(ztri[0]) || isnan(ztri[1]) || isnan(ztri[2]))
    			continue;  // out of range

    		if (plotType->currentIndex() == 0) {
				// mesh
				plcol0(15);
				plfill3(3, xtri, ytri, ztri);
				plcol1((tri[i]->val() - zmin) / (zmax - zmin));
			} else {
				// surface
				PLFLT colVal = (tri[i]->val() - zmin) / (zmax - zmin);
				plcol1(colVal);
				plfill3(3, xtri, ytri, ztri);
				plcol0(15);
			}
			
			if (tri[i]->edge() == -1)
				plpoly3(4, xtri, ytri, ztri, drawskirt1, false); //Sometimes the skirt of the plot can be seen. This draws the boundary for the skirt.
			else if (tri[i]->edge() == 1)
				plpoly3(4, xtri, ytri, ztri, drawskirt2, false);
				
			// draw edges
			if(plotType->currentIndex() == 0 || plotType->currentIndex() == 2 ) 
				plpoly3(4, xtri, ytri, ztri, draw, true);

		}
	}

	// clean up
	for(int k=0; k<tri.size(); k++){
		delete tri[k];
	}
}

// -------------------------------------

// Draws the contours.

void QsurfaceWindow::draw_cont_cartesian() {
		
    plwid(QtWidth2PlWidth(line_width_fld->value()));

    int np = 100;
	CONT_LEVEL *cont, *clev;
	CONT_LINE *cline;

	PLFLT clevel[np];
	PLINT nlevel = cont_lvl(clevel, np);

	if(nlevel > np) {
		nlevel = np;
		QsurfaceAppendTextDisplay("Contour interval too small, plot truncated");
	}
	if(nlevel == 0) {
		QsurfaceAppendTextDisplay("Contour interval too large, no contours");
		return;
	}

	/* prepare cont_store input */

	cgrid.nx = nr;
	cgrid.ny = ntheta;
	
	for (int i = 0; i < nr; i++) {
		for (int j = 0; j < ntheta; j++) {
			cgrid.xg[i][j] = r[i];
			cgrid.yg[i][j] = theta[j];
		}
	}

 	// Fill cont structure with contours. 
//****** x-axis in cartesian is ok if colourLog is NOT checked, so prob could be:	
	if(colourLog->isChecked())
	{
		cont_store((const PLFLT**)&(zLog[0]), nr, ntheta + full, 1, nr_end, 1, ntheta + full, clevel, nlevel, pltr2, (void *) &cgrid, &cont);
// SJS inserted above instead of next line
//		cont_store((const PLFLT**)&(zLog[nr_start]), nr_end - nr_start, ntheta + full, 1, nr_end - nr_start, 1, ntheta + full, clevel, nlevel, pltr2, (void *) &cgrid, &cont);
		//cont_store(&(zLog[nr_start]), nr_end - nr_start, ntheta + full, 1, nr_end - nr_start, 1, ntheta + full, clevel, nlevel, pltr2, (void *) &cgrid, &cont);
	}
	else {
		cont_store((const PLFLT**)&(z[0]), nr, ntheta + full, 1, nr_end, 1, ntheta + full, clevel, nlevel, pltr2, (void *) &cgrid, &cont);
		//cont_store(&(z[0]), nr, ntheta + full, 1, nr_end, 1, ntheta + full, clevel, nlevel, pltr2, (void *) &cgrid, &cont);
	}

	/* follow the contour levels and lines */
	clev = cont;

	int i=0;

	do { /* for each contour level */
		cline = clev->line;


		do { /* there are several lines that make up each contour */
			if (cline->npts > 0) {
				PLFLT zz[cline->npts];

				if (plotType->currentIndex() == 3) {
					for (int j = 0; j < cline->npts; j++)
						zz[j] = clev->level;

					plwid(QtWidth2PlWidth(line_width_fld->value()));
					plcol0(15);
					if(cline->npts > 2) plfill3(cline->npts, cline->x, cline->y, zz);
				} 
				else {
					for (int j = 0; j < cline->npts; j++)
						zz[j] = boxzmin;
				}

				plcol1((clev->level - zmin) / (zmax - zmin));
				plline3(cline->npts, cline->x, cline->y, zz);

			}
			cline = cline->next;
		} while (cline != NULL);

		clev = clev->next;
		i++;

	} while (clev != NULL);


	cont_clean_store(cont); /* now release the contour memory */

}

// --------------------------------------

void QsurfaceWindow::doPlotPolar()
{

	// slot invoked when Plot push button clicked
  	if(QuSlotDebug())
	{
		printf("QsurfaceWindow::doPlotPolar \n");
		fflush(stdout);
	}
		
	select_vpor(-1);

	plcol0(0);
	plschr(QPUI_CHAR_HEIGHT, 1.0*title_size_fld->value());

	pllab("", "", titleSlot->text().toLatin1().data());

	plschr(QPUI_CHAR_HEIGHT, 1.0*text_size_fld->value());

	select_vpor(0);

	// set colour table
	colourTable->getCurrentTable().apply(alphaSpin->value());

	if(plotType->currentIndex() == 4 || plotType->currentIndex() == 5){
		room = 1.4; // how much larger (factor) than the maximum radius of the plot should the 3d box be?
		bin_histogram_polar();
	}
	else {
		room = 1.3; //how much larger (factor) than the maximum radius of the plot should the 3d box be?
		surf_polar();
	}

	drawcolscale();


}

// -------------------------------------

void QsurfaceWindow::bin_histogram_polar(){

	int n = (nr_end - nr_start) * (nth_end - nth_start);

	PLFLT **zp = z;
	if(colourLog->isChecked()) zp = zLog;

	PLFLT xtop[4], ytop[4], ztop[4];
	vector <polygon*> top(n);

	int next=0;
	for(int i=nr_start; i<nr_end; i++){
		for(int j=nth_start; j<nth_end; j++){
		
			if( isnan(zp[i][j]) ) continue; // skip fill val data

			xtop[0] = r_minus[i]*cos(theta_minus[j]);
			xtop[1] =r_plus[i]*cos(theta_minus[j]);
			xtop[2] =r_plus[i]*cos(theta_plus[j]);
			xtop[3] =r_minus[i]*cos(theta_plus[j]);

			ytop[0] = r_minus[i]*sin(theta_minus[j]);
			ytop[1] =r_plus[i]*sin(theta_minus[j]);
			ytop[2] =r_plus[i]*sin(theta_plus[j]);
			ytop[3] =r_minus[i]*sin(theta_plus[j]);

			ztop[0] = ztop[1] = ztop[2] = ztop[3] = zp[i][j];
		
			top[next] = new polygon(xtop, ytop, ztop, 4, el, az);
			next++;
		}
	}

	if(next > 2) quicksortpoly(top, 0, next-1);

	PLFLT half_boxdia = sqrt(3.)/1.8;
    plwind( -half_boxdia, half_boxdia, -half_boxdia + 0.5*cos(el), half_boxdia + 0.5*cos(el));
    plcol0( 0 );

    plw3d( 1., 1., 1., -room*userRmax, room*userRmax, -room*userRmax, room*userRmax, boxzmin, zmax, el*180./M_PI, az*180./M_PI);

    draw_polar_axis();
    draw_angle(1.1);

	if(showCont->isChecked() )  draw_cont_polar();
	
	bool withEdges = false;
	if(plotType->currentIndex() == 4) withEdges = true;
	
    plot_hist_bar3(top, withEdges, next);

	// clean up
	for(int k=0; k<top.size(); k++){
		delete top[k];
	}
}

//Accepts an array of polygons - the top polygon of each histogram bar. Plots them. The points of the visible polygons must be in anticlockwise order.
void QsurfaceWindow::plot_hist_bar3(vector <polygon*> &pol, bool withEdges, int n){
	
    if(withEdges) plwid(QtWidth2PlWidth(line_width_fld->value()));
    
    int nsides=4;
	vector <polygon*> sides(nsides);
	
	for(int i=n-1; i > -1; i--){
		for(int j=0; j<nsides; j++){

			//set the four verticies of each side
			threevector sidepts[] = {pol[i]->getvert(j==3? j:j+1), pol[i]->getvert(j==3? 0:j), pol[i]->getvert(j==3? 0:j), pol[i]->getvert(j==3? j:j+1)};
			sidepts[2].setz(zmin);
			sidepts[3].setz(zmin);
			
			sides[j] = new polygon(sidepts, nsides, el, az);
		}
		
		quicksortpoly(sides, 0, 3);
		
		PLFLT val = sides[0]->getvert(0).getz();
		PLFLT col;
		if( val >= zmax) col = 1.0;
		else col = (sides[0]->getvert(0).getz() - zmin) / (zmax - zmin);
				
		for (int j = nsides; j > 0; j--) {
			if ( val >= zmin) {
				plcol1(col);
				plfill3(sides[j-1]);
				plcol0(0);
				if(withEdges) plpoly3(sides[j-1]);
			}
		}

		if (val >= zmin) {
			plcol1(col);
			plfill3(pol[i]);
			plcol0(0);
			if(withEdges) plpoly3(pol[i]);
		}
		
		// clean up
		for(int j=0; j<nsides; j++){
			delete sides[j];
		}
	}
}
// -------------------------------------

// Puts data into triangles, sorts and plots them.
void QsurfaceWindow::surf_polar(){

	// ensure r points are bin centres
	for (int i = 0; i < nr; i++) {
			r[i] = 0.5 *(r_plus[i] + r_minus[i]);
	}
	

	// ensure theta points are bin centres
	for (int j = 0; j < ntheta; j++) {
			theta[j] = 0.5 *(theta_plus[j] + theta_minus[j]);
	}

	int n = 2*(nr_end-nr_start-1)*(nth_end-nth_start-1 + full);
    if(nr_end == nr_start || nth_end <= nth_start - 1 + full) n=0;
    
	vector <triangle*> tri(n);
	
	PLFLT **zp = z;
	if(colourLog->isChecked()) zp = zLog;
	
	PLFLT xtri[4], ytri[4], ztri[4];
	threevector a, b, c, d;
	int k = 0;
	
	for (int i = nr_start; i < nr_end - 1; i++) {
		for (int j = nth_start; j < nth_end - 1; j++) {

			a.setx( r[i] * cos(theta[j]) );
			a.sety( r[i] * sin(theta[j]) );
			a.setz( zp[i][j] );

			b.setx( r[i + 1] * cos(theta[j]) );
			b.sety( r[i + 1] * sin(theta[j]) );
			b.setz( zp[i+1][j] );

			c.setx( r[i + 1] * cos(theta[j + 1]) );
			c.sety( r[i + 1] * sin(theta[j + 1]) );
			c.setz( zp[i+1][j+1] );

			d.setx( r[i] * cos(theta[j + 1]) );
			d.sety( r[i] * sin(theta[j + 1]) );
			d.setz( zp[i][j+1] );

			tri[k] = new triangle(a, b, c, el, az);
			tri[k + 1] = new triangle(c, d, a, el, az);

			if(i==nr_start)
				tri[k]->setedge(-1);
			else if(i==nr_end-2)
				tri[k + 1]->setedge(1);
			
			k += 2;
		}
		if(full)
		{
			a.setx( r[i] * cos(theta[ntheta-1]) );
			a.sety( r[i] * sin(theta[ntheta-1]) );
			a.setz( zp[i][ntheta-1] );

			b.setx( r[i + 1] * cos(theta[ntheta-1]) );
			b.sety( r[i + 1] * sin(theta[ntheta-1]) );
			b.setz( zp[i+1][ntheta-1] );

			c.setx( r[i + 1] * cos(theta[0]) );
			c.sety( r[i + 1] * sin(theta[0]) );
			c.setz( zp[i+1][0] );

			d.setx( r[i] * cos(theta[0]) );
			d.sety( r[i] * sin(theta[0]) );
			d.setz( zp[i][0] );

			tri[k] = new triangle(a, b, c, el, az);
			tri[k+1] = new triangle(c, d, a, el, az);

			if(i==0)
				tri[k]->setedge(-1);
			else if(i==nr-2)
				tri[k]->setedge(1);
			
			k += 2;
		}
	}

	PLFLT half_boxdia = sqrt(3.)/2.;
    plwind( -half_boxdia, half_boxdia, -half_boxdia + 0.4*cos(el), half_boxdia + 0.4*cos(el));

    plcol0( 0 );

    draw_polar_axis();

	// surface contours
    if(plotType->currentIndex() == 3)   
		draw_angle();

	// base contours or surface contours
    if(showCont->isChecked() || plotType->currentIndex() == 3)  
		draw_cont_polar();

	// draw angle label now if not drawn first
    if(plotType->currentIndex() != 3)
    	draw_angle(1.05);

    plschr( 0., 0.8*text_size_fld->value() );

	// do surface
    if(plotType->currentIndex() != 3) {

		if(n>2) quicksortpoly(tri, 0, n - 1);

		PLBOOL draw[3] = { true, true, true };
		PLBOOL drawskirt1[3] = { false, true, false };
		PLBOOL drawskirt2[3] = { false, true, false };

        plwid(QtWidth2PlWidth(line_width_fld->value()));

		for (int i = n - 1; i > -1; i--) {

			xtri[0] = tri[i]->get0().getx();
			ytri[0] = tri[i]->get0().gety();
			ztri[0] = tri[i]->get0().getz();

			xtri[1] = tri[i]->get1().getx();
			ytri[1] = tri[i]->get1().gety();
			ztri[1] = tri[i]->get1().getz();

			xtri[2] = tri[i]->get2().getx();
			ytri[2] = tri[i]->get2().gety();
			ztri[2] = tri[i]->get2().getz();

			xtri[3] = tri[i]->get0().getx();
			ytri[3] = tri[i]->get0().gety();
			ztri[3] = tri[i]->get0().getz();

    		if(ztri[0] < zmin || ztri[1] < zmin || ztri[2] < zmin || ztri[0] > zmax || ztri[1] > zmax || ztri[2] > zmax)
    			continue; // out of range
 
			if(isnan(ztri[0]) || isnan(ztri[1]) || isnan(ztri[2]) )
    			continue; // out of range

    		if (plotType->currentIndex() == 0) {
				// mesh
				plcol0(15);
				plfill3(3, xtri, ytri, ztri);
				plcol1((tri[i]->val() - zmin) / (zmax - zmin));
			} 
			else {
				// surface
				PLFLT colVal = (tri[i]->val() - zmin) / (zmax - zmin);
				plcol1(colVal);
				plfill3(3, xtri, ytri, ztri);
				plcol0(15);
			}

			if (tri[i]->edge() == -1)
				plpoly3(4, xtri, ytri, ztri, drawskirt1, false); //Sometimes the skirt of the plot can be seen. This draws the boundary for the skirt.
			else if (tri[i]->edge() == 1)
				plpoly3(4, xtri, ytri, ztri, drawskirt2, false);
			
			if(plotType->currentIndex() == 0 || plotType->currentIndex() == 2 ) plpoly3(4, xtri, ytri, ztri, draw, true);
			
		}
	}
	
	// clean up
	for(int k=0; k<tri.size(); k++){
		delete tri[k];
	}
}

// -------------------------------------

// Draws the contours.

void QsurfaceWindow::draw_cont_polar() {
 	
    plwid(QtWidth2PlWidth(line_width_fld->value()));
    
    int np = 100;
	CONT_LEVEL *cont, *clev;
	CONT_LINE *cline;

	PLFLT clevel[np];
	PLINT nlevel = cont_lvl(clevel, np);
	if(nlevel > np) {
		nlevel = np;
		QsurfaceAppendTextDisplay("Contour interval too small, plot truncated");
	}
	if(nlevel == 0) {
		QsurfaceAppendTextDisplay("Contour interval too large, no contours");
		return;
	}
	
	/* prepare cont_store input */

	cgrid.nx = nr;
	cgrid.ny = ntheta + full;

	for (int i = 0; i < nr; i++) {
		for (int j = 0; j < ntheta; j++) {
			cgrid.xg[i][j] = r[i] * cos(theta[j]);
			cgrid.yg[i][j] = r[i] * sin(theta[j]);
		}
		if (full) {
			cgrid.xg[i][ntheta] = r[i] * cos(theta[0]);
			cgrid.yg[i][ntheta] = r[i] * sin(theta[0]);
		}
	}

 	// Fill cont structure with contours. 
//**** this doesn't work if colourLog is not checked and rmin is userset.

	if(colourLog->isChecked())
	{ 
		cont_store((const PLFLT**)&(zLog[0]), nr, ntheta + full, 1, nr_end, 1, ntheta + full, clevel, nlevel, pltr2, (void *) &cgrid, &cont);
// SJS inserted above in place of next line	
//		cont_store((const PLFLT**)&(zLog[nr_start]), nr_end - nr_start, ntheta + full, 1, nr_end - nr_start, 1, ntheta + full, clevel, nlevel, pltr2, (void *) &cgrid, &cont);
		//cont_store(&(zLog[nr_start]), nr_end - nr_start, ntheta + full, 1, nr_end - nr_start, 1, ntheta + full, clevel, nlevel, pltr2, (void *) &cgrid, &cont);
	}
	else {
		cont_store((const PLFLT**)&(z[0]), nr, ntheta + full, 1, nr_end, 1, ntheta + full, clevel, nlevel, pltr2, (void *) &cgrid, &cont);
		//cont_store(&(z[0]), nr, ntheta + full, 1, nr_end, 1, ntheta + full, clevel, nlevel, pltr2, (void *) &cgrid, &cont);
	}



	/* follow the contour levels and lines */
	clev = cont;

	int i=0;

	do { /* for each contour level */
		cline = clev->line;


		do { /* there are several lines that make up each contour */
			int j = 0;
			if (cline->npts > 0) {
				PLFLT zz[cline->npts];

				if (plotType->currentIndex() == 3) {
					for (j = 0; j < cline->npts; j++)
						zz[j] = clev->level;

					plcol0(15);
					if(cline->npts > 2) plfill3(cline->npts, cline->x, cline->y, zz);
				} 
				else {
					for (j = 0; j < cline->npts; j++)
						zz[j] = boxzmin;
				}

				plcol1((clev->level - zmin) / (zmax - zmin));
				plline3(cline->npts, cline->x, cline->y, zz);

			}
			cline = cline->next;
		} while (cline != NULL);

		clev = clev->next;
		i++;

	} while (clev != NULL);

	cont_clean_store(cont); /* now release the contour memory */

}


// -------------------------------------

//Draws the angle labels.

void QsurfaceWindow::draw_angle(PLFLT offset) {

	plcol0(0);
	char angle[10];
	PLFLT Rposn = offset*userRmax;
	
	for (int j = 0; j < 12; j++) {
		int iomega = 30 * j;
		sprintf(angle, "%d", iomega);
		double aomega = (double) iomega * M_PI/180.;
		plptex3(Rposn * cos(aomega ), Rposn * sin(aomega),
				boxzmin, Rposn * cos(aomega), Rposn * sin(aomega), 0., 
				-Rposn * sin(aomega), Rposn * cos(aomega), 0., 0.0, angle);
	}

	char thetachar[] = "   ";
	string thetalabel = string(ThYLabSlot->text().toLatin1().data());

    PLFLT domega = M_PI/36.; //0.5*M_PI / strlen(thetalabel);
    PLFLT omega  = az + M_PI*1.8; // - thetalabel.size()*M_PI/72.;

	plschr(0, 1.0*text_size_fld->value());

    for ( int i=0; i<thetalabel.size(); i++ )
    {
		if(thetalabel[i] == '#'){
        	strncpy(&(thetachar[0]), &(thetalabel[i]), 3);
			i += 2;
		}
		else {
        	thetachar[0] = thetalabel[i];
			thetachar[1] = '\0';
		}
		plptex3( room*userRmax*sin(omega), room*userRmax*cos(omega), boxzmin,
				userRmax*cos(omega), -userRmax*sin(omega), 0., userRmax*sin(omega), userRmax*cos(omega), 0, 0.5, thetachar );
		omega += domega;
    }
}

// -------------------------------------

// Draws Double positive axis in log plot; Otherwise it just draws a box.
void QsurfaceWindow::draw_polar_axis() {
 
    plwid(QtWidth2PlWidth(axis_width_fld->value()));
  
 	plw3d( 1., 1., 1., -room*userRmax, room*userRmax, -room*userRmax, room*userRmax, boxzmin, zmax, el*180./M_PI, az*180./M_PI );
	
	plschr(0., 0.8*text_size_fld->value());
	if (RXLog->isChecked()) {

		if (colourLog->isChecked())
			plbox3("bu", RXLabSlot->text().toLatin1().data(), 0.0, 0, "bu", RXLabSlot->text().toLatin1().data(), 0.0, 0, "lbcdmnstuv", "", 0.0, 0);
		else
			plbox3("bu", RXLabSlot->text().toLatin1().data(), 0.0, 0, "bu", RXLabSlot->text().toLatin1().data(), 0.0, 0, "bcdmnstuv", "", 0.0, 0);		

		vector<PLFLT> ticks;
		logticks(SafeLog10(userRmin), userRmax + SafeLog10(userRmin), ticks);// userRmax is already logged

		PLFLT xline[2], yline[2], zline[2];
		PLFLT xtick[2], ytick[2], ztick[2];

		char tick_val[50];
		PLFLT pos;

		zline[0] = zline[1] = boxzmin;
		ztick[0] = ztick[1] = boxzmin;



		//Draw x axis
		if (az < M_PI*0.5 && az > -M_PI*0.5) {
			//draw axis at minimum y
			ytick[0] = -room * userRmax;

			//now drawing the ticks
			for (int i = 0; i < ticks.size(); i++) {
				if (ticks[i] > SafeLog10(userRmin)) {
					xtick[0] = xtick[1] = ticks[i] - SafeLog10(userRmin);
					ytick[1] = -room * userRmax * 0.97;

					if (ticks[i] == (int) ticks[i]) {
						pos = (xtick[0] + room * userRmax) / (2 * room * userRmax);
						ytick[1] = -room * userRmax * 0.95;
						sprintf(tick_val, "10#u%d", (int) ticks[i]);
						plmtex3("xp", 1.5, pos, 0.2, tick_val);

						if (i != 0) {
							pos = (room * userRmax - xtick[0]) / (2 * room * userRmax);
							plmtex3("xp", 1.5, pos, 0.2, tick_val);
						}
					}

					plline3(2, xtick, ytick, ztick);

					xtick[0] = xtick[1] = -ticks[i] + SafeLog10(userRmin);
					plline3(2, xtick, ytick, ztick);
				}
			}
		} 
		else if (az > M_PI*0.5 || az < -M_PI*0.5) {
			//draw axis at maximum y
			ytick[0] = room * userRmax;

			//now drawing the ticks
			for (int i = 0; i < ticks.size(); i++) {
				if (ticks[i] > SafeLog10(userRmin)) {
					xtick[0] = xtick[1] = ticks[i] - SafeLog10(userRmin);
					ytick[1] = room * userRmax * 0.97;

					if (ticks[i] == (int) ticks[i]) {
						pos = (xtick[0] + room * userRmax) / (2 * room * userRmax);
						ytick[1] = room * userRmax * 0.95;
						sprintf(tick_val, "10#u%d", (int) ticks[i]);
						plmtex3("xs", 1.5, pos, 0.2, tick_val);

						if (i != 0) {
							pos = (-xtick[0] + room * userRmax) / (2 * room * userRmax);
							plmtex3("xs", 1.5, pos, 0.2, tick_val);
						}
					}

					plline3(2, xtick, ytick, ztick);

					xtick[0] = xtick[1] = -ticks[i] + SafeLog10(userRmin);
					plline3(2, xtick, ytick, ztick);
				}
			}
		}

		//Draw y axis
		if (az > 0 && az < M_PI) {
			//draw axis at minimum x
			xtick[0] = -room * userRmax;

			//now drawing the ticks
			for (int i = 0; i < ticks.size(); i++) {
				if (ticks[i] > SafeLog10(userRmin)) {
					ytick[0] = ytick[1] = ticks[i] - SafeLog10(userRmin);
					xtick[1] = -room * userRmax * 0.97;

					if (ticks[i] == (int) ticks[i] && i != ticks.size() - 2) {
						pos = (ytick[0] + room * userRmax) / (2 * room * userRmax);
						xtick[1] = -room * userRmax * 0.95;
						sprintf(tick_val, "10#u%d", (int) ticks[i]);
						plmtex3("yp", 1.5, pos, 0.2, tick_val);

						if (i != 0) {
							pos = (-ytick[0] + room * userRmax) / (2 * room * userRmax);
							plmtex3("yp", 1.5, pos, 0.2, tick_val);
						}
					}

					plline3(2, xtick, ytick, ztick);

					ytick[0] = ytick[1] = -ticks[i] + SafeLog10(userRmin);
					plline3(2, xtick, ytick, ztick);
				}
			}
		} 
		else if (az < 0 && az > -M_PI) {
			//draw axis at maximum x
			xtick[0] = room * userRmax;

			for (int i = 0; i < ticks.size(); i++) {
				if (ticks[i] > SafeLog10(userRmin)) {
					ytick[0] = ytick[1] = ticks[i] - SafeLog10(userRmin);
					xtick[1] = room * userRmax * 0.97;

					if (ticks[i] == (int) ticks[i] && i != ticks.size() - 2) {
						pos = (ytick[0] + room * userRmax) / (2 * room * userRmax);
						xtick[1] = room * userRmax * 0.95;
						sprintf(tick_val, "10#u%d", (int) ticks[i]);
						plmtex3("ys", 1.5, pos, 0.2, tick_val);

						if (i != 0) {
							pos = (-ytick[0] + room * userRmax) / (2 * room * userRmax);
							plmtex3("ys", 1.5, pos, 0.2, tick_val);
						}
					}

					plline3(2, xtick, ytick, ztick);

					ytick[0] = ytick[1] = -ticks[i] + SafeLog10(userRmin);
					plline3(2, xtick, ytick, ztick);
				}
			}
		}

		//Drawing mid lines
		xline[0] = xline[1] = 0;
		yline[0] = -room*userRmax;
		yline[1] = userRmax;

		plline3(2, xline, yline, zline);
		plline3(2, yline, xline, zline);

	} 
	else {
		// linear Radius
		if (colourLog->isChecked())
			plbox3("bnstu", RXLabSlot->text().toLatin1().data(), 0.0, 0, "abnstu", RXLabSlot->text().toLatin1().data(), 0.0, 0, "lbcdmnstuv", "", 0.0, 0);
		else
			plbox3("bnstu", RXLabSlot->text().toLatin1().data(), 0.0, 0, "abnstu", RXLabSlot->text().toLatin1().data(), 0.0, 0, "bcdmnstuv", "", 0.0, 0);
	}
}



// -------------------------------

// Sorts a regular grid of r and theta in ascending order. Theta is set to 0-2*pi before the sorting is done. Output theta is in radians.
void QsurfaceWindow::sort_r_theta() {


	// make a single array out of z
	PLFLT w[ntheta * nr];

	for (int j = 0; j < ntheta; j++) {
		for (int i = 0; i < nr; i++) {
			w[j * nr + i] = z[i][j];
		}
	}

	// 2d array sorting in terms of theta.
	quick_sort_2d(theta, w, 0, ntheta - 1, nr, theta_plus, theta_minus);


	//Assigning w back to z
	for (int j = 0; j < ntheta; j++) {
		for (int i = 0; i < nr; i++) {
			z[i][j] = w[j * nr + i];
		}
	}

	// unpack again for r sort
	for (int i = 0; i < nr; i++) {
		for (int j = 0; j < ntheta; j++) {
			w[i * ntheta + j] = z[i][j];
		}
	}

	// 2d array sorting in terms of r
	quick_sort_2d(r, w, 0, nr - 1, ntheta, r_plus, r_minus);


	//Assigning w back to z
	for (int i = 0; i < nr; i++) {
		for (int j = 0; j < ntheta; j++) {
			z[i][j] = w[i * ntheta + j];
		}
	}
	
	
	rmax = r_plus[nr-1];
	rmin = r_minus[0];
	rminNonZero = rmin;
	if(rminNonZero < 1.e-30) {
		for(int i=0; i<nr-1; i++){
			// if -ve will be moved along to 0.0, so justy move up far enough (rmin is subtracted later)
			if(r[i] - rmin > 1.e-30) {
				rminNonZero = (r[i] + rmin) * 0.5;
				break;
			}
			else if(r_plus[i] - rmin > 1.e-30) {
				rminNonZero = (r_plus[i] + rmin) * 0.5;
				break;
			}
			else if(r_minus[i+1] - rmin > 1.e-30) {
				rminNonZero = (r_minus[i+1] + rmin) * 0.5;
				break;
			}
		}
	}
	
	thetamax = theta_plus[ntheta-1];
	thetamin = theta_minus[0];
	for(int i=0; i<ntheta-1; i++){
		if(theta_plus[i] > thetamax) thetamax = theta_plus[i];
		if(theta_minus[i] < thetamin) thetamin = theta_minus[i];
	}

	thetaminNonZero = thetamin;
	if(thetaminNonZero < 1.e-30) {
		for(int i=0; i<ntheta-1; i++){
			// if -ve will be moved along to 0.0, so just move up far enough (rmin is subtracted later)
			if(theta[i] - thetamin > 1.e-30) {
				thetaminNonZero = (theta[i] + thetamin) * 0.5;
				break;
			}
			else if(theta_plus[i] - thetamin > 1.e-30) {
				thetaminNonZero = (theta_plus[i] + thetamin) * 0.5;
				break;
			}
			else if(theta_minus[i+1] - thetamin > 1.e-30) {
				thetaminNonZero = (theta_minus[i+1] + thetamin) * 0.5;
				break;
			}
		}
	}

}

//----------------------------

//Sets up page and window.
void QsurfaceWindow::setwindow(bool forceSameWindow) {

	bool landscape;
    bool getNewWindow = ((plotWin_id->currentWindow == NULL )|| (newWindow->isChecked() && !forceSameWindow) );

    QPrinter::PageSize pageSize;
    switch(page_size->currentIndex())
    {
        case QPUI_A4:  default: case QPUI_DEFAULT:
        {
            pageSize = QPrinter::A4;
            break;
        }
        case QPUI_CUSTOM:
        {
            pageSize = QPrinter::Custom;
            break;
        }
        case QPUI_EXECUTIVE:
        {
            pageSize = QPrinter::Executive;
            break;
        }
        case QPUI_LEGAL:
        {
            pageSize = QPrinter::Legal;
            break;
        }
        case QPUI_LETTER:
        {
            pageSize = QPrinter::Letter;
        }
            
    }

    QPrinter::Orientation orientation;
	if(pageOrientation->currentIndex() == 0) {
		landscape = false;
		orientation = QPrinter::Portrait;
	}
	else {
		landscape = true;
		orientation = QPrinter::Landscape;
	}

    bool OK;
    double slotValue = pageWidth->text().toDouble(&OK);
    if(OK) width = slotValue;
    
    slotValue = pageHeight->text().toDouble(&OK);
    if(OK) height = slotValue;
    
	if(  !getNewWindow )
	{	
		// pick up existing window
        win = plotWin_id->focusWindow(width, height, DPI, pageSize, orientation, true);
        return;

	}


	if(width == height) width+=1; // There seems to be a bug when saving jpgs with the same width and height, the entire jpg turns black. KaHo

    win = plotWin_id->newPlotWindow(width, height, DPI, pageSize, orientation, true);
	
	win->setSmooth(true);
	if(forceSameWindow) win->move(0,0);


    elevation = win->addSlider(Qt::Vertical, 0, 90, 1);
    elevation->setValue(startElevation);
    azimuth = win->addSlider(Qt::Horizontal, -180, 180, 1);
    azimuth->setValue(startAzimuth);
    connect(elevation, SIGNAL(valueChanged(int)), this, SLOT(changeView()) );
    connect(azimuth, SIGNAL(valueChanged(int)), this, SLOT(changeView()) );

    QLabel* delayLabel = new QLabel("Animation delay: ");
    win->addWidget(delayLabel, 0, 0, Qt::AlignRight);
   
    delay = new QSpinBox();
    delay->setMinimum(0);
    delay->setMaximum(500);
    delay->setValue(0);
    delay->setSingleStep(10);
    delay->setToolTip("Delay in msec between plots in animations");
    win->addWidget(delay, 0, 1, Qt::AlignLeft);

    QLabel *space1 = new QLabel("    ");
    win->addWidget(space1, 0, 2, Qt::AlignCenter);
    
    animateBtn = new QPushButton("  Animate View  ");
    animateBtn->setMinimumSize(animateBtn->sizeHint());
    animateBtn->setFocusPolicy(Qt::StrongFocus);
    connect(animateBtn, SIGNAL(clicked()), this, SLOT(doAnimate()) );
    win->addWidget(animateBtn, 0, 3, Qt::AlignCenter);

    QLabel *space2 = new QLabel("    ");
    win->addWidget(space2, 0, 4, Qt::AlignCenter);
    
    runBtn = new QPushButton("  Animate Time  ",this);
    runBtn->setMinimumSize(runBtn->sizeHint());
    runBtn->setFocusPolicy(Qt::StrongFocus);
    connect(runBtn, SIGNAL(clicked()), this, SLOT(doRun()));
    win->addWidget(runBtn, 0, 5, Qt::AlignCenter);
    

	pladv(0); // this is done again later, but clears a bug with plwid being ignored

}



// ----------------------

// Sets the page tab feedback
void QsurfaceWindow::setPage()
{
	bool readonly = true;
	if(page_size->currentIndex() == QPUI_CUSTOM) readonly = false;
	
	switch(dpiPage->currentIndex())
	{
		case QPUI_72_DPI:
		{
			DPI = 72;
			break;
		}
		case QPUI_150_DPI:
		{
			DPI = 150;
			break;
		}
		case QPUI_300_DPI:
		{
			DPI = 300;
			break;
		}
		case QPUI_600_DPI:
		{
			DPI = 600;
			break;
		}
		case QPUI_1200_DPI:
		{
			DPI = 1200;
			break;
		}
		
	}
	
	switch(pageUnits->currentIndex())
	{
		case QPUI_PIXELS: 
		{
			conv = (double) DPI;
			break;
		}
		case QPUI_INCHES: 
		{
			conv = 1.0;
			break;
		}
		case QPUI_CM: 
		{
			conv = 2.54;
			break;
		}
		case QPUI_MM: 
		{
			conv = 25.4;
			break;
		}
		case QPUI_PSPOINTS: 
		{
			conv = (double) DPI;
			break;
		}
	}


	if(readonly)
	{
		switch(page_size->currentIndex()){
			case QPUI_A4: default:
			{
				width = 8.3 * conv;
				height = 11.694 * conv;
				break;
			}
			case QPUI_EXECUTIVE:
			{
				width = 7.5 * conv;
				height = 10.0 * conv;
				break;
			}
			case QPUI_LEGAL:
			{
				width = 8.5 * conv;
				height = 14 * conv;
				break;
			}
			case QPUI_LETTER:
			{
				width = 8.5 * conv;
				height = 11.0 * conv;
				break;
			}
			case QPUI_DEFAULT:
			{
				width = 8.3 * conv;
				height = 11.694 * conv;
				break;
			}
		}
		pageWidth->setText(QString().setNum(width));
		pageHeight->setText(QString().setNum(height));
		
	}
	pageWidth->setReadOnly(readonly);
	pageHeight->setReadOnly(readonly);
	
	pageWidth->setEnabled(!readonly);
	pageHeight->setEnabled(!readonly);
      
	
    for(int i=0; i<plotWin_id->size(); i++){
        plotWin_id->windowAt(i)->setPrintRes(DPI);
    }


}

// ---------------------------

// Method to quicksort x array and pseudo 2d arrays in ascending order according to the xarray.
// eg. for depend x[i], y[j] and data w[i*ny + j], accepts x and w.
// If wanted to sort y, need to re-order w to become w[j*nx + i]
void QsurfaceWindow::quick_sort_2d(PLFLT xarray[], PLFLT z[], int start, int end, int length, PLFLT x_plus[], PLFLT x_minus[]) {
	if (start < end) {
		PLFLT pivot = xarray[end];
		PLFLT pivot_plus = x_plus[end];
		PLFLT pivot_minus = x_minus[end];
		PLFLT w[length];
		int j;
		for (j = 0; j < length; j++) {
			w[j] = z[end * length + j];
		}
		int bottom = start - 1;
		int top = end;

		bool done = false;
		while (!done) {
			while (!done) {
				bottom += 1;

				if (bottom == top) {
					done = true;
					break;
				}
				if (xarray[bottom] > pivot) {
					xarray[top] = xarray[bottom];
					x_plus[top] = x_plus[bottom];
					x_minus[top] = x_minus[bottom];
					for (j = 0; j < length; j++) {
						z[top * length + j] = z[bottom * length + j];
					}
					break;
				}
			}
			while (!done) {
				top = top - 1;

				if (top == bottom) {
					done = true;
					break;
				}
				if (xarray[top] < pivot) {
					xarray[bottom] = xarray[top];
					x_plus[bottom] = x_plus[top];					
					x_minus[bottom] = x_minus[top];					
					for (j = 0; j < length; j++) {
						z[bottom * length + j] = z[top * length + j];
					}
					break;
				}
			}
		}
		xarray[top] = pivot;
		x_plus[top] = pivot_plus;
		x_minus[top] = pivot_minus;
		
		for (j = 0; j < length; j++) {
			z[top * length + j] = w[j];
		}

		quick_sort_2d(xarray, z, start, top - 1, length, x_plus, x_minus);
		quick_sort_2d(xarray, z, top + 1, end, length, x_plus, x_minus);
	}
}


//	near_nice_number() and nice_min() are used to find nice values to draw contour lines.
//	nice_min() rounds down zmin to a nice number. How much it gets rounded down depends on the stepsize of the contour lines.
//	near_nice_number() is used to round stepsize into nice numbers.
//	Hence each contour level sits on (nice min) + integer * (nice number)


//	find a nice number near to x. Nice means 1,2, or 5 times power of 10
//	The Nice number is always larger in magnitude than x
PLFLT QsurfaceWindow::near_nice_number(PLFLT x) {
	double xlog, r, temp;
	int pow_of_10, pos, inum;

	pos = (x > 0); // for x < 0, need to strip and keep sign info
	xlog = SaferLog10(fabs(x));
	pow_of_10 = (int) xlog; // find nearest power of 10; if < 0, dec by 1


	if (pow_of_10 <= 0)
		pow_of_10--;
	temp = xlog - pow_of_10; // temp is now log of number part only

	if (temp > 1) {
		temp -= 1;
		pow_of_10++;
	} // fudge for vals 0.1-0.9
	r = pow(10.0, temp);
	//   inum = (int) r;
	inum = (int) r;
	if (r / inum > 1.0)
		inum++;
/*	if (inum > 5)
		inum = 10;
	else if (inum > 2)
		inum = 5;
	else if (inum > 1)
		inum = 2;
	else
		inum = 1;*/
	r = inum * pow(10.0, pow_of_10);
	if (!pos)
		r *= -1;

	return (PLFLT) r;

}

PLFLT QsurfaceWindow::nice_min(PLFLT zmin, PLFLT stepsize) {
	int pos = (zmin > 0);
	PLFLT nice_min;
	int steppow_of_10 = (int)floor(SaferLog10(stepsize));

	int zminpow_of_10 = (int)floor(SaferLog10(abs(zmin)));

	PLFLT tmp = SaferLog10(abs(zmin)) - zminpow_of_10;

	if (zminpow_of_10 - steppow_of_10 != 0) {
		if (pos)
			nice_min = floor(pow(10.0, tmp - steppow_of_10 + zminpow_of_10 - 1));
		else
			nice_min = ceil(pow(10.0, tmp - steppow_of_10 + zminpow_of_10 - 1));
		nice_min *= pow(10.0, steppow_of_10 + 1);

	} else //if the step size is close to the value of zmin don't round down zmin too much.
	{
		if (pos)
			nice_min = floor(pow(10.0, tmp - steppow_of_10 + zminpow_of_10));
		else
			nice_min = ceil(pow(10.0, tmp - steppow_of_10 + zminpow_of_10));
		nice_min *= pow(10.0, steppow_of_10);
	}

	if (!pos)
		nice_min *= -1;

	return nice_min;
}




//Draws the colourscale. This is achived by drawing 2000 lines in a window so it looks continuous.
void QsurfaceWindow::drawcolscale() {

	bool OK;

	plwid(QtWidth2PlWidth(axis_width_fld->value()));

	select_vpor(5);
	
	if (pageOrientation->currentIndex() == 1) {
		
		// Landscape
		plwind(151.0, 161.0, zmin, zmax);
		PLFLT x[4], y[4];
	
		x[0]=151.0;
		x[1]=151.0;
		x[2]=161.0;
		x[3]=161.0;

		PLFLT zstep=(zmax-zmin)/200.;
		PLFLT i=0.;
		for(PLFLT col=zmin;col<zmax;col+=zstep, i+=1./200.)
		{
			y[0] = col;
			y[3] = col;
			y[1] = col+zstep;
			y[2] = y[1];
			if(i>=1.) i=0.999;
			plcol1(i);
			plfill(4, x, y);
		}

		plcol0(0);
		if (colourLog->isChecked())
			plbox("", 0.0, 0, "bnlstv", 0.0, 0);
		else
			plbox("", 0.0, 0, "bnstv", 0.0, 0);

		plmtex("l", 4.5, 0.5, 0.5, colLabSlot->text().toLatin1().data());

		//Drawing contour lines next to the box


		if (showContLabel->isChecked()) {

			char level[50];
			PLFLT clevel[100];
		
			int steps = cont_lvl(clevel);

			for (int i = 0; i < steps; i++) {
				pljoin(156, clevel[i], 161, clevel[i]);pljoin(clevel[i], 156, clevel[i], 161);
				if (colourLog->isChecked()) 	
					sprintf(level, "10#u%.2f", clevel[i]);
				else 
					sprintf(level, "%-.2g", clevel[i]);
				if (clevel[i] - zmin >= 0)
					plmtex("rv", 0.5, (clevel[i] - zmin) / (zmax - zmin), 0, level);
			}
		}

		//Other colours
		PLFLT xm[] = { 0, 0, 1, 1 };
		PLFLT ym[] = { 0, 1, 1, 0 };
		char max[50], min[50];

		if (colourLog->isChecked()) {
			sprintf(max, ">10#u%.2f", zmax);
			sprintf(min, "<10#u%.2f", zmin);
		} 
		else {
			sprintf(max, ">%.2e", zmax);
			sprintf(min, "<%.2e", zmin);
		}

		select_vpor(6);
		plwind(0, 1, 0, 1);
		plcol1(1.0);
		plfill(4, xm, ym);
		plcol0(0);
		plbox("bc", 0.0, 0, "bc", 0.0, 0);
		plmtex("t", 1.5, 0.5, 0.5, max);

		select_vpor(7);
		plwind(0, 1, 0, 1);
		plcol1(0.0);
		plfill(4, xm, ym);
		plcol0(0);
		plbox("bc", 0.0, 0, "bc", 0.0, 0);
		plmtex("b", 1.5, 0.5, 0.5, min);
	}
	else{				
		
		// Portrait
		// Put the colourbar horizontally so to maxmize use of space.
		
		PLFLT titleOffset = 0.0;
		if(showContLabel->isChecked()) titleOffset = 5;
		
		plwind(zmin, zmax, 151.0, 161.0);

		PLFLT x[4], y[4];
	
		y[0]=151.0;
		y[1]=151.0;
		y[2]=161.0;
		y[3]=161.0;

		PLFLT zstep=(zmax-zmin)/200.;
		PLFLT i=0.;
		for(PLFLT col=zmin;col<zmax;col+=zstep, i+=1./200.)
		{
			x[0] = col;
			x[3] = col;
			x[1] = col+zstep;
			x[2] = x[1];
			if(i>=1.) i=0.999;
			plcol1(i);
			plfill(4, x, y);
		}

		plschr(QPUI_CHAR_HEIGHT, 1.1*text_size_fld->value());

		plcol0(0);
		if (colourLog->isChecked())
			plbox("bnlstv", 0.0, 0, "", 0.0, 0);
		else
			plbox("bnstv", 0.0, 0, "", 0.0, 0);

		plmtex("t", 2.0+titleOffset, 0.5, 0.5, colLabSlot->text().toLatin1().data());


		// Drawing contour lines next to the box		
		if (titleOffset > 0.5) {
			// happens when contours are labelled on colour bar
			char level[50];
			PLFLT clevel[100];
		
			int steps = cont_lvl(clevel);
			for (int i = 0; i < steps; i++) {
				pljoin(clevel[i], 156, clevel[i], 161);
				if (colourLog->isChecked()) 	
					sprintf(level, "10#u%.2f", clevel[i]);
				else 
					sprintf(level, "%-.2g", clevel[i]);
				if (clevel[i] - zmin >= 0)
					plmtex("tv", 0.5, (clevel[i] - zmin) / (zmax - zmin), 0., level);
			}
		}

		//Other colours
		PLFLT xm[] = { 0, 0, 1, 1 };
		PLFLT ym[] = { 0, 1, 1, 0 };
		char max[50], min[50];

		if (colourLog->isChecked()) {
			sprintf(max, ">10#u%.2f", zmax);
			sprintf(min, "<10#u%.2f", zmin);
		} else {
			sprintf(max, ">%.2g", zmax);
			sprintf(min, "<%.2g", zmin);
		}

		plschr(QPUI_CHAR_HEIGHT, 1.0*text_size_fld->value());

		select_vpor(6);
		plwind(0, 1, 0, 1);
		plcol1(1.0);
		plfill(4, xm, ym);
		plcol0(0);
		plbox("bc", 0.0, 0, "bc", 0.0, 0);
		plmtex("b", 1.5, 0.0, 0.0, max);

		select_vpor(7);
		plwind(0, 1, 0, 1);
		plcol1(0.0);
		plfill(4, xm, ym);
		plcol0(0);
		plbox("bc", 0.0, 0, "bc", 0.0, 0);
		plmtex("b", 1.5, 0.0, 0.0, min);
		
		plschr(QPUI_CHAR_HEIGHT, 1.0*text_size_fld->value());

	}

}


// Sets up contour levels and puts it in an array. Returns the number of contour lines.
int QsurfaceWindow::cont_lvl(PLFLT clevel[], int nmax ) {

	bool OK;
	contStep = contSpace->text().toDouble(&OK);
	
	int steps = 0;
	if(!OK){

		PLFLT stepsize = (zmax - zmin) / 10;
		PLFLT nice_stepsize = near_nice_number(stepsize);
		PLFLT nice_zmin = nice_min(zmin, nice_stepsize);


		for (int i = 0; nice_stepsize * i + nice_zmin < zmax; i++) {
			if (nice_zmin + i * nice_stepsize > zmin) {
				clevel[steps] = nice_zmin + i * nice_stepsize;
				steps++;
			}
		}
		contSpace->setText(QString("%1").arg(nice_stepsize));
	}
	else{
					
		PLFLT nice_zmin = nice_min(zmin, contStep);
		for (int i = 0; contStep * i + nice_zmin < zmax && steps < nmax; i++) {
			if (nice_zmin + i * contStep > zmin) {
				clevel[steps] = nice_zmin + i * contStep;
				steps++;
			}
		}

	}
	return steps;
}



//Since lots of viewports are defined, only this part needs to be changed if the window size changes.
void QsurfaceWindow::select_vpor(int n)
{
	PLFLT x0, x1, y0, y1, xmean, ymean;

	PLFLT xcutoff;	//xcutoff, ycutoff is the one-sided fraction of the page that will be trimmed off if we cut a 1:sqrt(2) page out of the window.
	PLFLT ycutoff;

	bool landscape;
	
	// Getting plot option 

	if(pageOrientation->currentIndex() == 0) landscape = false;
	else landscape = true;

	//Try to fit the largest 1:sqrt(2) page into the window.

	xcutoff = (width*sqrt(2.) < (PLFLT) height? 0:((width-height/sqrt(2.))/2.)/(PLFLT)width);
	ycutoff = (width*sqrt(2.) > (PLFLT) height? 0:((height-width*sqrt(2.))/2.)/(PLFLT)height);

	if (landscape) {
		x0 = 0.1 + xcutoff * 0.6;
		y0 = 0.2 + ycutoff;
		y1 = 0.9 - ycutoff;
        x1 = x0 + (y1 - y0) * ((PLFLT) width / (PLFLT) height);
		xmean = (x0 + x1) / 2;
		ymean = (y0 + y1) / 2;

		switch (n) {
		case -1: //vpor for the title
			plvpor(0.02, 0.98, .8, y1+.03);
			break;
		case 0: //main vpor
			plvpor(x0, x1, y0+0.02, y1-0.02);
			break;
		case 1: //top right of main
			plvpor(xmean, x1, ymean, y1);
			break;
		case 2: //top left of main
			plvpor(x0, xmean, ymean, y1);
			break;
		case 3: //bottom left of main
			plvpor(x0, xmean, y0, ymean);
			break;
		case 4: //bottom right of main
			plvpor(xmean, x1, y0, ymean);
			break;
		case 5: //col scale
            plvpor(x1 + 0.05, x1 + 0.08, y0 + 0.07, y1 - 0.07);
			break;
		case 6: //max col
			plvpor(x1 + 0.05, x1 + 0.08, y1 - 0.06, y1 - 0.03);
			break;
		case 7: //min col
			plvpor(x1 + 0.05, x1 + 0.08, y0 + 0.03, y0 + 0.06);
			break;
		case 8: //axes label
			plvpor(xmean, x1, y0 - 0.1, y0);
			break;
		}
	} else {
		x0 = 0.05 + xcutoff;
		x1 = 0.95 - xcutoff;
		y0 = 0.60 - (x1 - x0) * 0.5 * ((PLFLT) width / (PLFLT) height);
		y1 = 0.60 + (x1 - x0) * 0.5 * ((PLFLT) width / (PLFLT) height);
		xmean = (x0 + x1) / 2;
		ymean = (y0 + y1) / 2;

		switch (n) {
		case -1: //vpor for the title
			plvpor(0.02, 0.98, 0, y1+.03);
			break;
		case 0: //main vpor
			plvpor(x0, x1, y0, y1);
			break;
		case 1: //top right of main
			plvpor(xmean, x1, ymean, y1);
			break;
		case 2: //top left of main
			plvpor(x0, xmean, ymean, y1);
			break;
		case 3: //bottom left of main
			plvpor(x0, xmean, y0, ymean);
			break;
		case 4: //bottom right of main
			plvpor(xmean, x1, y0, ymean);
			break;
		case 5: //col scale
			plvpor(x0 + 0.10, x1 - 0.10, 0.10, 0.14);
			break;
		case 6: //max col
			plvpor(x1 - 0.07, x1 - 0.03, 0.10, 0.14);
			break;
		case 7: //min col
			plvpor(x0 + 0.03, x0 + 0.07, 0.10, 0.14);
			break;
		case 8: //axes label
			plvpor(xmean, x1, y0 - 0.03, y0);
			break;
		}
	}
}



//Overloaded routines.
void plfill3(polygon *pol){

	PLFLT *x, *y, *z;

	int pts = pol->getpts();

	x = new PLFLT[pts];
	y = new PLFLT[pts];
	z = new PLFLT[pts];

	for(int i=0; i<pts; i++){
		x[i] = pol->getvert(i).getx();
		y[i] = pol->getvert(i).gety();
		z[i] = pol->getvert(i).getz();
	}

	plfill3(pts, x, y, z);
	
	delete x;
	delete y;
	delete z;
}

void plpoly3(polygon *pol){

	PLFLT *x, *y, *z;
	PLBOOL *draw;

	int pts = pol->getpts();

	x = new PLFLT[pts+1];
	y = new PLFLT[pts+1];
	z = new PLFLT[pts+1];
	draw = new PLBOOL[pts];

	for(int i=0; i<pts; i++){
		x[i] = pol->getvert(i).getx();
		y[i] = pol->getvert(i).gety();
		z[i] = pol->getvert(i).getz();
		draw[i] = true;
	}

	x[pts] = x[0];
	y[pts] = y[0];
	z[pts] = z[0];

	plpoly3(pts+1, x, y, z, draw, true);
	
	delete x;
	delete y;
	delete z;
	delete draw;

}


//A bit like quick_sort, but now we arrange everything in respect to projected depth on the x-y plane.
template <class T>

void quicksortpoly(vector <T*> &pol, int start, int end){
	if (start < end) {
		T* pivot = pol[end];
		int bottom = start - 1;
		int top = end;

		bool done = false;
		while (!done) {
			while (!done) {
				bottom += 1;

				if (bottom == top) {
					done = true;
					break;
				}
				if (pol[bottom]->getxydepth() > pivot->getxydepth()) {
					pol[top] = pol[bottom];
					break;
				}
			}
			while (!done) {
				top = top - 1;

				if (top == bottom) {
					done = true;
					break;
				}
				if (pol[top]->getxydepth() < pivot->getxydepth()) {
					pol[bottom] = pol[top];
					break;
				}
			}
		}
		pol[top] = pivot;

		//pivot = top;
		quicksortpoly(pol, start, top - 1);
		quicksortpoly(pol, top + 1, end);
	}
}


